(* Copyright (C) 2014, Daniel Wyckoff*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with  BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

Require Import FunctionalExtensionality.
Require Import ProofIrrelevance.

Definition type_eq_dec (T:Type) : Type :=
  forall (x y:T), {x = y} + {x <> y}.

Lemma type_eq_dec_eq :
  forall {T:Type}
         (pf pf' : type_eq_dec T),
    pf = pf'.
  intros T h1 h2.
  red in h1. red in h2.
  extensionality x. extensionality y.
  destruct (h1 x y) as [h3 | h3], (h2 x y) as [h4 | h4]; subst.
  f_equal. apply proof_irrelevance.
  contradict h4; auto. contradict h3; auto.
  f_equal. apply proof_irrelevance.
Qed.

Lemma impl_type_eq_dec_prod :
  forall {T U:Type},
    type_eq_dec T ->
    type_eq_dec U ->
    type_eq_dec (T*U).
  intros T U h1 h2.
  red in h1, h2. red.
  intros x y.
  destruct x as [x1 x2], y as [y1 y2]. 
  specialize (h1 x1 y1). specialize (h2 x2 y2).
  destruct h1 as [| h1], h2 as [| h2]; subst.
  left; auto.
  right. intro h3. inversion h3. contradiction.
  right. intro h3. inversion h3. contradiction.
  right. intro h3. inversion h3. contradiction.
Qed.


(*This doesn't seem to be that useful, but the next one is!*)
(*Wyckoff*)
Lemma match_eq_dec_sym :
  forall {T U:Type} (pf:type_eq_dec T) (x y:T)
         (bl br:U),
    match (pf x y) with
      | left _ => bl
      | right _ => br
    end =
    match (pf y x) with
      | left _ => bl
      | right _ => br
    end.
  intros T U h1 x y ? ?.
  destruct (h1 x y) as [h2 | h2], (h1 y x) as [h3 | h3]; auto; subst; try contradict h2; try contradict h3; auto.
Qed.


(* maybe I can twist this somewhere better.*)
(*
Lemma type_eq_dec_prod_impl :
  forall {T U:Type},
    type_eq_dec (T*U) ->
    {inhabited T} + {inhabited U} ->
    (type_eq_dec T)*(type_eq_dec U).
  intros T U h1 h3.
  destruct h3 as [h3 | h3].
  *)

Definition type_dep_eq_dec {T:Type} (P:T->Prop) : Type :=
  forall (a b:T), P a -> P b ->
                  {a = b} + {a <> b}.

(*Note so I don't try again : something like 
"Lemma type_dep_eq_dec_impl :
    forall {T:Type} (P:T->Prop),
      type_dep_eq_dec P ->
      excl_midP P -> type_eq_dec T."
can't be done.  But the opposite can:*)

Lemma type_dep_eq_dec_compat :
  forall {T:Type} (P:T->Prop),
    type_eq_dec T -> type_dep_eq_dec P.
  intros ? ? h1.
  red in h1; red.
  intros; apply h1.
Qed.

Definition dep_type_eq_dec {S} (T:S->Type) : Type :=
  forall s, type_eq_dec (T s).


Lemma sig_eq_dec :
  forall {T:Type} (pf:type_eq_dec T) (P:T->Prop),
    type_eq_dec {x:T | P x}.
  intros T pf P; red; red in pf.
  intros x y.
  destruct x as [x hx], y as [y hy].
  specialize (pf x y).
  destruct pf as [pf | pf]; subst.
  left; f_equal; apply proof_irrelevance.
  right; intro h1.
  apply (f_equal (@proj1_sig _ _)) in h1; simpl in h1; subst.
  contradict pf; auto.
Qed.

  

(*Hopefully one day I will be able to refactor all usages of this in bool2 to some 
consequence of Boolean deciable equality, and eliminate this axiom altogether. It's mainly used for decidable equality of arbitrarily infinite sets, which can perhaps be dealt with more ad hoc.*)
Axiom eq_dec : forall {T:Type}, type_eq_dec T.



(*This axiom is so far only "used" in an Ltac tactic, where you have to 
separate proof-terms by type.*)
(*
Axiom type_decidable : type_eq_dec Type.*)