(* Copyright (C) 2015-2019 Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.
 
You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*) 
     
Require Import SetUtilities.
Require Import TypeUtilities.
Require Import LogicUtilities.
Require Import FiniteMaps.
Require Import ArithUtilities.
Require Import ListUtilities.
Require Import FiniteBags.
Require Import NPeano.
Require Import FunctionProperties.
Require Import FunctionalExtensionality.
Require Import Description.
Require Import TypeUtilities2.
Require Import NaryFunctions.
Require Import DecidableDec.
Require Import SetUtilities2.
Require Import ProofIrrelevance.
Require Import ListUtilities2.
Require Import EqDec.
Require Import SetUtilities2_5.



Lemma fin_map_to_l_exp_seg_ex : 
  forall {T U:Type} (pfdt:type_eq_dec T) {A:Ensemble T} {B:Ensemble U} {def:U}
         (F:Fin_map A B def) (la:list T) (lb:list U),
    NoDup la ->
    list_to_set la = A -> list_to_set lb = B ->
    exists! f:{m:nat | m < (length la)} -> {y:U|In y lb},
               forall (x:T) (pf:In x la),
                 F |-> x = proj1_sig (f (exist (fun c=> c < length la) _ (lt_lind pfdt  la  _ pf))).
intros T U hdt A B def F la lb hn h3 h4. 
subst.
assert (h5:forall i:{m:nat | m < length la}, In (F |-> (nth_lt _ _ (proj2_sig i))) lb).
  intro i.
  rewrite list_to_set_in_iff. subst.
  apply fin_map_app_in.
  rewrite <- list_to_set_in_iff.
  pose proof (nth_lt_compat _ _ (proj2_sig i)) as h3.
  destruct h3 as [h3l h3r].
  apply in_nth_lt. 
exists (fun m=> (exist _ (F |-> (nth_lt _ _ (proj2_sig m)))
                       (h5 m))).
red; simpl. split.
intros x h6.
pose proof (lind_compat hdt _ _ h6) as h7. 
rewrite h7. reflexivity.
intros f h6. apply functional_extensionality.
intro i.
destruct i as [i h7]. simpl.
apply proj1_sig_injective. simpl. 
rewrite (h6 _ (in_nth_lt _ _ h7)).
f_equal. f_equal. apply proj1_sig_injective.
simpl.
eapply nth_lt_inj.
apply hn.
rewrite lind_compat.
reflexivity.
Qed.



Definition fin_map_to_l_exp_seg 
           {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
           (pfdt:type_eq_dec T)
           (F:Fin_map A B def) (la:list T) (lb:list U)
           (pfa:NoDup la) (pflsa:list_to_set la = A)
           (pflsb:list_to_set lb = B) : {m:nat | m < length la}->{y:U|In y lb} :=
  proj1_sig (constructive_definite_description _ (fin_map_to_l_exp_seg_ex pfdt F la lb pfa pflsa pflsb)).

Lemma fin_map_to_l_exp_seg_compat : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (pfdt:type_eq_dec T)
         (F:Fin_map A B def) (la:list T) (lb:list U)
         (pfa:NoDup la) (pflsa:list_to_set la = A)
         (pflsb:list_to_set lb = B),
    let f:=(fin_map_to_l_exp_seg pfdt F la lb pfa pflsa pflsb) in
    forall (x:T) (pf:In x la),
      F |-> x = proj1_sig (f ((exist (fun c=> c < length la) _ (lt_lind pfdt la x pf)))).
intros T U A B def h0 F la lb h1 h2 h3 f x h4.
unfold fin_map_to_l_exp_seg in f.
destruct constructive_definite_description as [f' h5]. simpl in f.
apply h5.
Qed.

Lemma fin_map_to_l_exp_bij : 
  forall {T U:Type} (pfdt:type_eq_dec T)
         (A:Ensemble T) (B:Ensemble U) (def:U)
         (la:list T) (lb:list U)
         (pfa:NoDup la) (pflsa:list_to_set la = A)
         (pflsb:list_to_set lb = B),
    bijective (fun f:Fin_map A B def=> fin_map_to_l_exp_seg pfdt f la lb pfa pflsa pflsb).
intros T U hdt A B def la lb h1 h2 h3. subst.
red. split.
red. intros F F' h2.
pose proof (fin_map_to_l_exp_seg_compat hdt F la lb h1 eq_refl eq_refl) as h3.
pose proof (fin_map_to_l_exp_seg_compat hdt F' la lb h1 eq_refl eq_refl) as h4.
apply fin_map_ext.
intros x h5.
rewrite <- list_to_set_in_iff in h5.

rewrite (h3 _ h5). rewrite (h4 _ h5).
f_equal. rewrite h2. 
reflexivity. 
red.
intro f.
pose [pr:T*U | exists pf:In (fst pr) la, snd pr = (proj1_sig (f (exist (fun c=> c < length la) _ (lt_lind hdt _ _ pf))))] as S. 
assert (h2:fun_well_defined S).
  red. split. intros x h2.
  destruct h2 as [h2]. destruct h2 as [y h2].
  exists y. red. split. split; auto. constructor. exists x. assumption.
  intros y' h3.
  destruct h3 as [h3l h3r].
  inversion h2 as [h4]. clear h2. simpl in h4. destruct h4 as [h4 h5].
  inversion h3r as [h6]. clear h3r. simpl in h6. destruct h6 as [h6 h7].
  pose proof (proof_irrelevance _ h4 h6); subst; auto.
  intros pr h2. destruct h2 as [h2]. destruct h2 as [h2 h3].
  split. constructor. exists (snd pr). constructor. simpl.
  exists h2. assumption.
  constructor. exists (fst pr). constructor. simpl.
  exists h2; auto.
red in h2.
assert (h3:dom_rel S = (list_to_set la)).
  apply Extensionality_Ensembles.
  red. split.
  red. intros x h3. destruct h3 as [h3]. destruct h3 as [y h3].
  inversion h3 as [h4]. clear h3. simpl in h4. destruct h4 as [h4].
  rewrite <- list_to_set_in_iff. assumption.
  red. intros x h3. constructor.
  rewrite <- list_to_set_in_iff in h3.
  exists (proj1_sig (f (exist (fun c => c < length la) _ (lt_lind hdt la x h3)))).
  constructor. simpl. exists h3; auto. 
rewrite h3 in h2. 
pose proof (list_to_set_finite la) as h4.
assert (h5:Finite (ran_rel S)).

  pose proof (fp_ran_rel_im' hdt h4  h2 def) as hran.
  rewrite hran.
  apply finite_image. apply h4.
pose (fin_map_intro _ _ def hdt h4 h5 _ h2) as F.
pose proof (list_to_set_finite lb) as h6.
assert (h7:Included (ran_rel S) (list_to_set lb)).
  red. intros y h8.
  destruct h8 as [h8]. destruct h8 as [x h8]. inversion h8 as [h9]. clear h8.
  simpl in h9. destruct h9 as [h9 ?]. subst.
  pose proof (proj2_sig (f (exist (fun c=> c < length la) _ (lt_lind hdt la x h9)))) as h10. simpl in h10.
  rewrite <- list_to_set_in_iff.
  assumption.
pose (fin_map_new_ran F h6 h7) as F'.
exists F'.
pose proof (fin_map_to_l_exp_seg_compat hdt F' la lb h1 eq_refl eq_refl) as h8.
apply functional_extensionality.
intro i. destruct i as [i h9].
apply proj1_sig_injective.
specialize (h8 (nth_lt la _  h9)).
pose proof (nth_lt_compat la _ h9) as h10.
pose proof (in_nth_lt la i h9) as h10'.
specialize (h8 h10').
unfold F' in h8.
rewrite <- fin_map_new_ran_compat in h8.
pose proof h10' as h10''.
rewrite list_to_set_in_iff in h10''.
pose proof (in_fin_map_to_fps F _ h10'') as h11.
rewrite h8 in h11.
unfold F in h11.
rewrite <- fin_map_to_fps_s_compat in h11.
inversion h11 as [h12]. simpl in h12. unfold F'.
destruct h12 as [h12 h13]. 
unfold F.
pose proof (lind_compat hdt la (nth_lt la  i h9) h10') as h14. 
apply nth_lt_inj in h14.
pose proof (proof_irrelevance _ h10' h12) as h15.
assert (h15':exist (fun c=> c < length la) _ (lt_lind hdt _ _  h12) = exist (fun m=> m < length la) i h9).
  apply proj1_sig_injective. simpl. 
  rewrite <- h15.  assumption.
rewrite h15' in h13. 
rewrite <- h13.
f_equal. f_equal. rewrite h15. rewrite h15'. 
reflexivity.
assumption.
rewrite <- list_to_set_in_iff.
assumption.
Qed.

 
Lemma card_l_exp_seg : 
  forall {T:Type} (l:list T) (n:nat), 
    NoDup l ->
    FiniteT_card _ (finite_l_exp_seg l n) = (length l)^n.
intros T l n h1.
pose proof (l_exp_seg_to_nprod_in_l_bij l n) as h2.
apply bijective_impl_invertible in h2.
pose proof (FiniteT_card_bijection).
pose proof (finite_l_exp_seg l n) as h3.
pose proof (FiniteT_card_bijection _ _ h3 _ h2) as h4.
pose proof (proof_irrelevance _ (finite_l_exp_seg l n) h3); subst.
rewrite <- h4.
pose proof (card_nprod_in_l _ h1 n) as h5.
pose proof (proof_irrelevance
              _ 
              (finite_nprod_in_l l n)
              (bij_finite ({m : nat | m < n} -> {x : T | In x l})
                          ({x : T | In x l} ^ n) (l_exp_seg_to_nprod_in_l l n)
                          (finite_l_exp_seg l n) h2)) as h6.
rewrite <- h6; auto.
Qed.



Lemma card_fin_maps : 
  forall {T U:Type} (pfdt:type_eq_dec T)
         (A:Ensemble T) (B:Ensemble U) (def:U)
         (pfa:Finite A) (pfb:Finite B),
    FiniteT_card (Fin_map A B def) (finitet_fin_maps _ _ def pfa pfb) =
    (card_fun1 B)^(card_fun1 A).
intros T U hdt A B def h1 h2.
pose proof (finite_set_list_no_dup _ h1) as h3. destruct h3 as [la h3].
destruct h3 as [h3l h3r].
pose proof (finite_set_list_no_dup _ h2) as h4. destruct h4 as [lb h4].
destruct h4 as [h4l h4r]. subst.
pose proof (fin_map_to_l_exp_bij hdt _ _ def _  lb h3r (eq_refl _) (eq_refl _)) as h5.
apply bijective_impl_invertible in h5.
pose proof (finitet_fin_maps _ _ def h1 h2) as h6.
pose proof (FiniteT_card_bijection _ _ h6 _ h5) as h7.
pose proof (proof_irrelevance
              _ h6
              (finitet_fin_maps (list_to_set la) (list_to_set lb) def h1 h2)); subst.
rewrite <- h7.
pose proof (card_l_exp_seg _  (length la) h4r) as h8.
pose proof (proof_irrelevance
              _ (finite_l_exp_seg lb (length la))
              (bij_finite (Fin_map (list_to_set la) (list_to_set lb) def)
                          ({m : nat | m < length la} -> {y : U | In y lb})
                          (fun f : Fin_map (list_to_set la) (list_to_set lb) def =>
                             fin_map_to_l_exp_seg hdt f la lb h3r eq_refl eq_refl)
                          (finitet_fin_maps (list_to_set la) (list_to_set lb) def h1 h2) h5)) as h9.
rewrite <- h9. clear h9.
rewrite h8.
pose proof (card_fun1_compat (list_to_set lb)) as h9.
pose proof (card_fun1_compat (list_to_set la)) as h10.
destruct h9 as [h9l h9r]. destruct h10 as [h10l h10r].
specialize (h9l h2). specialize (h10l h1).
pose proof (cardinal_length_compat la h3r) as h11.
pose proof (cardinal_length_compat lb h4r) as h12.
assert (h13:length la = card_fun1 (list_to_set la)).
  eapply cardinal_is_functional. apply h11. apply h10l. reflexivity.
assert (h14:length lb = card_fun1 (list_to_set lb)).
  eapply cardinal_is_functional. apply h12. apply h9l. reflexivity.
rewrite h13. rewrite h14.
reflexivity.
Qed. 


Lemma ex_same_im_subset_fin_map_inj :
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {def:U}
         (pfdt:type_eq_dec T) (F:Fin_map A B def),
    exists (A':Ensemble T) (pf:Included A' A),
      let F':=restriction F pf in
      im_fin_map F = im_fin_map F' /\
      inj F'.
intros T U A B def hdt F.
pose proof (fin_map_fin_dom F) as h1.
pose proof (ex_same_im_subset_sig_inj hdt h1 (fin_map_to_sig_fun F)) as h2.
destruct h2 as [A' [h2 [h3 h4]]].
destruct h3 as [h3]. destruct h3 as [h3a h3b].
exists A', h2.
simpl. split.
rewrite restriction_sig_fin_map_to_sig_fun in h3b.
do 2 rewrite im_fin_map_to_sig_fun in h3b.
pose proof (proof_irrelevance _ h2 h3a); subst; auto.
rewrite restriction_sig_fin_map_to_sig_fun in h4.
rewrite <- fin_map_inj_iff.
assumption.
Qed.




Lemma card_im_eq_inj :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (pfdt:type_eq_dec T)
         (F:Fin_map A B def),
    card_fun1 A  =
    card_fun1 (im_fin_map F) ->
    inj F.
intros T U A B def hdt F h1.
pose proof (ex_same_im_subset_fin_map_inj hdt F) as h5. simpl in h5.
destruct h5 as [A' [h5 [h6 h7]]].
rewrite h6 in h1.
pose proof (fin_map_fin_dom F) as hf.
pose proof (fin_map_card_im_le_dom' (restriction F h5)) as h8.
pose proof (incl_card_fun1 _ _ hf h5) as h9.
assert (h10:card_fun1 A' = card_fun1 A). omega.
apply incl_card_fun1_eq in h10; auto.
subst.
pose proof (restriction_trivial_eq F h5) as h10.
rewrite h10 in h7.
assumption.
Qed.


Lemma dom_ran_card_eq_bij : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (pfdt:type_eq_dec T) (F:Fin_map A B def),
    card_fun1 A = card_fun1 (im_fin_map F) ->
    card_fun1 B = card_fun1 (im_fin_map F) ->
    bij F.
intros T U A B def hdt F h1 h0.
pose proof (card_im_eq_inj hdt F h1) as h2.
symmetry in h0.
pose proof (im_ran_card_eq_surj F h0) as h3.
split; auto.
Qed.


(*This is for constructing a finite map from an explicit 
  list of pairs that conform to two easier to formulate 
  properties.*)
Lemma fps_to_f_eq_card_dom_rel_ex : 
  forall {T U:Type} (pfdt:type_eq_dec T)
         (pfdu:type_eq_dec U)
         (S:Ensemble (T*U)) (def:U),
    Finite S ->
    card_fun1 S = card_fun1 (dom_rel S) ->
    exists! (F:Fin_map (dom_rel S) (ran_rel S) def),
      S = [pr:T*U | snd pr = F |-> (fst pr) /\
                    Inn (dom_rel S) (fst pr)].
  intros T U hdt hdu S def h1 h2.
  pose proof (impl_type_eq_dec_prod hdt hdu) as hdtu.
pose proof (finite_inh_or_empty _ h1) as h0.
destruct h0 as [h0a | h0b].
destruct h0a as [s h0a].
assert (h3:forall x:sig_set S, Inn (dom_rel S) (fst (proj1_sig x))).
  intro p.
  destruct p as [p h4].
  simpl. constructor.  exists (snd p).
  rewrite <- surjective_pairing.
  assumption.
pose (fun x:sig_set S=>fst (proj1_sig x)) as f.
pose (sig_fun_to_fin_map hdtu f h1 (fst s)) as f'.
assert (h4: im_fin_map f' = dom_rel S). 
  apply Extensionality_Ensembles.
  red. split.
  intros x h4. destruct h4 as [x h4]. subst.
  pose proof (h3 (exist _ _ h4)) as h5.
  simpl in h5.
  unfold f'.
  rewrite (sig_fun_to_fin_map_compat hdtu f h1 (fst s) x h4) at 1.
  unfold f. simpl. assumption.
  red.
  intros x h4. destruct h4 as [h4]. destruct h4 as [y h4].
  unfold im_fin_map.
  apply Im_intro with (x, y); auto.
  unfold f'.
  rewrite (sig_fun_to_fin_map_compat hdtu f h1 (fst s) _ h4) at 1.
  unfold f. simpl.
  reflexivity.
rewrite <- h4 in h2.
pose proof @card_im_eq_inj.  
pose proof (card_im_eq_inj hdtu f' h2) as h5. 
pose (fin_map_left_inverse hdt f' h5 s) as f2.
assert (h6:Included (dom_rel S) (im_fin_map f')).
  rewrite h4. auto with sets.
pose (restriction f2 h6) as f3.
assert (h7:forall x:T, Inn (dom_rel S) x -> Inn (ran_rel S) (snd (f3 |-> x))).
  intros x h7.
  destruct h7 as [h7]. destruct h7 as [y h7].
  unfold f3. rewrite restriction_compat. 
  unfold f2. 
  assert (h8:Inn (im_fin_map f') x).
    rewrite h4. constructor. exists y; auto.
  destruct h8 as [x h8]. subst.
  rewrite fin_map_left_inverse_compat.
  constructor. exists (fst x). 
  rewrite <- surjective_pairing. assumption. assumption.
  constructor. exists y. assumption.
pose (fun x:sig_set (dom_rel S) => (snd (f3 |-> (proj1_sig x)))) as g.
pose (sig_fun_to_fin_map_ran hdt g  (dom_rel_finite S h1) def (ran_rel S)).
assert (h8:  Included (Im (full_sig (dom_rel S)) g) (ran_rel S)).
  red.
  intros x h9.
  destruct h9 as [x h9]. subst. clear h9.
  constructor. unfold g.
  destruct x as [x h8]. simpl.
  unfold f3. rewrite restriction_compat.
  unfold f2.
  rewrite <- h4 in h8.
  destruct h8 as [x h8]. subst.
  rewrite fin_map_left_inverse_compat.
  exists (fst x). rewrite <- surjective_pairing.
  assumption. assumption. assumption.
pose (sig_fun_to_fin_map_ran hdt g (dom_rel_finite S h1) def (ran_rel S) (ran_rel_finite S h1) h8) as f4.
exists f4.
red. split.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros pr h9.
constructor. split.
unfold f4. 
assert (h10:Inn (dom_rel S) (fst pr)).
  constructor. exists (snd pr). rewrite <- surjective_pairing.
  assumption.
pose proof (sig_fun_to_fin_map_ran_compat hdt g (dom_rel_finite S h1) def (ran_rel S) (ran_rel_finite S h1) h8 _ h10) as h11.
rewrite h11. 
unfold g. simpl.
unfold f3.
rewrite restriction_compat.
unfold f2.
assert (h12:fst pr = f' |-> pr).
  unfold f'.
  pose proof (sig_fun_to_fin_map_compat hdtu f h1 (fst s) pr h9) as h12.
  rewrite h12.
  unfold f. simpl.
  reflexivity.
rewrite h12.
rewrite fin_map_left_inverse_compat at 1.
reflexivity. assumption.
constructor.
exists (snd pr).
rewrite <- surjective_pairing.
assumption.
constructor.
exists (snd pr).
rewrite  <- surjective_pairing.
assumption.
(* >= *) 
red. intros p h9. 
destruct h9 as [h9]. destruct h9 as [h9 h10]. 
destruct p as [x y].
simpl in h9, h10.
rewrite <- h4 in h10.
destruct h10 as [p h10]. subst.
unfold f4.
destruct p as [x y].
assert (h13:Inn (dom_rel S) (f' |-> (x, y))).
  constructor.
  unfold f'.
  exists y.
  rewrite (sig_fun_to_fin_map_compat hdtu f h1 (fst s) _ h10).
  unfold f.
  simpl.
  assumption.
pose proof (sig_fun_to_fin_map_ran_compat hdt g (dom_rel_finite S h1) def (ran_rel S) (ran_rel_finite S h1) h8  _ h13) as h14.
rewrite h14 at 1.
unfold g, f3. simpl.
rewrite restriction_compat. unfold f2.
rewrite fin_map_left_inverse_compat.
simpl.
unfold f'.
pose proof (sig_fun_to_fin_map_compat hdtu f h1 (fst s) (x, y) h10) as h15.
rewrite h15.
unfold f.
simpl.
assumption.
assumption.
assumption.

intros k h9.
apply fin_map_ext.
intros x h10. 
unfold f4.
pose proof (sig_fun_to_fin_map_ran_compat hdt g (dom_rel_finite S h1) def (ran_rel S) (ran_rel_finite S h1) h8 _ h10) as h11.
rewrite h11 at 1.
unfold g, f3. simpl.
rewrite restriction_compat.
unfold f2.
pose h10 as h10'.
rewrite <- h4 in h10'.
destruct h10' as [p h10' ? h12].
rewrite h12.
rewrite fin_map_left_inverse_compat.
rewrite h9 in h10'.
destruct h10' as [h10'].
destruct h10' as [h10a h10b].
rewrite h10a.
f_equal.
unfold f'.
destruct p as [x' y'].
simpl in h10a, h10b. simpl.
assert (h13: Inn S (x', y')).
  rewrite h9. constructor. simpl. split; auto.
pose proof (sig_fun_to_fin_map_compat hdtu f h1 (fst s) (x', y') h13) as h14.
rewrite h14.
unfold f.
simpl.
reflexivity. assumption. assumption.
subst.
rewrite dom_rel_empty, ran_rel_empty.
exists (empty_map1 _ _ hdt def (Empty_set U) (Empty_is_finite U)).
red. split.
apply Extensionality_Ensembles.
red; split; auto with sets.
red.
intros x h3.
destruct h3 as [h3].
destruct h3; contradiction.
intros f h3.
symmetry.
pose proof (type_eq_dec_eq hdt (fin_map_type_eq_dec f)). subst.
apply empty_map1_compat.
Qed.

Definition fps_to_f_eq_card_dom_rel 
           {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
           (S:Ensemble (T*U)) (def:U)
           (pff:Finite S) (pfeq:card_fun1 S = card_fun1 (dom_rel S)) := proj1_sig (constructive_definite_description _ (fps_to_f_eq_card_dom_rel_ex pfdt pfdu S def pff pfeq)).

Lemma fps_to_f_eq_card_dom_rel_compat : 
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (S:Ensemble (T*U)) (def:U)
         (pff:Finite S) (pfeq:card_fun1 S = card_fun1 (dom_rel S)),
  let F := fps_to_f_eq_card_dom_rel pfdt pfdu S def pff pfeq in 
      S = [pr:T*U | snd pr = F |-> (fst pr) /\
                    Inn (dom_rel S) (fst pr)].
intros T U S hdt hdu def h1 h2 F.
unfold F, fps_to_f_eq_card_dom_rel.
destruct constructive_definite_description as [F' h3].
simpl.
assumption.
Qed.


Lemma fps_to_f_eq_card_dom_rel_compat' : 
  forall {T U:Type} (pfdt:type_eq_dec T)
         (pfdu:type_eq_dec U) (S:Ensemble (T*U)) (def:U)
         (pff:Finite S) (pfeq:card_fun1 S = card_fun1 (dom_rel S))
         (pr:T*U),
    Inn S pr ->
    fps_to_f_eq_card_dom_rel pfdt pfdu S def pff pfeq |-> (fst pr) = snd pr.
intros T U S def hdt hdu h1 h2 pr h3.
pose proof (fps_to_f_eq_card_dom_rel_compat S def hdt hdu h1 h2) as h4.
simpl in h4.
rewrite h4 in h3.
destruct h3 as [h3]. destruct h3 as [h3a h3b].
rewrite h3a.
reflexivity.
Qed.

 
Definition fin_map_compose_list_to_seg 
           {T U:Type} {l:list T} (pfdt:type_eq_dec T)
           {B:Ensemble U} {def:U}
           (a:Fin_map (list_to_set l) B def) :
  Fin_map (seg (length l)) B def. 
  pose proof (segT_injector_compat (list_to_segT_fun l)) as h1.
  simpl in h1.
  destruct h1 as [h1 h2].
  pose proof (im_list_to_segT_fun pfdt l) as hi.
  rewrite hi in h1.
  let P := type of h2 in match P with injective ?f' =>
                                      pose f' as f end.
  assert (h3:forall m:sig_set (seg (length l)),
               Inn (list_to_set l) (list_to_segT_fun l (st_conv_rev m))).
    intro m.
    pose proof (list_to_segT_fun_compat l (st_conv_rev m)) as h3.
    destruct h3 as [h3 h4].
    rewrite <- list_to_set_in_iff.
    assumption.
  pose (fun m:sig_set (seg (length l)) => a |-> (list_to_segT_fun l (st_conv_rev m))) as fc. 
  refine (sig_fun_to_fin_map_ran nat_eq_dec fc (finite_seg _) def B (fin_map_fin_ran a) _).
  red.
  intros x h4.
  destruct h4 as [x h4]. subst.
  unfold fc.
  eapply im_fin_map_ran.
  apply in_im_fin_map.
  apply h3.
Defined.



Lemma fin_map_compose_list_to_seg_compat :
  forall {T U:Type} {l:list T} {B:Ensemble U} {def:U}
         (pfdt:type_eq_dec T)
         (a:Fin_map (list_to_set l) B def),
    let f :=list_to_segT_fun l in
    let F := fin_map_compose_list_to_seg pfdt a in
    forall m:(length l) @,
      F |-> (proj1_sig m) = a |-> (f m).
  intros T U l B def hdt a f F m.
  unfold f.
  pose proof (list_to_segT_fun_compat l m) as h1.
  destruct h1 as [h1 h2].
  unfold F, fin_map_compose_list_to_seg.
  destruct (segT_injector_compat (list_to_segT_fun l)) as [h3  h4].
  match goal with |- sig_fun_to_fin_map_ran ?C ?D ?E ?F ?G ?H ?I |-> _ = _ => 
                  pose (sig_fun_to_fin_map_ran_compat C D E F G H I (proj1_sig m)) as h5 end.
  simpl in h5.
  assert (h6 :  Inn (seg (length l)) (proj1_sig m)).
    constructor. apply proj2_sig.
  specialize (h5 h6).
  rewrite h5 at 1.
  f_equal. f_equal.
  apply proj1_sig_injective; simpl.
  reflexivity.
Qed.


(*For use in functionality related to el_prod_lind.*)
Definition fin_map_compose_list_to_seg_lind
           {T U:Type} {l:list T} {B:Ensemble U} {def:U}
           (pfdt:type_eq_dec T)
           (a:Fin_map (list_to_set l) B def) (m:(length l) @) :
  Fin_map (segT_injector (list_to_segT_fun (firstn (proj1_sig m) l))) B def.
  pose proof (segT_injector_compat (list_to_segT_fun (firstn (proj1_sig m) l))) as h1.
  simpl in h1.
  destruct h1 as [h1 h2].
  pose proof (im_list_to_segT_fun pfdt (firstn (proj1_sig m) l)) as hi.
  rewrite hi in h1.
  let P := type of h2 in match P with injective ?f' =>
                                      pose f' as f end. 
  assert (h3 : forall c:sig_set  (segT_injector (list_to_segT_fun (firstn (proj1_sig m) l))),
               Inn (list_to_set l) (f c)).
    intro c. destruct c as [c hc]. destruct c as [c hc'].
    pose proof hc' as hc''.
    rewrite length_firstn in hc''.
    rewrite <- list_to_set_in_iff. 
    unfold f, restriction_fun.
    simpl.
    pose proof (list_to_segT_fun_compat (firstn (proj1_sig m) l)
                                         (exist (fun m0 : nat => m0 < length (firstn (proj1_sig m) l)) c hc')) as h4.
    simpl in h4.
    destruct h4 as [h4 h5].
    assert (h6:0 < proj1_sig m).
      destruct (zerop (proj1_sig m)) as [h6|]; auto.
      rewrite h6 in hc''. apply lt_n_0 in hc''. contradiction.
    eapply in_firstn.
    apply h4.
    apply lt_le_weak.
    apply proj2_sig.
  pose (fun  c:sig_set (segT_injector (list_to_segT_fun (firstn (proj1_sig m) l))) => a |-> (f c)) as fc.
  refine (sig_fun_to_fin_map_ran (A:= segT_injector (list_to_segT_fun (firstn (proj1_sig m) l)))  (segT_eq_dec _) fc (finite_segT_injector _) def B (fin_map_fin_ran a) _).
  red. intros x h4.
  destruct h4 as [x h4]. subst.
  unfold fc. 
  eapply im_fin_map_ran.
  apply in_im_fin_map.
  apply h3.
Defined.

Lemma fin_map_compose_list_to_seg_lind_compat :
  forall {T U:Type} {l:list T} {B:Ensemble U} {def:U}
         (pfdt:type_eq_dec T)
         (a:Fin_map (list_to_set l) B def) (m:(length l) @),
    let f := restriction_fun (list_to_segT_fun (firstn (proj1_sig m) l))
                             (segT_injector (list_to_segT_fun (firstn (proj1_sig m) l))) in
    let F := fin_map_compose_list_to_seg_lind pfdt a m in
    forall c:sig_set (segT_injector (list_to_segT_fun (firstn (proj1_sig m) l))),
      F |-> (proj1_sig c) = a |-> f c.
  intros T U l B def hdt a m f F c.
  unfold F, f, restriction_fun.
  pose proof (list_to_segT_fun_compat (firstn (proj1_sig m) l) (proj1_sig c)) as h1.
  destruct h1 as [h1 h2].
  unfold fin_map_compose_list_to_seg_lind.
  destruct (segT_injector_compat (list_to_segT_fun (firstn (proj1_sig m) l))) as [h3  h4].
  match goal with |- sig_fun_to_fin_map_ran ?C ?D ?E ?F ?G ?H ?I |-> _ = _ =>
                  pose (sig_fun_to_fin_map_ran_compat C D E F G H I (proj1_sig c)) as h5 end.
  simpl in h5.
  assert (h6 :  Inn
                (segT_injector (list_to_segT_fun (firstn (proj1_sig m) l)))
                (proj1_sig c)).
    apply proj2_sig.
  specialize (h5 h6).
  rewrite h5 at 1.
  f_equal.
Qed. 

 

(*For use in functionality related to el_prod_l *)
Definition fin_map_compose_list_to_seg_l
           {T U:Type} {l:list T} {B:Ensemble U} {def:U}
           (pfdt:type_eq_dec T)
           (a:Fin_map (list_to_set l) B def)  :
  Fin_map (segT_injector (list_to_segT_fun l)) B def. 
  pose proof (segT_injector_compat (list_to_segT_fun l)) as h1.
  simpl in h1.
  destruct h1 as [h1 h2].
  pose proof (im_list_to_segT_fun pfdt l) as hi.
  rewrite hi in h1. 
  let P := type of h2 in match P with injective ?f' =>
                                      pose f' as f end. 
  assert (h3:forall c:sig_set  (segT_injector (list_to_segT_fun l)),
               Inn (list_to_set l) (f c)).
    intro c.
    rewrite <- list_to_set_in_iff.
    unfold f.
    unfold restriction_fun.
    pose proof (list_to_segT_fun_compat l (proj1_sig c)) as h3.
    destruct h3; auto.
    pose (fun  c:sig_set (segT_injector (list_to_segT_fun l)) => a |-> (f c)) as fc.
    refine (sig_fun_to_fin_map_ran (segT_eq_dec _) fc (finite_segT_injector _) def B (fin_map_fin_ran a) _).
  red. intros x h4.
  destruct h4 as [x h4]. subst.
  unfold fc. 
  eapply im_fin_map_ran.
  apply in_im_fin_map.
  apply h3.
Defined.


Lemma fin_map_compose_list_to_seg_l_compat :
  forall  {T U:Type} {l:list T} {B:Ensemble U} {def:U}
          (pfdt:type_eq_dec T)
          (a:Fin_map (list_to_set l) B def),
    let f := restriction_fun (list_to_segT_fun l)
                             (segT_injector (list_to_segT_fun l)) in
    let F := fin_map_compose_list_to_seg_l pfdt a in
    forall c:sig_set (segT_injector (list_to_segT_fun l)),
      F |-> (proj1_sig c) = a |-> f c.
  intros T U l B def hdt a f F c.
  unfold F, f.
  unfold restriction_fun.
  pose proof (list_to_segT_fun_compat l (proj1_sig c)) as h1.
  destruct h1 as [h1 h2].
  unfold fin_map_compose_list_to_seg_l.
  destruct (segT_injector_compat (list_to_segT_fun l)) as [h3  h4].
  match goal with |- sig_fun_to_fin_map_ran ?C ?D ?E ?F ?G ?H ?I |-> _ = _ =>
                  pose (sig_fun_to_fin_map_ran_compat C D E F G H I (proj1_sig c)) as h5 end.
  simpl in h5.
  assert (h6 :  Inn (segT_injector (list_to_segT_fun l))
                (proj1_sig c)).
    apply proj2_sig.  
  specialize (h5 h6).
  rewrite h5 at 1.
  f_equal.
Qed.



Lemma fin_map_seg_transfer_ex : 
  forall {U:Type} {def:U} {B:Ensemble U} {m n:nat}
         (f:Fin_map (seg m) B def)
         (pf:m = n),
  exists! F:Fin_map (seg n) B def,
    forall b:n @,
      F |-> (proj1_sig b) = f |-> (proj1_sig (segT_transfer_r pf b)).
  intros U def B m n f h1.
  subst.
  exists f.
  red.
  split.
  intro b.
  rewrite segT_transfer_r_eq_refl.
  reflexivity.
  intros f' h1.
  apply fin_map_ext.
  intros a h2.
  destruct h2 as [h2].
  specialize (h1 (exist (fun c => c < n) _ h2)).
  simpl in h1.
  rewrite h1; auto.
Qed.

Definition fin_map_seg_transfer
           {U:Type} {def:U} {B:Ensemble U} {m n:nat}
           (f:Fin_map (seg m) B def)
           (pf:m = n) :
  Fin_map (seg n) B def :=
  proj1_sig (constructive_definite_description _ (fin_map_seg_transfer_ex f pf)).

Lemma fin_map_seg_transfer_compat :
  forall {U:Type} {def:U} {B:Ensemble U} {m n:nat}
         (f:Fin_map (seg m) B def)
         (pf:m = n),
    let F := fin_map_seg_transfer f pf in
    forall b:n @,
      F |-> (proj1_sig b) = f |-> proj1_sig (segT_transfer_r pf b).
  unfold fin_map_seg_transfer; intros; destruct constructive_definite_description; auto.
Qed.

Lemma fin_map_seg_transfer_compat' :
  forall {U:Type} {def:U} {B:Ensemble U} {m n:nat}
         (f:Fin_map (seg m) B def)
         (pf:m = n) (b:nat),
    b < n ->
    fin_map_seg_transfer f pf |-> b = f |-> b.
  intros U def B m n f h1 b h2.
  pose proof (fin_map_seg_transfer_compat f h1) (exist (fun c => c < n) _ h2) as h3.
  simpl in h3.
  rewrite h3.
  f_equal.
  rewrite segT_transfer_r_compat.
  simpl.
  reflexivity.
Qed.

Lemma fin_map_seg_transfer_eq_refl :
  forall {U:Type} {def:U} {B:Ensemble U} {m:nat}
         (f:Fin_map (seg m) B def),
    fin_map_seg_transfer f eq_refl =
    f.
  intros; apply fin_map_ext; auto; intros; rewrite fin_map_seg_transfer_compat'; auto.
  destruct H; auto.
Qed.


Lemma fin_map_seg_transfer0 :
  forall {U:Type} {def:U} {B:Ensemble U} {m:nat}
         (F:Fin_map (seg 0) B def),
    F = fin_map_dom_subst (eq_sym seg_0_eq) (empty_map1 _ _ nat_eq_dec _ _ (fin_map_fin_ran F)).
  intros;  apply fin_map_ext; intros; destruct H; omega.
Qed.


Lemma fin_map_seg_one_transfer_ex : 
  forall {U:Type} {def:U} {B:Ensemble U} {m n:nat}
         (f:Fin_map (Full_set (seg_one_t m)) B def)
         (pf:m = n),
  exists! F:Fin_map (Full_set (seg_one_t n)) B def,
    forall b:seg_one_t n,
      F |-> b = f |-> sot_transfer_r pf b.
  intros U def B m n f h1.
  subst.
  exists f.
  red.
  split.
  intro b.
  rewrite sot_transfer_r_eq_refl.
  reflexivity.
  intros f' h1.
  apply fin_map_ext.
  intros a h2. clear h2.
  specialize (h1 a).
  rewrite h1.
  rewrite sot_transfer_r_eq_refl.
  reflexivity.
Qed. 

Definition fin_map_seg_one_transfer
           {U:Type} {def:U} {B:Ensemble U} {m n:nat}
           (f:Fin_map (Full_set (seg_one_t m)) B def)
           (pf:m = n) : Fin_map (Full_set (seg_one_t n)) B def :=
  proj1_sig (constructive_definite_description _ (fin_map_seg_one_transfer_ex f pf)).

Lemma fin_map_seg_one_transfer_compat :
  forall {U:Type} {def:U} {B:Ensemble U} {m n:nat}
         (f:Fin_map (Full_set (seg_one_t m)) B def)
         (pf:m = n),
    let F := fin_map_seg_one_transfer f pf in
    forall b:seg_one_t n,
      F |-> b = f |-> sot_transfer_r pf b.
  unfold fin_map_seg_one_transfer; intros; destruct constructive_definite_description; auto.
Qed.



Lemma fin_map_seg_pred_ex' :
  forall {U:Type} {def:U} {B:Ensemble U} {n:nat}
         (f:Fin_map (Full_set ((S n) @)) B def),
  exists! f' : Fin_map (Full_set (n @)) B def,
    forall m:n @,
      f' |->  m = f |-> (segT_S m).
  intros U def B n f.
  pose (fun m:sig_set (Full_set (n @)) => f |-> (segT_S (proj1_sig m))) as g.
  pose (sig_fun_to_fin_map_ran (segT_eq_dec _) g (finite_full_set_segT  _) def _ (fin_map_fin_ran f)) as F. 
  let P := type of F in match P with ?hyp -> _ => assert (h1:hyp) end.
    red. intros x h1.
    destruct h1 as [x h1]. subst. unfold g.
    apply fin_map_app_in. constructor.
  pose (F h1) as F'.
  exists F'.
  red. split.
  unfold F', F.
  intro m.
  pose proof (sig_fun_to_fin_map_ran_compat (segT_eq_dec _) g (finite_full_set_segT _) def _ (fin_map_fin_ran f) h1 m (Full_intro _ _)) as h2.
  rewrite h2.
  unfold g. simpl.
  reflexivity.
  intros G h2.
  apply fin_map_ext.
  intros m h3.
  specialize (h2 m).
  rewrite h2.
  unfold F', F.
  pose proof (sig_fun_to_fin_map_ran_compat (segT_eq_dec _) g (finite_full_set_segT _) def _ (fin_map_fin_ran f) h1 m (Full_intro _ _)) as h4.
  rewrite h4.
  unfold g.
  simpl.
  reflexivity.
Qed.

Definition fin_map_seg_pred'
           {U:Type} {def:U} {B:Ensemble U} {n:nat}
           (f:Fin_map (Full_set ((S n) @)) B def) :=
  proj1_sig (constructive_definite_description _ (fin_map_seg_pred_ex' f)).

Lemma fin_map_seg_pred_compat' :
   forall {U:Type} {def:U} {B:Ensemble U} {n:nat}
          (f:Fin_map (Full_set ((S n) @)) B def),
     let f' := fin_map_seg_pred' f in
     forall m:n @,
       f' |->  m = f |-> (segT_S m).
  unfold fin_map_seg_pred'; intros; destruct constructive_definite_description; auto.
Qed.



Lemma fin_map_seg_pred_ex :
  forall {U:Type} {def:U} {B:Ensemble U} {n:nat}
         (f:Fin_map (seg (S n)) B def),
  exists! f' : Fin_map (seg n) B def,
    forall m:nat, m < n ->
      f' |->  m = f |-> S m.
  intros U def B n f.
  pose (fun m:sig_set (seg n) => f |-> S (proj1_sig m)) as g.
  pose (sig_fun_to_fin_map_ran nat_eq_dec g (finite_seg  _) def _ (fin_map_fin_ran f)) as F. 
  let P := type of F in match P with ?hyp -> _ => assert (h1:hyp) end.
    red. intros x h1.
    destruct h1 as [x h1].  subst. clear h1.
    destruct x as [x h1]. unfold g.
    apply fin_map_app_in. constructor.
    simpl.
    apply lt_n_S. destruct h1; auto.
  pose (F h1) as F'.
  exists F'.
  red. split.
  unfold F', F.
  intros m hi.
  assert (hi':Inn (seg n) m).
    constructor; auto.
  pose proof (sig_fun_to_fin_map_ran_compat nat_eq_dec g (finite_seg  _) def _ (fin_map_fin_ran f) h1 m hi') as h2. 
  rewrite h2.
  unfold g. simpl.
  reflexivity.
  intros G h2. 
  apply fin_map_ext.
  intros m h3.
  pose proof h3 as h3'.
  destruct h3' as [h3'].
  specialize (h2 m).
  rewrite h2.
  unfold F', F.
  assert (hi':Inn (seg n) m).
  constructor; auto.
  pose proof (sig_fun_to_fin_map_ran_compat nat_eq_dec g (finite_seg  _) def _ (fin_map_fin_ran f) h1 m h3) as h4.
  rewrite h4.
  unfold g.
  simpl.
  reflexivity.
  assumption.
Qed.




Definition fin_map_seg_pred
           {U:Type} {def:U} {B:Ensemble U} {n:nat}
           (f:Fin_map (seg (S n)) B def) :=
  proj1_sig (constructive_definite_description _ (fin_map_seg_pred_ex f)).

Lemma fin_map_seg_pred_compat :
   forall {U:Type} {def:U} {B:Ensemble U} {n:nat}
          (f:Fin_map (seg (S n)) B def),
     let f' := fin_map_seg_pred f in
     forall m:nat,
       m < n  ->
       f' |->  m = f |-> (S m).
  unfold fin_map_seg_pred; intros; destruct constructive_definite_description; auto.
Qed.

(*strong inequality*)
Lemma fin_map_seg_rest_ex :
  forall {U:Type} {def:U} {B:Ensemble U} {n:nat}
         (f:Fin_map (seg n) B def) (m:nat) (pf:m < n),
  exists! F:Fin_map (seg m) B def,
    forall r:nat,
      r < m ->
      F |-> r = f |-> r.
  intros U def B n f m h1.
  exists (restriction f (incl_lt_seg _ _ h1)).
  red. split.
  intros r h2. 
  apply restriction_compat. constructor; auto.
  intros f' h2.
  apply fin_map_ext.
  intros r h3. destruct h3 as [h3].
  specialize (h2 _ h3).
  rewrite h2.
  apply restriction_compat. constructor; auto.
Qed.


Definition fin_map_seg_rest {U:Type} {def:U} {B:Ensemble U} {n:nat}
           (f:Fin_map (seg n) B def) (m:nat) (pf:m < n) :
  Fin_map (seg m) B def :=
  proj1_sig (constructive_definite_description _ (fin_map_seg_rest_ex f m pf)).

Lemma fin_map_seg_rest_compat :
  forall {U:Type} {def:U} {B:Ensemble U} {n:nat}
         (f:Fin_map (seg n) B def) (m:nat) (pf:m < n)
         (r:nat),
    let F := fin_map_seg_rest f m pf in  
    r < m ->
    F |-> r = f |-> r.
  unfold fin_map_seg_rest; intros; destruct constructive_definite_description; auto.
Qed.

(*weak inequality*)
Lemma fin_map_seg_rest_ex' :
  forall {U:Type} {def:U} {B:Ensemble U} {n:nat}
         (f:Fin_map (seg n) B def) (m:nat) (pf:m <= n),
  exists! F:Fin_map (seg m) B def,
    forall r:nat,
      r < m ->
      F |-> r = f |-> r.
  intros U def B n f m h1.
  apply le_lt_eq_dec in h1.
  destruct h1 as [h1|]; subst.
  exists (fin_map_seg_rest f m h1).
  red; split.
  intros; rewrite fin_map_seg_rest_compat; auto.
  intros f' h2.
  apply fin_map_ext.
  intros ? h3; rewrite h2, fin_map_seg_rest_compat; auto.
  destruct h3; auto.
  destruct h3; auto.
  exists f.
  red; split.
  intros; auto.
  intros f' h1.
  apply fin_map_ext; intros; auto.
  rewrite h1; auto.
  destruct H; auto.
Qed.

Definition fin_map_seg_rest' {U:Type} {def:U} {B:Ensemble U} {n:nat}
           (f:Fin_map (seg n) B def) (m:nat) (pf:m <= n) :
  Fin_map (seg m) B def :=
  proj1_sig (constructive_definite_description _ (fin_map_seg_rest_ex' f m pf)).

Lemma fin_map_seg_rest_compat' :
  forall {U:Type} {def:U} {B:Ensemble U} {n:nat}
         (f:Fin_map (seg n) B def) (m:nat) (pf:m <= n)
         (r:nat),
    let F := fin_map_seg_rest' f m pf in  
    r < m ->
    F |-> r = f |-> r.
  unfold fin_map_seg_rest'; intros; destruct constructive_definite_description; auto.
Qed.

Lemma fin_map_seg_rest_compat'' :
  forall {U:Type} {def:U} {B:Ensemble U} {n:nat}
         (f:Fin_map (seg n) B def) (m:nat) (pf:m < n)
         (r:nat),
    let pf' := lt_n_Sm_le _ _ (lt_trans _ _ _ pf (lt_n_Sn _)) in
    let F := fin_map_seg_rest f m pf in
    let F' := fin_map_seg_rest' f m pf' in
    r < m ->
    F |-> r = F' |-> r.
  unfold fin_map_seg_rest', fin_map_seg_rest; intros; do 2 destruct constructive_definite_description; simpl; auto.
  rewrite e, e0; auto.
Qed.
