(*Copyright (C) 2015-2019 Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

Require Import Setoid.
Require Import Basics.
Require Import NaryFunctions.
Require Import Infinite_sets.
Require Import NPeano.
Require Import ProofIrrelevance.
Require Import Description.

Require Import SetUtilities.
Require Import SetUtilities1_5.
Require Import ArithUtilities.
Require Import TypeUtilities.
Require Import FiniteMaps.
Require Import ListUtilities.
Require Import DecidableDec.
Require Import LogicUtilities.
Require Import Infinite_sets.
Require Import FunctionalExtensionality.
Require Import FunctionProperties. 
Require Import WellOrders.
Require Import Classical_Wf.
Require Import FiniteBags.
Require Import RelationClasses.
Require Import WellOrders.
Require Import EqDec.

(*Version 0.3 bool2 -- certified -- Version 8.4 pl4 Coq.*)

Section Transfer.

Lemma extends_sig1_transfer_fun_sig_iff :
  forall {T U:Type} {A A':Ensemble T} (pf:A = A')
         (f:T->U) (g:sig_set A->U),
    extends_sig1 f g <-> extends_sig1 f (transfer_fun_sig pf g).
  intros T U A A' h1 f g. subst.
  rewrite transfer_fun_sig_eq_refl.
  tauto.
Qed.


Lemma extends_sig_sig_transfer_fun_compat :
  forall {T U:Type} {A B B':Ensemble T} {E:Ensemble (sig_set A)}
         (f:sig_set B->U) (g:sig_set E->U)
         (pf:B = B') (pfs: sig_set B = sig_set B'),
    extends_sig_sig f g <->
    extends_sig_sig (transfer_fun pfs f) g.
  intros T U A B B' E f h1 h2 h3.
  subst.
  pose proof (proof_irrelevance _ h3 (eq_refl _)); subst.
  rewrite transfer_fun_eq_refl.
  tauto.
Qed.


Lemma extends_sig_sig_transfer_fun_sig1_iff :
  forall {T U:Type} {A B B':Ensemble T} {E:Ensemble (sig_set A)}
         (pf: B = B') (f:sig_set B->U) (g:sig_set E->U),
    extends_sig_sig f g <-> extends_sig_sig (transfer_fun_sig pf f) g.
  intros T U A B B' E h1 f g. subst.
  rewrite transfer_fun_sig_eq_refl.
  tauto.
Qed.

Lemma extends_sig_sig_transfer_fun_sig2_iff :
  forall {T U:Type} {A B:Ensemble T} {E E':Ensemble (sig_set A)}
         (pf: E = E') (f:sig_set B->U) (g:sig_set E->U),
    extends_sig_sig f g <-> extends_sig_sig f (transfer_fun_sig pf g).
  intros T U A B B' E h1 f g. subst.
  rewrite transfer_fun_sig_eq_refl.
  tauto.
Qed.


Lemma proj1_sig_snd_transfer_dep_incl :
  forall {T:Type} {A A':Ensemble T}
         (pf:A = A') (f:T*{C:Ensemble T | Included C A}->
                        T*{C:Ensemble T | Included C A})
         (pr:T*{C:Ensemble T | Included C A'}),
    proj1_sig (snd (transfer_dep
                      (U:=fun D => T*{C:Ensemble T | Included C D}->
                                   T*{C:Ensemble T | Included C D})
                      pf f pr)) =
    proj1_sig (snd (f (transfer_dep_r
                         (U :=
                            (fun D =>
                               (T*{C:Ensemble T | Included C D})%type))
                         pf pr))).
  intros; subst; rewrite transfer_dep_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.


Lemma proj1_sig_snd_transfer_dep_incl' :
  forall {T:Type} {A A':Ensemble T}
         (pf:A = A') (f:T*{C:Ensemble T | Included C A}->
                        T*{C:Ensemble T | Included C A})
         (pr:T*{C:Ensemble T | Included C A}),
    proj1_sig (snd (transfer_dep (U := (fun D =>
                                          (T*{C:Ensemble T | Included C D})%type)) pf (f pr))) =
    proj1_sig (snd (f pr)).
  intros; subst; rewrite transfer_dep_eq_refl; auto.
Qed.


Lemma proj1_sig_snd_transfer_dep_incl_r :
  forall {T:Type} {A A':Ensemble T}
         (pf:A' = A) (f:T*{C:Ensemble T | Included C A}->
                        T*{C:Ensemble T | Included C A})
         (pr:T*{C:Ensemble T | Included C A'}),
    proj1_sig (snd (transfer_dep_r
                      (U:=fun D => T*{C:Ensemble T | Included C D}->
                                   T*{C:Ensemble T | Included C D})
                      pf f pr)) =
    proj1_sig (snd (f (transfer_dep
                         (U := (fun D =>
                                  (T*{C:Ensemble T | Included C D})%type))
                         pf pr))).
  intros; subst; rewrite transfer_dep_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.


Lemma proj1_sig_snd_transfer_dep_incl_r' :
  forall {T:Type} {A A':Ensemble T}
         (pf:A' = A) (f:T*{C:Ensemble T | Included C A}->
                        T*{C:Ensemble T | Included C A})
         (pr:T*{C:Ensemble T | Included C A}),
    proj1_sig (snd (transfer_dep_r
                      (U :=
                         (fun D =>
                            (T*{C:Ensemble T | Included C D})%type))
                      pf (f pr))) =
    proj1_sig (snd (f pr)).
  intros; subst; rewrite transfer_dep_r_eq_refl; auto.
Qed.

Lemma im_full_sig_functional :
  forall {T U:Type} {A B:Ensemble T}
         (f:sig_set A -> U) (pf:A = B),
    Im (full_sig A) f =
    Im (full_sig B) (transfer_fun_sig pf f).
  intros T U A B f h1.
  subst.
  rewrite transfer_fun_sig_eq_refl; auto.
Qed.

Lemma im_ext_full_hetero :
  forall {T U V:Type} (pf:T = U) (f:T->V) (g:U->V),
    (forall x:T, f x = g (transfer pf x)) ->
    Im (Full_set T) f =
    Im (Full_set U) g.
  intros T U V h1 f g h2; subst.
  apply im_ext_in.
  intros x h3.
  rewrite h2.
  rewrite transfer_eq_refl; auto.
Qed.

Lemma im_ext_full_hetero_set1 :
  forall {T U:Set} {V:Type} (pf:T = U) (f:T->V) (g:U->V),
    (forall x:T, f x = g (s_transfer pf x)) ->
    Im (Full_set T) f =
    Im (Full_set U) g.
  intros T U V h1 f g h2; subst.
  apply im_ext_in.
  intros x h3.
  rewrite h2.
  rewrite s_transfer_eq_refl; auto.
Qed.


Lemma segT_functional_id :
  forall {m n:nat} (pf:m = n)
         (x:m @),
    let pfeq := segT_functional _ _ pf in
    proj1_sig x = proj1_sig (s_transfer pfeq x).
  intros m n h1 x; subst; simpl.
  destruct x as [x hx].
  simpl.
  rewrite segT_functional_eq_refl.
  rewrite s_transfer_eq_refl; auto.
Qed.

Lemma segT_functional_id_r :
  forall {m n:nat} (pf:n = m)
         (x:m @),
    let pfeq := segT_functional _ _ pf in
    proj1_sig x = proj1_sig (s_transfer_r pfeq x).
  intros m n h1 x; subst; simpl.
  destruct x as [x hx].
  simpl.
  rewrite segT_functional_eq_refl.
  rewrite s_transfer_r_eq_refl; auto.
Qed.



Lemma proj1_segT_transfer :
  forall {m n:nat} (pf:m = n)
         (x:m @)
         (pfeq:m @ = n @),
    proj1_sig x = proj1_sig (s_transfer pfeq x).
  intros m n h1 x h2; subst; simpl.
  rewrite (segT_functional_id (eq_refl)).
  repeat f_equal.
  apply proof_irrelevance.
Qed.

Lemma proj1_segT_transfer_r :
  forall {m n:nat} (pf:n = m)
         (x:m @)
         (pfeq:n @ = m @),
    proj1_sig x = proj1_sig (s_transfer_r pfeq x).
  intros m n h1 x h2; subst; simpl.   
  rewrite (segT_functional_id_r (eq_refl)). 
  repeat f_equal.
  apply proof_irrelevance.
Qed.



Lemma seg_one_t_functional_id :
  forall {m n:nat} (pf:m = n)
         (x:seg_one_t m),
    let pfeq := seg_one_t_functional _ _ pf in
    proj1_sig x = proj1_sig (s_transfer pfeq x).
  intros m n h1 x; subst; simpl.
  rewrite seg_one_t_functional_eq_refl; auto.
Qed.

Lemma proj1_seg_one_t_transfer :
  forall {m n:nat} (pf:m = n)
         (x:seg_one_t m)
         (pfeq:seg_one_t m = seg_one_t n),
    proj1_sig x = proj1_sig (s_transfer pfeq x).
  intros m n h1 x h2; subst; simpl.
  rewrite (seg_one_t_functional_id (eq_refl)).
  repeat f_equal.
  apply proof_irrelevance.
Qed.




  
End Transfer.



Section Card.
Lemma card_fun1_full_sig :
  forall {T:Type} (S:Ensemble T),
    Finite S ->
    card_fun1 (full_sig S) = card_fun1 S.
  intros T S h0.
  pose proof (im_proj1_sig_full_sig S) as h2.
  rewrite h2 at 3.
  eapply injective_preserves_card_fun1.
  red; intros; apply proj1_sig_injective; auto.
  rewrite <- finite_full_sig_iff.
  assumption.
Qed.

Lemma card_fun1_im_proj2_sig :
  forall {T:Type} {P:T->Prop} (A:Ensemble T)
         (pfx:(forall x:T, Inn A x -> P x)),
    Finite A ->
    card_fun1 A = card_fun1 (im_proj2_sig A pfx).
intros T P A h1 h2. 
rewrite <- card_fun1_full_sig; auto.
apply injective_preserves_card_fun1.
red. intros x y h3.
apply exist_injective in h3.
apply proj1_sig_injective; auto.
rewrite <- finite_full_sig_iff; auto.
Qed.


Lemma le_im_card' :
  forall {T U:Type} {A:Ensemble T} 
         (f:sig_set A-> U),
    Finite A -> 
    card_fun1 (Im (full_sig A) f) <= card_fun1 A.
  intros T U A f h1 .
  pose proof h1 as h1'.
  rewrite finite_full_sig_iff in h1'.
  pose proof (le_im_card  _ f h1') as h3.
  rewrite card_fun1_full_sig in h3; auto.
Qed.


Lemma preserves_card_fun1_injective :
  forall {T U:Type} {A:Ensemble T} 
         (f:sig_set A -> U),
    Finite A ->
    card_fun1 A = card_fun1 (Im (full_sig A) f) ->
    injective f.
intros T U A f h0 h1.
red; intros x y h4.
apply NNPP. intro ha. 
destruct x as [x h5], y as [y h6].
assert (h7 : Included (Couple x y) A).
  red; intros a h7; destruct h7; auto.
pose proof (decompose_setminus_inc _ _ h7) as h8.
pose proof (setminus_inc A (Couple x y)) as h9.
pose proof (f_equal card_fun1 h8) as h8'.
rewrite card_disj_union'' in h8'.
rewrite card_fun1_couple in h8'.
pose proof (im_full_sig_union_eq' f h7 h9 h8) as h10.
termr h10.
redterm0 y0.
rename c into S.
redterm1 y0.
rename c into R.

assert (h11 : R = Singleton (f (exist _ _ h5))).
  unfold R, restriction_sig.
  apply Extensionality_Ensembles; red; split; red; intros a h12.
  destruct h12 as [a h12]. clear h8. subst. clear h12.
  destruct a as [a h11]. destruct h11; simpl.
  pose proof (proof_irrelevance _ h5 (h7 x (Couple_l T x y)));subst.
  constructor.
  rewrite h4 at 1.
  pose proof (proof_irrelevance _ h6 (h7 y (Couple_r T x y))); subst.
  constructor.
  destruct h12.
  apply Im_intro with (exist _ _ (Couple_l _ x y)). constructor.
  f_equal. apply proj1_sig_injective; auto.
unfold R in h11.
rewrite h11 in h10.
destruct (classic_dec (Inn S (f (exist (Inn A) x h5)))) as [h13 | h13].
pose proof (inclusion_iff_union (Singleton (f (exist (Inn A) x h5))) S)
  as h14.
assert (h15 : Included (Singleton (f (exist (Inn A) x h5))) S).
  red; intros a h16.
  destruct h16.
  apply h13.
rewrite h14 in h15.
unfold S in h15.  
rewrite h15 in h10.
pose proof (f_equal card_fun1 h10) as h16.
rewrite h16 in h1.
rewrite h8' in h1.
symmetry in h1.
pose proof (finite_setminus A (Couple x y) h0) as h17.
redterm0 S.
rename S into W.
redterm0 W.
rename c0 into g.
pose proof (le_im_card' g h17) as h18.
unfold g in h18.
omega.
pose proof (f_equal card_fun1 h10) as h16.
rewrite card_disj_union'' in h16.
rewrite card_fun1_sing in h16.
rewrite h16 in h1.
rewrite h8' in h1.
pose proof (finite_setminus A (Couple x y) h0) as h17.
rename S into W.
redterm0 W.
redterm0 W.
rename c0 into g.
pose proof (le_im_card' g h17) as h19.
unfold g in h19.
omega.
apply Singleton_is_finite.
apply finite_image.
rewrite <- finite_full_sig_iff.
apply finite_setminus; auto.
apply Extensionality_Ensembles; red; split; red; intros c h17.
destruct h17 as [c h17 h18].
destruct h17.
clear h8.
inversion h18 as [c' h19 ? h20].
subst.
destruct c' as [c' h21].
unfold restriction_sig in h20.
simpl in h20.
contradiction.
destruct h17.
intro.
clear h8. subst. 
contradict ha.
apply proj1_sig_injective; auto.
apply finite_couple.
apply finite_setminus; auto.  
apply Extensionality_Ensembles; red; split; red; intros c h17.
destruct h17 as [c h17 h18].
destruct h17.
destruct h18 as [h18 h19].
contradict h19.
left.
destruct h18 as [h18 h19].
contradict h19.
right.
destruct h17.
Qed.

 

Lemma surj_impl_inj_sig_set_eq_card :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         (f:sig_set A -> sig_set B),
    Finite A -> Finite B ->
    card_fun1 A = card_fun1 B ->
    surjective f -> injective f.
intros T U A B f h1 h2 h3 h4. 
pose proof h4 as h4'.  
rewrite surj_iff' in h4'.
unfold im_proj1_sig in h4'. rewrite im_im in h4'.
rewrite <- h4' in h3.
red; intros x y h5.
destruct x as [x h6].
destruct y as [y h7].
pose proof h2 as h2'.
rewrite <- h4' in h2'.
pose proof (proj2_sig (f (exist _ _ h6))) as h8; simpl in h8.
rewrite <- h4' in h8 at 1.
pose proof (proj2_sig (f (exist _ _ h7))) as h9; simpl in h9.
rewrite <- h4' in h9 at 1.
pose proof (card_fun1_eq_sub
              _ _ h1 h2' h3 x
              (proj1_sig (f (exist _ _ h6))) h6 h8) as h10. 
pose proof (preserves_card_fun1_injective
              (fun x : sig_set A => proj1_sig (f x)) h1 h3) as h11.
red in h11.
pose proof (f_equal (@proj1_sig _ _) h5) as h5'.
specialize (h11 (exist _ _ h6) (exist _ _ h7) h5').
assumption.
Qed.


Lemma surj_impl_bij_sig_set_eq_card :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         (f:sig_set A -> sig_set B),
    Finite A -> Finite B ->
    card_fun1 A = card_fun1 B ->
    surjective f -> bijective f.
intros; split; try apply surj_impl_inj_sig_set_eq_card; auto.
Qed.


Lemma inj_impl_surj_sig_set_eq_card :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         (f:sig_set A -> sig_set B),
    Finite A -> Finite B ->
    card_fun1 A = card_fun1 B ->
    injective f -> surjective f.
intros T U A B f h1 h2 h3 h4.
rewrite surj_iff'.
unfold im_proj1_sig. rewrite im_im.
pose proof (injective_preserves_card_fun1 (fun x => proj1_sig (f x))) as hi.
hfirst hi.
red; intros x y h5. 
destruct x as [x h6], y as [y h7]. simpl in h5.
apply proj1_sig_injective in h5.
red in h4. apply h4 in h5.
assumption.
apply incl_card_fun1_eq; auto.
red; intros y h6. destruct h6 as [y h6]. subst. apply proj2_sig.
pose proof h1 as h1'.
rewrite finite_full_sig_iff in h1'.
specialize (hi hf).
pose proof (injective_preserves_card_fun1 _ hf (full_sig A) h1') as h6.
rewrite <- h6.
rewrite card_fun1_full_sig; auto.
Qed.


Lemma inj_impl_bij_sig_set_eq_card :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         (f:sig_set A -> sig_set B),
    Finite A -> Finite B ->
    card_fun1 A = card_fun1 B ->
    injective f -> bijective f.
intros; split; try apply inj_impl_surj_sig_set_eq_card; auto.
Qed.


  
Lemma card_pairwise_disjoint : 
   forall {T:Type} (F:Family T) (pff:Finite F),
    (forall S:Ensemble T, Inn F S -> Finite S) ->
    pairwise_disjoint F ->
    card_fun1 (FamilyUnion F) =
    plus_bag_nat (im_set_bag nat_eq_dec F pff card_fun1).
intros T F h1.
pose proof (finite_set_list_no_dup _ h1) as h2. 
destruct h2 as [l h2].
destruct h2 as [h2l h2r].
generalize dependent F. revert h2r.
induction l as [|A l h1]; simpl.
intros h1 F h2 h3 h4 h5. subst.
pose proof (im_set_bag_empty nat_eq_dec (@card_fun1 T)) as him.
pose proof (proof_irrelevance
              _ h2 
              (Empty_is_finite (Ensemble T))); subst. 
rewrite im_set_bag_empty. rewrite plus_bag_nat_empty.
rewrite empty_family_union. apply card_fun1_empty.
intros h2 F h4 h5 hin h6. 
pose proof (no_dup_cons_nin _ _ h2) as h3.
pose proof (f_equal (fun S=>Subtract S A) h5) as h7. simpl in h7.
rewrite list_to_set_in_iff in h3.
rewrite sub_add_same_nin in h7; auto.
pose proof (no_dup_cons _ _ h2) as h8.
pose proof (subtract_preserves_finite F A h4) as h9.
pose proof (incl_subtract F A) as h11.
specialize (h1 h8 _ h9 h7).
hfirst h1.
intros S h10. 
specialize (hin _ (h11 _ h10)); auto.
specialize (h1 hf
               (pairwise_disjoint_incl _ _ _ h6 h11)).
subst. 
rewrite family_union_add.
pose proof (sub_add_same_nin _ _ h3) as heq.
rewrite <- heq at 1; auto. 
pose proof (hin _ (Add_intro2 _ (list_to_set l) A)) as h13.
pose proof (Finite_Finite_Union (list_to_set l)) as h16.
hfirst h16.
intros S h15. apply hin; auto. left; auto.
specialize (h16 hf0).
hfirst h16.
eapply Finite_downward_closed. apply h4. auto with sets.
specialize (h16 hf1).
pose proof (sub_add_same_nin _ _ h3) as h17.
rewrite h17.
pose proof (proof_irrelevance _ h4 (Add_preserves_Finite _ _ _ hf1)); subst.
rewrite plus_bag_nat_im_set_bag_add.
pose proof (pairwise_disjoint_add_family_union _ _ h3 h6) as h18.
red in h18. rewrite Intersection_commutative in h18.
pose proof (card_disj_union' _ _ h13 h16 h18) as h19.
pose proof (card_fun_card_fun1_compat _ (Union_preserves_Finite _ _ _ h13 h16)) as h21. 
rewrite <- h21.
rewrite <- h19.
simpl in h17. simpl in h1. rewrite h17 in h1 at 1.
pose proof (subsetT_eq_compat _ _ _ _ h9 hf1 h17) as h22.
dependent rewrite -> h22 in h1.
rewrite <- h1.
rewrite (card_fun_card_fun1_compat _ h13).
rewrite (card_fun_card_fun1_compat _ h16).
ring.
assumption.
Qed.



(*See subsequent lemma for easing the finitude to just one set.*)
Lemma bij_ex_impl_eq_card : 
  forall {T U:Type} (pfdt:type_eq_dec T) (A:Ensemble T) (B:Ensemble U) 
         (pfa:Finite A) (pfb:Finite B) (P:T->U->Prop),
    (forall x:T, Inn A x -> 
                 exists! y:U, Inn B y /\ P x y) ->
    (forall y:U, Inn B y -> 
                 exists! x:T, Inn A x /\ P x y) ->
    card_fun1 A = card_fun1 B.
intros T U hdt A B h1 h2 P h3 h4.
destruct (finite_inh_or_empty_dec _ h2) as [hinh | h5].

destruct hinh as [d hinh]. 
pose [pr:T*U | exists (pfa:Inn A (fst pr)),
               exists (pfb:Inn B (snd pr)),
                 proj1_sig (constructive_definite_description _ (h3 _ pfa)) = snd pr /\ 
                 proj1_sig (constructive_definite_description _ (h4 _ pfb)) = fst pr] as S.
assert (h5:fun_well_defined S).
  red. split.
  intros x h5. destruct h5 as [h5]. destruct h5 as [y h5].
  exists y. red. split. split; auto. 
  constructor. exists x. assumption.
  intros y' h6. destruct h6 as [h6l h6r].
  destruct h5 as [h5]. destruct h5 as [h5l h5r]. destruct h5r as [h5a h5b]. 
  simpl in h5l. simpl in h5b. simpl in h5a.
  destruct constructive_definite_description as [b h8].
  destruct constructive_definite_description as [a h9].
  simpl in h5b. destruct h5b; subst.
  destruct h9 as [h9l h9r]. destruct h8 as [h10l h10r].
  destruct h6r as [h6r]. destruct h6r as [h11 h12]. destruct h12 as [h12 h13].
  destruct constructive_definite_description as [b' h8'].
  destruct constructive_definite_description as [a' h9'].
  simpl in h13. destruct h13; subst. simpl in h9'. simpl in h8'.
  destruct h9' as [h9a' h9b']. destruct h8' as [h8a' h8b'].
  pose proof (h3 _ h5l) as h10.
  destruct h10 as [y'' h10]. red in h10.
  destruct h10 as [h10a h10b]. destruct h10a as [h10a' h10b'].
  pose proof (h10b _ (conj h8a' h8b')). subst.
  pose proof (h10b _ (conj h10l h10r)). subst.
  reflexivity.
  intros pr h5. destruct h5 as [h5]. destruct h5 as [h5 h6].
  destruct h6 as [h6 h7].
  destruct constructive_definite_description as [b h8].
  destruct constructive_definite_description as [a h9].
  simpl in h7. destruct h7; subst.
  split. constructor. exists (snd pr). constructor. simpl.
  exists h5. exists h6. split.
  destruct constructive_definite_description as [y' h10]. simpl.
  destruct h10 as [h10l h10r].
  destruct h9 as [h9l h9r].
  pose proof (h3 _ h5) as h11. destruct h11 as [y h11].
  red in h11. destruct h11 as [h11l h11r].
  destruct h11l as [h11a h11b].
  pose proof (h11r _ (conj h10l h10r)). subst.
  pose proof (h11r _ h8). assumption.
  destruct constructive_definite_description as [a h10]. simpl.
  pose proof (h4 _ h6) as h11.
  destruct h11 as [a' h11]. red in h11. destruct h11 as [h11a h11b].
  pose proof (h11b _ h10). subst.
  pose proof (h11b _ h9). assumption.
  constructor. exists (fst pr).
  constructor.
  simpl. exists h5. exists h6.
  destruct constructive_definite_description as [b h10].
  destruct constructive_definite_description as [a h11].
  simpl. split.
  pose proof (h3 _ h5) as h12. destruct h12 as [b' h12r].
  red in h12r. destruct h12r as [h12a h12b].
  pose proof (h12b _ h10); subst.
  pose proof (h12b _ h8); subst. reflexivity.
  pose proof (h4 _ h6) as h12. destruct h12 as [a' h12r].
  red in h12r. destruct h12r as [h12a h12b].
  pose proof (h12b _ h11); subst.
  pose proof (h12b _ h9); subst. reflexivity.
red in h5.
assert (h6:dom_rel S = A).
  apply Extensionality_Ensembles.
  red. split. red. intros x h6.
  destruct h6 as [h6]. destruct h6 as [y h6].
  destruct h6 as [h6]. simpl in h6. destruct h6 as [h6]. assumption.
  red. intros x h6.
  constructor.
  exists  (proj1_sig
             (constructive_definite_description
                (fun y : U => Inn B y /\ P x y)
                (h3 _ h6))).
  constructor. simpl. exists h6.
  destruct constructive_definite_description as [y h7].
  simpl.
  destruct h7 as [h7l h7r].
  exists h7l. split; auto.
  destruct constructive_definite_description as [x' h8].
  simpl.
  destruct h8 as [h8l h8r].
  pose proof (h4 _ h7l) as h8.
  destruct h8 as [x'' h8]. red in h8.
  destruct h8 as [h8a h8b].
  pose proof (h8b _ (conj h8l h8r)); subst.
  pose proof (h8b _ (conj h6 h7r)); subst.
  reflexivity. 
rewrite h6 in h5.
assert (h7:ran_rel S = B).
  apply Extensionality_Ensembles.
  red. split.
  red.
  intros y h7. destruct h7 as [h7]. destruct h7 as [x h7].
  destruct h7 as [h7]. destruct h7 as [h7 h8]. destruct h8 as [h8 h9].
  simpl in h8. assumption.
  red. intros y h7. constructor.
  exists (proj1_sig
             (constructive_definite_description
                (fun x : T => Inn A x /\ P x y)
                (h4 _ h7))).
  constructor. simpl.
  destruct constructive_definite_description as [x h8]. simpl.
  destruct h8 as [h8l h8r].
  exists h8l. exists h7. split.
  destruct constructive_definite_description as [y' h9].
  simpl. 
  pose proof (h3 _ h8l) as h10.
  destruct h10 as [y'' h10]. red in h10.
  destruct h10 as [h10l h10r].
  pose proof (h10r _ (conj h7 h8r)). 
  pose proof (h10r _ h9). congruence.
  destruct constructive_definite_description as [x' h9].
  simpl.
  pose proof (h4 _ h7) as h10.
  destruct h10 as [x'' h10]. red in h10.
  destruct h10 as [h10l h10r].
  pose proof (h10r _ h9). 
  pose proof (h10r _ (conj h8l h8r)). congruence.
rewrite h7 in h5.
pose (fin_map_intro _ _ d hdt h1 h2 _ h5) as F.
apply bij_dom_ran_card_eq' with F.
red. split.
red.
intros x x' h8 h9 h10.
pose proof (in_fin_map_to_fps F x  h8) as h11.
pose proof (in_fin_map_to_fps F x' h9) as h12.
unfold F in h11. unfold F in h12.
rewrite <- fin_map_to_fps_s_compat in h11.
rewrite <- fin_map_to_fps_s_compat in h12.
unfold F in h10. rewrite h10 in h11.
destruct h11 as [h11]. simpl in h11. destruct h11 as [h11a h11b].
destruct h11b as [h11b h11c]. 
destruct constructive_definite_description as [b h13].
destruct (constructive_definite_description _ 
          (h4 (fin_fps_to_f hdt h1 S h5 d x') h11b)) as [a h14].
simpl in h11c. destruct h11c as [h11d h11e].
destruct h13 as [h13l h13r].
pose proof (h4 _ h13l) as h15.
destruct h15 as [x'' h15]. red in h15. destruct h15 as [h15l h15r].
pose proof (h15r _ (conj h8 h13r)).
rewrite <- h11d in h14.
pose proof (h15r _ h14).
inversion h12 as [h16]. clear h12. simpl in h16.
destruct h16 as [h16 h17]. destruct h17 as [h17 h18].
destruct constructive_definite_description as [b' h19].
destruct (constructive_definite_description _
          (h4 (fin_fps_to_f hdt h1 S h5 d x') h17)) as [a' h20].
simpl in h18. destruct h18 as [h18l h18r].
rewrite <- h11d in h20.
pose proof (h15r _ h20). congruence.
red.
intros b h8.
exists (proj1_sig (constructive_definite_description _ (h4 _ h8))).
split.
destruct constructive_definite_description as [a h9].
simpl. destruct h9; auto.
destruct constructive_definite_description as [a h9].
simpl.
destruct h9 as [h9l h9r].
pose proof (fin_fps_to_f_s_compat hdt h1 h5 d _ h9l) as h10.
destruct h10 as [h10]. simpl in h10.
destruct h10 as [h10 h11]. destruct h11 as [h11 h12].
destruct constructive_definite_description as [y h13].
destruct (constructive_definite_description _ 
           (h4 (fin_fps_to_f hdt h1 S h5 d a) h11)) as [x h14].
simpl in h12. destruct h12 as [h12l h12r].
pose proof (h3 _ h10) as h15.
destruct h15 as [y' h15]. red in h15.
destruct h15 as [h15l h15r].
rewrite <- h12r in h15r.
destruct h14 as [h14l h14r]. 
pose proof (h15r _ (conj h11 h14r)).
rewrite <- h12r in h13.
pose proof (h15r _ h13).
rewrite <- h12r in h9r.
pose proof (h15r _ (conj h8 h9r)). congruence.
subst.
pose proof (eq_empty A) as h7.
hfirst h7.
red; intros a h8.
specialize (h3 _ h8).
destruct h3 as [y h3].
red in h3; destruct h3 as [h3 h9].
destruct h3; contradiction.
apply h7 in hf; subst.
do 2 rewrite card_fun1_empty; auto.
Qed.


(*Uses LEM*)
(*Like above function, but with only one required finitude.*)
Lemma bij_ex_impl_eq_card' : 
  forall {T U:Type} (pfdt:type_eq_dec T)
         (A:Ensemble T) (B:Ensemble U) 
         (pfa:Finite A) (P:T->U->Prop),
    (forall x:T, Inn A x -> 
                 exists! y:U, Inn B y /\ P x y) ->
    (forall y:U, Inn B y -> 
                 exists! x:T, Inn A x /\ P x y) ->
    card_fun1 A = card_fun1 B.
intros T U hdt A B h1 P h2 h3.
assert (h4:Finite B).
apply NNPP.
intro h4.
assert (h5:~FiniteT (sig_set B)).
  intro h5. apply FiniteT_sig_Finite in h5. contradiction.
pose (fun x:sig_set B => 
        (proj1_sig (constructive_definite_description _ (h3 _ (proj2_sig x))))) as f.
rewrite <- Finite_FiniteT_iff in h5.
assert (h6:Finite (Im (Full_set (sig_set B)) f)).
  eapply Finite_downward_closed.
  apply h1. red. intros x h6.
  destruct h6 as [x h6]. subst.
  unfold f. destruct constructive_definite_description as [a h7].
  simpl. destruct h7; assumption.
pose proof (Pigeonhole_bis _ _ _ _ h5 h6) as h7.
contradict h7.
red.
intros x y h7.
destruct x as [x h8]. destruct y as [y h9]. unfold f in h7.
destruct constructive_definite_description as [a h10].
destruct constructive_definite_description as [a' h11].
simpl in h7. subst. simpl in h10. simpl in h11.
pose proof (h3 _ h9) as h12.
destruct h12 as [a h12]. red in h12.
destruct h12 as [h12l h12r].
pose proof (h12r _ h11). subst.
destruct h10 as [h10l h10r].
specialize (h2 _ h10l).
destruct h2 as [b h2]. red in h2. destruct h2 as [h2l h2r].
destruct h12l as [h12a h12b].
pose proof (h2r _ (conj h8 h10r)). subst.
pose proof (h2r _ (conj h9 h12b)). subst.
apply proj1_sig_injective.
simpl. reflexivity.
apply (bij_ex_impl_eq_card hdt  _ _  h1 h4 _ h2 h3).
Qed.

Lemma bij_eq_card : 
  forall {T U:Type} (pfdt:type_eq_dec T)
         (A:Ensemble T) (B:Ensemble U)
         (pfa:Finite A)
         (f:sig_set A -> sig_set B),
    bijective f ->
    card_fun1 A = card_fun1 B.
  intros T U h1 A B h2 f h3.
  red in h3.
  destruct h3 as [h3 h4].
  red in h3, h4.
  pose (fun (x:T) (y:U) =>  exists pf:Inn A x,
                                 y = proj1_sig (f (exist _ _ pf))) as P.
  apply (bij_ex_impl_eq_card' h1 A B h2 P) ; auto.
  intros x h5.
  exists (proj1_sig (f (exist _ _ h5))).
  red. split. split.
  apply proj2_sig.
  unfold P.
  exists h5.
  reflexivity.
  intros y h6.
  destruct h6 as [h6 h7].
  unfold P in h7.
  destruct h7 as [h7 h8]; subst.
  f_equal. f_equal. f_equal. apply proof_irrelevance.
  intros y h5.
  specialize (h4 (exist _ _ h5)).
  destruct h4 as [x h4].
  pose proof (f_equal (@proj1_sig _ _) h4) as h4'; simpl in h4'; subst.
  destruct x as [x hx].
  exists x.
  red; split. split.
  assumption.
  unfold P.
  exists hx.
  reflexivity.
  intros x' h6.
  destruct h6 as [h6 h7].
  unfold P in h7.
  destruct h7 as [h7 h8].
  apply proj1_sig_injective in h8.
  apply h3 in h8.
  apply exist_injective in h8; auto.
Qed.


End Card.

Section Sort.

(*These [pop_ens] lemmas are kind of deprecated, since you can achieve 
the same effect with [can_set_nat_choice].*)
Lemma pop_ens_iter_aux_functional :
  forall {T:Type} (S S':Ensemble T)
         (pff:Finite S) (pff':Finite S')
         (pfi:Inhabited S) (pfi':Inhabited S')
         (R:T->T->Prop) (pfso:sortable R)
         (def:T)
         (pfeq:S = S'),
    pop_ens_iter_aux S pff pfi R pfso def =
    transfer_dep_r (U:=fun C => T*{A:Ensemble T | Included A C}->
                                T*{A:Ensemble T | Included A C}) pfeq (pop_ens_iter_aux S' pff' pfi' R pfso def).
  intros T S S' h1 h2  h3 h4 R h5 h6 h7.
  subst.
  pose proof (proof_irrelevance _ h1 h2); subst.
  rewrite transfer_dep_r_eq_refl; auto.
Qed.

Lemma fst_pop_ens_iter_aux_functional :
  forall {T:Type} (A A':Ensemble T)
         (pff:Finite A) (pff':Finite A')
         (pfi:Inhabited A) (pfi':Inhabited A')
         (R:T->T->Prop) (pfso:sortable R)
         (pft:Transitive R) (pfir:irrefl R) (def:T)
         (pfeq:A = A')
         (pr:(T*{C:Ensemble T | Included C A})),
    fst (pop_ens_iter_aux A pff pfi R pfso def pr) =
    fst (pop_ens_iter_aux A' pff' pfi' R pfso def
                          (transfer_dep (U:=fun C:Ensemble T => (T*{B:Ensemble T | Included B C})%type) pfeq pr)).
  intros T A A' h1 h2 h3 h4 R h5 h6 h7 def h8 pr.
  subst.
  rewrite transfer_dep_eq_refl.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
Qed.




Definition pop_l_iter_aux {T:Type} (l:list T) (pf:l <> nil)
           (R:T->T->Prop) (pfso:sortable R) (def:T)  :=
  pop_ens_iter_aux (list_to_set l) (list_to_set_finite _) (not_nil_inhabited _  pf) R pfso def.


Lemma pop_l_iter_aux_eq :
  forall {T:Type} (l:list T) (pf:l <> nil)
         (R:T->T->Prop) (pfso:sortable R)
         (pft:Transitive R) (pfir:irrefl R) (def:T),
    pop_l_iter_aux l pf R pfso def =
    pop_ens_iter_aux (list_to_set l) (list_to_set_finite _) (not_nil_inhabited _  pf) R pfso def.
  intros? l; induction l; try reflexivity; unfold pop_l_iter_aux; auto.
Qed.


Lemma pop_ens_iter_sing_S :
  forall {T:Type} (a:T) (pff:Finite (Singleton a)) (pfi:Inhabited (Singleton a))
         (R:T->T->Prop) (pfso:sortable R) (def:T) (n:nat),
    pop_ens_iter (Singleton a) pff pfi R pfso def (S n) = def.
  intros T a h1 h0 R h3 def n.   
  unfold pop_ens_iter.
  rewrite pop_ens_iter_aux_sing.  
  gterml.
  redterm0 x.
  redterm1 c.
  redterm0 c0.
  simpl.  
  lr_match_goal'; simpl; auto.
  destruct n as [|n]. simpl in hr.
  rewrite min_set_sortable_sing, subtract_sing in hr.
  pose proof (In_singleton _ a) as h6.
  rewrite <- hr in h6.  
  contradiction.
  terml hr.
  redterm0 x0.
  redterm0 c2.
  redterm2 c3.
  redterm1 c3.
  pose (snd c5) as E.
  pose proof (snd_iter_uo_const c4 E) as h4; unfold c4, E in h4.
  hfirst h4.
  intro pr.
  
  lr_match_goal'; apply proj1_sig_injective; simpl.
  pose proof (min_set_sortable_compat (Singleton a) h1 h0 _ h3) as h5;
    simpl in h5.
  destruct h5 as [h5 h6].
  inversion h5 as [h7].
  rewrite h7 at 1.  
  rewrite subtract_sing; auto.
  pose proof (min_set_sortable_compat (Singleton a) h1 h0 _ h3) as h5; 
  simpl in h5.
  destruct h5 as [h5 h6].
  inversion h5 as [h7].
  rewrite h7 at 1.
  rewrite subtract_sing; auto.
  specialize (h4 hf c5 n).
  unfold c5 in h4.
  rewrite h4 in hr; simpl in hr.
  rewrite min_set_sortable_sing, subtract_sing in hr.
  symmetry in hr.
  apply sing_not_empty in hr; contradiction.
Qed.

(*The below is usually better than the above!*)
(*This is to choose a canonical (least) finite set of natural numbers 
  from a finite family.*)
Definition can_set_nat_choice
           (F:Family nat) (pff:Finite F) (pfi:Inhabited F)
           (pfs:forall S:Ensemble nat, Inn F S -> Finite S)
: Ensemble nat :=
  let f := (fun S => sort_set_nat (proj1_sig S) (pfs _ (proj2_sig S))) in
  list_to_set (min_set_sortable
                 (Im (full_sig F) f)
                 (finite_image _ _ _ f (iff1 (finite_full_sig_iff F) pff))
                 (inh_im (full_sig F) f (inh_full_sig F pfi))
                 lt_list_nat lt_list_nat_sortable).


Lemma in_can_set_nat_choice :
  forall (F:Family nat) (pff:Finite F) (pfi:Inhabited F)
         (pfs:forall S:Ensemble nat, Inn F S -> Finite S),
    Inn F (can_set_nat_choice F pff pfi pfs).
  intros F h1 h2 h3. unfold can_set_nat_choice.
  gterm0.
  redterm0 c.
  redterm4 c0.
  redterm3 c0.
  redterm2 c0.
  redterm1 c0.
  redterm0 c0.
  pose proof (min_set_sortable_compat c1 c2 c3 c4 c5) as h4.
  unfold c1, c2, c3, c4, c5 in h4. simpl in h4.
  destruct h4 as [h4 h5].
  inversion h4 as [F' h8 h6 h7]; subst. clear h8.
  destruct F' as [F' h8].
  simpl in h7.
  pose proof (in_list_reps_sort_set_nat F' (h3 F' h8)) as h9.
  destruct h9 as [h9]. destruct h9 as [h9 h10].
  rewrite <- h7 in h9.
  subst.
  assumption.
Qed.


Lemma can_set_nat_choice_sing :
  forall (A:Ensemble nat) (pff:Finite (Singleton A))
         (pfi:Inhabited (Singleton A))
         (pfs:forall S:Ensemble nat, Inn (Singleton A) S -> Finite S),
    can_set_nat_choice (Singleton A) pff pfi pfs = A.
  intros A h1 h2 h3.
  unfold can_set_nat_choice.
  match goal with |- list_to_set (min_set_sortable _  ?pf1 ?pf2 _ _ ) =
                     _ => generalize pf1 pf2 end.
  rewrite full_sig_sing.
  match goal with |- forall _ _,
                       list_to_set (min_set_sortable ?C _ _ _ _) = _ =>
                  assert (h4:C = Singleton (sort_set_nat A (h3 A (In_singleton (Ensemble nat) A)))) end.
  rewrite im_singleton; auto.
  intros h5 h6.
  rewrite (min_set_sortable_functional _ _ h5 (Singleton_is_finite _ _) h6 (inh_sing _) _ _  lt_list_nat_sortable h4).
  rewrite min_set_sortable_sing.
  apply sort_set_nat_compat.
Qed.



Definition min_ens_list_nat (F:Ensemble (list nat)) (pff:Finite F)
           (pfi:Inhabited F) : list nat :=
  min_set_sortable F pff pfi lt_list_nat lt_list_nat_sortable.


Lemma ens_ex_dec :
  forall {T:Type} (P:T->Prop) (R:Relation T) (A:Ensemble T),
    sort_dec R ->
    Finite A ->
    pred_dec P ->
    {exists x:T, Inn A x /\ P x} + {~exists x:T, Inn A x/\ P x}.
  intros T P R A hr hfa h1. 
  destruct (l_ex_dec (sort_dec_eq_dec _ hr) P
                     (sort_set R hr _ hfa) h1) as [h3 | h3].
  left.
  destruct h3 as [x [h3 h4]].
  rewrite <- in_sort_set_iff in h3.
  exists x; split; auto.
  right.
  contradict h3.
  destruct h3 as [x [h3 h4]].
  exists x.
  rewrite <- in_sort_set_iff.
  split; auto.
Qed.


Definition inv_im_min
           {T U:Type} (R:Relation T) (pfwo:well_order R) (f:T->U) (y:U)
           (pfs:surjective f)
: T :=
  min_set _ _ pfwo (inv_im1 f y) (inh_surj f y pfs).


Definition inv_im_min_nat
           {U:Type} (f:nat->U) (y:U)
           (pf:surjective f)
: nat :=
  inv_im_min lt wo_nat f y pf.


Lemma inv_im_min_compat :
  forall {T U:Type} (R:Relation T) (pfwo:well_order R) (f:T->U) (y:U)
         (pfs:surjective f) (x:T) (pf:Inn (inv_im1 f y) x),
    f x = y /\
    Inn (inv_im1 f y) (inv_im_min R pfwo f y pfs) /\
    ~ R x (inv_im_min R pfwo f y pfs).
  unfold inv_im_min; intros T U R h1 f y h2 x h3.
  pose proof (min_set_compat _ R h1 _ (inh_surj f y h2)) as h4.
  simpl in h4.
  destruct h4 as [h4 h5].
  split; auto.
  pose proof (min_set_compat T R h1 (inv_im1 f y) (inh_surj f y h2)) as h6.
  simpl in h6.
  destruct h6 as [h6 h7].
  inversion h6.
  destruct h3; auto.
Qed.

 
Lemma inv_im_min_compat' :
  forall {T U:Type} (R:Relation T) (pfwo:well_order R) (f:T->U) (y:U)
         (pfs:surjective f),
    f (inv_im_min R pfwo f y pfs) = y.
  intros.
  unfold inv_im_min.
  pose proof (min_set_compat T R pfwo (inv_im1 f y) (inh_surj f y pfs)) as h1.
  simpl in h1.
  destruct h1 as [h1 h2].
  inversion h1; auto.
Qed.

         
Lemma in_inv_im_min :
  forall {T U:Type} (R:Relation T) (pfwo:well_order R) (f:T->U) (y:U)
         (pfs:surjective f),
    Inn (inv_im1 f y) (inv_im_min R pfwo f y pfs).
  intros T U R h1 f y h2.
  econstructor.
  apply inv_im_min_compat'.
Qed.


Lemma inv_im_min_nat_compat : 
  forall {U:Type} (f:nat->U) (y:U)
         (pfs:surjective f) (n:nat) (pf:Inn (inv_im1 f y) n),
    f n = y /\
    Inn (inv_im1 f y) (inv_im_min_nat f y pfs) /\
    ~ n < (inv_im_min_nat f y pfs).
  unfold inv_im_min; intros; apply inv_im_min_compat; auto.
Qed.

Lemma inv_im_min_nat_compat' :
  forall {U:Type} (f:nat->U) (pf:surjective f) (y:U),
    f (inv_im_min_nat f y pf) = y.
  unfold inv_im_min_nat; intros; apply inv_im_min_compat'.
Qed.



Lemma in_inv_im_min_nat :
  forall {U:Type} (f:nat->U) (pf:surjective f) (y:U),
    Inn (inv_im1 f y) (inv_im_min_nat f y pf).
  intros; constructor; apply inv_im_min_nat_compat'.
Qed.


Lemma le_inv_im_min :
  forall {T U:Type} (R:Relation T) (pfwo:well_order R)
         (f:T->U) (pf:surjective f) (x:T),
    Rw R (inv_im_min R pfwo f (f x) pf) x.
  intros T U R h1 f h2 x.
  pose proof (inv_im_min_compat R h1 f (f x) h2 x) as h3.
  hfirst h3.
  econstructor; auto.
  specialize (h3 hf).
  destruct h3 as [h3 [h4 h5]]. 
  destruct h1 as [h1 h6].
  red in h6.
  pose proof (h6  (inv_im_min R {| wo_well_founded := h1; wo_total_strict_order := h6 |} f
                              (f x) h2) x) as h7.
  destruct h7 as [h7 | [h7 | h7]].
  left; auto.
  right; auto.
  contradiction.
Qed.


Lemma le_inv_im_min_nat :
  forall {U:Type} (f:nat->U) (pf:surjective f) (i:nat),
    inv_im_min_nat f (f i) pf <= i.
  unfold inv_im_min_nat; intros.
  rewrite <- Rw_le_iff.
  eapply le_inv_im_min.
Qed.



Lemma lind_map_seg_fun_to_seg_inv_im_min_nat :
  forall {U:Type} (pfdu:type_eq_dec U) 
         (f:nat->U) (pfs:surjective f) (i:nat)
         (pf:In (f (inv_im_min_nat f (f i) pfs)) (map_seg (fun_to_seg f (S i)))),
    lind pfdu (map_seg (fun_to_seg f (S i)))
         (f (inv_im_min_nat f (f i) pfs)) pf =
    inv_im_min_nat f (f i) pfs.
  intros U hdu f h1 i h2.
  rewrite lind_eq_iff.
  split.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg.
  rewrite fun_to_seg_compat.
  reflexivity.
  intro h3.
  rewrite in_firstn_iff in h3.
  destruct h3 as [m [h3 h4]].
  erewrite nth_lt_functional3 in h4. 
  rewrite nth_lt_map_seg, fun_to_seg_compat in h4.
  rewrite inv_im_min_nat_compat' in h4.
  pose proof (inv_im_min_nat_compat f (f i) h1 m) as h5.
  hfirst h5.
  econstructor; auto.
  specialize (h5 hf).
  destruct h5 as [? [? ?]]; omega.

  Grab Existential Variables.
  eapply lt_le_trans.
  apply h3.
  eapply le_trans.
  apply le_inv_im_min_nat.
  auto with arith.
  rewrite length_map_seg.
  eapply le_trans.
  apply le_inv_im_min_nat.
  auto with arith.
  red.
  apply le_n_S.
  apply  le_inv_im_min_nat.
  rewrite length_map_seg.
  red.
  apply le_n_S.
  apply  le_inv_im_min_nat.
Qed.
  
  
Definition surj_nat_to_fun
           {U V:Type}
           (f:nat->U) (g:nat->V) (pfs:surjective f) : U -> V :=
  fun y => g (inv_im_min_nat f y pfs).


Lemma surj_nat_to_fun_compat :
  forall {U V:Type} (pfdu:type_eq_dec U) (pfdv:type_eq_dec V)
         (f:nat->U)
         (g:nat->V)
         (pfs:surjective f)
         (pfs':surjective g),
    fun_linds_eq_nat pfdu pfdv f g ->
    forall i:nat,
      surj_nat_to_fun f g pfs (f i) =
      g i.  
  intros U V h1 h2 f g h3 h3' h4 i.
  unfold surj_nat_to_fun.

  pose proof (lind_map_seg_fun_to_seg_inv_im_min_nat h1 f h3 i) as h5.
  hfirst h5.
  rewrite in_map_seg_iff.
  exists (inv_im_min_nat f (f i) h3).
  ex_goal.
  red.
  apply le_n_S.
  apply le_inv_im_min_nat.
  exists hex.
  unfold fun_to_seg; auto.
  specialize (h5 hf).

  pose proof (lind_map_seg_fun_to_seg_inv_im_min_nat h2 g h3' i) as h5'.
  hfirst h5'.
  rewrite in_map_seg_iff.
  exists (inv_im_min_nat g (g i) h3').
  ex_goal.
  red.
  apply le_n_S.
  apply le_inv_im_min_nat.
  exists hex.
  unfold fun_to_seg; auto.
  specialize (h5' hf0).
  red in h4.
  specialize (h4 (S i)).
  pose proof (linds_map_seg_eq_lind_occ
                h1 h2 f g i 0
                (S i) (lt_n_Sn _)
                h4) as h6.
  hfirst h6.
  apply gt_lt.
  rewrite <- (count_occ_In h1).
  rewrite in_map_seg_iff.
  exists i, (lt_n_Sn _).
  rewrite fun_to_seg_compat; auto.
  specialize (h6 hf1).
  hfirst h6.
  apply in_map_seg.
  specialize (h6 hf2).
  hfirst h6.
  apply in_map_seg.
  specialize (h6 hf3).
  rewrite <- h5.

  pose proof (inv_im_min_nat_compat f (f i) h3 (inv_im_min_nat f (f i) h3)) as h7.
  hfirst h7.
  apply in_inv_im_min_nat.
  specialize (h7 hf4).
  destruct h7 as [h7 [h8 h9]].
  generalize hf.
  rewrite h7.
  intro h10.
  pose proof (proof_irrelevance _ h10 hf2); subst.
  unfold lind.
  rewrite h6.
  unfold lind in h5'.

  pose proof (inv_im_min_nat_compat g (g i) h3' (inv_im_min_nat g (g i) h3')) as h7'.
  hfirst h7'.
  apply in_inv_im_min_nat.
  specialize (h7' hf5).
  destruct h7' as [h7' [h8' h9']].
  pose proof (lind_functional h2 (map_seg (fun_to_seg g (S i))) (g i) (g (inv_im_min_nat g (g i) h3')) hf3 hf0) as h10.
  hfirst h10. 
  rewrite inv_im_min_nat_compat'; auto.
  rewrite <- h7'.
  f_equal.
  erewrite <- lind_map_seg_fun_to_seg_inv_im_min_nat.
  unfold lind.
  apply lind_occ_functional; auto.

  Grab Existential Variables.
  rewrite in_map_seg_iff.
  exists (inv_im_min_nat g (g i) h3').
  ex_goal.
  red.
  apply le_n_S.
  apply le_inv_im_min_nat.
  exists hex.
  rewrite fun_to_seg_compat; auto.
Qed.


Definition surj_nat_to_fun_sig
           {U V:Type} {D:Ensemble U} {E:Ensemble V}
           (f:nat->sig_set D) (g:nat->sig_set E)
           (pfs:surjective f) : sig_set D -> V :=
  fun y => proj1_sig (surj_nat_to_fun f g pfs y).



Lemma surj_nat_to_fun_sig_compat :
  forall {U V:Type} {D:Ensemble U} {E:Ensemble V}
         (pfdu:type_eq_dec U) (pfdv:type_eq_dec V)
         (f:nat->sig_set D)
         (g:nat->sig_set E)
         (pfs:surjective f)
         (pfs':surjective g),
    let pfdu' := sig_set_eq_dec pfdu D in
    let pfdv' := sig_set_eq_dec pfdv E in
    fun_linds_eq_nat pfdu' pfdv' f g ->
    forall i:nat,
      surj_nat_to_fun_sig f g pfs (f i) =
      proj1_sig (g i).
  intros U V D E hdu hdv hf g hs hs' hdu' hdv' h2 i.
  unfold surj_nat_to_fun_sig.
  f_equal.
  eapply surj_nat_to_fun_compat; auto.
  apply h2.
Qed.


Lemma lind_map_seg_fun_to_seg_inv_im_seg_min :
  forall {U:Type} (pfdu:type_eq_dec U) 
         (f:nat->U) (n:nat)
         (y:sig_set (Im (seg n) f))
         (pf:In (f (inv_im_seg_min f n y))
                (map_seg (fun_to_seg f n))),
    lind pfdu (map_seg (fun_to_seg f n))
         (f (inv_im_seg_min f n y)) pf =
    inv_im_seg_min f n y.
  intros U hdu f n y h1.
  rewrite lind_eq_iff.
  split. 
  rewrite nth_lt_map_seg'.
  rewrite fun_to_seg_compat; auto.
  intro h2.
  rewrite in_firstn_iff in h2.
  destruct h2 as [m [h2 h3]].
  rewrite nth_lt_map_seg' in h3.
  rewrite fun_to_seg_compat in h3.
  unfold inv_im_seg_min in h2.
  let P := type of h2 in
  match P with
      m < SetUtilities.min_set_nat ?x1 ?x2 ?x3 =>
      pose proof (SetUtilities.min_set_nat_compat x1 x2 x3) as h4 end.
  simpl in h4.
  destruct h4 as [h4 h5].
  destruct h4 as [h4].
  destruct h4 as [h4 h6].
  specialize (h5 m).
  hfirst h5.
  constructor.
  constructor.
  ex_goal.
  omega.
  pose proof (inv_im_seg_min_compat f n y) as h7.
  exists hex.
  rewrite fun_to_sig_seg_compat.
  apply proj1_sig_injective.
  rewrite <- h7 at 1.
  rewrite <- h3.
  simpl; auto.
  rewrite in_sing_iff.
  intro; omega.
  specialize (h5 hf).
  omega.
   
  Grab Existential Variables.
  rewrite length_map_seg.
  apply lt_impl_le.
  apply lt_inv_im_seg_min.
  rewrite length_map_seg. 
  apply lt_inv_im_seg_min.
Qed.


Lemma nat_sig_to_fun_compat : 
  forall {U V:Type}
         {D:Ensemble U} {E:Ensemble V}
         (pfdu:type_eq_dec U)
         (pfdv:type_eq_dec V)
         (f:nat->sig_set D)
         (g:nat->sig_set E)
         (i n:nat)
         (pf:i < n),
    linds pfdu (map_seg (fun_to_seg (proj1_sig_fun f) n)) =
    linds pfdv (map_seg (fun_to_seg (proj1_sig_fun g) n)) ->
    nat_sig_to_fun f g n
                   (exist _ _ (Im_intro _ _ (seg n) (proj1_sig_fun f) i (in_seg _ _ pf) _ eq_refl)) =
    proj1_sig (g i).
  intros U V D E hdu hdv f g i n hlt hle.
  unfold nat_sig_to_fun.
  f_equal.
   
  pose proof (lind_map_seg_fun_to_seg_inv_im_seg_min
                hdu (proj1_sig_fun f)
                n) as h5.
  pose (Im_intro _ _ (seg n) (proj1_sig_fun f) _ (in_seg _ _ hlt) _
                 eq_refl) as h6.
  specialize (h5 (exist _ _ h6)).
  hfirst h5.
  rewrite in_map_seg_iff.
  exists (inv_im_seg_min (proj1_sig_fun f) n
                         (exist (Inn (Im (seg n) (proj1_sig_fun f)))
                                (proj1_sig_fun f i) h6)).
  ex_goal.
  apply lt_inv_im_seg_min.
  exists hex.
  unfold fun_to_seg; auto.
  specialize (h5 hf).

  pose proof (lind_map_seg_fun_to_seg_inv_im_seg_min
                hdv (proj1_sig_fun g)
                n) as h5'.
  pose proof (Im_intro _ _ (seg n) (proj1_sig_fun g) _ (in_seg _ _ hlt) _ eq_refl) as h6'.
  specialize (h5' (exist _ _ h6')).
  hfirst h5'.
  rewrite in_map_seg_iff.
  exists (inv_im_seg_min (proj1_sig_fun g) n
          (exist (Inn (Im (seg n) (proj1_sig_fun g))) (proj1_sig_fun g i) h6')).
  ex_goal.
  apply lt_inv_im_seg_min.
  exists hex.
  unfold fun_to_seg; auto.
  specialize (h5' hf0).
  fold h6. 
  rewrite <- h5. 

  
  pose proof (linds_map_seg_eq_lind_occ
                hdu hdv (proj1_sig_fun f) (proj1_sig_fun g) i 0
                n hlt hle) as h7.
  hfirst h7.
  apply gt_lt.
  rewrite <- (count_occ_In hdu).
  rewrite in_map_seg_iff.
  exists i, hlt.
  rewrite fun_to_seg_compat; auto.
  specialize (h7 hf1).
  hfirst h7.
  apply in_map_seg.
  specialize (h7 hf2).
  hfirst h7.
  apply in_map_seg.
  specialize (h7 hf3).

  pose proof h6 as h8.
  apply in_proj1_sig_fun in h8.
 
  pose proof (inv_im_seg_min_sig_compat f n (exist _ _ h6)) as h9. 
  pose proof (f_equal (@proj1_sig _ _) h9) as h10.
  generalize hf.
  rewrite proj1_sig_fun_compat, h10.
  intro h11; simpl.
  unfold lind.
  pose proof (proof_irrelevance _ hf2 h11); subst.
  rewrite h7.

  pose proof (inv_im_seg_min_sig_compat g n (exist _ _ h6')) as h9'.
  pose proof (f_equal (@proj1_sig _ _) h9') as h10'.
  simpl in h10'.
  rewrite proj1_sig_fun_compat in h10'.
  apply proj1_sig_injective.
  unfold proj1_sig_fun.
  rewrite <- h10'.
  repeat f_equal; auto.
  rewrite <- h5'.
  unfold lind; apply lind_occ_functional.
  reflexivity.
  reflexivity.
  auto.
Qed.


End Sort.

Section Finite.


Lemma Finite_Finite_Union_rev : 
  forall {T:Type} (F:Family T), 
    Finite (FamilyUnion F) -> Finite F.
intros T F h2.
apply NNPP.
intro h3.
assert (h4:Approximant _ F [S:Ensemble T | Inn F S /\ Included S (FamilyUnion F)]).
constructor.
eapply Finite_downward_closed.
eapply power_set_finite. apply h2.
red. intros A h4.
destruct h4 as [h4].
destruct h4 as [h4l h4r].
constructor. assumption.
red. intros S h4.
destruct h4 as [h4]. destruct h4; auto.
pose proof (make_new_approximant _ _ _ h3 h4) as h5.
assert (h6: (Setminus F [S : Ensemble T | Inn F S /\ Included S (FamilyUnion F)]) = Empty_set _).
  apply Extensionality_Ensembles; red; split; auto with sets.
  red. intros A h6.
  destruct h6 as [h6 h7].
  contradict h7.
  constructor. split; auto.
  red. intros x h7. apply family_union_intro with A; auto.
rewrite h6 in h5.
destruct h5. contradiction.
Qed.


Lemma card_cart_prod : 
  forall {T U:Type} 
         (A:Ensemble T) (B:Ensemble U),
    Finite A -> Finite B ->
    card_fun1 (cart_prod A B) = (card_fun1 A) * (card_fun1 B).
  intros T U  A B h1 h2.
destruct (eq_dec A (Empty_set _)) as [hae | hnae].
subst. rewrite cart_prod_empty. do 2 rewrite card_fun1_empty.
ring.
destruct (eq_dec B (Empty_set _)) as [hbe | hnbe].
subst. rewrite cart_prod_empty'. do 2 rewrite card_fun1_empty.
ring.
pose proof (not_empty_Inhabited _ _ hnbe) as h11.
destruct h11 as [b h11].
pose proof (not_empty_Inhabited _ _ hnae) as h12.
destruct h12 as [a h12].
pose proof (cart_prod_fin _ _ h1 h2) as h3.
rewrite cart_prod_eq. rewrite cart_prod_eq in h3.
pose proof (Finite_Finite_Union_rev _  h3) as h4.
term0 h4.
rename c into F.
assert (h6: forall S, Inn F S -> Finite S).
  intros S h7. destruct h7 as [h7]. destruct h7 as [y h7].
  destruct h7 as [h7l h7r]. subst. 
  apply cart_prod_sing_fin; auto.
assert (hpw: pairwise_disjoint F).
  red. intros X Y h7 h8 h9.
  destruct h8 as [h8]. destruct h8 as [y h8]. destruct h8 as [h8l h8r].
  destruct h9 as [h9]. destruct h9 as [y' h9]. destruct h9 as [h9l h9r].
 subst.
 red.
 apply Extensionality_Ensembles; red; split; auto with sets.
 red. intros pr h10. destruct h10 as [pr h10l h10r].
 destruct h10l as [h10l]. destruct h10l as [h10a h10b].
 destruct h10r as [h10r]. destruct h10r as [h10c h10d].
 destruct h10b; destruct h10d; subst.
 contradict h7. reflexivity.
pose proof (card_pairwise_disjoint _ h4 h6 hpw) as h5.
rewrite h5.
assert (h7:card_fun1 F = card_fun1 B).
pose [pr:(Ensemble (T*U))*U |
      Inn B (snd pr) /\ fst pr = cart_prod A (Singleton (snd pr))] as P.
  assert (h8:dom_rel P = [S : Ensemble (T * U)
     | exists y : U, Inn B y /\ S = cart_prod A (Singleton y)]).
    apply Extensionality_Ensembles.
    red. split. 
    red. intros S h9.
    destruct h9 as [h9]. destruct h9 as [y h9].
    destruct h9 as [h9]. simpl in h9.
    constructor. exists y. assumption.
    red. intros S h9.
    destruct h9 as [h9]. constructor. destruct h9 as [y h9].
    exists y. constructor. simpl. assumption.
  assert (h9:ran_rel P = B).
    apply Extensionality_Ensembles.
    red. split.
    red. intros y h9. destruct h9 as [h9]. 
    destruct h9 as [x h9]. destruct h9 as [h9]. simpl in h9.
    destruct h9; assumption.
    red. intros y h9. constructor.
    exists (cart_prod A (Singleton y)). constructor. simpl.
    split; auto.
  assert (h10:fun_well_defined P).
    red. rewrite h8. rewrite h9.
    constructor.
    intros S h10.
    destruct h10 as [h10]. destruct h10 as [y h10].
    destruct h10 as [h10l h10r].
    exists y. red. split. split; auto. constructor. simpl.
    split; auto.
    intros y' h11'. destruct h11' as [h11l h11r].
    destruct h11r as [h11']. destruct h11' as [h11a h11b].
    simpl in h11b. 
    rewrite h10r in h11b.
    assert (h13:Inn (cart_prod A (Singleton y)) (a, y)).
      constructor; simpl. split; auto; try constructor.
    rewrite h11b in h13. inversion h13 as [h14]. clear h13.
    simpl in h14. destruct h14 as [? h14]. destruct h14; auto.
    intros pr h10. destruct h10 as [h10]. 
    destruct h10 as [h10l h10r]. 
    split. constructor. exists (snd pr). split; auto. assumption.

    red in h10. rewrite h8 in h10. rewrite h9 in h10.
  rename F into S.
  pose (fin_map_intro _ _ b (ens_eq_dec (T*U))  h4 h2 P h10) as F.
  assert (h12':bij F).
    red. split.
    red. intros X Y h12' h13 h14.
    do 2 rewrite fin_map_app_compat in h14.
    pose proof (fin_fps_to_f_s_compat
                  (ens_eq_dec (T*U)) h4 h10 b _ h12') as h15.
    pose proof (fin_fps_to_f_s_compat
                  (ens_eq_dec (T*U)) h4 h10 b _ h13) as h16. 
    assert (h17:P = fin_map_to_fps F).
    pose proof (fin_map_to_fps_s_compat
                  (ens_eq_dec (T*U)) _ _ b h4 h2 P h10) as h17.
      rewrite h17. f_equal.
    assert (h18:fin_fps_to_f (ens_eq_dec (T*U)) h4 P h10 b X = fin_fps_to_f (ens_eq_dec (T*U)) h4 (fin_map_to_fps F) (fp_fin_map_to_fps F) b X). 
      pose proof (subsetT_eq_compat _ _ _ _ h10 (fp_fin_map_to_fps F) h17) as h19.
      dependent rewrite -> h19. reflexivity.
      erewrite fin_fps_to_f_functional in h18.
      symmetry in h18.
    rewrite h14 in h18 at 1.
    erewrite fin_fps_to_f_functional in h15.
    rewrite h18 in h15 at 1.
    inversion h15 as [h19]. clear h15.
    simpl in h19. destruct h19 as [h19l h19r].
    inversion h16 as [h20]. clear h16. simpl in h20.
    destruct h20 as [h20l h20r].
    pose proof (subsetT_eq_compat _ _ _ _ h10 (fp_fin_map_to_fps F) h17)
      as h19.
    assert (h21:(fin_fps_to_f (ens_eq_dec (T*U)) h4 (fin_map_to_fps F) (fp_fin_map_to_fps F) b Y) =  (fin_fps_to_f (ens_eq_dec (T*U)) h4 P h10 b Y)).
      dependent rewrite -> h19. reflexivity.
   erewrite fin_fps_to_f_functional in h19r.
   rewrite h21 in h19r at 1.
   rewrite <- h19r in h20r. rewrite h20r.
   reflexivity. reflexivity. assumption. assumption.
   red. intros y h12'.
   exists (cart_prod A (Singleton y)). constructor. constructor.
   exists y. split; auto.
   assert (h14:Inn (fin_map_to_fps F) (cart_prod A (Singleton y), y)).
     unfold F. rewrite <- fin_map_to_fps_s_compat.
     constructor. simpl. split; auto.
   pose proof (fin_map_to_fps_fin_map_app_compat F ((cart_prod A (Singleton y)), y) h14) as h13. simpl in h13. simpl. assumption.
eapply bij_dom_ran_card_eq'. apply h12'.
rewrite <- h7.
assert (h8:forall S, Inn F S -> card_fun1 S = card_fun1 A).
  intros S h9. destruct h9 as [h9]. destruct h9 as [y h9].
  destruct h9 as [h9l h9r].
  subst.
apply card_cart_prod_sing'; auto.
pose proof (plus_bag_nat_im_set_bag_eq_f _ h4 card_fun1 (card_fun1 A) h8) as h9.
rewrite h9.
unfold F.
ring.
Qed.


End Finite.

Section Random.


Definition tuple_prod_ens {T U n} (A:T->Ensemble U) (x:T^n) :
  Ensemble (U^n) :=
  [y:U^n | forall i (pfi:i < n),
             Inn (A (nth_prod i pfi x)) (nth_prod i pfi y)].


Definition tuple_prod_ens_dep  {T n} {U:T->Type}
           (A:forall x:T, Ensemble (U x)) (x:T^n) :
  Ensemble (tuple_prod U x) :=
  [y:tuple_prod U x |
   (forall i (pfi:i < n),
      Inn (A (nth_prod i pfi x)) (nth_tuple_prod y i pfi))].



Lemma in_nth_tuple_prod_ens_dep :
  forall {T n} {U:T->Type}
         {A:forall t:T, Ensemble (U t)}
         {w:T^n}
         {t:tuple_prod U w}
         (pfi:Inn (tuple_prod_ens_dep A w) t)
         i (pfi:i < n),
    Inn (A (nth_prod i pfi w))
        (nth_tuple_prod t i pfi).
  intros ? ? ? ? ? ? h1; destruct h1; auto.
Qed.

Lemma in_transfer_tuple_prod1 :
  forall {T} {U:T->Type}
         {A:forall t:T, Ensemble (U t)}
         {x:T}
         (a:U x),
    Inn (A x) a ->
    Inn (tuple_prod_ens_dep A (x, tt) (n := 1))
        (transfer_to_tuple_prod1 a).
  intros T U A x a h1.
  constructor.
  intros i hi.
  pose proof (lt1 _ hi); subst.
  simpl; auto.
Qed.

Lemma in_transfer_to_tuple_prod1_iff :
    forall {T} {U:T->Type}
           {A:forall t, Ensemble (U t)}
           {s:T}
           {t:T^1}
           (pfe:t = (s, tt))
           (u:U s),
      Inn (tuple_prod_ens_dep A (s, tt) (n := 1))
          (transfer_to_tuple_prod1 u) <->
      Inn (tuple_prod_ens_dep A t)
          (transfer_dep_r pfe (transfer_to_tuple_prod1 u)).
  intros; subst; rewrite transfer_dep_r_eq_refl; tauto.
Qed.
    
                            


  
Definition tuple_prod_sys {T U X} (A:T->Ensemble U)
           (v:X->T) :
  X -> Ensemble U :=
  fun x => A (v x).

Definition tuple_prod_ens_sys {T U X n}
           (A:T->Ensemble U) (v:X->T)
           (x:X^n) : Ensemble (U^n) :=
  [y:U^n | forall i (pfi:i < n),
             Inn (tuple_prod_sys A v (nth_prod i pfi x))
                 (nth_prod i pfi y)].


Lemma in_tuple_prod_ens_compat :
  forall {T U n} (A:T->Ensemble U) (x:T^n)
         (y:U^n),
    Inn (tuple_prod_ens A x) y ->
    Inn (tuple_prod_ens_dep A x)
        (transfer_to_tuple_prod_const _ y).
  intros T U n A x y h1; destruct h1 as [h1]; constructor;
  intros i hi; specialize (h1 i hi);
  rewrite nth_tuple_prod_transfer_to_tuple_prod_const; auto.
Qed.

Lemma in_tuple_prod_ens_dep_const_compat :
  forall {T U n} {x:T^n} (A:T->Ensemble U)
         (p:tuple_prod (fun _:T => U) x),
    Inn (tuple_prod_ens_dep A x) p ->
    Inn (tuple_prod_ens A x)
        (transfer_tuple_prod_const p).
  intros T U n x A p h1; destruct h1 as [h1]; constructor;
  intros i hi; specialize (h1 i hi);
  rewrite nth_prod_transfer_tuple_prod_const; auto.
Qed.

Lemma in_tuple_prod_ens_nth :
  forall {T U n} {A:T->Ensemble U}
         {x:T^n}
         {y:U^n},
    Inn (tuple_prod_ens A x) y ->
    forall i (pfi:i < n),
      Inn (A (nth_prod i pfi x)) (nth_prod i pfi y).
  intros ? ? ? ? ? ? h1; destruct h1; auto.
Qed.


Lemma in_tuple_prod_ens_dep_nth_iff :
  forall {T n} {U:T->Type} {x:T^n}
         (A:forall x:T, Ensemble (U x))
         (y:tuple_prod U x),
    Inn (tuple_prod_ens_dep A x) y <->
    forall i (pfi:i < n),
      Inn (A (nth_prod i pfi x)) (nth_tuple_prod y i pfi).
  intros ? ? ? ? ? ?; split; intro h1.
  intros i hi.
  destruct h1; auto.
  constructor; auto.
Qed.


Definition sig_tuple_prod_ens_dep_const_to 
           {T U n} (A:T->Ensemble U) (w:T^n) :
  sig_set (tuple_prod_ens_dep A w) ->
  sig_set (tuple_prod_ens A w) :=
  fun p' =>
    let (p, pfp) := p' in
    let pfp' := in_tuple_prod_ens_dep_const_compat A p pfp in
    exist (fun c => Inn (tuple_prod_ens A w) c)
          (transfer_tuple_prod_const p)
          pfp'.

Lemma inj_sig_tuple_prod_ens_dep_const_to : 
  forall {T U n} (A:T->Ensemble U) (w:T^n)
         (x x':sig_set (tuple_prod_ens_dep A w)),
    sig_tuple_prod_ens_dep_const_to A w x =
    sig_tuple_prod_ens_dep_const_to A w x' ->
    x = x'.
  unfold sig_tuple_prod_ens_dep_const_to.
  intros ? ? ? ? ? x x' h1.
  destruct x, x'.
  apply exist_injective in h1.
  apply inj_transfer_tuple_prod_const in h1; subst; f_equal;
  apply proof_irrelevance.
Qed.


Definition transfer_tuple_prod_ens
           {T U n} (A:T->Ensemble U) (x:T^n) :
  Ensemble (tuple_prod (fun _:T => U) x) :=
  Im (tuple_prod_ens A x)
     (transfer_to_tuple_prod_const x).

Definition transfer_tuple_prod_ens_dep
           {T U n} (A:T->Ensemble U) (x:T^n) :
  Ensemble (U^n) :=
  Im (tuple_prod_ens_dep A x) (transfer_tuple_prod_const (x:=x)).

Lemma transfer_tuple_prod_ens_compat : 
  forall {T U n} (A:T->Ensemble U) (x:T^n),
    transfer_tuple_prod_ens A x =
    tuple_prod_ens_dep A x.
  intros T U n A x; apply Extensionality_Ensembles; red; split; red;
  intros y hy; destruct hy as [hy]; subst; 
  [destruct H; constructor; intros i hi;
  rewrite
    nth_tuple_prod_transfer_to_tuple_prod_const; auto|
   apply Im_intro with (transfer_tuple_prod_const y);
     [constructor; intros i hi; specialize (hy i hi);
      rewrite nth_prod_transfer_tuple_prod_const; auto|
      rewrite transfer_to_tuple_prod_const_compat; auto]].
Qed.


Lemma transfer_tuple_prod_ens_dep_compat :
  forall {T U n} (A:T->Ensemble U) (x:T^n),
    tuple_prod_ens A x =
    transfer_tuple_prod_ens_dep A x.
  unfold transfer_tuple_prod_ens_dep;
  intros T U n A x; apply Extensionality_Ensembles; red; split; red;
  intros y hy; destruct hy as [hy];
  [apply Im_intro with (transfer_to_tuple_prod_const _ y);
    try constructor;
    [intros i hi; specialize (hy i hi);
     rewrite nth_tuple_prod_transfer_to_tuple_prod_const; auto|
     rewrite transfer_tuple_prod_const_to_compat; auto]|
   subst; destruct H; constructor; intros i hi; specialize (H i hi);
   rewrite nth_prod_transfer_tuple_prod_const; auto].
Qed.


Lemma tuple_prod_ens0 :
  forall {T U} (A:T->Ensemble U) (x:T^0),
    tuple_prod_ens A x = Singleton tt.
  intros T U A x.
  destruct x; simpl.
  apply Extensionality_Ensembles; red; split; red; intros y h1.
  destruct h1.
  destruct y.
  constructor.
  destruct y.
  constructor.
  intros; omega.
Qed.

Lemma tuple_prod_ens_dep0 :
  forall {T} (U:T->Type)
         (A:forall x:T, Ensemble (U x))
         (x:T^0),
    tuple_prod_ens_dep A x = Singleton tt.
  intros T U A x; destruct x; simpl.
  apply Extensionality_Ensembles; red; split; red; intros x h1;
  destruct h1.
  destruct x.
  constructor.
  constructor.
  intros; omega.
Qed.



Lemma tuple_prod_ens_n0 :
  forall {T U n} (A:T->Ensemble U) (x:T^n)
         (pfeq:n = 0),
    tuple_prod_ens A x = Singleton (transfer_tt pfeq U).
  intros; subst; rewrite transfer_tt_eq_refl; apply tuple_prod_ens0.
Qed.


Lemma tuple_prod_ens_dep_n0 :
  forall {T n} (U:T->Type)
         (A:forall x:T, Ensemble (U x))
         (x:T^n)
         (pfeq:n = 0),
    tuple_prod_ens_dep A x = Singleton (transfer_tt_tuple_prod U pfeq x).
  intros; subst; rewrite transfer_tt_tuple_prod_eq_refl;
  apply tuple_prod_ens_dep0.
Qed.



Lemma in_tt_tuple_prod_ens_dep :
  forall {T} (U:T->Type)
         (A:forall x:T, Ensemble (U x))
         (x:T^0),
    Inn (tuple_prod_ens_dep A x) tt.
  intros; rewrite tuple_prod_ens_dep0; constructor; auto.
Qed.

Lemma in_tt_tuple_prod_ens_dep' :
  forall {T} (U:T->Type)
         (A:forall x:T, Ensemble (U x))
         (x:T^0),
    Inn (tuple_prod_ens_dep A x) x.
  intros; rewrite tuple_prod_ens_dep0; simpl in x; destruct x;
  constructor; auto.
Qed.


Lemma in_transfer_tt_tuple_prod_ens_dep :
  forall {T n} (U:T->Type)
         (A:forall x:T, Ensemble (U x))
         (x:T^n)
         (pfeq:n = 0),
    Inn (tuple_prod_ens_dep A x)
        (transfer_tt_tuple_prod U pfeq x).
  intros; subst; simpl;
  rewrite tuple_prod_ens_dep0; simpl in x;
  destruct x;
  constructor; auto.
Qed.




Lemma tuple_prod_ens_S :
  forall {T U n} (A:T->Ensemble U) (x:T^S n),
    tuple_prod_ens A x =
    [t:U^(S n) | Inn (A (fst x)) (fst t) /\
                 Inn (tuple_prod_ens A (snd x))
                     (snd t)].
  intros T U n A x; apply Extensionality_Ensembles; red; split; red;
  intros y h1; destruct h1 as [h1]; constructor.
  split.
  specialize (h1 0 (lt_0_Sn _)).
  destruct x, y; simpl; simpl in h1; auto.
  destruct x as [? x], y as [? y]; simpl.
  constructor.
  intros i hi.
  specialize (h1 (S i) (lt_n_S _ _ hi)); simpl in h1; auto.
  pose proof (proof_irrelevance _ hi (lt_S_n _ _ (lt_n_S _ _ hi))) as h2.
  rewrite h2; auto.
  destruct h1 as [h1 h2].
  destruct x as [? x], y as [? y]; simpl in h1, h2.
  destruct h2.
  intros i hi.
  destruct i as [|i].
  simpl; auto.
  specialize (H i (lt_S_n _ _ hi)).
  simpl; auto.
Qed.


Lemma tuple_prod_ens_dep_S :
  forall {T n} {U:T->Type}
         (A:forall t:T, Ensemble (U t))
         (x:T^S n),
    tuple_prod_ens_dep A x =
    [t:tuple_prod U x |
     Inn (A (fst x)) (fst t) /\
     Inn (tuple_prod_ens_dep A (snd x))
         (snd t)].
  intros T n U A x; apply Extensionality_Ensembles; red; split; red;
  intros y h1; destruct h1 as [h1]; constructor.
  destruct x as [? x]; simpl.
  destruct y as [u y]; simpl in u; simpl; simpl in y.
  split.
  specialize (h1 0 (lt_0_Sn _)); simpl in h1; auto.
  constructor.
  intros i hi.
  specialize (h1 (S i) (lt_n_S _ _ hi)); simpl in h1; auto.
  pose proof (proof_irrelevance _ hi (lt_S_n _ _ (lt_n_S _ _ hi))) as h2.
  rewrite h2; auto.
  destruct h1 as [h1 h2].
  destruct x as [? x], y as [? y]; simpl in h1, h2.
  destruct h2.
  intros i hi.
  destruct i as [|i].
  simpl; auto.
  specialize (H i (lt_S_n _ _ hi)).
  simpl; auto.
Qed.


  


Lemma finite_mem_rel_classes_im_rel_sig_fun : 
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U),
    Finite A ->
    finite_mem (rel_classes_im_rel_sig_fun f).
  intros T U A f h1.
  red; intros S h2.
  destruct h2 as [S h2]. subst.
  destruct h2 as [x h2]. subst.
  clear h2.
  apply finite_image.
  eapply Finite_downward_closed.
  rewrite <- finite_full_sig_iff; auto.
  red; intros; constructor.
Qed.

Lemma couple_eq_add_iff :
  forall {T:Type} (pfdt:type_eq_dec T)
         (A:Ensemble T) (a b c:T),
    Couple a b = Add A c <->
    ((a = b /\ a = c /\ (A = Singleton a \/ A = Empty_set _)) \/
     (a <> b /\ ((a = c /\ (A = Singleton b \/ A = Couple a b)) \/
                 (b = c /\ (A = Singleton a \/ A = Couple a b))))).
  intros T hdt A a b c; split; intro h1.
  pose proof (finite_couple a b) as hf.
  rewrite h1 in hf.
  pose proof (add_downward_closed _ _ hf) as hf'.
  pose proof (Add_intro2 _ A c) as h2.
  rewrite <- h1 in h2.
  destruct h2.
  pose proof (Couple_r _ a b) as h2. 
  rewrite h1 in h2.
  destruct h2 as [x h2 | x h2]; subst. 
  destruct (hdt x a) as [|h3]; subst.
  rewrite in_add_eq, couple_singleton in h1; auto.
  right.
  split; auto.
  left.
  split; auto.
  destruct (in_fin_ens_dec hdt A hf' a) as [hi | hi].
  right.
  rewrite in_add_eq in h1; auto.
  left.
  rewrite couple_comm in h1.
  rewrite couple_add_add in h1.
  apply add_nin_inj_eq in h1; subst.
  rewrite add_empty_sing; auto.
  rewrite add_empty_sing.
  rewrite in_sing_iff; auto.
  assumption.
  destruct h2.
  rewrite couple_singleton in h1.
  apply sing_eq_add in h1.
  destruct h1 as [? h2].
  left; tauto.
  destruct (hdt a b) as [|h3]; subst.
  rewrite couple_singleton in h1.
  apply sing_eq_add in h1.
  destruct h1 as [? h1].
  left; tauto.
  right.
  split; auto.
  right.
  split; auto.
  pose proof (Couple_l _ a b) as h4.
  rewrite h1 in h4.
  destruct h4 as [a h4 | a h4]; subst.
  destruct (in_fin_ens_dec hdt A hf' b) as [hi | hi].
  right.
  rewrite in_add_eq in h1; auto.
  rewrite couple_add_add in h1.
  apply add_nin_inj_eq in h1; subst.
  left.
  rewrite add_empty_sing; auto.
  rewrite add_empty_sing.
  rewrite in_sing_iff; auto.
  assumption.
  destruct h4.
  contradict h3; auto.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 [h2 h3]]; subst.
  subst.
  rewrite couple_singleton.
  destruct h3; subst.
  rewrite in_add_eq; auto.
  constructor.
  rewrite add_empty_sing; auto.
  destruct h1 as [h1 h2].
  destruct h2 as [h2 | h2].
  destruct h2 as [h2 h3]; subst.
  destruct h3; subst.
  rewrite couple_comm, couple_add_sing; auto.
  rewrite in_add_eq; auto.
  left.
  destruct h2 as [? h2]; subst.
  destruct h2; subst.
  rewrite couple_add_sing; auto.
  rewrite in_add_eq; auto.
  right.
Qed.


Lemma couple_eq_add_add_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (A:Ensemble T)
         (a b x y:T),
    Couple a b = Add (Add A x) y <->
    (a = b /\ x = y /\ a = x /\ (A = Empty_set _ \/ A = Singleton a)) \/
    (a <> b /\
     ((x = y /\ (((x = a /\
                 (A = Couple a b \/ A = Singleton b)) \/
                 (x = b /\
                  (A = Couple a b \/ A = Singleton a))))) \/
      (x <> y /\ (x = a /\ y = b \/ x = b /\ y = a ) /\
       (A = Empty_set _ \/
        A = Singleton a \/
        A = Singleton b \/
        A = Couple a b)))).
  intros T hdt A a b x y.
  pose proof (finite_couple a b) as hf.
  split; intro h1.
  rewrite h1 in hf.
  pose proof (add_downward_closed _ _ hf) as hf'.
  pose proof (add_downward_closed _ _ hf') as hf''.
  pose proof h1 as h1'.
  rewrite couple_eq_add_iff in h1'; auto.
  destruct h1' as [h1' | h1'].
  destruct h1' as [h1' [h2 [h3 | h3]]]; subst; subst.
  symmetry in h3.
  apply sing_eq_add in h3.
  destruct h3 as [? [? | ?]]; subst; tauto.
  apply add_not_empty in h3; contradiction.
  destruct h1' as [h1' h2].
  destruct h2 as [h2 | h2].
  destruct h2 as [? h2]; subst.
  destruct h2 as [h2 | h2].
  symmetry in h2.
  apply sing_eq_add in h2. 
  destruct h2 as [? [? | ?]]; subst. 
  right.
  split; auto.
  right.
  split; auto.
  right.
  split; auto.
  tauto.
  right.
  split; auto.
  destruct (hdt x y) as [h3 | h3]; subst.
  rewrite in_add_eq in h1.
  left. 
  split; auto. 
  pose proof (Couple_r _ y b) as h3. 
  rewrite h1 in h3.
  destruct h3 as [b h3 | b h3].
  destruct (in_fin_ens_dec hdt A hf'' y) as [h4 | h4].
  rewrite in_add_eq in h2; auto.
  rewrite couple_comm, couple_add_add in h1.
  apply add_nin_inj_eq in h1; auto; subst.
  rewrite add_empty_sing; tauto.
  rewrite add_empty_sing, in_sing_iff; auto.
  destruct h3.
  tauto.
  right; constructor.
  pose proof (Couple_l _ y b) as h4. 
  right. 
  split; auto.
  rewrite <- h2 in h4.
  destruct h4 as [y h4 |y h4]; subst.
  rewrite in_add_eq in h1.
  pose proof (Add_intro2 _ A x) as h5.
  rewrite <- h1 in h5.
  destruct h5.
  contradict h3; auto.
  split; auto.
  destruct (in_fin_ens_dec hdt _ hf'' b) as [h5 | h5].
  rewrite in_add_eq in h1; subst; auto.
  rewrite couple_add_add in h1.
  apply add_nin_inj_eq in h1; auto.
  rewrite add_empty_sing in h1; subst; tauto.
  rewrite add_empty_sing, in_sing_iff; auto.
  left; auto.
  destruct h4.
  contradict h3; auto.
  destruct h2 as [? [h3 | h4]]; subst.
  right.
  symmetry in h3.
  apply sing_eq_add in h3.
  destruct h3 as [? [? | ?]]; subst.
  tauto.
  tauto.
  pose proof (Add_intro2 _ A x) as h5.
  rewrite h4 in h5.
  destruct h5.
  pose proof (Couple_r _ a y) as h5.
  rewrite <- h4 in h5.
  destruct h5 as [y h5 | y h5].
  rewrite in_add_eq in h1.
  destruct (in_fin_ens_dec hdt _ hf'' a) as [h6 | h6].
  rewrite in_add_eq in h4; subst; tauto.
  rewrite couple_comm, couple_add_add in h1.
  apply add_nin_inj_eq in h1; auto.
  rewrite add_empty_sing in h1; subst; tauto.
  rewrite add_empty_sing, in_sing_iff; auto.
  left; auto.
  destruct h5.
  contradict h1'; auto.
  rewrite in_add_eq in h1.
  pose proof (Couple_l _ a y) as h5.
  rewrite <- h4 in h5.
  destruct h5 as [a h5 | a h5]; subst.
  destruct (in_fin_ens_dec hdt _ hf'' y) as [h6 | h6].
  rewrite in_add_eq in h1; subst; tauto.
  rewrite couple_add_add in h1.
  apply add_nin_inj_eq in h1; auto.
  rewrite add_empty_sing in h1; subst; tauto.
  rewrite add_empty_sing, in_sing_iff; auto.
  destruct h5.
  contradict h1'; auto.
  right; constructor.

  destruct h1 as [h1 | h1].
  destruct h1 as [? [? [? [? | ?]]]]; subst; subst.
  rewrite add_empty_sing, in_add_eq.
  rewrite couple_singleton; auto.
  constructor.
  do 2 rewrite in_add_eq.
  rewrite couple_singleton; auto.
  constructor.
  constructor.
  constructor.
  destruct h1 as [h1 h2].
  destruct h2 as [h2 | h2].
  destruct h2 as [h2 h3]; subst.
  destruct h3 as [h3 | h3].
  destruct h3 as [? [? | ?]]; subst.
  rewrite in_add_eq.
  rewrite in_add_eq; auto.
  left.
  right.
  constructor.
  rewrite in_add_eq, <- couple_add_sing, couple_comm; auto.
  right; constructor.
  destruct h3 as [? [? | ?]]; subst.
  rewrite in_add_eq.
  rewrite in_add_eq; auto.
  right.
  right.
  constructor.
  rewrite in_add_eq, couple_add_sing; auto.
  right.
  constructor.
  destruct h2 as [h2 h3].
  destruct h3 as [h3 h4]. 
  destruct h3 as [[? ?] | [? ?]], h4 as [? | [? | [? | ?]]]; subst.
  rewrite couple_add_add; auto.
  rewrite (in_add_eq (Singleton a) a), couple_add_add; auto.
  rewrite <- couple_add_add, <- couple_add_sing; auto.
  constructor.
  rewrite in_add_eq, couple_comm, <- couple_add_sing; auto.
  left; constructor.
  do 2 rewrite in_add_eq; auto.
  left.
  right.
  left.
  rewrite couple_comm, couple_add_add; auto.
  rewrite in_add_eq, <- couple_add_sing; auto.
  left; constructor.
  rewrite (in_add_eq (Singleton b) b), <- couple_add_sing, couple_comm; auto.
  constructor.
  rewrite in_add_eq, in_add_eq; auto.
  right.
  left.
  left.
Qed.


End Random.

 


    
Lemma infinite_nat : 
  ~Finite (Full_set nat).
pose proof nat_infinite as h1.
rewrite Finite_FiniteT_iff.
assumption.
Qed.
 
Lemma infinite_pos_evens : 
  ~Finite pos_evens.
intro h1.
rewrite pos_evens_eq in h1.
pose proof (Pigeonhole_bis _ _ _ _ infinite_nat h1) as h2.
contradict h2.
red.
intros; omega.
Qed.

Lemma infinite_pos_odds : 
  ~Finite pos_odds.
intro h1.
rewrite pos_odds_eq in h1.
pose proof (Pigeonhole_bis _ _ _ _ infinite_nat h1) as h2.
contradict h2.
red.
intros; omega.
Qed.
 


Definition segT_transfer {m n:nat} (pf:m = n) (a:m @) :
  n @ :=
  s_transfer_dep pf a.

Definition segT_transfer_r {m n:nat} (pf:n = m) (a:m @) :
  n @ :=
  s_transfer_dep_r pf a.

Lemma segT_transfer_transfer_r_compat : 
  forall {m n:nat} (pf:m = n) (a:m @),
    segT_transfer pf a  =
    segT_transfer_r (eq_sym pf) a.
  intros m n h1 a. subst.
  unfold segT_transfer, segT_transfer_r.
  rewrite s_transfer_dep_eq_refl.
  pose proof (proof_irrelevance _ (@eq_sym nat n n (@eq_refl nat n))
                                (eq_refl)) as h1.
  rewrite h1.
  rewrite s_transfer_dep_r_eq_refl.
  reflexivity.
Qed.


Lemma segT_transfer_r_transfer_compat : 
  forall {m n:nat} (pf:n = m) (a:m @),
    segT_transfer_r pf a  =
    segT_transfer (eq_sym pf) a.
  intros m n h1 a. subst.
  unfold segT_transfer, segT_transfer_r.
  rewrite s_transfer_dep_r_eq_refl.
  pose proof (proof_irrelevance _ (@eq_sym nat m m (@eq_refl nat m))
                                (eq_refl _)); subst.
  rewrite s_transfer_dep_eq_refl; auto.
Qed.

Lemma segT_transfer_eq_refl :
  forall {m:nat} (pf:m = m) (a:m @),
    segT_transfer pf a = a.
  intros m h1 a.
  unfold segT_transfer.
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  rewrite s_transfer_dep_eq_refl; auto.
Qed.

Lemma segT_transfer_r_eq_refl :
  forall {m:nat} (pf:m = m) (a:m @),
    segT_transfer_r pf a = a.
  intros m h1 a.
  unfold segT_transfer_r.
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  rewrite s_transfer_dep_r_eq_refl; auto.
Qed.


Lemma segT_transfer_undoes_segT_transfer_r :
  forall {m n:nat} (pf:m = n) (a:n @),
    segT_transfer pf (segT_transfer_r pf a) = a.
  intros m n h1 a. subst.
  rewrite segT_transfer_eq_refl, segT_transfer_r_eq_refl.
  reflexivity.
Qed.


Lemma segT_transfer_r_undoes_segT_transfer :
  forall {m n:nat} (pf:m = n) (a:m @),
    segT_transfer_r pf (segT_transfer pf a) = a.
  intros m n h1 a. subst.
  rewrite segT_transfer_eq_refl, segT_transfer_r_eq_refl.
  reflexivity.
Qed.

Lemma segT_transfer_cond :
  forall {m n:nat} (pf:m = n) (a:m @),
    proj1_sig a < n.
  intros m n h1 a.
  subst.
  apply proj2_sig.
Qed.


Lemma segT_transfer_compat :
  forall {m n:nat} (pf:m = n) (a:m @),
    segT_transfer pf a =
    exist (fun a => a < n) _ (segT_transfer_cond pf a). 
  intros m n h1 a.
  unfold segT_transfer.
  destruct h1.
  rewrite transfer_dep_eq_refl.
  apply proj1_sig_injective; auto.
Qed.


Lemma segT_transfer_r_compat :
  forall {m n:nat} (pf:m = n) (a:n @),
    segT_transfer_r pf a =
    exist (fun a => a < m)  _ (segT_transfer_cond (eq_sym pf) a). 
  intros m n h1 a.
  unfold segT_transfer_r.
  destruct h1.
  rewrite transfer_dep_r_eq_refl.
  apply proj1_sig_injective; auto.
Qed.


Definition transfer_segT_fun1 {n m U} (pf:n = m) (f:n @->U)
: m @ -> U :=
  fun x => f (segT_transfer (eq_sym pf) x).

Definition transfer_segT_fun2 {n m T} (pf:n = m) (f:T -> n @)
: T -> m @ :=
  fun x => segT_transfer pf (f x).


Lemma transfer_segT_fun1_eq_refl :
  forall {U n} (pf:n = n) (f:n @ -> U),
    transfer_segT_fun1 pf f = f.
  intros U n h1 f; apply functional_extensionality;
  unfold transfer_segT_fun1, eq_rect_r, eq_rect;  intro.
  pose proof (proof_irrelevance _ (eq_sym h1) (eq_refl _)) as h2;
    rewrite h2; auto.
Qed.


Lemma transfer_segT_fun2_eq_refl :
  forall {T n} (pf:n = n) (f:T -> n @),
    transfer_segT_fun2 pf f = f.
  intros U n h1 f; apply functional_extensionality;
  unfold transfer_segT_fun1, eq_rect_r, eq_rect;  intro.
  pose proof (proof_irrelevance _ h1 (eq_refl _))  as h2; 
  rewrite h2; auto.
Qed.



Lemma transfer_segT_fun1_compat :
  forall {U n m} (pf:n = m) (f:n @ -> U) (x:m @),
    transfer_segT_fun1 pf f x  =
    f (segT_transfer (eq_sym pf) x).
  intros; subst; rewrite transfer_segT_fun1_eq_refl,
                 segT_transfer_eq_refl; auto.
Qed.

Lemma transfer_segT_fun2_compat :
  forall {T n m} (pf:n = m) (f:T -> n @) (x:T),
    transfer_segT_fun2 pf f x =
    segT_transfer pf  (f x).
  intros; subst; rewrite transfer_segT_fun2_eq_refl,segT_transfer_eq_refl; auto.
Qed. 


Definition segT_transfer_ens {m n:nat} (pf:m = n) (S:Ensemble (m @))
  : Ensemble (n @) :=
  transfer_dep (U := (fun a:nat => Ensemble (segT a))) pf S.


Definition segT_transfer_ens_r {m n:nat} (pf:m = n) (S:Ensemble (n @)) : Ensemble (m @)
  :=
  transfer_dep_r (U := (fun a:nat => Ensemble (segT a))) pf S.


Definition segT_transfer_fam {m n:nat} (pf:m = n) (S:Family (m @))
  : Family (n @) :=
  transfer_dep (U := (fun a:nat => Family (segT a))) pf S.


Definition segT_transfer_fam_r {m n:nat} (pf:m = n) (S:Family (n @)) : Family (m @)
  :=
  transfer_dep_r (U := (fun a:nat => Family (a @))) pf S.

Lemma segT_transfer_ens_eq_refl :
  forall (m:nat) (S:Ensemble (m @)),
    segT_transfer_ens eq_refl S = S.
  intros.
  unfold segT_transfer_ens.
  rewrite transfer_dep_eq_refl.
  reflexivity.
Qed.

Lemma segT_transfer_ens_r_eq_refl :
  forall (m:nat) (S:Ensemble (m @)),
    segT_transfer_ens_r eq_refl S = S.
  intros.
  unfold segT_transfer_ens_r.
  rewrite transfer_dep_r_eq_refl.
  reflexivity.
Qed.

Lemma segT_transfer_fam_eq_refl :
  forall (m:nat) (S:Family (m @)),
    segT_transfer_fam eq_refl S = S.
  intros.
  unfold segT_transfer_fam.
  rewrite transfer_dep_eq_refl.
  reflexivity.
Qed.

Lemma segT_transfer_fam_r_eq_refl :
  forall (m:nat) (S:Family (m @)),
    segT_transfer_fam_r eq_refl S = S.
  intros.
  unfold segT_transfer_fam_r.
  rewrite transfer_dep_r_eq_refl.
  reflexivity.
Qed.

Lemma in_segT_transfer_ens :
  forall {m n:nat} (pf:m = n) (S:Ensemble (m @)) (x:m @),
    Inn S x <->
    Inn (segT_transfer_ens pf S) (segT_transfer pf x).
  intros m n h1 S x.
  subst.
  rewrite segT_transfer_ens_eq_refl, segT_transfer_eq_refl.
  tauto.
Qed.

Lemma in_segT_transfer_ens_r :
  forall {m n:nat} (pf:m = n) (S:Ensemble (n @)) (x:n @),
    Inn S x <->
    Inn (segT_transfer_ens_r pf S) (segT_transfer_r pf x).
  intros m n h1 S x.
  subst.
  rewrite segT_transfer_ens_r_eq_refl.
  rewrite segT_transfer_r_eq_refl.
  tauto.
Qed.


Lemma in_segT_transfer_ens_hetero :
  forall {m n:nat} (pf:m = n) (S:Ensemble (m @)) (x:n @),
    Inn (segT_transfer_ens pf S) x <->
    Inn S (segT_transfer_r pf x).
  intros m n h1 S x.
  subst.
  rewrite segT_transfer_ens_eq_refl.
  rewrite segT_transfer_r_eq_refl.
  tauto.
Qed.

Lemma in_segT_transfer_ens_r_hetero :
  forall {m n:nat} (pf:m = n) (S:Ensemble (n @)) (x:m @),
    Inn (segT_transfer_ens_r pf S) x <->
    Inn S (segT_transfer pf x).
  intros m n h1 S x.
  subst.
  rewrite segT_transfer_ens_r_eq_refl.
  rewrite segT_transfer_eq_refl.
  tauto.
Qed.


Lemma in_segT_transfer_fam :
  forall {m n:nat} (pf:m = n) (F:Family (m @)) (S:Ensemble (m @)),
    Inn F S <->
    Inn (segT_transfer_fam pf F) (segT_transfer_ens pf S).
  intros m n h1 F S.
  subst.
  rewrite segT_transfer_fam_eq_refl, segT_transfer_ens_eq_refl.
  tauto.
Qed.


Lemma in_segT_transfer_fam_r :
  forall {m n:nat} (pf:m = n) (F:Family (n @)) (S:Ensemble (n @)),
    Inn F S <->
    Inn (segT_transfer_fam_r pf F) (segT_transfer_ens_r pf S).
  intros m n h1 F S.
  subst.
  rewrite segT_transfer_fam_r_eq_refl, segT_transfer_ens_r_eq_refl.
  tauto.
Qed.


Lemma in_segT_transfer_fam_hetero :
  forall {m n:nat} (pf:m = n) (F:Family (m @)) (S:Ensemble (n @)),
    Inn (segT_transfer_fam pf F) S <->
    Inn F (segT_transfer_ens_r pf S).
  intros m n h1 F S.
  subst.
  rewrite segT_transfer_fam_eq_refl, segT_transfer_ens_r_eq_refl.
  tauto.
Qed.

Lemma in_segT_transfer_fam_r_hetero :
  forall {m n:nat} (pf:m = n) (F:Family (n @)) (S:Ensemble (m @)),
    Inn (segT_transfer_fam_r pf F) S <->
    Inn F (segT_transfer_ens pf S).
  intros m n h1 F S.
  subst.
  rewrite segT_transfer_fam_r_eq_refl.
  rewrite segT_transfer_ens_eq_refl.
  tauto.
Qed.


(*Leftover from seg_set_one_t days.  Might phase out eventually.*)
Definition sot_transfer {m n:nat} (pf:m = n) (a:seg_one_t m) :
  seg_one_t n :=
  s_transfer_dep pf a.

Definition sot_transfer_r {m n:nat} (pf:n = m) (a:seg_one_t m) :
  seg_one_t n :=
  s_transfer_dep_r pf a.

Lemma sot_transfer_transfer_r_compat : 
  forall {m n:nat} (pf:m = n) (a:seg_one_t m),
    sot_transfer pf a  =
    sot_transfer_r (eq_sym pf) a.
  intros m n h1 a. subst.
  unfold sot_transfer, sot_transfer_r.
  rewrite s_transfer_dep_eq_refl.
  pose proof (proof_irrelevance
                _ (@eq_sym nat n n (@eq_refl nat n))
                (eq_refl _)) as h1.
  rewrite h1; auto.
Qed.


Lemma sot_transfer_r_transfer_compat : 
  forall {m n:nat} (pf:n = m) (a:seg_one_t m),
    sot_transfer_r pf a  =
    sot_transfer (eq_sym pf) a.
  intros m n h1 a. subst.
  unfold sot_transfer, sot_transfer_r.
  rewrite s_transfer_dep_r_eq_refl.
  pose proof (proof_irrelevance
                _ (@eq_sym nat m m (@eq_refl nat m))
                (eq_refl _)) as h1.
  rewrite h1; auto.
Qed.


Lemma sot_transfer_eq_refl :
  forall {m:nat} (a:seg_one_t m),
    sot_transfer eq_refl a = a.
  intros m a.
  unfold sot_transfer.
  rewrite s_transfer_dep_eq_refl.
  reflexivity.
Qed.

Lemma sot_transfer_r_eq_refl :
  forall {m:nat} (a:seg_one_t m),
    sot_transfer_r eq_refl a = a.
  intros m a.
  unfold sot_transfer_r.
  rewrite s_transfer_dep_r_eq_refl.
  reflexivity.
Qed.


Lemma sot_transfer_undoes_sot_transfer_r :
  forall {m n:nat} (pf:m = n) (a:seg_one_t n),
    sot_transfer pf (sot_transfer_r pf a) = a.
  intros m n h1 a. subst.
  rewrite sot_transfer_eq_refl, sot_transfer_r_eq_refl.
  reflexivity.
Qed.


Lemma sot_transfer_r_undoes_sot_transfer :
  forall {m n:nat} (pf:m = n) (a:seg_one_t m),
    sot_transfer_r pf (sot_transfer pf a) = a.
  intros m n h1 a. subst.
  rewrite sot_transfer_eq_refl, sot_transfer_r_eq_refl.
  reflexivity.
Qed.

Lemma sot_transfer_cond :
  forall {m n:nat} (pf:m = n) (a:seg_one_t m),
    1 <= (proj1_sig a) <= n.
  intros m n h1 a.
  subst.
  apply proj2_sig.
Qed.


Lemma sot_transfer_eq :
  forall {m n:nat} (pf:m = n) (a:seg_one_t m),
    sot_transfer pf a =
    exist _ _ (sot_transfer_cond pf a). 
  intros m n h1 a.
  unfold sot_transfer.
  destruct h1.
  rewrite transfer_dep_eq_refl.
  apply proj1_sig_injective; auto.
Qed.


Lemma sot_transfer_r_eq :
  forall {m n:nat} (pf:m = n) (a:seg_one_t n),
    sot_transfer_r pf a =
    exist _ _ (sot_transfer_cond (eq_sym pf) a). 
  intros m n h1 a.
  unfold sot_transfer_r.
  destruct h1.
  rewrite transfer_dep_r_eq_refl.
  apply proj1_sig_injective; auto.
Qed.


Section FiniteSets.

(*We'll be using lists as sets directly, transparently, 
with set equality the following predicate, and some [Setoid] 
functionality.*)
Inductive set_eq {T} (l l':list T) : Prop :=
| set_eq_intro : incl l  l' ->
                 incl l' l ->
                 set_eq l l'.


Lemma refl_set_eq :
  forall T (l:list T),
    set_eq l l.
  intros; constructor; auto; apply incl_refl.
Qed.
   
Lemma sym_set_eq :
  forall T, Symmetric (set_eq (T:=T)).
  intro; red; intros; inversion H; constructor; auto.
Qed.

Lemma trans_set_eq :
  forall T, Transitive (set_eq (T:=T)).
  intro; red; intros ? ? ? h1 h2; destruct h1, h2; constructor; auto;
  [eapply trans_incl; try apply H; try assumption |
   eapply trans_incl; try apply H2; try assumption].
Qed.
End FiniteSets.

Add Parametric Relation
    (T:Type) : (list T) (@set_eq T)
    reflexivity proved by (@refl_set_eq T)
    symmetry proved by (@sym_set_eq T)
    transitivity proved by (@trans_set_eq T)
      as list_eq.


      
Infix "==" := set_eq (at level 70).

Notation "x <=> y" := (not (x == y)) (at level 70).

Section FiniteSets'.
Lemma not_set_eq_neq :
  forall {T} (A B:list T),
    A <=> B ->
    A <> B.
  intros ? ? ? h1; contradict h1; subst. reflexivity.
Qed.


Add Parametric Relation
    (T:Type) : (list T) (@incl T)
    reflexivity proved by (@incl_refl T)
    transitivity proved by (@trans_incl T)
      as Fincl.

Lemma list_ext :
  forall {T} (E F:list T),
    incl E F ->
    incl F E ->
    E == F.
  intros; constructor; auto.
Qed.

Inductive full_fin_set {T} (A:list T) : Prop :=
| full_fin_set_intro : (forall x:T, In x A) -> full_fin_set A.

Add Parametric Morphism {T} :
  full_fin_set
    with signature (@set_eq T ==> iff)
      as full_fin_set_eq.
Proof.
  intros ? ? h1; destruct h1.
  split; intro h1; destruct h1; constructor; intros;
  [eapply H; auto |
   eapply H0; auto].
Qed.

Add Parametric Morphism {T} :
  full_fin_set
    with signature (@incl T ++> impl)
      as full_fin_set_incl.
Proof.
  intros l l' h1; red; intro h2.
  constructor.
  intro x.
  destruct h2 as [h2].
  specialize (h2 x).
  apply h1 in h2; auto.
Qed.


Add Parametric Morphism {T} :
  full_fin_set
    with signature (flip (@incl T) --> impl)
      as full_fin_set_incl'.
Proof.
  intros l l' h1; red; intro h2; destruct h2; constructor; intro x;
  specialize (H x); apply h1 in H; auto.
Qed.


Lemma in_addl_eq :
  forall {T} (A:list T) (a:T),
    In a A ->
    a :: A == A.
  intros T A a h1; constructor; red; intros x h2.
  destruct h2 as [|h2]; subst; auto.
  right; auto.
Qed.

Lemma set_eq_nil :
  forall {T} (A:list T),
    A == nil ->
    A = nil.
  intros T A h1.
  inversion h1 as [h2].
  apply incl_nil' in h2; auto.
Qed.


  
Lemma disc_set_eq : 
  forall {T} (A:list T) (a:T),
    a :: A <=> nil.
  intros T A a h1.
  inversion h1.
  apply incl_nil' in H; discriminate.
Qed.

Add Parametric Morphism {T} (a:T) :
    (In a)
    with signature (@incl T ++> impl)
      as in_set_incl_impl.
  intros A B h1; red; intro h2; apply h1 in h2; auto.
Qed.


Add Parametric Morphism {T} (a:T) :
    (In a)
    with signature (set_eq ==> impl)
      as in_set_impl.
  intros A B h1; inversion h1 as [h2 h3]; red; intro h4.
  apply h2 in h4; auto.
Qed.

Add Parametric Morphism {T} (a:T) :
    (In a)
    with signature (set_eq ==> flip impl)
      as in_set_flip_impl.
  intros A B h1; inversion h1 as [h2 h3]; red; intro h4.
  apply h3 in h4; auto.
Qed.


Add Parametric Morphism {T} (a:T) :
    (In a)
    with signature (set_eq ==> iff)
      as in_set_iff.
  intros l l' h1; destruct h1 as [h2 h3]; split; intro h4;
  [apply h2 in h4; auto |
   apply h3 in h4; auto].
Qed.
                    

Add Parametric Morphism {T} (a:T) :
  (cons a)
    with signature (set_eq ==> set_eq)
      as cons_set_eq.
  intros A B h1. 
  constructor; red; intros c h2; destruct h2 as [|h2]; subst.
  left; auto.
  rewrite h1 in h2.
  right; auto.
  left; auto.
  rewrite <- h1 in h2.
  right; auto.
Qed.



Add Parametric Morphism {T} (pfdt:type_eq_dec T) (a:T) :
  (remove pfdt a)
    with signature (@incl T ==> @incl T)
      as remove_incl.
  intros A B h1; red in h1; red; intros x hx; rewrite in_remove_iff in hx; destruct hx as [h2 h3]; subst.
  rewrite in_remove_iff.
  split; auto.
Qed.



Add Parametric Morphism {T} (pfdt:type_eq_dec T) (a:T) :
  (remove pfdt a)
    with signature (set_eq ==> set_eq)
      as remove_set_eq.
  intros ? ? h1.
  destruct h1; constructor; apply remove_incl; auto.
Qed.

Lemma remove_sig_set_eq :
  forall {T} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pfn : ~In a l),
    map (@proj1_sig _ _) (remove_sig_list pfdt pfn) == remove pfdt a l.  
  intros T hdt l a h1.
  constructor; red; intros x h2;  
  [rewrite in_map_iff in h2;
    rewrite in_remove_iff;
    destruct h2 as [x' [h2 h3]];
    destruct x' as [x' hx']; subst;
    unfold remove_sig_list in h3;
    rewrite in_map_in_dep_iff in h3;
    destruct h3 as [x [h3 h4]];
    lr_match' h4;
    subst; try contradiction;
    try destruct x' as [x' h5];
    apply exist_injective in h4; subst; simpl; auto
  |
  rewrite in_remove_iff in h2;
    rewrite in_map_iff;
    destruct h2 as [h2 h3];
    exists (exist (fun c => c <> a) _ h3);
    split; auto;
    unfold remove_sig_list;
    rewrite in_map_in_dep_iff;
    exists x, h2;
    lr_match_goal';
    try contradiction;
    f_equal;
    apply proof_irrelevance].
Qed.


Definition cardl {T} (pfdt:type_eq_dec T)
           (E:list T) :=
  length (list_singularize pfdt E).

Lemma addl_eq_iff :
  forall {T} (pfdt:type_eq_dec T)
         (A B:list T) a,
    ~In a A ->
    (B == a :: A <->
     A == (remove pfdt a B) /\ In a B).
  intros T hdt A B a h1; split; intro h2.
  split.
  constructor; red; intros c h3.
  rewrite in_remove_iff.
  pose proof (in_cons a _ _ h3) as h4.
  rewrite <- h2 in h4.
  split; auto; intro; subst; contradiction.
  rewrite in_remove_iff in h3.
  destruct h3 as [h3 h4].
  rewrite h2 in h3. 
  destruct h3 as [|h3]; subst.
  contradict h4; auto.
  assumption. 
  rewrite h2.
  left; auto.
  destruct h2 as [h2 hi].
  constructor; red; intros c h3.
  destruct (hdt c a) as [|h5]; subst.
  left; auto.
  pose proof (in_remove_iff hdt B c a) as h6.
  hr h6. split; auto.
  rewrite <- h6 in hr.
  rewrite <- h2 in hr.
  right; auto.
  destruct (hdt c a) as [|h4]; subst; auto.
  destruct h3 as [|h3]; subst.
  contradict h4; auto.
  rewrite h2 in h3.
  rewrite in_remove_iff in h3.
  destruct h3; auto.
Qed.  

Lemma addl_inj_iff :
  forall {T} (pfdt:type_eq_dec T)
         (A B:list T) a b,
    ~In a A ->
    ~In a B ->
    ~In b A ->
    ~In b B ->
    (b :: B == a :: A <->
     (a = b /\ A == B )).
  intros T hdt A B a b; intros;
  split; intro h1.
  inversion h1 as [h2 h3]. 
  red in h2, h3. 
  destruct (hdt a b) as [|h6]; subst.
  split; auto.
  constructor; red; intros x h4.
  destruct (hdt x b) as [|h5]; subst.
  contradiction. 
  specialize (h3 _ (in_cons _ _ _ h4)).
  destruct h3; subst; auto; contradict h5; auto.

  pose proof (in_cons b _ _ h4) as h5.
  rewrite h1 in h5.
  destruct h5 as [|h5]; subst.
  contradiction.
  assumption.
  pose proof (in_eq b B) as h7.
  rewrite h1 in h7.
  destruct h7 as [|h7]; subst.
  contradict h6; auto.
  contradiction.
  destruct h1 as [? h1]; subst.
  rewrite h1; reflexivity.
Qed.

Lemma addl_inj_iff' :
  forall {T} (pfdt:type_eq_dec T)
         (A B:list T) a b,
    a <> b -> 
    ~In a A ->
    ~In b B ->
    In a B ->
    In b A ->
    (b :: B == a :: A <->
     (remove pfdt b A ==
      remove pfdt a B)).
  intros T hdt A B a b; intros; split; intro h1.
  inversion h1 as [h2 h3].
  red in h2, h3.
  constructor; red; intros x h4.
  rewrite in_remove_iff in h4.
  destruct h4 as [h4 h5].
  destruct (hdt x a) as [|h6]; subst.
  contradiction.
  specialize (h3 _ (in_cons a _ _ h4)). 
  destruct h3 as [|h3]; subst.
  contradict h5; auto.
  rewrite in_remove_iff.
  split; auto.
  rewrite in_remove_iff in h4.
  destruct h4 as [h4 h5].
  destruct (hdt x b) as [|h6]; subst.
  contradiction.
  specialize (h2 _ (in_cons b _ _ h4)).
  destruct h2 as [|h2]; subst.
  contradict h5; auto.
  rewrite in_remove_iff; split; auto.
  inversion h1 as [h2 h3].
  red in h2, h3.
  constructor; red; intros x h4;
  destruct
    (hdt x a)
    as [|he],
       (hdt x b) as
        [|he']; subst.
  left; auto.
  left; auto.
  right; auto.
  destruct h4 as [|h4]; subst.
  contradict he'; auto.
  right.
  specialize (h3 x).
  hfirst h3.
  rewrite in_remove_iff; split; auto.
  specialize (h3 hf).
  rewrite in_remove_iff in h3.
  destruct h3; auto.
  subst.
  left; auto.
  right; auto.
  left; auto.
  right.
  destruct h4 as [|h4]; subst.
  contradict he; auto.
  specialize (h2 x).
  hfirst h2.
  rewrite in_remove_iff; split; auto.
  specialize (h2 hf).
  rewrite in_remove_iff in h2; destruct h2; auto.
Qed.

  

Add Parametric Morphism {T} (pfdt:type_eq_dec T) :
  (list_singularize pfdt)
    with signature (@incl T ++> @incl T)
      as list_singularize_incl.
  intros A B h1; red in h1; red; intros a h2.
  rewrite <- list_singularize_in_iff in h2.
  apply h1 in h2.
  rewrite <- list_singularize_in_iff; auto.
Qed.

Add Parametric Morphism {T} (pfdt:type_eq_dec T) :
  (list_singularize pfdt)
    with signature (@set_eq T ==> @set_eq T)
      as list_singularize_set_eq.
  intros ? ? h1;
  destruct h1 as [h1 h2]; red in h1, h2; constructor; intros c h3;
  rewrite <- list_singularize_in_iff in h3;
  rewrite <- list_singularize_in_iff;
  [apply h1 in h3; auto |
   apply h2 in h3; auto].
Qed.

Add Parametric Morphism {T} (pfdt:type_eq_dec T) :
  (fun l => length (list_singularize pfdt l))
    with signature (@set_eq T ==> @eq nat)
      as length_list_singularize_set_eq.
  intros l l' h1.
  apply list_to_sets_eq_length.
  destruct h1.
  apply anti_sym_incl_list_to_set; auto.
Qed.


Add Parametric Morphism {T} (pfdt:type_eq_dec T) :
  (cardl pfdt)
    with signature (@set_eq T ==> @eq nat)
      as cardl_eq.
  unfold cardl; intro l.
  induction l as [|a l ih]; simpl.
  intros ? h2.
  symmetry in h2.
  apply set_eq_nil in h2; subst; simpl; auto.
  intro l'.
  destruct l' as [|a' l'].
  intro h2.
  apply set_eq_nil in h2; discriminate.
  intro h2.
  destruct (pfdt a a') as [|h3]; subst.
  destruct (in_dec pfdt a' l) as [h4 | h4],
                               (in_dec pfdt a' l') as [h5 | h5].
  rewrite list_singularize_in_cons; auto.
  rewrite list_singularize_in_cons; auto.
  apply ih.
  do 2 rewrite in_addl_eq in h2; auto.
  rewrite in_addl_eq in h2.
  pose proof (ih (a'::l') h2) as h6.
  simpl in h6.
  rewrite list_singularize_in_cons; auto.
  assumption.
  rewrite list_singularize_nin; auto.
  rewrite list_singularize_in_cons; auto.
  simpl.
  rewrite (in_addl_eq l') in h2.
  symmetry in h2.
  rewrite (addl_eq_iff pfdt) in h2.
  destruct h2 as [h2 h6].
  specialize (ih (remove pfdt a' l')).
  hfirst ih.
  assumption.
  specialize (ih hf).
  rewrite list_singularize_remove,
  length_remove_no_dup in ih.
  pose proof h5 as h5'.
  rewrite (list_singularize_in_iff pfdt) in h5'.
  apply in_not_nil in h5'.
  apply not_nil_lt in h5'.
  omega.
  apply no_dup_list_singularize.
  rewrite <- list_singularize_in_iff; auto.
  assumption.
  assumption.
  assumption.
  rewrite list_singularize_nin; auto.
  rewrite list_singularize_nin; auto; simpl.
  f_equal; apply ih.
  rewrite addl_inj_iff in h2; auto.
  destruct h2; symmetry; auto.
  destruct
    (in_dec pfdt a l)
    as [h4|h4],
       (in_dec pfdt a' l') as [h5|h5].
  rewrite list_singularize_in_cons.
  rewrite list_singularize_in_cons.
  do 2 rewrite in_addl_eq in h2; auto.
  assumption.
  assumption.
  rewrite list_singularize_in_cons; auto.
  rewrite list_singularize_nin; auto.
  rewrite in_addl_eq in h2.
  specialize (ih (a'::l')).
  apply ih in h2.
  rewrite list_singularize_nin in h2; auto.
  assumption.
  rewrite list_singularize_nin; auto.
  rewrite list_singularize_in_cons; auto.
  simpl.
  rewrite (in_addl_eq l') in h2.
  symmetry in h2.
  rewrite (addl_eq_iff pfdt) in h2.
  destruct h2 as [h2 h6].
  specialize (ih (remove pfdt a l') h2).
  rewrite ih,
  list_singularize_remove,
  length_remove_no_dup.
  rewrite (list_singularize_in_iff pfdt) in h5.
  apply in_not_nil in h5.
  apply not_nil_lt in h5.
  omega.
  apply no_dup_list_singularize.
  rewrite <- list_singularize_in_iff; auto.
  assumption.
  assumption.
  assumption. 
  rewrite list_singularize_nin; auto.
  rewrite list_singularize_nin; auto. 
  simpl; f_equal.
  pose proof (in_eq a l) as h6.
  rewrite h2 in h6.
  destruct h6 as [|h6]; subst.
  contradict h3; auto.
  pose proof (in_eq a' l') as h6'.
  rewrite <- h2 in h6'.
  destruct h6' as [|h6']; subst.
  contradiction. 
  specialize (ih (remove pfdt a (a'::l'))). 
  hfirst ih.
  rewrite <- h2.
  rewrite remove_cons_eq, nin_remove_eq; auto.
  reflexivity.
  specialize (ih hf).
  rewrite list_singularize_remove, list_singularize_nin,
  length_remove_no_dup in ih; auto.
  constructor.
  rewrite <- list_singularize_in_iff; auto.
  apply no_dup_list_singularize.
  right.
  rewrite <- list_singularize_in_iff; auto.
  right; auto.
Qed.


Definition proj1_sig_set_eq {T} {P Q:T->Prop}
           (l: list {x:T | P x}) 
           (l':list {x:T | Q x}) : Prop :=
  map (@proj1_sig _ P) l ==
  map (@proj1_sig _ Q) l'.


Lemma proj1_sig_set_nil1 :
  forall {T} (P Q:T->Prop)
         (l': list {x:T | P x}),
    proj1_sig_set_eq nil l' (P:=P) ->
    l' = nil.
  unfold proj1_sig_set_eq; intros ? ? ? l.
  destruct l; auto; simpl.
  intro h1.
  symmetry in h1.
  apply disc_set_eq in h1; contradiction.
Qed.

Lemma proj1_sig_set_nil2 :
  forall {T} (P Q:T->Prop)
         (l: list {x:T | P x}),
    proj1_sig_set_eq l nil (Q:=Q) ->
    l = nil.
  unfold proj1_sig_set_eq; intros ? ? ? l.
  destruct l; auto; simpl.
  intro h1.
  apply disc_set_eq in h1; contradiction.
Qed.


End FiniteSets'.

