(*Copyright (C) 2017, Daniel Wyckoff*)

(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*Version 0.3 bool2 -- certified -- Version 8.4 pl4 Coq.*)


Require Import SetUtilities.
Require Import TypeUtilities.
Require Import LogicUtilities.
Require Import ProofIrrelevance.
Require Import EqDec.
Require Import ListUtilities.
Require Import FunctionalExtensionality.
Require Import SetUtilities1_5.
Require Import SetUtilities2.
Require Import ListUtilities2.
Require Import ArithUtilities.
Require Import FunctionProperties.
Require Import Description.
Require Import WellOrders.


Section Injectors.
 
  
Lemma finite_rel_classes_im_rel_sig_fun :
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U),
    Finite A -> Finite (rel_classes_im_rel_sig_fun f).
intros T U A f h1.
rewrite rel_classes_im_rel_eq_inv_im_singletons'.
rewrite im_im.
rewrite finite_full_sig_iff in h1.
unfold rel_classes_im_rel_sig_fun.
apply finite_image; auto.
Qed.


Lemma ex_same_im_subset_sig_inj : 
  forall {T U:Type} {A:Ensemble T}
         (pfdt:type_eq_dec T),
    Finite A ->
    forall (f:sig_set A->U),
    exists (B:Ensemble T) (pf:Included B A),
      Inn (same_im_subsets_sig f) B /\
         Injective (restriction_sig f B pf). 
  intros T U A hdt h1 f.

  pose proof (finite_rel_classes_im_rel_sig_fun f h1) as h2.

pose proof (finite_set_list_no_dup _ h2) as h3.
destruct h3 as [l h3]. destruct h3 as [h3a h3b].
revert A h1 f h2 h3a h3b.
induction l as [|B l h1]. 
intros A h1 f h2 h3 h4. 
simpl in h3. apply empty_rel_classes_im_rel_eq_inv_im_singletons in h3.
 subst.
exists (Empty_set _).  exists (inclusion_reflexive _).
split. constructor. exists (inclusion_reflexive _).
f_equal.
apply functional_extensionality.
intro x. contradict (proj2_sig x).
red. intro x. contradict (proj2_sig x).
intros A h2 f h3 h4 h5.
pose proof (finite_mem_rel_classes_im_rel_sig_fun f h2) as hfm.
assert (h16:In B (B::l)). auto with datatypes.
rewrite list_to_set_in_iff in h16.
rewrite <- h4 in h16. 
pose proof (partition_fam_rel_classes_im_rel_sig_fun f) as h17.
red in h17. destruct h17 as [h17 h18].
assert (h19:Included B A).
  rewrite <- h18.
  red. intros x h19. 
  destruct h16 as [S h16]. clear h18. subst.
  apply family_union_intro with (Im S (@proj1_sig _ _)).
  unfold rel_classes_im_rel_sig_fun.
  apply Im_intro with S; auto. assumption.
pose proof (setminus_inc A B) as h6.
assert (h7:Finite (Setminus A B)).
  apply Finite_downward_closed with A; auto.
specialize (h1 (Setminus A B) h7 (restriction_sig f _ h6)).
pose proof (finite_rel_classes_im_rel_sig_fun 
              (restriction_sig f (Setminus A B) h6) h7) as h8.
specialize (h1 h8).
pose proof (no_dup_cons_nin _ _ h5) as h9.
pose proof (no_dup_cons _ _ h5) as h10. 
assert (h11: rel_classes_im_rel_sig_fun (restriction_sig f (Setminus A B) h6) =
       list_to_set l). 
  pose proof (rel_classes_im_rel_sig_fun_setminus A B f h16) as h12.
  pose proof (proof_irrelevance _ h6 (setminus_inc A B)) as h13.
  clear h18. subst. rewrite h12.
  rewrite h4.
  red in hfm.
  simpl in h4.
  rewrite h4 in hfm.
  rewrite (eq_refl (list_to_set (B::l))) in hfm at 1.
  rewrite <- p_mem_l_iff in hfm at 1.
  pose proof (subtract_remove_ens_compat hdt (B::l) B (Finite_downward_closed _ _ h2 _ h19) hfm) as hs.
  simpl in hs; rewrite sub_add_eq in hs.
  destruct fin_ens_eq_dec as [hf | hf].
  pose proof @remove_dep_not_in.
  pose proof (p_mem_l_cons _ _ _ hfm) as hfm'.
  pose proof (remove_dep_not_in (@Finite T) (@fin_ens_eq_dec T hdt) _ B (Finite_downward_closed _ _ h2 _ h19) hfm' h9) as hr.
  rewrite <- hr .
  simpl. rewrite sub_add_eq.
  rewrite subtract_nin; auto.
  intro h6.
  rewrite <- list_to_set_in_iff in h6.
  rewrite hr in h6.
  contradiction.
  pose proof h12 as h12'. clear h12.
  rewrite h4 in h12'.
  rewrite sub_add_eq in h12'.
  rewrite subtract_nin in h12'.
  simpl. rewrite sub_add_same_nin; auto.
  rewrite <- list_to_set_in_iff; auto.
specialize (h1 h11 h10).
destruct h1 as [C h1].
destruct h1 as [h12 [h13 h14]]. 
pose proof (inh_rel_classes_im_rel_sig_fun f _ h16) as h20.
destruct h20 as [b h20].
exists (Add C b).   
assert (h21:Included (Add C b) A).
  red. intros x h21. destruct h21 as [x h21l | x h21r].
  auto with sets. destruct h21r. auto with sets.
exists h21. 
split.
constructor. exists h21.
destruct h13 as [h13]. destruct h13 as [h13 h22].
rewrite im_full_sig_add.
unfold restriction_sig. simpl. unfold restriction_sig in h22.
simpl in h22.
assert (h24: (fun x : sig_set C =>
           f
             (exist (Inn A) (proj1_sig x)
                (h6 (proj1_sig x) (h13 (proj1_sig x) (proj2_sig x))))) = 
             (fun x : sig_set C =>
         f
           (exist (Inn A) (proj1_sig x)
              (h21 (proj1_sig x) (incl_add C b (proj1_sig x) (proj2_sig x)))))).
  apply functional_extensionality.
  intro x. f_equal. apply proj1_sig_injective. simpl. reflexivity.
rewrite h24 in h22.
rewrite <- h22.
unfold restriction_sig in h14. simpl in h14.
pose proof (full_sig_decompose_setminus_incl _ _ h19) as h25.
rewrite h25 at 1.
pose proof (im_elt_of_rel_classes_im_rel_sig_fun f B h19 h16 _ h20) as h26.
rewrite im_union. do 2 rewrite im_im.
rewrite h26 at 1.
unfold Add.
f_equal. f_equal.
apply functional_extensionality.
intro x. f_equal. apply proj1_sig_injective. simpl.
reflexivity.
f_equal. f_equal. apply proj1_sig_injective. simpl.
reflexivity.
unfold restriction_sig in h14. simpl in h14.
unfold restriction_sig. simpl.
red.
intros x y h22. red in h14.
destruct x as [x h23], y as [y h24]. simpl in h22.
apply proj1_sig_injective. simpl.
destruct h23 as [x h23a | x h23b], h24 as [y h24a | y h24b].
specialize (h14 (exist _ _ h23a) (exist _ _ h24a)). simpl in h14.
assert (h25:h6 x (h12 x h23a) = h21 x (Union_introl T C (Singleton b) x h23a)). apply proof_irrelevance.
assert (h26:h6 y (h12 y h24a) = h21 y (Union_introl T C (Singleton b) y h24a)). apply proof_irrelevance.
rewrite h25, h26 in h14.
specialize (h14 h22). apply exist_injective in h14. assumption.
destruct h24b.
assert (h24:Inn B x).
  symmetry in h22.
  pose proof (rel_classes_im_rel_sig_fun_consistent_membership f B h16 _ _ h22 h20) as h24.
  simpl in h24.
  assumption.
pose proof (h12 _ h23a) as h25.
destruct h25. contradiction.
destruct h23b.
assert (h24:Inn B y).
  pose proof (rel_classes_im_rel_sig_fun_consistent_membership f B h16 _ _ h22 h20) as h24.
  simpl in h24.
  assumption.
pose proof (h12 _ h24a) as h26.
destruct h26.
contradiction.
destruct h23b, h24b. reflexivity.
Qed.




Lemma finite_inhabited_sig_fun_injectors : 
  forall {T U:Type} {A:Ensemble T} 
         (pfd:type_eq_dec T) (f:sig_set A->U),
    Finite A ->
    Inhabited (sig_fun_injectors f).
intros T U A h0 f h1.
unfold sig_fun_injectors.
pose proof (ex_same_im_subset_sig_inj h0 h1 f) as h2. 
destruct h2 as [B [h2 [h3 h4]]].
apply Inhabited_intro with B.
constructor. exists h2.
split; auto.
Qed.







Lemma finite_sig_fun_injectors :
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U),
    Finite A ->
    Finite (sig_fun_injectors f).
  intros T U A f h1.
  eapply Finite_downward_closed.
  eapply power_set_finite.
  apply h1.
  red; intros S h2.
  constructor.
  destruct h2 as [h2].
  destruct h2; auto.
Qed.

Lemma finite_sig_fun_injectors_fam :
  forall {T U:Type} {A:Ensemble T} (f:sig_set A->U),
    Finite A ->
    forall S:Ensemble T,
      Inn (sig_fun_injectors f) S ->
      Finite S.
  intros T U A f h1 S h2.
  destruct h2 as [h2].
  destruct h2 as [h2 [h3 h4]].
  destruct h4 as [h4].
  destruct h4 as [h4 h5].
  eapply Finite_downward_closed.
  apply h1.
  assumption.
Qed.

Definition nat_sig_fun_injector {U:Type} {A:Ensemble nat}
           (f:sig_set A->U) (pf:Finite A) : Ensemble nat :=
  can_set_nat_choice (sig_fun_injectors f)
                     (finite_sig_fun_injectors f pf)
                     (finite_inhabited_sig_fun_injectors nat_eq_dec f pf)
                     (finite_sig_fun_injectors_fam f pf).


Definition in_nat_sig_fun_injector :
  forall {U:Type} {A:Ensemble nat}
         (f:sig_set A->U) (pf:Finite A),
    Inn (sig_fun_injectors f) (nat_sig_fun_injector f pf).
  intros U A f h1.
  unfold nat_sig_fun_injector.
  apply (in_can_set_nat_choice (sig_fun_injectors f)
        (finite_sig_fun_injectors f h1)
        (finite_inhabited_sig_fun_injectors nat_eq_dec f h1)
        (finite_sig_fun_injectors_fam f h1)).
Qed.

Lemma incl_nat_sig_fun_injector :
  forall {U:Type} {A:Ensemble nat}
         (f:sig_set A->U) (pf:Finite A),
    Included (nat_sig_fun_injector f pf) A.
  intros U A f h1.
  pose proof (in_nat_sig_fun_injector f h1) as h2.
  destruct h2 as [h2].
  destruct h2; auto.
Qed.

Lemma finite_nat_sig_fun_injector :
  forall {U:Type} {A:Ensemble nat}
         (f:sig_set A->U) (pf:Finite A),
    Finite (nat_sig_fun_injector f pf).
  intros U A f h1.
  unfold nat_sig_fun_injector.
  unfold can_set_nat_choice.
  apply list_to_set_finite.
Qed.



Lemma nat_sig_fun_injector_zero :
  forall {U:Type} (p:sig_set (seg 1)->U),
    nat_sig_fun_injector p (finite_seg 1) =
    seg 1.
  intros U p.
  pose proof seg_1_eq as h1.
  rewrite h1.
  unfold nat_sig_fun_injector, can_set_nat_choice.
  match goal with |- list_to_set (min_set_sortable ?x1 ?x2 ?x3 ?x4 ?x5) = _ => pose proof (min_set_sortable_compat x1 x2 x3 x4 x5) as h2 end.
  simpl in h2.
  destruct h2 as [h2 h3].
  inversion h2. subst.
  rewrite H0 at 1. 
  unfold sort_set_nat, sort_set.
  match goal with |- list_to_set (min_set_sortable ?x1 ?x2 ?x3 ?x4 ?x5) = _ => pose proof (min_set_sortable_compat x1 x2 x3 x4 x5) as h5 end.
  simpl in h5.
  destruct h5 as [h5 h6].
  inversion h5. destruct H1 as [h7 h8].
  rewrite <- h7 at 1.
  destruct x as [x h9].
  simpl.
  destruct h9 as [h9 h10].
  simpl in h7.
  destruct h9 as [h11 h12]. 
  pose proof h11 as h11'. rewrite h1 in h11'.
  apply singleton_inc in h11'.
  destruct h11'; subst.
  symmetry in h7. apply empty_set_nil in h7.
  simpl in h5, h7, h8, h6.
  destruct h12 as [h12 h13].
  destruct h13 as [h13].
  destruct h13 as [h13 h14].
  clear h8 h6 h5 h7 H0 H.
  assert (h15:Inn (seg 1) 0).
    constructor; auto.
  assert (h16:Inn (Im (full_sig (seg 1)) p) (p (exist _ _ h15))).
    apply Im_intro with (exist _ _ h15).
    constructor.
    reflexivity.
  rewrite h14 in h16.
  inversion h16. subst.
  destruct x. contradiction.
  reflexivity.
Qed.


Lemma nat_sig_fun_injector_one :
  forall {U:Type} (p:sig_set (seg_one 1)->U),
    nat_sig_fun_injector p (finite_seg_one 1) =
    seg_one 1.
  intros U p.
  assert (h1:seg_one 1 = Singleton 1).
    apply Extensionality_Ensembles; red; split; red; intros x h1.
    destruct h1 as [h1].
    assert (h2:x = 1). omega.
    subst.
    constructor.
  destruct h1.
  constructor. auto with arith.
  rewrite h1.
  unfold nat_sig_fun_injector.
  unfold can_set_nat_choice.
  match goal with |- list_to_set (min_set_sortable ?x1 ?x2 ?x3 ?x4 ?x5) = _ => pose proof (min_set_sortable_compat x1 x2 x3 x4 x5) as h2 end.
  simpl in h2.
  destruct h2 as [h2 h3].
  inversion h2. subst.
  rewrite H0 at 1.
  unfold sort_set_nat, sort_set.
  match goal with |- list_to_set (min_set_sortable ?x1 ?x2 ?x3 ?x4 ?x5) = _ => pose proof (min_set_sortable_compat x1 x2 x3 x4 x5) as h5 end.
  simpl in h5.
  destruct h5 as [h5 h6].
  inversion h5. destruct H1 as [h7 h8].
  rewrite <- h7 at 1.
  destruct x as [x h9].
  simpl.
  destruct h9 as [h9 h10].
  simpl in h7.
  destruct h9 as [h11 h12]. 
  pose proof h11 as h11'. rewrite h1 in h11'.
  apply singleton_inc in h11'.
  destruct h11'; subst.
  symmetry in h7. apply empty_set_nil in h7.
  simpl in h5, h7, h8, h6.
  destruct h12 as [h12 h13].
  destruct h13 as [h13].
  destruct h13 as [h13 h14].
  clear h8 h6 h5 h7 H0 H.
  assert (h15:Inn (seg_one 1) 1).
    constructor; auto.
  assert (h16:Inn (Im (full_sig (seg_one 1)) p)(p (exist _ _ h15))).
    apply Im_intro with (exist _ _ h15).
    constructor.
    reflexivity.
  rewrite h14 in h16.
  inversion h16. subst.
  destruct x. contradiction.
  reflexivity.
Qed.


Definition segT_injector {U:Type} {n:nat} (f:n ---> U) :
  Ensemble (n @).
  pose (fun_st_conv f) as f'.
  pose (nat_sig_fun_injector f' (finite_seg _)) as S.
  pose proof (incl_nat_sig_fun_injector f' (finite_seg _)) as h1.
  pose (im_proj2_sig _ h1) as S'.
  refine (ens_ss_conv S').
Defined.


Lemma incl_segT_injector :
  forall {U:Type} {n:nat} (f:n ---> U),
    Included (segT_injector f) (Full_set _).
  intros; constructor.
Qed.


Lemma segT_injector_compat :
  forall {U:Type} {n:nat} (f:n ---> U),
    let S := segT_injector f in 
    Im (Full_set (n @)) f =
    Im S f /\
    injective (restriction_fun f S).
  intros U n f S.
  pose (segT_injector f) as f'.
  unfold segT_injector in f'.
  pose proof (in_nat_sig_fun_injector (fun_st_conv f) (finite_seg n)) as h1.
  destruct h1 as [h1].
  destruct h1 as [h1 [h2 h3]].
  destruct h3 as [h3].
  destruct h3 as [h3 h4].
  assert (h1 = h3). apply proof_irrelevance. subst.
  split.
  apply Extensionality_Ensembles; red; split; red; intros x h5.
  destruct h5 as [x h5 ? h6]. subst. clear h5.
  destruct x as [m h5]. 
  assert (h6:Inn  (Im (full_sig (seg n)) (fun_st_conv f)) (f (exist (fun a => a < n) _ h5))).
    assert (h7:Inn (seg n) m).
      constructor; auto.
    apply Im_intro with (exist _ _ h7). constructor.
    unfold fun_st_conv.
    f_equal. apply proj1_sig_injective; auto.
  unfold S, segT_injector, ens_ss_conv.
  rewrite im_im.
  unfold im_proj2_sig. rewrite im_im.
  simpl.
  let P := type of h4 in match P with _ = ?x => pose x as A end.
  match goal with |- Inn ?S _ => pose S as B end.
  assert (h7:A = B).
    unfold A, B.
    apply im_ext_in.
    intros x h8.
    unfold restriction_sig, fun_st_conv.
    f_equal.
    apply proj1_sig_injective; auto.
  fold B.
  rewrite <- h7.
  rewrite h4 in h6.
  assumption.

  destruct h5 as [x h5]. subst.
  apply Im_intro with x. constructor.
  reflexivity.

  red. intros x y h1.
  red in h2.
  unfold S, segT_injector, ens_ss_conv in x, y.
  unfold im_proj2_sig in x, y.
  destruct x as [x h5], y as [y h6].
  apply proj1_sig_injective; simpl.
  unfold restriction_fun in h1.
  simpl in h1.
  rewrite im_im in h5, h6.
  destruct h5 as [x h5], h6 as [y h6]; subst.
  apply proj1_sig_injective; simpl. simpl in h1.
  specialize (h2 x y).
  unfold restriction_sig, fun_sot_conv in h2. simpl in h2.
  assert (h7:h3 (proj1_sig x) (proj2_sig x)=  incl_nat_sig_fun_injector (fun_st_conv f)
                (finite_seg n) (proj1_sig x) 
                (proj2_sig x)).
    apply proof_irrelevance.

  assert (h8:h3 (proj1_sig y) (proj2_sig y)=  incl_nat_sig_fun_injector (fun_st_conv f)
                (finite_seg n) (proj1_sig y) 
                (proj2_sig y)).
    apply proof_irrelevance.
  rewrite <- h7, <- h8 in h1 .
  specialize (h2 h1).
  subst.
  reflexivity.
Qed.


Lemma finite_segT_injector :
  forall {U:Type} {n:nat} (f:n ---> U),
    Finite (segT_injector f).
  intros U n f.
  unfold segT_injector, ens_ss_conv, im_proj2_sig.
  rewrite im_im. simpl.
  eapply finite_image.
  pose proof (finite_nat_sig_fun_injector (fun_st_conv f) (finite_seg n)) as h1.
  apply (iff1 (finite_full_sig_iff
                 (nat_sig_fun_injector (fun_st_conv f) (finite_seg n))) h1).
Qed.


Lemma segT_injector_zero :
  forall {U:Type} (p:1 @->U),
    segT_injector p =
    Singleton st01.
  intros U p.
  pose proof (segT_injector_compat p) as h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct x as [x h3].
  assert (h4:x = 0). omega.
  subst.
  unfold st01.
  assert (h4: h3 = st01_pf). apply proof_irrelevance. 
  subst.
  constructor.
  destruct x as [x h3].
  assert (h4: x = 0). omega.
  subst.
  simpl in h1.
  unfold segT_injector, ens_ss_conv, im_proj2_sig.
  rewrite im_im.
  simpl.
  pose proof (nat_sig_fun_injector_zero (fun_st_conv p)) as h4.
  assert (h5:Inn (seg 1) 0).
    constructor. auto with arith.
  rewrite <- h4 in h5.
  apply Im_intro with (exist _ _ h5).
  constructor.
  apply proj1_sig_injective; simpl.
  reflexivity.
Qed.


Definition seg_one_t_injector {U:Type} {n:nat} (f:seg_one_t n->U) :
  Ensemble (seg_one_t n).
  pose (fun_sot_conv f) as f'.
  pose (nat_sig_fun_injector f' (finite_seg_one n)) as S.
  pose proof (incl_nat_sig_fun_injector f' (finite_seg_one n)) as h1.
  pose (im_proj2_sig _ h1) as S'.
  refine (ens_sso_conv S').
Defined.

Lemma incl_seg_one_t_injector :
  forall {U:Type} {n:nat} (f:seg_one_t n -> U),
    Included (seg_one_t_injector f) (Full_set _).
  intros; constructor.
Qed.


Lemma seg_one_t_injector_compat :
  forall {U:Type} {n:nat} (f:seg_one_t n->U),
    let S := seg_one_t_injector f in 
    Im (Full_set (seg_one_t n)) f =
    Im S f /\
    injective (restriction_fun f S).
  intros U n f S.
  pose (seg_one_t_injector f) as f'.
  unfold seg_one_t_injector in f'.
  pose proof (in_nat_sig_fun_injector (fun_sot_conv f) (finite_seg_one n)) as h1.
  destruct h1 as [h1].
  destruct h1 as [h1 [h2 h3]].
  destruct h3 as [h3].
  destruct h3 as [h3 h4].
  assert (h1 = h3). apply proof_irrelevance. subst.
  split.
  apply Extensionality_Ensembles; red; split; red; intros x h5.
  destruct h5 as [x h5 ? h6]. subst. clear h5.
  destruct x as [m h5]. 
  assert (h6:Inn  (Im (full_sig (seg_one n)) (fun_sot_conv f)) (f (exist _ _ h5))).
    assert (h7:Inn (seg_one n) m).
      constructor; auto.
    apply Im_intro with (exist _ _ h7). constructor.
    unfold fun_sot_conv.
    f_equal. apply proj1_sig_injective; auto.
  unfold S, seg_one_t_injector, ens_sso_conv.
  rewrite im_im.
  unfold im_proj2_sig. rewrite im_im.
  simpl.
  let P := type of h4 in match P with _ = ?C =>
                                      match goal with |- Inn ?D _ =>
                                                      assert (h7:C = D) end end.
    apply im_ext_in.
    intros x h8.
    unfold restriction_sig, fun_sot_conv. 
    f_equal.
    apply proj1_sig_injective; auto.
  rewrite <- h7.
  rewrite h4 in h6.
  assumption.

  destruct h5 as [x h5]. subst.
  apply Im_intro with x. constructor.
  reflexivity.

  red. intros x y h1.
  red in h2.
  unfold S, seg_one_t_injector, ens_sso_conv in x, y.
  unfold im_proj2_sig in x, y.
  destruct x as [x h5], y as [y h6].
  apply proj1_sig_injective; simpl.
  unfold restriction_fun in h1.
  simpl in h1.
  rewrite im_im in h5, h6.
  destruct h5 as [x h5], h6 as [y h6]; subst.
  apply proj1_sig_injective; simpl. simpl in h1.
  specialize (h2 x y).
  unfold restriction_sig, fun_sot_conv in h2. simpl in h2.
  hfirst h2. 
  match goal with |- ?x = _ => let Q := type of h1 in match Q with ?y = _ => assert (h7:x = y) end end.
  f_equal. f_equal. apply proof_irrelevance.
  match goal with |- _ = ?x => let Q := type of h1 in match Q with _ = ?y => assert (h8:x = y) end end.
      f_equal. f_equal. apply proof_irrelevance.
  rewrite h7, h8; auto.
  specialize (h2 hf).
  subst.
  reflexivity.
Qed.

Lemma finite_seg_one_t_injector :
  forall {U:Type} {n:nat} (f:seg_one_t n->U),
    Finite (seg_one_t_injector f).
  intros U n f.
  unfold seg_one_t_injector, ens_sso_conv, im_proj2_sig.
  rewrite im_im. simpl.
  eapply finite_image.
  pose proof (finite_nat_sig_fun_injector (fun_sot_conv f) (finite_seg_one n)) as h1.
  apply (iff1 (finite_full_sig_iff
                 (nat_sig_fun_injector (fun_sot_conv f) (finite_seg_one n))) h1).
Qed.


Lemma seg_one_t_injector_one :
  forall {U:Type} (p:seg_one_t 1->U),
    seg_one_t_injector p =
    Singleton sot_1.
  intros U p.
  pose proof (seg_one_t_injector_compat p) as h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct x as [x h3].
  assert (h4:x = 1). omega.
  subst.
  unfold sot_1.
  assert (h4: h3 = conj (le_n 1) (le_n 1)). apply proof_irrelevance.
  subst.
  constructor.
  destruct x as [x h3].
  assert (h4: x = 1). omega.
  subst.
  simpl in h1.
  unfold seg_one_t_injector, ens_sso_conv, im_proj2_sig.
  rewrite im_im.
  simpl.
  pose proof (nat_sig_fun_injector_one (fun_sot_conv p)) as h4.
  assert (h5:Inn (seg_one 1) 1).
    constructor. auto with arith.
  rewrite <- h4 in h5.
  apply Im_intro with (exist _ _ h5).
  constructor.
  apply proj1_sig_injective; simpl.
  reflexivity.
Qed.


Lemma seg_one_t_2_injector_eq_pos_evens_fun_unite_finite_odds :
  forall {U:Type} (f:seg_one_t 2->U),
    f sot12 = f sot22 ->
    seg_one_t_injector f = Singleton sot12 \/
    seg_one_t_injector f = Singleton sot22.
  intros U f h1.
  pose proof (seg_one_t_injector_compat f) as h2.
  simpl in h2.
  destruct h2 as [h2 h3].
  assert (h4 : Im (Full_set (seg_one_t 2))f  = Couple (f sot12) (f sot22)).
  rewrite full_set_seg_one_t_2.
  rewrite im_couple.
  reflexivity.
  rewrite h2 in h4.
  rewrite h1 in h4.
  rewrite couple_singleton in h4.
  pose proof (f_equal card_fun1 h4) as h5.
  rewrite card_fun1_sing in h5.
  rewrite <- h2 in h5.
  pose proof (card_fun1_compat (Im (Full_set (seg_one_t 2)) f)) as ha.
  simpl in ha. destruct ha as [ha h7]. clear h7.
  assert (h6:  Finite (Im (Full_set (seg_one_t 2)) f)).
    apply finite_image. apply finite_full_set_seg_one_t.
  specialize (ha h6).
  pose proof (card_fun1_compat (seg_one_t_injector f)) as h6'.
  simpl in h6'. destruct h6' as [h7 h8]. clear h8.
  specialize (h7 (finite_seg_one_t_injector f)).
  rewrite h5 in ha.
  pose proof (injective_preserves_card_fun1 _ h3  (Full_set (sig_set (seg_one_t_injector f)))) as h9.
  assert (h8 :  Finite (Full_set (sig_set (seg_one_t_injector f)))).
    rewrite Finite_FiniteT_iff.
    apply Finite_ens_type.
    apply finite_seg_one_t_injector.
  specialize (h9 h8).
  assert (h11 : Im (Full_set (sig_set (seg_one_t_injector f)))
                   (restriction_fun f (seg_one_t_injector f)) =
                Im (seg_one_t_injector f) f).
    apply Extensionality_Ensembles; red; split; red; intros x h11.
    destruct h11 as [x h11].
    subst.
    unfold restriction_fun.
    apply Im_intro with (proj1_sig x). apply proj2_sig.
    reflexivity.
    destruct h11 as [x h11]. subst.
    apply Im_intro with (exist _ _ h11).
    constructor.
    unfold restriction_fun. simpl. reflexivity.
  pose proof h9 as h9'.
  rewrite h11 in h9'.
  rewrite h2 in h5.
  rewrite  h5 in h9' at 1.
  pose proof (card_fun1_full_sig (seg_one_t_injector f) (finite_seg_one_t_injector _)) as h12.
  unfold full_sig in h12.
  rewrite h9' in h12.
  symmetry in h12. apply card_fun1_1 in h12.
  destruct h12 as [x h12].
  destruct x as [x h13].
  destruct h13 as [h13 h14].
  pose proof h13 as h13'. apply le_lt_eq_dec in h13'.
  destruct h13' as [h15 | h15].
  assert (h16: x = 2). auto with arith.
  subst.
  right.
  rewrite h12.
  f_equal. apply proj1_sig_injective; auto.
  subst.
  left.
  rewrite h12.
  f_equal. apply proj1_sig_injective; auto.
Qed.


End Injectors.



(*In this section I develop two ways to get constructive 
description on finite and countable types without depending on
an indefinite description.  I'm sure the finite case fits the 
bill.  It remains to be seen if any of the prerequisites for this
use indefinite description.*)
Section CountableDefiniteDescription.

Lemma bij_seg_inh_iff :
  forall {T:Type} (A:Ensemble T),
    Finite A ->
    forall (f:sig_set (seg (card_fun1 A)) ->
              sig_set A),
      bijective f ->
      forall (P:sig_set A->Prop),
        (Inhabited [x:sig_set A | P x] <->
         Inhabited [n : nat
                   | exists pf : Inn (seg (card_fun1 A)) n,
            P (f (exist _ _ pf))]).
intros T A h1 f h2 P.
destruct h2 as [h2a h2b].
pose proof (card_fun1_full_sig A h1) as h3.
symmetry in h3.
pose proof (f_equal seg h3) as h4.
pose (transfer_fun (sig_set_eq _ _ h4) f) as f'.
pose (fun x => exist _ _ (Full_intro _ (f' x))) as f''.

assert (h5:bijective f'').
  unfold f''.
  red. split.
  red. intros x y h6.
  apply exist_injective in h6.
  unfold f' in h6.
  rewrite <- (transfer_undoes_transfer_r
             (sig_set_eq (seg (card_fun1 A))
                         (seg (card_fun1 (full_sig A))) h4)
             x) in h6.
  rewrite <- (transfer_undoes_transfer_r
             (sig_set_eq (seg (card_fun1 A))
                         (seg (card_fun1 (full_sig A))) h4) y) in h6.
  do 2 rewrite transfer_fun_compat in h6.
  do 2 rewrite transfer_r_sig_set_eq in h6 at 1.
  apply h2a in h6.
  apply exist_injective in h6.
  apply proj1_sig_injective in h6.
  assumption.
  red. intro y.
  destruct y as [y h8].
  pose proof (h2b y) as h9.
  destruct h9 as [x h9]. subst.
  exists (transfer (sig_set_eq _ _ h4) x).
  apply proj1_sig_injective. simpl.
  pose proof (transfer_sig_set_eq _ _ h4
                                  (sig_set_eq (seg (card_fun1 A))
                                              (seg (card_fun1 (full_sig A))) h4) x) as h9.
  rewrite h9 at 1.
  unfold f'.
  rewrite transfer_fun_eq; f_equal.
  rewrite (transfer_r_sig_set_eq _ _ h4 (sig_set_eq (seg (card_fun1 A))
           (seg (card_fun1 (full_sig A))) h4)) at 1.
  simpl.
  apply proj1_sig_injective; auto.
pose proof (bij_seg_inh_full_iff f'' h5 (fun x:sig_set A=>P x)) as h6.
rewrite h6 at 1.
split.
intro h7.
destruct h7 as [a h7].
destruct h7 as [h7].
apply Inhabited_intro with a.
constructor.
destruct h7 as [h7 h8].
unfold full_sig in h4.
pose proof h7 as h7'.
rewrite  <- h4 in h7'.
exists h7'.
unfold f'' in h8. simpl in h8.
unfold f' in h8.
rewrite transfer_fun_eq in h8.
erewrite f_equal.
apply h8.
rewrite (transfer_r_sig_set_eq _ _ h4).
repeat f_equal; apply proof_irrelevance.

intro h7.
destruct h7 as [n h7].
destruct h7 as [h7].
apply Inhabited_intro with n. constructor.
destruct h7 as [h7 h8].
pose proof h7 as h7'.
rewrite h4 in h7'.
exists h7'.
unfold f'', f'.
simpl.
pose proof (transfer_undoes_transfer_r
              (sig_set_eq (seg (card_fun1 A))
                          (seg (card_fun1 (full_sig A))) h4)
              (exist _ _ h7')) as h9.
rewrite <- h9 at 1.
rewrite transfer_fun_compat.
rewrite (transfer_r_sig_set_eq _ _ h4 (sig_set_eq (seg (card_fun1 A))
                 (seg (card_fun1 (full_sig A))) h4)).
simpl.

match goal with |- P (f ?x) =>
                let D := type of h8 in
                match D with P (f ?y) =>
                             assert (h10: x = y)
                end
end.
  apply proj1_sig_injective; auto.
rewrite  h10.
assumption.
Grab Existential Variables.
f_equal. assumption.
f_equal. assumption.
Qed.


Lemma bij_seg_finite_iff :
  forall {T:Type} (A:Ensemble T),
    Finite A ->
    forall (f:sig_set (seg (card_fun1 A)) ->
              sig_set A),
      bijective f ->
      forall (P:sig_set A->Prop),
        (Finite [x:sig_set A | P x] <->
         Finite [n : nat
                   | exists pf : Inn (seg (card_fun1 A)) n,
            P (f (exist _ _ pf))]).
intros T A h1 f h2 P.
destruct h2 as [h2a h2b].
pose proof (card_fun1_full_sig A h1) as h3.
symmetry in h3.
pose proof (f_equal seg h3) as h4.
pose (transfer_fun (sig_set_eq _ _ h4) f) as f'.
pose (fun x => exist _ _ (Full_intro _ (f' x))) as f''.

assert (h5:bijective f'').
  unfold f''.
  red. split.
  red. intros x y h6.
  apply exist_injective in h6.
  unfold f' in h6.
  rewrite <- (transfer_undoes_transfer_r
             (sig_set_eq (seg (card_fun1 A))
                         (seg (card_fun1 (full_sig A))) h4)x) in h6.
  rewrite <- (transfer_undoes_transfer_r
                (sig_set_eq (seg (card_fun1 A))
                            (seg (card_fun1 (full_sig A))) h4) y) in h6.
  do 2 rewrite transfer_fun_compat in h6.
  do 2 rewrite transfer_r_sig_set_eq in h6 at 1.
  apply h2a in h6.
  apply exist_injective in h6.
  apply proj1_sig_injective in h6.
  assumption.
  red. intro y.
  destruct y as [y h8].
  pose proof (h2b y) as h9.
  destruct h9 as [x h9]. subst.
  exists (transfer (sig_set_eq _ _ h4) x).
  apply proj1_sig_injective. simpl.
  pose proof (transfer_sig_set_eq _ _ h4
                                  (sig_set_eq (seg (card_fun1 A))
                                              (seg (card_fun1 (full_sig A))) h4) x) as h9.
  rewrite h9 at 1.
  unfold f'.
  rewrite transfer_fun_eq,
  (transfer_r_sig_set_eq _ _ h4
                         (sig_set_eq (seg (card_fun1 A))
                                     (seg (card_fun1 (full_sig A))) h4)).
  f_equal.
  apply proj1_sig_injective; auto.
pose proof (bij_seg_finite_full_iff f'' h5 (fun x:sig_set A=>P x)) as h6.
split.
intro h0.
rewrite h6 in h0.
eapply Finite_downward_closed.
apply h0.
red.
intros x h7.
destruct h7 as [h7].
constructor.
destruct h7 as [h7 h8].
pose proof h7 as h7'.
rewrite h4 in h7'.
exists h7'.
unfold f'', f'.
simpl.
rewrite transfer_fun_eq, (transfer_r_sig_set_eq _ _ h4).
simpl.
erewrite f_equal.
apply h8.
repeat f_equal.
apply proof_irrelevance.

intro h7.
rewrite h6.
eapply Finite_downward_closed.
apply h7.
red; intros x h8.
destruct h8 as [h8].
constructor.
destruct h8 as [h8 h9].
pose proof h8 as h7'.
unfold full_sig in h4.
rewrite <- h4 in h7' at 1.
exists h7'.
unfold f'' in h9. simpl in h9.
unfold f' in h9.
rewrite transfer_fun_eq, (transfer_r_sig_set_eq _ _ h4) in h9.
simpl in h9.
erewrite f_equal.
apply h9.
repeat f_equal; apply proof_irrelevance.

Grab Existential Variables.
f_equal; auto.
f_equal; auto.
Qed.


Lemma finite_constructive_definite_description :
  forall {T:Type} (A:Ensemble T)
         (pf:Finite A) (f:sig_set (seg (card_fun1 A)) ->
                          sig_set A),
         bijective f ->
  forall (P:sig_set A->Prop),
    (exists x:sig_set A, P x) ->
    {x:sig_set A| P x}.
intros T A h1 f h2 P h3.
pose proof (ex_inh _  h3) as h4.
pose proof (bij_seg_inh_iff _ h1 f h2 ) as h5.
rewrite h5 in h4.
pose  [n : nat
         | exists pf : Inn (seg (card_fun1 A)) n,
             P (f (exist (Inn (seg (card_fun1 A))) n pf))] as S.
pose proof (bij_seg_finite_iff _ h1 f h2) as h6.
assert (ha:Finite [x : sig_set A | P x]).
  eapply Finite_downward_closed.
  rewrite <- finite_full_sig_iff.
  assumption.
  red; intros; constructor.
specialize (h6 P).
rewrite h6 in ha.
pose (max_set_nat _ ha h4).
pose proof (max_set_nat_compat _ ha h4) as h10.
simpl in h10.
destruct h10 as [h10 h11].
destruct h10 as [h10].
pose proof (destruct_exists_pf_set _ _ _ h10) as h12.
simpl in h12.
refine (exist _ _ h12).
Defined.

    
Lemma finite_constructive_definite_description_alt :
  forall {T:Type} (A:Ensemble T)
         (pf:Finite A) (f:(card_fun1 A) @->
                          sig_set A),
         bijective f ->
  forall (P:sig_set A->Prop),
    (exists x:sig_set A, P x) ->
    {x:sig_set A| P x}.
  intros T A h1 f h2 P h3.
  eapply finite_constructive_definite_description; auto.
  rewrite  bij_st_conv_iff in h2 at 1.
  apply h2.
Defined.


(*This is to be used in situations that don't allow propositional 
case analysis, like in a function.*)
Lemma surj_fin_inv :
  forall {T U:Type} {A:Ensemble T}
    (f:sig_set (seg (card_fun1 A)) ->
       sig_set A),
    Finite A ->
    bijective f ->
    forall (g:sig_set A->U),
      surjective g ->
      forall (y:U),
        Inn (Im (Full_set _) g) y ->
        {x:sig_set A | g x = y}.
  intros T U A f h1 h0 g h2 y h3.
  apply Im_inv in h3.
  pose proof (finite_constructive_definite_description A h1 f h0 _ h3) as h4.
  simpl in h4.
  destruct h4 as [x h4]. destruct h4 as [? h4].
  refine (exist _ _ h4).
Defined.


Lemma bij_seg_inh_iff' : 
  forall {T:Type} (A:Ensemble T),
    Finite A ->
    forall (f:sig_set (seg (card_fun1 A)) -> 
              sig_set A),
      bijective f ->
      forall (P:T->Prop),
        (Inhabited [x:sig_set A | P (proj1_sig x)] <-> 
         Inhabited [n : nat
                   | exists pf : Inn (seg (card_fun1 A)) n,
            P
              (proj1_sig
                 (f (exist _ _ pf)))]).
intros T A h1 f h2 P.
destruct h2 as [h2a h2b].
pose proof (card_fun1_full_sig A h1) as h3.
symmetry in h3.
pose proof (f_equal seg h3) as h4.
pose (transfer_fun (sig_set_eq _ _ h4) f) as f'.
pose (fun x => exist _ _ (Full_intro _ (f' x))) as f''.
 
assert (h5:bijective f'').
  unfold f''.
  red. split.
  red. intros x y h6.
  apply exist_injective in h6.
  unfold f' in h6.  
  rewrite <- (transfer_undoes_transfer_r    
             (sig_set_eq (seg (card_fun1 A))
                         (seg (card_fun1 (full_sig A))) h4)
             x) in h6.
  rewrite <- (transfer_undoes_transfer_r    
             (sig_set_eq (seg (card_fun1 A))
                         (seg (card_fun1 (full_sig A))) h4) y) in h6.
  do 2 rewrite transfer_fun_compat in h6.
  do 2 rewrite transfer_r_sig_set_eq in h6 at 1.
  apply h2a in h6.
  apply exist_injective in h6.
  apply proj1_sig_injective in h6.
  assumption.
  red. intro y.
  destruct y as [y h8]. 
  pose proof (h2b y) as h9.  
  destruct h9 as [x h9]. subst.
  exists (transfer (sig_set_eq _ _ h4) x).
  apply proj1_sig_injective. simpl.
  pose proof (transfer_sig_set_eq _ _ h4 
                                  (sig_set_eq (seg (card_fun1 A))
                                              (seg (card_fun1 (full_sig A))) h4) x) as h9.
  rewrite h9 at 1.
  unfold f'.
  rewrite transfer_fun_eq, 
  (transfer_r_sig_set_eq _ _ h4). 
  simpl.
  f_equal.
  apply proj1_sig_injective; auto.
pose proof (bij_seg_inh_full_iff f'' h5 (fun x:sig_set A=>P (proj1_sig x))) as h6.
rewrite h6 at 1.
split.
intro h7.
destruct h7 as [a h7].
destruct h7 as [h7].
apply Inhabited_intro with a.
constructor.
destruct h7 as [h7 h8].
unfold full_sig in h4.
pose proof h7 as h7'. 
rewrite  <- h4 in h7'.
exists h7'.
unfold f'' in h8. simpl in h8.
unfold f' in h8.
rewrite transfer_fun_eq,
(transfer_r_sig_set_eq _ _ h4) in h8; simpl in h8.
erewrite f_equal.
apply h8.
repeat f_equal; apply proof_irrelevance.

intro h7.
destruct h7 as [n h7].
destruct h7 as [h7].
apply Inhabited_intro with n. constructor.
destruct h7 as [h7 h8].
pose proof h7 as h7'.
rewrite h4 in h7'.
exists h7'.
unfold f'', f'.
simpl.
pose proof (transfer_undoes_transfer_r
              (sig_set_eq (seg (card_fun1 A))
                          (seg (card_fun1 (full_sig A))) h4) (exist _ _ h7')) as h9.
rewrite <- h9 at 1.
rewrite transfer_fun_compat.
rewrite (transfer_r_sig_set_eq _ _ h4 (sig_set_eq (seg (card_fun1 A))
                 (seg (card_fun1 (full_sig A))) h4)).
simpl.
match goal with |- P (proj1_sig (f ?x)) =>
                let D := type of h8 in
                match D with P (proj1_sig (f ?y)) =>
                             assert (h10: x = y)
                end
end.
f_equal. apply proof_irrelevance.
rewrite h10.
assumption.
Grab Existential Variables.
f_equal. assumption.
f_equal. assumption.
Qed.


Lemma bij_seg_finite_iff' :
  forall {T:Type} (A:Ensemble T),
    Finite A ->
    forall (f:sig_set (seg (card_fun1 A)) -> 
              sig_set A),
      bijective f ->
    forall (P:T->Prop),
        (Finite [x:sig_set A | P (proj1_sig x)] <-> 
         Finite [n : nat
                   | exists pf : Inn (seg (card_fun1 A)) n,
            P
              (proj1_sig
                 (f (exist _ _ pf)))]).
intros T A h1 f h2 P.
destruct h2 as [h2a h2b].
pose proof (card_fun1_full_sig A h1) as h3.
symmetry in h3.
pose proof (f_equal seg h3) as h4.
pose (transfer_fun (sig_set_eq _ _ h4) f) as f'.
pose (fun x => exist _ _ (Full_intro _ (f' x))) as f''.

assert (h5:bijective f'').
  unfold f''.
  red. split.
  red. intros x y h6.
  apply exist_injective in h6.
  unfold f' in h6.  
  rewrite <- (transfer_undoes_transfer_r    
             (sig_set_eq (seg (card_fun1 A))
                         (seg (card_fun1 (full_sig A))) h4) x) in h6.
  rewrite <- (transfer_undoes_transfer_r    
             (sig_set_eq (seg (card_fun1 A))
                         (seg (card_fun1 (full_sig A))) h4)
             y) in h6.
  do 2 rewrite transfer_fun_compat in h6.
  do 2 rewrite transfer_r_sig_set_eq in h6 at 1.
  apply h2a in h6.
  apply exist_injective in h6.
  apply proj1_sig_injective in h6.
  assumption.
  red. intro y.
  destruct y as [y h8]. 
  pose proof (h2b y) as h9.  
  destruct h9 as [x h9]. subst.
  exists (transfer (sig_set_eq _ _ h4) x).
  apply proj1_sig_injective. simpl.
  pose proof (transfer_sig_set_eq _ _ h4 
                                  (sig_set_eq (seg (card_fun1 A))
                                              (seg (card_fun1 (full_sig A))) h4) x) as h9.
  rewrite h9 at 1.
  unfold f'.
  match goal with |- transfer_fun _ _ ?x = _ =>
                  pose proof ((transfer_undoes_transfer_r
                                 (sig_set_eq (seg (card_fun1 A))
                                             (seg (card_fun1 (full_sig A))) h4)) x) as h10 end.
  rewrite <- h10 at 1.
  rewrite transfer_fun_compat.
  rewrite (transfer_r_sig_set_eq _ _ h4 (sig_set_eq (seg (card_fun1 A))
           (seg (card_fun1 (full_sig A))) h4)) at 1.
  simpl.
  f_equal.
  apply proj1_sig_injective; auto.
pose proof (bij_seg_finite_full_iff f'' h5 (fun x:sig_set A=>P (proj1_sig x))) as h6.
split.
intro h0.
rewrite h6 in h0.
eapply Finite_downward_closed.
apply h0.
red.
intros x h7.
destruct h7 as [h7].
constructor.
destruct h7 as [h7 h8].
pose proof h7 as h7'.
rewrite h4 in h7'.
exists h7'.
unfold f'', f'.
simpl.
match goal with |- P (proj1_sig (transfer_fun _ _ ?x)) =>
                pose proof (transfer_undoes_transfer_r
                              (sig_set_eq (seg (card_fun1 A))
                                          (seg (card_fun1 (full_sig A))) h4) x) as h9 end.
rewrite <- h9.
rewrite transfer_fun_compat.
rewrite (transfer_r_sig_set_eq _ _ h4).
simpl.
match goal with |- P (proj1_sig (f ?x)) =>
                let D := type of h8 in
                match D with P (proj1_sig (f ?y)) =>
                             assert (h10: x = y)
                end
end.
  apply proj1_sig_injective; auto.
rewrite <- h10 in h8.
assumption.

intro h7.
rewrite h6.
eapply Finite_downward_closed.
apply h7.
red; intros x h8.
destruct h8 as [h8].
constructor.
destruct h8 as [h8 h9].
pose proof h8 as h7'. 
unfold full_sig in h4.
rewrite <- h4 in h7' at 1.
exists h7'.
unfold f'' in h9. simpl in h9.
unfold f' in h9.
let C := type of h9 in match C with
                           P (proj1_sig (transfer_fun _ _ ?c)) =>
                           pose proof (
                                  transfer_undoes_transfer_r
                                    (sig_set_eq (seg (card_fun1 A))
                                                (seg (card_fun1 (full_sig A))) h4)
                                    c)
                             as h9' end.
rewrite <- h9' in h9.
rewrite transfer_fun_compat in h9.
rewrite (transfer_r_sig_set_eq _ _ h4) in h9.
simpl in h9.
match goal with |- P (proj1_sig (f ?x)) =>
                let D := type of h9 in
                match D with P (proj1_sig (f ?y)) =>
                             assert (h10: x = y)
                end
end.
  apply proj1_sig_injective; auto.
rewrite h10.
assumption.
Grab Existential Variables.
f_equal; auto.
f_equal; auto.
Qed.


Lemma finite_constructive_definite_description' :
  forall {T:Type} (A:Ensemble T)
         (pf:Finite A) (f:sig_set (seg (card_fun1 A)) ->
                          sig_set A),
         bijective f ->
  forall (P:T->Prop),
    (exists x:sig_set A, P (proj1_sig x)) ->
    {x:T| P x}.
intros T A h1 f h2 P h3.
pose proof (ex_inh _  h3) as h4.
pose proof (bij_seg_inh_iff' _ h1 f h2 ) as h5.
rewrite h5 in h4.
pose  [n : nat
         | exists pf : Inn (seg (card_fun1 A)) n,
             P
               (proj1_sig
                  (f (exist (Inn (seg (card_fun1 A))) n pf)))] as S.
pose proof (bij_seg_finite_iff' _ h1 f h2) as h6.
assert (ha:Finite [x : sig_set A | P (proj1_sig x)]).
  eapply Finite_downward_closed.
  rewrite <- finite_full_sig_iff.
  assumption.
  red; intros; constructor.
specialize (h6 P).
rewrite h6 in ha.
pose (max_set_nat _ ha h4).
pose proof (max_set_nat_compat _ ha h4) as h10.
simpl in h10.
destruct h10 as [h10 h11].
destruct h10 as [h10].
pose proof (destruct_exists_pf_set _ _ _ h10) as h12.
simpl in h12.
refine (exist _ _ h12).
Defined.



Lemma inh_im_seg_one_iff : 
  forall {T:Type} {n:nat} (p:seg_one_t n->T),
    Inhabited (Im (Full_set _) p) <->
    Inhabited [m:seg_one_t n | Inn (Im (Full_set _) p) (p m)].
intros T n p. split.
intro h1. destruct h1 as [x h1].
destruct h1 as [m h1]. subst.
apply Inhabited_intro with m.
constructor.
apply Im_intro with m. constructor. reflexivity.
intro h1.
destruct h1 as [m h1]. destruct h1 as [h1].
apply Inhabited_intro with (p m).
apply Im_intro with m; auto; constructor.
Qed.


Lemma inh_im_seg_one_iff' : 
  forall {T:Type} {n:nat} (p:seg_one_t n->T),
    Inhabited (Im (Full_set _) p) <->
    Inhabited [m:nat | exists pf:1 <= m <= n,
                                 Inn (Im (Full_set _) p) (p (exist _ _ pf))].
intros T n p.
rewrite inh_im_seg_one_iff.
split.
intro h1. destruct h1 as [m h1]. destruct m as [m h2].
destruct h1 as [h1].
apply Inhabited_intro with m.
constructor. exists h2.
assumption.
intro h1. destruct h1 as [m h1]. destruct h1 as [h1].
destruct h1 as [h1 h2].
apply Inhabited_intro with (exist (fun s => 1 <= s <= n) _ h1).
constructor.
assumption.
Qed.



Lemma finite_index_assoc_fun_finite_aux : 
  forall {T:Type} {n:nat} (p:seg_one_t n->T),
         Finite [m : nat
         | exists pf : 1 <= m <= n,
             Inn (Im (Full_set (seg_one_t n)) p)
               (p (exist (fun m0 : nat => 1 <= m0 <= n) m pf))].
intros T n p.
eapply Finite_downward_closed.
apply (finite_seg_one n).
red; intros m h1. 
destruct h1 as [h1].
destruct h1 as [h1].
constructor; auto.
Qed.

Lemma finite_index_assoc_fun_in_aux : 
  forall {T:Type} {n:nat} (p:seg_one_t n->T) (a:nat),
    Inn [m : nat
         | exists pf : 1 <= m <= n,
                       Inn (Im (Full_set (seg_one_t n)) p)
                                    (p (exist (fun m0 : nat => 1 <= m0 <= n) m pf))] a ->
    1 <= a <= n.
intros T n p a h1.
destruct h1 as [h1].
destruct h1; auto.
Qed.


Definition finite_index_assoc_fun 
           {T U:Type} {n:nat} (p:seg_one_t n->T)
           (q:seg_one_t n->U) :
  let S := seg_one_t_injector p in
  sig_set (Im S p)->U.
  intros S y.
  pose proof (seg_one_t_injector_compat p) as h1. simpl in h1.
  destruct h1 as [h1 h2].
  pose (left_inverse _ h2  (full_sig (seg_one_t_injector p))) as pi.
  refine (q _).
  pose proof (im_full_sig_restriction (seg_one_t_injector p)
                                      p) as h3.
  pose (transfer_fun (sig_set_eq _ _ h3) pi) as pi'.
  refine (proj1_sig (pi' y)).
Defined.


Lemma finite_index_assoc_fun_compat : 
  forall {T U:Type} {n:nat} (p:seg_one_t n->T)
         (q:seg_one_t n->U) (m:seg_one_t n)
         (pf:Inn (seg_one_t_injector p) m),
    finite_index_assoc_fun p q (exist _ _ (Im_intro _ _ _ p m pf _ (eq_refl _)))  = q m.
  intros T U n p q m h1.
  unfold finite_index_assoc_fun.
  destruct (seg_one_t_injector_compat p) as [h2 h3].
  pose proof (left_inverse_compat (restriction_fun p (seg_one_t_injector p)) h3
                                  (full_sig (seg_one_t_injector p)) (exist _ _ h1) (Full_intro _ _)) as h4.
  pose proof (f_equal (@proj1_sig _ _) h4) as h5. simpl in h5.
  rewrite <- h5 at 5.
  f_equal.
  f_equal.
  f_equal.
  match goal with
      |- transfer_fun ?x _ ?z = _ =>
      rewrite <- (transfer_undoes_transfer_r x z) end.
  rewrite transfer_fun_compat.
  f_equal.
  match goal with
      |- transfer_r ?z _ = _ => 
     rewrite (transfer_r_sig_set_eq _ _  (im_full_sig_restriction (seg_one_t_injector p) p) z) end. 
   apply proj1_sig_injective; auto.
Qed.

(*This is if both the given seg_one_T functions have different 
  integer bound variables, yet are equal.*)
Definition finite_index_assoc_fun' 
           {T U:Type} {n n':nat} (p:seg_one_t n->T)
           (q:seg_one_t n'->U) (pf:n = n') :
  let S := seg_one_t_injector p in
  sig_set (Im S p)->U.
  subst.
  refine (finite_index_assoc_fun p q).
Defined.

Lemma finite_index_assoc_fun_compat' : 
  forall {T U:Type} {n n':nat} (p:seg_one_t n->T)
         (q:seg_one_t n'->U) (m:seg_one_t n)
         (pfeq:n = n')
         (pf:Inn (seg_one_t_injector p) m),
    finite_index_assoc_fun' p q pfeq (exist _ _ (Im_intro _ _ _ p m pf _ (eq_refl _)))  = q (s_transfer (seg_one_t_functional _ _ pfeq)  m).
  intros T U n n' p q m h1 h2.
  subst.
  pose proof (proof_irrelevance
                _ 
                (seg_one_t_functional n' n' eq_refl)
                (eq_refl _)) as h3.
  rewrite h3.
  rewrite s_transfer_eq_refl.
  unfold finite_index_assoc_fun', eq_rect_r.
  simpl.
  apply finite_index_assoc_fun_compat.
Qed.


(*Countable*)
Lemma bij_nat_inh_iff :
  forall {T:Type} {A:Ensemble T}
         (f:nat->sig_set A) (P:sig_set A->Prop),
    bijective f ->
    (Inhabited [x:sig_set A | P x] <->
     Inhabited [n : nat | P (f n)]).
intros T A f P h1.
split.
intro h2.
destruct h2 as [x h2]. destruct h2 as [h2].
pose proof h1 as h1'.
apply bijective_impl_invertible in h1'.
pose (proj1_sig (function_inverse _ h1')) as f'.
apply Inhabited_intro with (f' x).
constructor.
unfold f', function_inverse.
destruct constructive_definite_description as [g h3].
simpl.
destruct h3 as [h3 h4].
rewrite h4.
assumption.
intro h2.
destruct h2 as [n h2]. destruct h2 as [h2].
apply Inhabited_intro with (f n).
constructor.
assumption.
Qed.



Lemma bij_nat_set_inh_iff :
  forall {T:Type} {A:Ensemble T} {S:Ensemble nat}
         (f:sig_set S->sig_set A) (P:sig_set A->Prop),
    bijective f ->
    ~Finite S ->
    (Inhabited [x:sig_set A | P x] <->
     Inhabited [n : sig_set S | P (f n)]).
intros T A S f P h1 h2.
split.
intro h3.
destruct h3 as [x ha]. destruct ha as [ha].
pose proof h1 as h1'.
apply bijective_impl_invertible in h1'.
pose (proj1_sig (function_inverse _ h1')) as f'.
apply Inhabited_intro with (f' x).
constructor.
unfold f', function_inverse.
destruct constructive_definite_description as [g h3].
simpl.
destruct h3 as [h3 h4].
rewrite h4.
assumption.
intro hb.
destruct hb as [n hb]. destruct hb as [hb].
apply Inhabited_intro with (f n).
constructor.
assumption.
Qed.


Lemma countable_constructive_definite_description :
  forall {T:Type} {A:Ensemble T}
         (f:nat->sig_set A) (P:sig_set A->Prop),
    bijective f ->
    (exists x:sig_set A, P x) ->
    {x:sig_set A| P x}.
intros T A f P h1 h2.
pose proof wo_nat as h3.
pose proof (ex_inh _  h2) as h4.
rewrite (bij_nat_inh_iff f P h1) in h4. 
pose (min_set _ _ h3 _ h4) as m.
pose proof (min_set_compat _ _ h3 _ h4) as h5. simpl in h5.
destruct h5 as [h5 h6]. destruct h5 as [h5].
refine (exist _ _ h5).
Defined.
    

Lemma countable_set_constructive_definite_description :
  forall {T:Type} {S:Ensemble nat} {A:Ensemble T}
         (f:sig_set S->sig_set A) (P:sig_set A->Prop),
    ~Finite S ->
    bijective f ->
    (exists x:sig_set A, P x) ->
    {x:sig_set A| P x}.
intros T S A f P h1 h2 h3.
pose proof wo_nat as h8.
pose proof (well_order_exist_rel _ h8 (fun a => Inn S a)) as h8'.
pose proof (ex_inh _ h3) as h9.
rewrite (bij_nat_set_inh_iff f P h2 h1) in h9. 
pose (min_set _ _ h8' _ h9) as n. 
pose proof (min_set_compat _ _ h8' _ h9) as hn.
simpl in hn.
destruct hn as [h10 h11].
inversion h10 as [h12].
refine (exist _ _ h12).
Defined.

End CountableDefiniteDescription.


