(* Copyright (C) 2014-2019, Daniel Wyckoff, except for the portions so labeleled 
which I got from Daniel Schepler, in his Zorn's Lemma package.*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)


(* Some of these lemmas are copied and pasted from Daniel Schepler's
  "Zorns Lemma" package.  Such lemmas are notated "Schepler."
 Some tactics come from Benjamen Peirce's "Software Foundations" and are 
marked with "SF."  All others are labeled "Wyckoff". *)

(*Version 0.3 bool2 -- certified -- Coq 8.4pl4*)

Require Import Specif.
Require Import ProofIrrelevance.
Require Import FunctionalExtensionality.
Require Import Description.
Require Import ClassicalChoice.
Require Import EqDec.
Require Import FunctionProperties.
Require Import Equality.

Lemma eq_JMeq :
  forall {T} {x y:T},
    x = y -> x ~= y.
  intros; subst; constructor.
Qed.
    

(*Wyckoff*)
Lemma neq_sym : forall {T:Type} (a b:T), a <> b -> b <> a.
intros ? ? ? h1. intro. subst. contradict h1. apply eq_refl.
Qed.

(*Wyckoff*)
Lemma neq_sym_iff : forall {T:Type} (a b:T), a <> b <-> b <>a.
intros; split; apply neq_sym.
Qed.

(*Wyckoff*)
Lemma eq_sym_iff : forall {T:Type} (a b:T), a = b <-> b = a.
intros; split; auto.
Qed.

(*Wyckoff*)
Lemma eq_iff :
  forall {T:Type} {P:T->Prop} (a b:T),
    a = b -> (P a <-> P b).
  intros; subst. tauto.
Qed.

(*Wyckoff*)
Lemma s_eq_iff :
  forall {T:Set} {P:T->Prop} (a b:T),
    a = b -> (P a <-> P b).
  intros; subst. tauto.
Qed.

(*Wyckoff*)
Lemma P_subst :
  forall {T:Type} {P:T->Prop} {x y:T} (pf:P x),
    x = y ->
    P y.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma eq2 : forall {T:Type} {x y y':T}, 
                  (x = y) -> (x = y') -> (y = y').
intros; subst; auto.
Defined.

(*Wyckoff*)
Lemma eq2' : forall {T:Type} {x x' y:T}, 
                  (x = y) -> (x' = y) -> (x = x').
intros; subst; auto.
Defined.


(*Wyckoff*)
Lemma eq_trans : forall {T:Type} {x y z:T}, 
                  (x = y) -> (y = z) -> (x = z).
intros; subst; auto.
Defined.


(*Wyckoff*)
Lemma eq_trans' : forall {T:Type} {x y z:T}, 
                    (y = x) -> (z = y) -> (z = x).
intros; subst; auto.
Defined.

(*Wyckoff*)
Lemma neq_eq_trans :
  forall {T} {x y z:T},
    x <> y ->
    y = z ->
    x <> z.
  intros; intro; subst; contradict H; auto.
Qed.

(*Wyckoff*)
Lemma eq_neq_trans :
  forall {T} (a b c:T),
    a = b -> b <> c -> a <> c.
  intros; subst; intro; subst; contradict H0; auto.
Qed.

(*Wyckoff*)
Lemma eq_dec_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (a:T)
         (pf:{a = a} + {a <> a}),
    pf = left (eq_refl a).
  intros ? ? ? h1; destruct h1; f_equal.
  apply proof_irrelevance.
  contradict n; auto.
Qed.


(*Wyckoff*)
Lemma eq_dec_neq_r :
  forall {T:Type} (pfdt:type_eq_dec T) (a a':T)
         (pfneq:a <> a') (pf: {a' = a} + {a' <> a}),
    pf = right (neq_sym _ _ pfneq).
  intros T h1 a a' h2 h3.
  destruct h3 as [|h3]; subst.
  contradict h2; auto.
  f_equal.
  apply proof_irrelevance.
Qed.


(*These switch functions are a bit out of place, but have to be
in this file for dependency reasons.*)
(*Wyckoff*)
Definition switch_eq_dec {T:Type} (a b:T) (pf:{a = b} + {a <> b}) :
  {b = a} + {b <> a} :=
  match pf with
    | left pf => left (eq_sym pf)
    | right pf => right (neq_sym _ _ pf)
  end.

(*Wyckoff*)
Lemma switch_eq_dec_eq :
  forall {T:Type} (pf:type_eq_dec T) (a b:T),
    pf a b = switch_eq_dec _ _ (pf b a).
  intros T hdt a b; unfold switch_eq_dec.
  destruct (hdt a b) as [|h1], (hdt b a) as [|h2]; f_equal;
  try apply proof_irrelevance; try contradict h1; try contradict h2; auto.
Qed.


(*Wyckoff*)
Lemma dec_not_l :
  forall {P Q:Prop},
    {P} + {Q} -> ~P ->
    Q.
  tauto.
Qed.

(*Wyckoff*)
Lemma dec_not_r :
  forall {P Q:Prop},
    {P} + {Q} -> ~Q ->
    P.
  tauto.
Qed.

(*Wyckoff*)
Lemma eq_eq_refl :
  forall {T} {x:T} (pf:x = x),
    pf = eq_refl _.
  intros; apply proof_irrelevance.
Qed.



Section utilities.
Variable P : Prop.
Variable T : Type.
Variable Q : T -> Prop.

(*Wyckoff*)
Lemma prop_dis_forall : 
  (forall (x:T), P \/ (Q x)) -> 
  (P \/ (forall (y:T), (Q y))).
intro h1.
assert (h2: P \/ ~P).
  tauto.
case h2.
tauto.
intro h3.
right.
intro y.
assert (h4: P \/ Q y).
  apply h1.
tauto.
Qed.

End utilities.


Section exist_fact.
(*Wyckoff*)
Lemma existTexist : forall (A:Type) (P:A->Prop) (x y:A) (p:(P x)) (q:(P y)),
  existT P x p = existT P y q -> exist P x p = exist P y q.
intros A P x y p q h1.
assert (h2: sig_of_sigT (existT P x p) = sig_of_sigT (existT P y q)).
rewrite h1. reflexivity.
apply h2.
Qed.


(*Wyckoff*)
Lemma exist_injective : forall {U:Type} (P:U->Prop) (x y:U) (p:P x) (q:P y),
  exist _ x p = exist _ y q -> x = y.
intros U P x y p q h1.
pose proof (f_equal (@proj1_sig _ _) h1) as h2.
simpl in h2.
assumption.
Qed.


(*Schepler*)
Lemma proj1_sig_injective: forall {A:Type} (P:A->Prop)
  (a1 a2:{x:A | P x}), proj1_sig a1 = proj1_sig a2 -> a1 = a2.
Proof.
intros.
destruct a1.
destruct a2.
simpl in H.
apply existTexist.
apply subsetT_eq_compat; trivial.
Qed.

(*Schepler*)
Lemma subset_eq_compatT: forall (U:Type) (P:U->Prop) (x y:U)
  (p:P x) (q:P y), x = y -> exist P x p = exist P y q.
Proof.
intros.
destruct H.
f_equal.
apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma proj1_sig_eq_iff :
  forall {T} {P:T->Prop} (x:{t:T | P t}) (y:T),
    proj1_sig x = y <->
    exists pf:P y,
      x = exist _ y pf.
  intros T P x y; split; intro h1; subst.
  exists (proj2_sig _).
  apply proj1_sig_injective; simpl; auto.
  destruct h1; subst; auto.
Qed.

  
(*Schepler*)
Lemma choice_on_dependent_type: forall {A:Type} {B:A->Type}
  (R:forall a:A, B a -> Prop),
  (forall a:A, exists b:B a, R a b) ->
  exists f:(forall a:A, B a), forall a:A, R a (f a).
Proof.
intros.
destruct (choice (fun (a:A) (b:{a:A & B a}) =>
  match b with existT a' b0 => a=a' /\ R a' b0 end))
as [choice_fun].
intro a.
destruct (H a) as [b].
exists (existT (fun a:A => B a) a b).
split; trivial.
assert (f0:forall a:A, {b:B a | R a b}).
intro.
pose proof (H0 a).
destruct (choice_fun a) as [a' b].
destruct H1.
destruct H1.
exists b; trivial.
exists (fun a:A => proj1_sig (f0 a)).
intro.
destruct (f0 a) as [b].
exact r.
Qed.



(*Wyckoff*)
Lemma unfold_sig : forall {T:Type} (P:T->Prop) (x:{t:T | P t}), 
  x = exist _ (proj1_sig x) (proj2_sig x).
intros T P x.
destruct x.
simpl. reflexivity.
Qed.

(*Wyckoff*)
Lemma proj1_sig_of_exist :
  forall {T:Type} {P:T->Prop} (x:T) (pf:P x),
    x = proj1_sig (exist P x pf).
  intros; auto.
Qed.



(*Wyckoff*)  
Lemma propositional_ex_unq :
  forall (P:Prop) (R:P->Prop),
    (exists pf:P,
       R pf) ->
    exists! pf:P,
        R pf.
  intros P R h1.
  destruct h1 as [h2 h3].
  exists h2.
  red; split; auto.
  intros; apply proof_irrelevance.
Qed.

(*Wyckoff*)  
Definition constructive_definite_proof
           {P:Prop} {R:P->Prop}
           (pfe:exists pfp:P, R pfp) : {pfp:P | R pfp} :=
  let x := constructive_definite_description
             _ 
             (propositional_ex_unq _ _ pfe) in
  exist _ (proj1_sig x) (proj2_sig x).


(*Wyckoff*)
Lemma destruct_exists_pf_set_aux : 
  forall {T:Type} (P:T->Prop) (x:T) (Q:P x->Prop),
    (exists pf:P x, Q pf) ->
    P x.
intros T P x Q h1.
destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma destruct_exists_pf_set : 
  forall {T:Type} (P:T->Prop) (x:T) (Q:P x->Prop)
         (pf:exists pf:P x, Q pf),
    Q (destruct_exists_pf_set_aux P x Q pf).
intros T P x Q h1.
destruct h1 as [h1 h2].
pose proof (proof_irrelevance
              _ h1
              (destruct_exists_pf_set_aux
                 P x Q
                 (ex_intro (fun pf : P x => Q pf) h1 h2))) as h3.
rewrite <- h3.
assumption.
Qed.

(*Wyckoff*)
Lemma unq_ex_impl_ex : 
  forall {T:Type} (P:T->Prop),
    (exists! x:T, P x) ->
    exists x:T, P x.
intros T P h1.
destruct h1 as [x h1]. red in h1.
destruct h1 as [h1 h2].
exists x; auto.
Qed.

End exist_fact.


(*Wyckoff*)
Lemma conditional_correspondence : 
  forall (P Q:Prop) (h1:{P}+{~P}) (h2:{Q}+{~Q}),
    (if h1 then true else false) =
    (if h2 then true else false) -> (P<->Q).
intros P Q h1 h2 h3.
assert (h4:forall (P' Q':Prop) (h1:{P'}+{~P'}) (h2:{Q'}+{~Q'}),
             (if h1 then true else false) =
             (if h2 then true else false) -> (P'->Q')).
  intros P' Q' h1' h2' h3'.
  destruct h1' as [h1a | h1b]; destruct h2' as [h2a | h2b].
  auto. discriminate. discriminate.
  tauto.
red. split.
apply (h4 _ _ h1 h2 h3).
symmetry in h3.
apply (h4 _ _ h2 h1 h3). 
Qed.

(*Wyckoff*)
Definition iff1 {P Q:Prop} (pf:P <-> Q) (pfp:P) : Q.
rewrite pf in pfp. refine pfp.
Defined.

(*Wyckoff*)
Definition iff2 {P Q:Prop} (pf:P <-> Q) (pfq:Q) : P.
rewrite <- pf in pfq. refine pfq.
Defined.

(*Wyckoff*)
Definition iffn1 {P Q:Prop} (pf:P <-> Q) (pfp: ~P) : ~Q.
rewrite pf in pfp. refine pfp.
Defined.

(*Wyckoff*)
Definition iffn2 {P Q:Prop} (pf:P <-> Q) (pfq: ~Q) :~ P.
rewrite <- pf in pfq. refine pfq.
Defined.



(*Wyckoff*)
Lemma contrapos : forall (P Q:Prop), (P -> Q) -> (~Q -> ~P).
intros; tauto.
Qed.

(*Wyckoff*)
Definition simpl_prop (P:Prop) : Prop.
simpl P. refine P.
Defined.

(*Wyckoff*)
Definition simpl_pf {P:Prop} (pf:P) : (simpl_prop P).
simpl pf. refine pf.
Defined.

(*Section Tactics.  Top Level.*)

(*Wyckoff*)
Ltac hfirst H :=
  let h := fresh "hf" in
  let P := type of H in
  match P with
      ?hyp -> _ =>
      assert (h:hyp)
  end.

(*Wyckoff*)  
Ltac hfirst_dec H :=
  let h := fresh "hfd" in
      let P := type of H in
      match P with
          ?hyp -> _ => 
          assert (hfd:{hyp} + {~hyp})
      end.

(*Wyckoff*)  
Ltac nhfirst_dec H :=
  let h := fresh "hfd" in
  let P := type of H in
  match P with
      ~ ?hyp -> _ => 
      assert (hfd:{hyp} + {~hyp})
  end.

(*Wyckoff*)
Ltac hneg H :=
  let h := fresh "hn" in
  let hyp := type of H in 
  assert (h:~hyp).

(*Wyckoff*)
Ltac negh H :=
  let h := fresh "nh" in
  let P := type of H in
  match P with
      ~?hyp => assert (h:hyp)
  end.
  
(*Wyckoff*)
Ltac hl H :=
  let h := fresh "hl" in
  let P := type of H in
  match P with
      ?hyp <-> _ =>
      assert (h:hyp)
  end.

(*Wyckoff*)
Ltac hr H :=
  let h := fresh "hr" in
  let P := type of H in
  match P with
      _ <-> ?hyp =>
      assert (h:hyp)
  end.


(*Wyckoff*)
Ltac nhl H :=
  let P := type of H in
  match P with
      ?hyp <-> _ =>
      let h := fresh "hl" in
      assert (h: ~hyp)
  end.

(*Wyckoff*)
Ltac nhr H :=
  let h := fresh "hr" in
  let P := type of H in
  match P with
      _ <-> ?hyp =>
      assert (h: ~hyp)
  end.

(*Wyckoff*)
Ltac ghl :=
  let h := fresh "hl" in
  match goal with
      |- ?hyp <-> _ =>
      assert (h:hyp)
  end.

(*Wyckoff*)
Ltac ghr :=
  let h := fresh "hr" in
  match goal with
      |- _ <-> ?hyp =>
      assert (h:hyp)
  end.

(*Wyckoff*)
Ltac gnhl :=
  let h := fresh "hl" in
  match goal with
      |- ?hyp <-> _ =>
      assert (h: ~hyp)
  end.

(*Wyckoff*)
Ltac gnhr :=
  let h := fresh "hr" in
  match goal with
      |- _ <-> ?hyp =>
      assert (h: ~hyp)
  end.


(*Wyckoff*)
Ltac ex_goal :=
  match goal with |- exists h:?pf, _ => 
                  let x := fresh "hex" in
                  assert (x:pf) end.


(*Wyckoff*)
Ltac terml h :=
  let P := type of h in
  match P with
      ?x = _ =>
      let x' := fresh "x" in
      pose x as x' end.

(*Wyckoff*)
Ltac termr h :=
  let P := type of h in
  match P with
      _ = ?y  =>
      let y' := fresh "y" in
      pose y as y' end.

(*Wyckoff*)
Ltac gterml :=
  match goal with 
      |- ?x = _ =>
      let x' := fresh "x" in
      pose x as x' end.


(*Wyckoff*)
Ltac gtermr :=
  match goal with 
      |- _ = ?y  =>
      let y' := fresh "y" in
      pose y as y' end.


(*Wyckoff*)
Ltac nterml h :=
  let P := type of h in
  match P with
      ?x <> _ =>
      let x' := fresh "x" in
      pose x as x' end.

(*Wyckoff*)
Ltac ntermr h :=
  let P := type of h in
  match P with
      _ <> ?y  =>
      let y' := fresh "y" in
      pose y as y' end.

(*Wyckoff*)
Ltac ngterml :=
  match goal with 
      |- ?x <> _ =>
      let x' := fresh "x" in
      pose x as x' end.


(*Wyckoff*)
Ltac ngtermr :=
  match goal with 
      |- _ <> ?y  =>
      let y' := fresh "y" in
      pose y as y' end.


(*Poses the offset term in a predicate or the left hand side of an 
equality. (Do a [symmetry] for right-hand match. As Below So Above.*)
Ltac gtermO0 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ ?c'] => pose c' as c 
    | [|- _ _ ?c'] => pose c' as c 
    | [|- _ _ _ ?c'] => pose c' as c 
    | [|- _ _ _ _ ?c'] => pose c' as c 
    | [|- _ _ _ _ _ ?c'] => pose c' as c 
    | [|- _ _ _ _ _ _ ?c']=> pose c' as c 
    | [|- _ _ _ _ _ _ _ ?c']=> pose c' as c 
    | [|- _ _ _ _ _ _ _ _ ?c'] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ ?c'] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ _ ?c'] => pose c' as c
  end.

(*Wyckoff*)
Ltac gtermO1 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ ?c' _] => pose c' as c 
    | [|- _ _ ?c' _] => pose c' as c 
    | [|- _ _ _ ?c' _] => pose c' as c 
    | [|- _ _ _ _ ?c' _] => pose c' as c 
    | [|- _ _ _ _ _ ?c' _] => pose c' as c 
    | [|- _ _ _ _ _ _ ?c' _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ ?c' _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ _ ?c' _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ ?c' _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ _ ?c' _] => pose c' as c
  end.

(*Wyckoff*)
Ltac gtermO2 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ ?c' _ _] => pose c' as c 
    | [|- _ _ ?c' _ _] => pose c' as c 
    | [|- _ _ _ ?c' _ _] => pose c' as c 
    | [|- _ _ _ _ ?c' _ _] => pose c' as c 
    | [|- _ _ _ _ _ ?c' _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ ?c' _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ ?c' _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ _ ?c' _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ ?c' _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ _ ?c' _ _] => pose c' as c
  end.


(*Wyckoff*)
Ltac gtermO3 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ ?c' _ _ _] => pose c' as c 
    | [|- _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ _ _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ ?c' _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ ?c' _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ _ ?c' _ _ _] => pose c' as c
  end.

(*Wyckoff*)
Ltac gtermO4 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ ?c' _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ ?c' _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _] => pose c' as c
  end.

(*Wyckoff*)
Ltac gtermO5 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ ?c' _ _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ ?c' _ _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _] => pose c' as c
  end.

(*Wyckoff*)
Ltac gtermO6 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ ?c' _ _ _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ ?c' _ _ _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _] => pose c' as c
  end.

(*Wyckoff*)
Ltac gtermO7 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ ?c' _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c
  end.

(*Wyckoff*)
Ltac gtermO8 :=
  let c := fresh "c" in
  match goal with
    | [|- _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c
  end.

(*Wyckoff*)
Ltac gtermO9 :=
  let c := fresh "c" in
  match goal with
    | [|- _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c
  end.


(*Wyckoff*)
Ltac gtermA0 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ -> _ ?c'] => pose c' as c 
    | [|- _ -> _ _ ?c'] => pose c' as c 
    | [|- _ -> _ _ _ ?c'] => pose c' as c 
    | [|- _ -> _ _ _ _ ?c'] => pose c' as c 
    | [|- _ -> _ _ _ _ _ ?c'] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ ?c']=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ ?c']=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ ?c'] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?c'] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?c'] => pose c' as c
end.

(*Wyckoff*)
Ltac gtermA1 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ -> _ ?c' _] => pose c' as c 
    | [|- _ -> _ _ ?c' _] => pose c' as c 
    | [|- _ -> _ _ _ ?c' _] => pose c' as c 
    | [|- _ -> _ _ _ _ ?c' _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ ?c' _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ ?c' _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ ?c' _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ ?c' _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?c' _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?c' _] => pose c' as c 
  end.

(*Wyckoff*)
Ltac gtermA2 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ -> _ ?c' _ _] => pose c' as c 
    | [|- _ -> _ _ ?c' _ _] => pose c' as c 
    | [|- _ -> _ _ _ ?c' _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ ?c' _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ ?c' _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ ?c' _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ ?c' _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ ?c' _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?c' _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _] => pose c' as c 
  end.

(*Wyckoff*)
Ltac gtermA3 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ -> _ ?c' _ _ _] => pose c' as c 
    | [|- _ -> _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ ?c' _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ ?c' _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _] => pose c' as c 
  end.

(*Wyckoff*)
Ltac gtermA4 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ -> _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ ?c' _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ ?c' _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _] => pose c' as c 
  end.


(*Wyckoff*)
Ltac gtermA5 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ -> _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ ?c' _ _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _] => pose c' as c 
  end.


(*Wyckoff*)
Ltac gtermA6 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ -> _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ ?c' _ _ _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _] => pose c' as c 
  end.


(*Wyckoff*)
Ltac gtermA7 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ -> _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ ?c' _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _] => pose c' as c 
  end.

(*Wyckoff*)
Ltac gtermA8 :=
  let c := fresh "c" in
  match goal with 
    | [|- _ -> _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _] => pose c' as c 
  end.

(*Wyckoff*)
Ltac gtermA9 :=
  let c := fresh "c" in
  match goal with
    | [|- _ -> _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _]=> pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _] => pose c' as c 
  end.

(*Wyckoff*)
Ltac gterm0 :=
  first [gtermO0 | gtermA0].

(*Wyckoff*)
Ltac gterm1 :=
  first [gtermO1 | gtermA1].

(*Wyckoff*)
Ltac gterm2 :=
  first [gtermO2 | gtermA2].

(*Wyckoff*)
Ltac gterm3 :=
  first [gtermO3 | gtermA3].

(*Wyckoff*)
Ltac gterm4 :=
  first [gtermO4 | gtermA4].

(*Wyckoff*)
Ltac gterm5 :=
  first [gtermO5 | gtermA5].

(*Wyckoff*)
Ltac gterm6 :=
  first [gtermO6 | gtermA6].

(*Wyckoff*)
Ltac gterm7 :=
  first [gtermO7 | gtermA7].

(*Wyckoff*)
Ltac gterm8 :=
  first [gtermO8 | gtermA8].

(*Wyckoff*)
Ltac gterm9 :=
  first [gtermO0 | gtermA9].


(*Wyckoff*)
Ltac termO0 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ ?c' => pose c' as c 
    | _ _ ?c' => pose c' as c 
    | _ _ _ ?c' => pose c' as c 
    | _ _ _ _ ?c' => pose c' as c 
    | _ _ _ _ _ ?c' => pose c' as c 
    | _ _ _ _ _ _ ?c' => pose c' as c 
    | _ _ _ _ _ _ _ ?c'=> pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' => pose c' as c
  end.

(*Wyckoff*)
Ltac termO1 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ ?c' _  => pose c' as c 
    | _ _ ?c' _ => pose c' as c 
    | _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ => pose c' as c
  end.

(*Wyckoff*)
Ltac termO2 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ ?c' _ _ => pose c' as c 
    | _ _ ?c' _ _ => pose c' as c 
    | _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac termO3 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ ?c' _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac termO4 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ ?c' _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac termO5 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac termO6 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac termO7 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac termO8 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac termO9 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac termA0 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ -> _ ?c' => pose c' as c 
    | _ -> _ _ ?c' => pose c' as c 
    | _ -> _ _ _ ?c' => pose c' as c 
    | _ -> _ _ _ _ ?c' => pose c' as c 
    | _ -> _ _ _ _ _ ?c' => pose c' as c 
    | _ -> _ _ _ _ _ _ ?c' => pose c' as c 
    | _ -> _ _ _ _ _ _ _ ?c'=> pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ ?c' => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ ?c' => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ _ ?c' => pose c' as c
  end.

(*Wyckoff*)
Ltac termA1 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ -> _ ?c' _  => pose c' as c 
    | _ -> _ _ ?c' _ => pose c' as c 
    | _ -> _ _ _ ?c' _ => pose c' as c 
    | _ -> _ _ _ _ ?c' _ => pose c' as c 
    | _ -> _ _ _ _ _ ?c' _ => pose c' as c 
    | _ -> _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ => pose c' as c
  end.


(*Wyckoff*)
Ltac termA2 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ -> _ ?c' _ _ => pose c' as c 
    | _ -> _ _ ?c' _ _ => pose c' as c 
    | _ -> _ _ _ ?c' _ _ => pose c' as c 
    | _ -> _ _ _ _ ?c' _ _ => pose c' as c 
    | _ -> _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac termA3 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ -> _ ?c' _ _ _ => pose c' as c 
    | _ -> _ _ ?c' _ _ _ => pose c' as c 
    | _ -> _ _ _ ?c' _ _ _ => pose c' as c 
    | _ -> _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac termA4 k :=
  let c := fresh "c" in
  let P := type of k in
  match P with
    | _ -> _ ?c' _ _ _ _ => pose c' as c 
    | _ -> _ _ ?c' _ _ _ _ => pose c' as c 
    | _ -> _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac termA5 k :=
  let c := fresh "c" in
  let P := type of k in
  match P with
    | _ -> _ ?c' _ _ _ _ _ => pose c' as c 
    | _ -> _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac termA6 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ -> _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac termA7 k :=
  let c := fresh "c" in
  let P := type of k in
  match k with
    | _ -> _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c
end.

(*Wyckoff*)
Ltac termA8 k :=
  let c := fresh "c" in
  let P := type of k in 
  match P with
    | _ -> _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac termA9 k :=
  let c := fresh "c" in
  let P := type of k in
  match P with
    | _ -> _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ -> _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac term0 k :=
  first [termO0 k| termA0 k].

(*Wyckoff*)
Ltac term1 k :=
  first [termO1 k| termA1 k].

(*Wyckoff*)
Ltac term2 k :=
  first [termO2 k| termA2 k].

(*Wyckoff*)
Ltac term3 k :=
  first [termO3 k| termA3 k].

(*Wyckoff*)
Ltac term4 k :=
  first [termO4 k| termA4 k].

(*Wyckoff*)
Ltac term5 k :=
  first [termO5 k| termA5 k].

(*Wyckoff*)
Ltac term6 k :=
  first [termO6 k| termA6 k].

(*Wyckoff*)
Ltac term7 k :=
  first [termO7 k| termA7 k].

(*Wyckoff*)
Ltac term8 k :=
  first [termO8 k| termA8 k].

(*Wyckoff*)
Ltac term9 k :=
  first [termO0 k| gtermA9 k].



(*Wyckoff*)
Ltac redterm0 k :=
  let c := fresh "c" in
  let k' := eval red in k in
  match k' with
    | _ ?c' => pose c' as c 
    | _ _ ?c' => pose c' as c 
    | _ _ _ ?c' => pose c' as c 
    | _ _ _ _ ?c' => pose c' as c 
    | _ _ _ _ _ ?c' => pose c' as c 
    | _ _ _ _ _ _ ?c' => pose c' as c 
    | _ _ _ _ _ _ _ ?c'=> pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' => pose c' as c
  end.

(*Wyckoff*)
Ltac redterm1 k :=
  let c := fresh "c" in
  let k' := eval red in k in
  match k' with
    | _ ?c' _  => pose c' as c 
    | _ _ ?c' _ => pose c' as c 
    | _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ => pose c' as c
  end.

(*Wyckoff*)
Ltac redterm2 k :=
  let c := fresh "c" in
  let k' := eval red in k in
  match k' with
    | _ ?c' _ _ => pose c' as c 
    | _ _ ?c' _ _ => pose c' as c 
    | _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac redterm3 k :=
  let c := fresh "c" in
  let k' := eval red in k in
  match k' with
    | _ ?c' _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac redterm4 k :=
  let c := fresh "c" in
  let k' := eval red in k in
  match k' with
    | _ ?c' _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac redterm5 k :=
  let c := fresh "c" in
  let k' := eval red in k in
  match k' with
    | _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac redterm6 k :=
  let c := fresh "c" in
  let k' := eval red in k in
  match k' with
    | _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac redterm7 k :=
  let c := fresh "c" in
  let k' := eval red in k in
  match k' with
    | _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac redterm8 k :=
  let c := fresh "c" in
  let k' := eval red in k in
  match k' with
    | _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ => pose c' as c
  end.

(*Wyckoff*)
Ltac redterm9 k :=
  let c := fresh "c" in
  let k' := eval red in k in 
  match k' with
    | _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c 
    | _ _ _ _ _ _ _ _ _ _ ?c' _ _ _ _ _ _ _ _ _ => pose c' as c
  end.


(*Wyckoff*)
Ltac ltl k :=
  let c := fresh "xlt" in
  let P := type of k in
  match P with
    | ?c' < _ => pose c' as c
  end.

(*Wyckoff*)
Ltac ltr k :=
  let c := fresh "xlt" in
  let P := type of k in
  match P with
    | _ < ?c' => pose c' as c
  end.

(*Wyckoff*)
Ltac gtl k :=
  let c := fresh "xgt" in
  let P := type of k in
  match P with
    | ?c' > _ => pose c' as c
  end.

(*Wyckoff*)
Ltac gtr k :=
  let c := fresh "xgt" in
  let P := type of k in
  match P with
    | _ > ?c' => pose c' as c
  end.

(*Wyckoff*)
Ltac gltl :=
  let c := fresh "xlt" in
  match goal with
    |- ?c' < _ => pose c' as c
  end.

(*Wyckoff*)
Ltac gltr :=
  let c := fresh "xlt" in
  match goal with
    |- _ < ?c' => pose c' as c
  end.

(*Wyckoff*)
Ltac ggtl :=
  let c := fresh "xgt" in
  match goal with
    |- ?c' > _ => pose c' as c
  end.

(*Wyckoff*)
Ltac ggtr :=
  let c := fresh "xgt" in
  match goal with
    |- _ > ?c' => pose c' as c
  end.


(*Wyckoff*)
Ltac lel k :=
  let c := fresh "xle" in
  let P := type of k in
  match P with
    | ?c' <= _ => pose c' as c
  end.

(*Wyckoff*)
Ltac ler k :=
  let c := fresh "xle" in
  let P := type of k in
  match P with
    | _ <= ?c' => pose c' as c
  end.

(*Wyckoff*)
Ltac gel k :=
  let c := fresh "xge" in
  let P := type of k in
  match P with
    | ?c' >= _ => pose c' as c
  end.

(*Wyckoff*)
Ltac ger k :=
  let c := fresh "xge" in
  let P := type of k in
  match P with
    | _ >= ?c' => pose c' as c
  end.

(*Wyckoff*)
Ltac glel :=
  let c := fresh "xle" in
  match goal with
    |- ?c' <= _ => pose c' as c
  end.

(*Wyckoff*)
Ltac gler :=
  let c := fresh "xle" in
  match goal with
    |- _ <= ?c' => pose c' as c
  end.

(*Wyckoff*)
Ltac ggel :=
  let c := fresh "xge" in
  match goal with
    |- ?c' >= _ => pose c' as c
  end.

(*Wyckoff*)
Ltac gger :=
  let c := fresh "xge" in
  match goal with
    |- _ >= ?c' => pose c' as c
  end.



(*Wyckoff*)
(*This is used to eliminate dependencies of [let]-posed propositions 
by posing such as a proof and clearing it.*)
Ltac redterm_cp h :=
  let pf := eval red in h in
      let h0 := fresh "hc" in
      pose proof h as h0; clear h.


(*Wyckoff*) 
Ltac match_match_goal :=
  match goal with |- context 
                       [match ?mtch' with
                          | 0  => _
                          | S _ => _
                        end] =>
                  let y := fresh "mtch" in
                  pose mtch' as y
  end.

(*Wyckoff*) 
Ltac match_match_match_goal :=
  match goal with |- context 
                       [match ?mtch' with
                          | 0  => _
                          | S _ => 
                              match _ with
                                | 0 => _
                                | S _ => _
                               end
                        end] =>
                  let y := fresh "mtch" in
                  pose mtch' as y 
  end.

(*Wyckoff*)
Ltac minusl pf :=
  match pf with ?a' - _ => 
      let y := fresh "ma" in
      pose a' as y end.

(*Wyckoff*)
Ltac minusr pf :=
  match pf with _ - ?b' =>
      let y := fresh "mb" in
      pose b' as y end.



(*This isolates the discriminee in a sumbool match in the 
given hypotheseis [h]*)
(*An analogous version of SF's [desruct_head_match] but for hypotheses,
and easier for me to understand!*)

(*Wyckoff*)
Ltac lr_match h :=
  let P := type of h in
  let y := fresh "hlr" in 
  match P with context
                 [match ?x with
                    | left _ => _ 
                    | right _  => _
                  end] =>
               pose x as y
  end.

(*Wyckoff*)
Ltac lr_match_goal :=
  let y := fresh "hlr" in 
  match goal with |- context
                       [match ?x with
                      | left _ => _ 
                      | right _  => _
                        end] =>
                  pose x as y
  end.

(*Wyckoff*)
Ltac lr_match_sum h :=
  let y := fresh "hlr" in
    let P := type of h in
    match P with context
                   [match ?x with
                      | inleft _ => _ 
                      | inright _  => _
                    end] =>
                 pose x as y
    end.

(*Wyckoff*)
Ltac lr_match_sum_goal :=
  let y := fresh "hlr" in
  match goal with |- context
                       [match ?x with
                          | inleft _ => _ 
                          | inright _  => _
                        end] =>
                  pose x as y
  end.



(*Sometimes lr_match doesn't work on ifs, and so . . . *)
(*Wyckoff*)  
Ltac lr_if h :=
  let y := fresh "hlr" in 
  let P := type of h in
  match P with context
                 [if ?x then _ else _] =>
               pose x as y
  end.

(*Wyckoff*)
Ltac lr_if_goal :=
  let y := fresh "hlr" in 
  match goal with |- context
                       [if ?x then _ else _] =>
                  pose x as y
  end.

(*Wyckoff*)
Ltac match_eq_refl h  :=
  let pfef := fresh "hef" in
  let P := type of h in
  match P with context
                 [match ?e with 
                    | eq_refl => _
                  end] =>
               pose e as pfef
 end.


(*Wyckoff*)
Ltac match_eq_refl_goal :=
  let pfef := fresh "hef" in
  match goal with
        |- context
             [match ?e with 
                | eq_refl => _
              end] =>
        pose e as pfef
  end.


(*Wyckoff*)
(*The [_x]-ed versions of the above 2 generalize and clear matched term.*)
Ltac match_eq_refl_x h  :=
  let pfef := fresh "hef" in
  let P := type of h in
  match P with context
                 [match ?e with 
                    | eq_refl => _
                  end] =>
               pose e as pfef; fold pfef in h;
               generalize h pfef; clear pfef;
               intros h pfef
 end.

Ltac match_eq_refl_goal_x :=
  let pfef := fresh "hef" in
  match goal with
      |- context
           [match ?e with 
              | eq_refl => _
            end] =>
      pose e as pfef; fold pfef;
      generalize pfef; clear pfef;
      intros pfef
  end.



(*These primed versions are for when you want the default behavior, 
and a one-step solution.  The unprimed versions are more useful when 
you want more control (like folding a matched expression in 
other expressions).*)
Ltac lr_match' h :=
  let y := fresh "hlr" in
  let hl' := fresh "hl" in
  let hr' := fresh "hr" in
  let P := type of h in
  match P with context
                 [match ?x with
                    | left _ => _ 
                    | right _  => _
                    end] =>
               pose x as y; fold y in h; destruct y as [hl'|hr']
  end.



(*Wyckoff*)
Ltac lr_match_goal' :=
  let y := fresh "hlr" in
  let hl' := fresh "hl" in
  let hr' := fresh "hr" in
  match goal with |- context
                       [match ?x with
                          | left _ => _ 
                          | right _  => _
                        end] =>
                  pose x as y; fold y; destruct y as [hl'|hr']
  end.



(*Wyckoff*)  
Ltac lr_if' h :=
  let y := fresh "hlr" in
  let hl' := fresh "hl" in
  let hr' := fresh "hr" in
  let P := type of h in
  match P with context
                 [if ?x then _ else _] =>
               pose x as y; fold y in h; destruct y as [hl'|hr']
  end.

(*Wyckoff*)
Ltac lr_if_goal' :=
  let y := fresh "hlr" in
  let hl' := fresh "hl" in
  let hr' := fresh "hr" in
  match goal with |- context
                       [if ?x then _ else _] =>
                  pose x as y; fold y; destruct y as [hl'|hr']
  end.

(*Wyckoff*)
Ltac lr_match_sum' h :=
  let y := fresh "hlr" in
  let hl' := fresh "hl" in
  let hr' := fresh "hr" in
  let P := type of h in
  match P with context
                 [match ?x with
                    | inleft _ => _ 
                    | inright _  => _
                    end] =>
               pose x as y; fold y in h; destruct y as [hl'|hr']
  end.




(*Wyckoff*)
Ltac lr_match_sum_goal' :=
  let y := fresh "hlr" in
  let hl' := fresh "hl" in
  let hr' := fresh "hr" in
  match goal with |- context
                       [match ?x with
                          | inleft _ => _ 
                          | inright _  => _
                        end] =>
                  pose x as y; fold y; destruct y as [hl'|hr']
  end.


(*Wyckoff*)
Ltac match_eq_refl' h :=
  let pfef := fresh "hef" in
  let pfef' := fresh "hef" in
  let P := type of h in
  match P with
      context
        [match ?e in  (_ = _) return (_ _) with
           | eq_refl => _
         end] =>
      pose e as pfef; fold pfef; destruct pfef as [pfef']
  end.



(*Wyckoff*)
Ltac match_eq_refl_imp' h :=
  let pfef := fresh "hef" in
  let pfef' := fresh "hef" in
  let P := type of h in
  match P with
      context
        [match ?e in  (_ = _) return (_ -> _) with
           | eq_refl => _
         end] =>
      pose e as pfef; fold pfef; destruct pfef as [pfef']
  end.


(*Wyckoff*)
Ltac match_eq_refl_goal' :=
  let pfef := fresh "hef" in
  let pfef' := fresh "hef" in
   match goal with
      |- context
           [match ?e in  (_ = _) return (_ _) with
              | eq_refl => _
            end] =>
      pose e as pfef; fold pfef; destruct pfef as [pfef']
end.


(*Wyckoff*)
Ltac match_eq_refl_imp_goal' :=
  let pfef := fresh "hef" in
  let pfef' := fresh "hef" in
   match goal with
      |- context
           [match ?e in  (_ = _) return (_ -> _) with
              | eq_refl => _
            end] =>
      pose e as pfef; fold pfef; destruct pfef as [pfef']
end.


(*The pattern for these (eventually?) 1000 generalization tactics is to match 
  0-9 hypotheses with gen[O-I][n][O-I].  *)
(*Adding too many generalizations is of marginal utility, so 
it's not quite 1000 generalizations, although I prepared 
incomplete templates for them in the Preview file (the general file for half-worked 
stuff that won't make the next release).*)
(*Wyckoff*)
Ltac genO0O :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ _ _ _ _ ?h = _] => generalize h
    | [|- _ _ _ _ ?h = _] => generalize h
    | [|- _ _ _ ?h = _] => generalize h
    | [|- _ _ ?h = _] => generalize h
    | [|- _ ?h = _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ _ _ _ _ _ ?h] => generalize h
    | [|- _ _ _ _ _ ?h] => generalize h
    | [|- _ _ _ _ ?h] => generalize h
    | [|- _ _ _ ?h] => generalize h
    | [|- _ _ ?h] => generalize h
    | [|- _ ?h] => generalize h
  end.

(*Wyckoff*)
Ltac genO1O :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ _ _ _ ?h _ = _] => generalize h
    | [|- _ _ _ ?h _ = _] => generalize h
    | [|- _ _ ?h _ = _] => generalize h
    | [|- _ ?h _ = _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ _ _ _ _ ?h _] => generalize h
    | [|- _ _ _ _ ?h _] => generalize h
    | [|- _ _ _ ?h _] => generalize h
    | [|- _ _ ?h _] => generalize h
    | [|- _ ?h _] => generalize h
  end.

(*Wyckoff*)
Ltac genO2O :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ _ _ ?h _ _ = _] => generalize h
    | [|- _ _ ?h _ _ = _] => generalize h
    | [|- _ ?h _ _ = _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ _ _ _ ?h _ _] => generalize h
    | [|- _ _ _ ?h _ _] => generalize h
    | [|- _ _ ?h _ _] => generalize h
    | [|- _ ?h _ _] => generalize h
end.

(*Wyckoff*)
Ltac genO3O :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ _ ?h _ _ _ = _] => generalize h
    | [|- _ ?h _ _ _ = _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ _ _ ?h _ _ _] => generalize h
    | [|- _ _ ?h _ _ _] => generalize h
    | [|- _ ?h _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genO4O :=
  match goal with
    | [|- _ _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ _ ?h _ _ _ _] => generalize h
    | [|- _ ?h _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genO5O :=
  match goal with
    | [|- _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ ?h _ _ _ _ = _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ _ ?h _ _ _ _] => generalize h
    | [|- _ ?h _ _ _ _] => generalize h
end.


(*Wyckoff*)
Ltac genO6O :=
  match goal with
    | [|- _ _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ ?h _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genO7O :=
  match goal with
    | [|- _ _ _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genO8O :=
  match goal with
    | [|- _ _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genO9O :=
  match goal with
    | [|- _ _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ _] => generalize h
  end.



(*Wyckoff*)
Ltac genA0O :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _?h = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _?h = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _?h = _] => generalize h
    | [|- _ -> _ _ _ _ _ _?h = _] => generalize h
    | [|- _ -> _ _ _ _ _?h = _] => generalize h
    | [|- _ -> _ _ _  _?h = _] => generalize h
    | [|- _ -> _ _ _ ?h = _] => generalize h
    | [|- _ -> _ _ ?h = _] => generalize h
    | [|- _ -> _ ?h = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _?h] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _?h]=> generalize h
    | [|- _ -> _ _ _ _ _ _ _?h]=> generalize h
    | [|- _ -> _ _ _ _ _ _?h]  => generalize h
    | [|- _ -> _ _ _ _ _?h] => generalize h
    | [|- _ -> _ _ _  _?h] => generalize h
    | [|- _ -> _ _ _ ?h] => generalize h
    | [|- _ -> _ _ ?h] => generalize h
    | [|- _ -> _ ?h] => generalize h
  end.


(*Wyckoff*)
Ltac genA1O :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ _ ?h _ = _] => generalize h
    | [|- _ -> _ ?h _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ _ _ ?h _] => generalize h
    | [|- _ -> _ _ ?h _] => generalize h
    | [|- _ -> _ ?h _] => generalize h
  end.


(*Wyckoff*)
Ltac genA2O :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ ?h _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ _ ?h _ _] => generalize h
    | [|- _ -> _ ?h _ _] => generalize h
end.


(*Wyckoff*)
Ltac genA3O :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ ?h _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genA4O :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ ?h _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genA5O :=
  match goal with
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _] => generalize h
end.

(*Wyckoff*)
Ltac genA6O :=
  match goal with
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genA7O :=
  match goal with
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genA8O :=
  match goal with
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genA9O :=
  match goal with
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genB0O :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ ?h = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ _ ?h] => generalize h
    | [|- _ -> _ -> _ ?h] => generalize h
  end.



(*Wyckoff*)
Ltac genB1O :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ ?h _] => generalize h
  end.



(*Wyckoff*)
Ltac genB2O :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ ?h _ _] => generalize h
end.



(*Wyckoff*)
Ltac genB3O :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genB4O :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genB5O :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _] => generalize h
end.

(*Wyckoff*)
Ltac genB6O :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genB7O :=
  match goal with
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genB8O :=
  match goal with
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genB9O :=
  match goal with
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genC0O :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ ?h] => generalize h
  end.



(*Wyckoff*)
Ltac genC1O :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _] => generalize h
  end.



(*Wyckoff*)
Ltac genC2O :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _] => generalize h
end.



(*Wyckoff*)
Ltac genC3O :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genC4O :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genC5O :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _] => generalize h
end.

(*Wyckoff*)
Ltac genC6O :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _]  => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genC7O :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genC8O :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genC9O :=
  match goal with
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genD0O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h] => generalize h
  end.


(*Wyckoff*)
Ltac genD1O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _]=> generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _] => generalize h
  end.



(*Wyckoff*)
Ltac genD2O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _] => generalize h
end.



(*Wyckoff*)
Ltac genD3O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genD4O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genD5O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _] => generalize h
end.

(*Wyckoff*)
Ltac genD6O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genD7O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genD8O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genD9O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genE0O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h] => generalize h
  end.

(*Wyckoff*)
Ltac genE1O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _] => generalize h
  end.



(*Wyckoff*)
Ltac genE2O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _] => generalize h
end.



(*Wyckoff*)
Ltac genE3O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genE4O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genE5O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _] => generalize h
end.

(*Wyckoff*)
Ltac genE6O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genE7O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genE8O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _] => generalize h

    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genE9O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genF0O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h] => generalize h
  end.



(*Wyckoff*)
Ltac genF1O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _] => generalize h
  end.



(*Wyckoff*)
Ltac genF2O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _] => generalize h
end.


(*Wyckoff*)
Ltac genF3O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genF4O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genF5O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _] => generalize h
end.

(*Wyckoff*)
Ltac genF6O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genF7O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genF8O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genF9O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genG0O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h] => generalize h
  end.



(*Wyckoff*)
Ltac genG1O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _] => generalize h
  end.


(*Wyckoff*)
Ltac genG2O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _] => generalize h
end.



(*Wyckoff*)
Ltac genG3O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genG4O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genG5O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _] => generalize h
end.


(*Wyckoff*)
Ltac genG6O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genG7O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genG8O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genG9O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genH0O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h] => generalize h
  end.



(*Wyckoff*)
Ltac genH1O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _]  => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _]  => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _] => generalize h
  end.



(*Wyckoff*)
Ltac genH2O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _] => generalize h
end.



(*Wyckoff*)
Ltac genH3O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genH4O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genH5O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _] => generalize h
end.

(*Wyckoff*)
Ltac genH6O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genH7O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genH8O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genH9O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genI0O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h] => generalize h
  end.



(*Wyckoff*)
Ltac genI1O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _] => generalize h
  end.



(*Wyckoff*)
Ltac genI2O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _]=> generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _] => generalize h
end.



(*Wyckoff*)
Ltac genI3O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genI4O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genI5O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _] => generalize h
end.

(*Wyckoff*)
Ltac genI6O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genI7O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genI8O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _] => generalize h
  end.

(*Wyckoff*)
Ltac genI9O :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _] => generalize h
  end.


(*Wyckoff*)
Ltac genO0A :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ _ _ ?h = _ -> _] => generalize h
    | [|- _ _ ?h = _ -> _] => generalize h
    | [|- _ ?h = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ _ _ _ ?h -> _] => generalize h
    | [|- _ _ _ ?h -> _] => generalize h
    | [|- _ _ ?h -> _] => generalize h
    | [|- _ ?h -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO1A :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ _ ?h _ = _ -> _] => generalize h
    | [|- _ ?h _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ _ _ ?h _ -> _] => generalize h
    | [|- _ _ ?h _ -> _] => generalize h
    | [|- _ ?h _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO2A :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ ?h _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ _ ?h _ _ -> _] => generalize h
    | [|- _ ?h _ _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genO3A :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ ?h _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genO4A :=
  match goal with
    | [|- _ _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO5A :=
  match goal with
    | [|- _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ -> _] => generalize h
end.


(*Wyckoff*)
Ltac genO6A :=
  match goal with
    | [|- _ _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO7A :=
  match goal with
    | [|- _ _ _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO8A :=
  match goal with
    | [|- _ _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO9A :=
  match goal with
    | [|- _ _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genA0A :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ ?h = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ _ ?h -> _] => generalize h
    | [|- _ -> _ ?h -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genA1A :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ ?h _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genA2A :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ -> _] => generalize h
end.


(*Wyckoff*)
Ltac genA3A :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA4A :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA5A :=
  match goal with
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genA6A :=
  match goal with
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA7A :=
  match goal with
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA8A :=
  match goal with
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA9A :=
  match goal with
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB0A :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ ?h -> _] => generalize h
  end.

 

(*Wyckoff*)
Ltac genB1A :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genB2A :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genB3A :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genB4A :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB5A :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genB6A :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB7A :=
  match goal with
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB8A :=
  match goal with
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB9A :=
  match goal with
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genC0A :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genC1A :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genC2A :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genC3A :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genC4A :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genC5A :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genC6A :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _]  => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genC7A :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genC8A :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genC9A :=
  match goal with
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genD0A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genD1A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _]=> generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genD2A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genD3A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genD4A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genD5A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genD6A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genD7A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genD8A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genD9A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE0A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE1A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genE2A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genE3A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genE4A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE5A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genE6A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE7A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE8A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h

    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE9A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genF0A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genF1A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genF2A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ -> _] => generalize h
end.


(*Wyckoff*)
Ltac genF3A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genF4A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genF5A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genF6A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genF7A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genF8A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genF9A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genG0A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genG1A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genG2A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genG3A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genG4A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genG5A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _] => generalize h
end.


(*Wyckoff*)
Ltac genG6A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genG7A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genG8A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genG9A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genH0A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genH1A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _]  => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _]  => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genH2A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genH3A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genH4A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genH5A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genH6A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genH7A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genH8A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genH9A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genI0A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genI1A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genI2A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genI3A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genI4A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genI5A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genI6A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genI7A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genI8A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genI9A :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genO0B :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ ?h = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ _ ?h -> _ -> _] => generalize h
    | [|- _ ?h -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO1B :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ ?h _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO2B :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genO3B :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genO4B :=
  match goal with
    | [|- _ _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO5B :=
  match goal with
    | [|- _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ -> _ -> _] => generalize h
end.


(*Wyckoff*)
Ltac genO6B :=
  match goal with
    | [|- _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO7B :=
  match goal with
    | [|- _ _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO8B :=
  match goal with
    | [|- _ _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO9B :=
  match goal with
    | [|- _ _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genA0B :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ ?h -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genA1B :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genA2B :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ -> _ -> _] => generalize h
end.


(*Wyckoff*)
Ltac genA3B :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genA4B :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA5B :=
  match goal with
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genA6B :=
  match goal with
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA7B :=
  match goal with
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA8B :=
  match goal with
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA9B :=
  match goal with
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB0B :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genB1B :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genB2B :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ -> _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genB3B :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genB4B :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB5B :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genB6B :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB7B :=
  match goal with
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB8B :=
  match goal with
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB9B :=
  match goal with
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genC0B :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genC1B:=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genC2B :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ -> _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genC3B :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genC4B :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genC5B :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genC6B :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _]  => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genC7B :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genC8B :=
  match goal with
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genC9B :=
  match goal with
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genD0B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genD1B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _ -> _]=> generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genD2B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ -> _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genD3B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genD4B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genD5B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genD6B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genD7B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genD8B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genD9B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE0B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE1B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genE2B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ -> _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genE3B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genE4B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE5B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genE6B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE7B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genE8B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h

    | [|- _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genE9B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genF0B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genF1B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genF2B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ -> _ -> _] => generalize h
end.


(*Wyckoff*)
Ltac genF3B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genF4B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genF5B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genF6B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genF7B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genF8B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genF9B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genG0B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genG1B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genG2B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ -> _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genG3B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _ -> _] => generalize h
  end.


(*The lower-case [genNX]s traverse a string of implications from 
the right-most implication to the left-most, in accord with their 
eponymous indices, wherease the
upper-case [GenNX]s traverses name indices from 
left to right.*)

Ltac gen0O :=
  first [genO0O | genA0O | genB0O | genC0O | genD0O |
         genE0O | genF0O | genG0O | genH0O | genI0O].

Ltac gen0A :=
  first [genO0A | genA0A | genB0A | genC0A | genD0A |
         genE0A | genF0A | genG0A | genH0A | genI0A].

Ltac gen0B :=
  first [genO0B | genA0B | genB0B | genC0B | genD0B |
         genE0B | genF0B | genG0B].

Ltac gen1O :=
  first [genO1O | genA1O | genB1O | genC1O | genD1O |
         genE1O | genF1O | genG1O | genH1O | genI1O].

Ltac gen1A :=
  first [genO1A | genA1A | genB1A | genC1A | genD1A |
         genE1A | genF1A | genG1A | genH1A | genI1A].

Ltac gen1B :=
  first [genO1B | genA1B | genB1B | genC1B | genD1B |
         genE1B | genF1B | genG1B].

Ltac gen2O :=
  first [genO2O | genA2O | genB2O | genC2O | genD2O |
         genE2O | genF2O | genG2O | genH2O | genI2O].

Ltac gen2A :=
  first [genO2A | genA2A | genB2A | genC2A | genD2A |
         genE2A | genF2A | genG2A | genH2A | genI2A].

Ltac gen2B :=
  first [genO2B | genA2B | genB2B | genC2B | genD2B |
         genE2B | genF2B | genG2B].

Ltac gen3O :=
  first [genO3O | genA3O | genB3O | genC3O | genD3O |
         genE3O | genF3O | genG3O | genH3O | genI3O].

Ltac gen3A :=
  first [genO3A | genA3A | genB3A | genC3A | genD3A |
         genE3A | genF3A | genG3A | genH3A | genI3A].

Ltac gen3B :=
  first [genO3B | genA3B | genB3B | genC3B | genD3B |
         genE3B | genF3B | genG3B].

Ltac gen4O :=
  first [genO4O | genA4O | genB4O | genC4O | genD4O |
         genE4O | genF4O | genG4O | genH4O | genI4O].

Ltac gen4A :=
  first [genO4A | genA4A | genB4A | genC4A | genD4A |
         genE4A | genF4A | genG4A | genH4A | genI4A].

Ltac gen4B :=
  first [genO4B | genA4B | genB4B | genC4B | genD4B |
         genE4B | genF4B].

Ltac gen5O :=
  first [genO5O | genA5O | genB5O | genC5O | genD5O |
         genE5O | genF5O | genG5O | genH5O | genI5O].

Ltac gen5A :=
  first [genO5A | genA5A | genB5A | genC5A | genD5A |
         genE5A | genF5A | genG5A | genH5A | genI5A].

Ltac gen5B :=
  first [genO5B | genA5B | genB5B | genC5B | genD5B |
         genE5B | genF5B].

Ltac gen6O :=
  first [genO6O | genA6O | genB6O | genC6O | genD6O |
         genE6O | genF6O | genG6O | genH6O | genI6O].

Ltac gen6A :=
  first [genO6A | genA6A | genB6A | genC6A | genD6A |
         genE6A | genF6A | genG6A | genH6A | genI6A].

Ltac gen6B :=
  first [genO6B | genA6B | genB6B | genC6B | genD6B |
         genE6B | genF6B].

Ltac gen7O :=
  first [genO7O | genA7O | genB7O | genC7O | genD7O |
         genE7O | genF7O | genG7O | genH7O | genI7O].

Ltac gen7A :=
  first [genO7A | genA7A | genB7A | genC7A | genD7A |
         genE7A | genF7A | genG7A | genH7A | genI7A].

Ltac gen7B :=
  first [genO7B | genA7B | genB7B | genC7B | genD7B |
         genE7B | genF7B].

Ltac gen8O :=
  first [genO8O | genA8O | genB8O | genC8O | genD8O |
         genE8O | genF8O | genG8O | genH8O | genI8O].

Ltac gen8A :=
  first [genO8A | genA8A | genB8A | genC8A | genD8A |
         genE8A | genF8A | genG8A | genH8A | genI8A].

Ltac gen8B :=
  first [genO8B | genA8B | genB8B | genC8B | genD8B |
         genE8B | genF8B].

Ltac gen9O :=
  first [genO9O | genA9O | genB9O | genC9O | genD9O |
         genE9O | genF9O | genG9O | genH9O | genI9O].

Ltac gen9A :=
  first [genO9A | genA9A | genB9A | genC9A | genD9A |
         genE9A | genF9A | genG9A | genH9A | genI9A].

Ltac gen9B :=
  first [genO9B | genA9B | genB9B | genC9B | genD9B |
         genE9B | genF9B].



Ltac gen0 := first [gen0O | gen0A | gen0B].
Ltac gen1 := first [gen1O | gen1A | gen1B].
Ltac gen2 := first [gen2O | gen2A | gen2B].
Ltac gen3 := first [gen3O | gen3A | gen3B].
Ltac gen4 := first [gen4O | gen4A | gen4B].
Ltac gen5 := first [gen5O | gen5A | gen5B].
Ltac gen6 := first [gen6O | gen6A | gen6B].
Ltac gen7 := first [gen7O | gen7A | gen7B].
Ltac gen8 := first [gen8O | gen8A | gen8B].
Ltac gen9 := first [gen9O | gen9A | gen9B].

Ltac GenO0 :=
  first [genO0O | genO0A |genO0B].

Ltac GenO1 :=
  first [genO1O | genO1A |genO1B].

Ltac GenO2 :=
  first [genO2O | genO2A |genO2B].

Ltac GenO3 :=
  first [genO3O | genO3A |genO3B].

Ltac GenO4 :=
  first [genO4O | genO4A |genO4B].

Ltac GenO5 :=
  first [genO5O | genO5A |genO5B].

Ltac GenO6 :=
  first [genO6O | genO6A |genO6B].

Ltac GenO7 :=
  first [genO7O | genO7A |genO7B].

Ltac GenO8 :=
  first [genO8O | genO8A |genO8B].

Ltac GenO9 :=
  first [genO9O | genO9A |genO9B].

Ltac GenA0 :=
  first [genA0O | genA0A |genA0B].

Ltac GenA1 :=
  first [genA1O | genA1A |genA1B].

Ltac GenA2 :=
  first [genA2O | genA2A |genA2B].

Ltac GenA3 :=
  first [genA3O | genA3A |genA3B].

Ltac GenA4 :=
  first [genA4O | genA4A |genA4B].

Ltac GenA5 :=
  first [genA5O | genA5A |genA5B].

Ltac GenA6 :=
  first [genA6O | genA6A |genA6B].

Ltac GenA7 :=
  first [genA7O | genA7A |genA7B].

Ltac GenA8 :=
  first [genA8O | genA8A |genA8B].

Ltac GenA9 :=
  first [genA9O | genA9A |genA9B].

Ltac GenB0 :=
  first [genB0O | genB0A |genB0B].

Ltac GenB1 :=
  first [genB1O | genB1A |genB1B].

Ltac GenB2 :=
  first [genB2O | genB2A |genB2B].

Ltac GenB3 :=
  first [genB3O | genB3A |genB3B].

Ltac GenB4 :=
  first [genB4O | genB4A |genB4B].

Ltac GenB5 :=
  first [genB5O | genB5A |genB5B].

Ltac GenB6 :=
  first [genB6O | genB6A |genB6B].

Ltac GenB7 :=
  first [genB7O | genB7A |genB7B].

Ltac GenB8 :=
  first [genB8O | genB8A |genB8B].

Ltac GenB9 :=
  first [genB9O | genB9A |genB9B].

Ltac GenC0 :=
  first [genC0O | genC0A |genC0B].

Ltac GenC1 :=
  first [genC1O | genC1A |genC1B].

Ltac GenC2 :=
  first [genC2O | genC2A |genC2B].

Ltac GenC3 :=
  first [genC3O | genC3A |genC3B].

Ltac GenC4 :=
  first [genC4O | genC4A |genC4B].

Ltac GenC5 :=
  first [genC5O | genC5A |genC5B].

Ltac GenC6 :=
  first [genC6O | genC6A |genC6B].

Ltac GenC7 :=
  first [genC7O | genC7A |genC7B].

Ltac GenC8 :=
  first [genC8O | genC8A |genC8B].

Ltac GenC9 :=
  first [genC9O | genC9A |genC9B].

Ltac GenD0 :=
  first [genD0O | genD0A |genD0B].

Ltac GenD1 :=
  first [genD1O | genD1A |genD1B].

Ltac GenD2 :=
  first [genD2O | genD2A |genD2B].

Ltac GenD3 :=
  first [genD3O | genD3A |genD3B].

Ltac GenD4 :=
  first [genD4O | genD4A |genD4B].

Ltac GenD5 :=
  first [genD5O | genD5A |genD5B].

Ltac GenD6 :=
  first [genD6O | genD6A |genD6B].

Ltac GenD7 :=
  first [genD7O | genD7A |genD7B].

Ltac GenD8 :=
  first [genD8O | genD8A |genD8B].

Ltac GenD9 :=
  first [genD9O | genD9A |genD9B].

Ltac GenE0 :=
  first [genE0O | genE0A |genE0B].

Ltac GenE1 :=
  first [genE1O | genE1A |genE1B].

Ltac GenE2 :=
  first [genE2O | genE2A |genE2B].

Ltac GenE3 :=
  first [genE3O | genE3A |genE3B].

Ltac GenE4 :=
  first [genE4O | genE4A |genE4B].

Ltac GenE5 :=
  first [genE5O | genE5A |genE5B].

Ltac GenE6 :=
  first [genE6O | genE6A |genE6B].

Ltac GenE7 :=
  first [genE7O | genE7A |genE7B].

Ltac GenE8 :=
  first [genE8O | genE8A |genE8B].

Ltac GenE9 :=
  first [genE9O | genE9A |genE9B].

Ltac GenF0 :=
  first [genF0O | genF0A |genF0B].

Ltac GenF1 :=
  first [genF1O | genF1A |genF1B].

Ltac GenF2 :=
  first [genF2O | genF2A |genF2B].

Ltac GenF3 :=
  first [genF3O | genF3A |genF3B].

Ltac GenF4 :=
  first [genF4O | genF4A |genF4B].

Ltac GenF5 :=
  first [genF5O | genF5A |genF5B].

Ltac GenF6 :=
  first [genF6O | genF6A |genF6B].

Ltac GenF7 :=
  first [genF7O | genF7A |genF7B].

Ltac GenF8 :=
  first [genF8O | genF8A |genF8B].

Ltac GenF9 :=
  first [genF9O | genF9A |genF9B].

Ltac GenG0 :=
  first [genG0O | genG0A |genG0B].

Ltac GenG1 :=
  first [genG1O | genG1A |genG1B].

Ltac GenG2 :=
  first [genG2O | genG2A |genG2B].

Ltac GenG3 :=
  first [genG3O | genG3A |genG3B].

Ltac GenG4 :=
  first [genG4O | genG4A].

Ltac GenG5 :=
  first [genG5O | genG5A].

Ltac GenG6 :=
  first [genG6O | genG6A].

Ltac GenG7 :=
  first [genG7O | genG7A].

Ltac GenG8 :=
  first [genG8O | genG8A].

Ltac GenG9 :=
  first [genG9O | genG9A].

Ltac GenH0 :=
  first [genH0O | genH0A].

Ltac GenH1 :=
  first [genH1O | genH1A].

Ltac GenH2 :=
  first [genH2O | genH2A].

Ltac GenH3 :=
  first [genH3O | genH3A].

Ltac GenH4 :=
  first [genH4O | genH4A].

Ltac GenH5 :=
  first [genH5O | genH5A].

Ltac GenH6 :=
  first [genH6O | genH6A].

Ltac GenH7 :=
  first [genH7O | genH7A].

Ltac GenH8 :=
  first [genH8O | genH8A].

Ltac GenH9 :=
  first [genH9O | genH9A].

Ltac GenI0 :=
  first [genI0O | genI0A].

Ltac GenI1 :=
  first [genI1O | genI1A].

Ltac GenI2 :=
  first [genI2O | genI2A].

Ltac GenI3 :=
  first [genI3O | genI3A].

Ltac GenI4 :=
  first [genI4O | genI4A].

Ltac GenI5 :=
  first [genI5O | genI5A].

Ltac GenI6 :=
  first [genI6O | genI6A].

Ltac GenI7 :=
  first [genI7O | genI7A].

Ltac GenI8 :=
  first [genI8O | genI8A].

Ltac GenI9 :=
  first [genI9O | genI9A].

Ltac Gen0 :=
  first [GenO0 | GenA0 | GenB0 | GenC0 | GenD0 |
         GenE0 | GenF0 | GenG0 | GenH0 | GenI0].

Ltac Gen1 :=
  first [GenO1 | GenA1 | GenB1 | GenC1 | GenD1 |
         GenE1 | GenF1 | GenG1 | GenH1 | GenI1].

Ltac Gen2 :=
  first [GenO2 | GenA2 | GenB2 | GenC2 | GenD2 |
         GenE2 | GenF2 | GenG2 | GenH2 | GenI2].

Ltac Gen3 :=
  first [GenO3 | GenA3 | GenB3 | GenC3 | GenD3 |
         GenE3 | GenF3 | GenG3 | GenH3 | GenI3].

Ltac Gen4 :=
  first [GenO4 | GenA4 | GenB4 | GenC4 | GenD4 |
         GenE4 | GenF4 | GenG4 | GenH4 | GenI4].

Ltac Gen5 :=
  first [GenO5 | GenA5 | GenB5 | GenC5 | GenD5 |
         GenE5 | GenF5 | GenG5 | GenH5 | GenI5].

Ltac Gen6 :=
  first [GenO6 | GenA6 | GenB6 | GenC6 | GenD6 |
         GenE6 | GenF6 | GenG6 | GenH6 | GenI6].

Ltac Gen7 :=
  first [GenO7 | GenA7 | GenB7 | GenC7 | GenD7 |
         GenE7 | GenF7 | GenG7 | GenH7 | GenI7].

Ltac Gen8 :=
  first [GenO8 | GenA8 | GenB8 | GenC8 | GenD8 |
         GenE8 | GenF8 | GenG8 | GenH8 | GenI8].

Ltac Gen9 :=
  first [GenO9 | GenA9 | GenB9 | GenC9 | GenD9 |
         GenE9 | GenF9 | GenG9 | GenH9 | GenI9].

(*You get the idea, the next part of adding the appropriate amount of 
  implications to these templates to corresopnd with the last character 
  will be left as a ritual of sleep-deprivation and eager commits!
  I don't want to write a script to do this.*)


Ltac genG4B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genG5B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _ -> _] => generalize h
end.


(*Wyckoff*)
Ltac genG6B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genG7B:=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genG8B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genG9B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genH0B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genH1B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _]  => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _]  => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genH2B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ -> _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genH3B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genH4B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genH5B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genH6B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genH7B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genH8B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genH9B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genI0B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genI1B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genI2B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ -> _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genI3B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genI4B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genI5B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genI6B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genI7B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genI8B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genI9B :=
  match goal with
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genO0C :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ ?h -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO1C :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO2C :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ -> _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genO3C :=
  match goal with
    | [|- _ _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ -> _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genO4C :=
  match goal with
    | [|- _ _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO5C :=
  match goal with
    | [|- _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
end.


(*Wyckoff*)
Ltac genO6C :=
  match goal with
    | [|- _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO7C :=
  match goal with
    | [|- _ _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO8C :=
  match goal with
    | [|- _ _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genO9C :=
  match goal with
    | [|- _ _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genA0C :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h -> _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genA1C :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ -> _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genA2C :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ -> _ -> _ -> _] => generalize h
end.


(*Wyckoff*)
Ltac genA3C :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ -> _ -> _ -> _] => generalize h
  end.


(*Wyckoff*)
Ltac genA4C :=
  match goal with
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA5C :=
  match goal with
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ _ ?h _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ -> _ -> _ -> _] => generalize h
end.

(*Wyckoff*)
Ltac genA6C :=
  match goal with
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ _ ?h _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA7C :=
  match goal with
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ _ ?h _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA8C :=
  match goal with
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ _ ?h _ _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genA9C :=
  match goal with
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ ?h _ _ _ _ _ _ _ _ _ -> _ -> _ -> _] => generalize h
  end.

(*Wyckoff*)
Ltac genB0C :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h -> _ -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genB1C :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ -> _ -> _ -> _] => generalize h
  end.



(*Wyckoff*)
Ltac genB2C :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ -> _ -> _ -> _] => generalize h
end.



(*Wyckoff*)
Ltac genB3C :=
  match goal with
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ = _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ _ ?h _ _ _ -> _ -> _ -> _] => generalize h
    | [|- _ -> _ -> _ ?h _ _ _ -> _ -> _ -> _] => generalize h
  end.

(*SF*)
Inductive Boxer : Type :=
| boxer : forall {A:Type}, A -> Boxer.

(*Wyckoff*)
Definition unbox (b:Boxer) : match b with
                               | boxer T _ => T
                             end:=
  match b with
    | boxer _ x => x
  end.
  
Require Import List.

(*SF*)
Notation "'>>'" :=
  (@nil Boxer)
  (at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1" :=
  ((boxer v1)::nil)
  (at level 0, v1 at level 0).
(*e  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2" :=
  ((boxer v1)::(boxer v2)::nil)
  (at level 0, v1 at level 0, v2 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3 v4" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::(boxer v4)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0,
   v4 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3 v4 v5" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::(boxer v4)::(boxer v5)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0,
   v4 at level 0, v5 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3 v4 v5 v6" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::(boxer v4)::(boxer v5)
   ::(boxer v6)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0,
   v4 at level 0, v5 at level 0, v6 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3 v4 v5 v6 v7" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::(boxer v4)::(boxer v5)
   ::(boxer v6)::(boxer v7)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0,
   v4 at level 0, v5 at level 0, v6 at level 0, v7 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3 v4 v5 v6 v7 v8" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::(boxer v4)::(boxer v5)
   ::(boxer v6)::(boxer v7)::(boxer v8)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0,
   v4 at level 0, v5 at level 0, v6 at level 0, v7 at level 0,
   v8 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3 v4 v5 v6 v7 v8 v9" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::(boxer v4)::(boxer v5)
   ::(boxer v6)::(boxer v7)::(boxer v8)::(boxer v9)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0,
   v4 at level 0, v5 at level 0, v6 at level 0, v7 at level 0,
   v8 at level 0, v9 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3 v4 v5 v6 v7 v8 v9 v10" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::(boxer v4)::(boxer v5)
   ::(boxer v6)::(boxer v7)::(boxer v8)::(boxer v9)::(boxer v10)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0,
   v4 at level 0, v5 at level 0, v6 at level 0, v7 at level 0,
   v8 at level 0, v9 at level 0, v10 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::(boxer v4)::(boxer v5)
   ::(boxer v6)::(boxer v7)::(boxer v8)::(boxer v9)::(boxer v10)
   ::(boxer v11)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0,
   v4 at level 0, v5 at level 0, v6 at level 0, v7 at level 0,
   v8 at level 0, v9 at level 0, v10 at level 0, v11 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11 v12" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::(boxer v4)::(boxer v5)
   ::(boxer v6)::(boxer v7)::(boxer v8)::(boxer v9)::(boxer v10)
   ::(boxer v11)::(boxer v12)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0,
   v4 at level 0, v5 at level 0, v6 at level 0, v7 at level 0,
   v8 at level 0, v9 at level 0, v10 at level 0, v11 at level 0,
   v12 at level 0).
(*  : ltac_scope.*)

(*SF*)
Notation "'>>' v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 v11 v12 v13" :=
  ((boxer v1)::(boxer v2)::(boxer v3)::(boxer v4)::(boxer v5)
   ::(boxer v6)::(boxer v7)::(boxer v8)::(boxer v9)::(boxer v10)
   ::(boxer v11)::(boxer v12)::(boxer v13)::nil)
  (at level 0, v1 at level 0, v2 at level 0, v3 at level 0,
   v4 at level 0, v5 at level 0, v6 at level 0, v7 at level 0,
   v8 at level 0, v9 at level 0, v10 at level 0, v11 at level 0,
   v12 at level 0, v13 at level 0).
(*  : ltac_scope.*)
 
(*Wyckoff*)
Ltac larg_list :=
  match goal with
    | |- ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11 ?x12 ?x13 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12)
    | |- ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11 ?x12 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12)
    | |- ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11)
    | |- ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10)
    | |- ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9)
    | |- ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8)
    | |- ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7)
    | |- ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 = _ => constr:(>> x1 x2 x3 x4 x5 x6)
    | |- ?f ?x1 ?x2 ?x3 ?x4 ?x5 = _ => constr:(>> x1 x2 x3 x4 x5)
    | |- ?f ?x1 ?x2 ?x3 ?x4 = _ => constr:(>> x1 x2 x3 x4)
    | |- ?f ?x1 ?x2 ?x3 = _ => constr:(>> x1 x2 x3)
    | |- ?f ?x1 ?x2 = _ => constr:(>> x1 x2)
    | |- ?f ?x1 = _ => constr:(>> x1)
    | |- _ => idtac
  end.

(*Wyckoff*)
Ltac rarg_list :=
  match goal with
    | |- _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11 ?x12 ?x13 => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12)
    | |- _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11 ?x12  => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12)
    | |- _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11  => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11)
    | |- _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10  => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10)
    | |- _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9  => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9)
    | |- _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8  => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8)
    | |- _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7  => constr:(>> x1 x2 x3 x4 x5 x6 x7)
    | |- _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6  => constr:(>> x1 x2 x3 x4 x5 x6)
    | |- _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5  => constr:(>> x1 x2 x3 x4 x5)
    | |- _ = ?f ?x1 ?x2 ?x3 ?x4 => constr:(>> x1 x2 x3 x4)
    | |- _ = ?f ?x1 ?x2 ?x3 => constr:(>> x1 x2 x3)
    | |- _ = ?f ?x1 ?x2  => constr:(>> x1 x2)
    | |- _ = ?f ?x1  => constr:(>> x1)
    | |- _ => idtac
  end.


(*Wyckoff*)
Ltac larg_list_h h :=
  let P := type of h in
  match P with
    |  ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11 ?x12 ?x13 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12)
    |  ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11 ?x12 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12)
    |  ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11)
    |  ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10)
    |  ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9)
    |  ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8)
    |  ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 = _ => constr:(>> x1 x2 x3 x4 x5 x6 x7)
    |  ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 = _ => constr:(>> x1 x2 x3 x4 x5 x6)
    |  ?f ?x1 ?x2 ?x3 ?x4 ?x5 = _ => constr:(>> x1 x2 x3 x4 x5)
    |  ?f ?x1 ?x2 ?x3 ?x4 = _ => constr:(>> x1 x2 x3 x4)
    |  ?f ?x1 ?x2 ?x3 = _ => constr:(>> x1 x2 x3)
    |  ?f ?x1 ?x2 = _ => constr:(>> x1 x2)
    |  ?f ?x1 = _ => constr:(>> x1)
    |  _ => idtac
  end.


(*Wyckoff*)
Ltac rarg_list_h h :=
  let P := type of h in
  match P with
    | _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11 ?x12 ?x13 => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12)
    |  _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11 ?x12  => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12)
    |  _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10 ?x11  => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11)
    |  _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9 ?x10  => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9 x10)
    |  _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8 ?x9  => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8 x9)
    |  _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7 ?x8  => constr:(>> x1 x2 x3 x4 x5 x6 x7 x8)
    |  _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6 ?x7  => constr:(>> x1 x2 x3 x4 x5 x6 x7)
    |  _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5 ?x6  => constr:(>> x1 x2 x3 x4 x5 x6)
    |  _ = ?f ?x1 ?x2 ?x3 ?x4 ?x5  => constr:(>> x1 x2 x3 x4 x5)
    |  _ = ?f ?x1 ?x2 ?x3 ?x4 => constr:(>> x1 x2 x3 x4)
    |  _ = ?f ?x1 ?x2 ?x3 => constr:(>> x1 x2 x3)
    |  _ = ?f ?x1 ?x2  => constr:(>> x1 x2)
    |  _ = ?f ?x1  => constr:(>> x1)
    |  _ => idtac
  end.


(*SF*)
Ltac get_fun_arg E :=
  match E with
    | ?X1 ?X2 ?X3 ?X4 ?X5 ?X6 ?X7 ?X8 ?X9 ?X10 ?X11 ?X => constr:((X1 X2 X3 X4 X5 X6 X7 X8 X9 X10 X11,X))
    | ?X1 ?X2 ?X3 ?X4 ?X5 ?X6 ?X7 ?X8 ?X9 ?X10 ?X => constr:((X1 X2 X3 X4 X5 X6 X7 X8 X9 X10,X))
    | ?X1 ?X2 ?X3 ?X4 ?X5 ?X6 ?X7 ?X8 ?X9 ?X => constr:((X1 X2 X3 X4 X5 X6 X7 X8 X9,X))
    | ?X1 ?X2 ?X3 ?X4 ?X5 ?X6 ?X7 ?X8 ?X => constr:((X1 X2 X3 X4 X5 X6 X7 X8,X))
    | ?X1 ?X2 ?X3 ?X4 ?X5 ?X6 ?X7 ?X => constr:((X1 X2 X3 X4 X5 X6 X7,X))
    | ?X1 ?X2 ?X3 ?X4 ?X5 ?X6 ?X => constr:((X1 X2 X3 X4 X5 X6,X))
    | ?X1 ?X2 ?X3 ?X4 ?X5 ?X => constr:((X1 X2 X3 X4 X5,X))
    | ?X1 ?X2 ?X3 ?X4 ?X => constr:((X1 X2 X3 X4,X))
    | ?X1 ?X2 ?X3 ?X => constr:((X1 X2 X3,X))
    | ?X1 ?X2 ?X => constr:((X1 X2,X))
    | ?X1 ?X => constr:((X1,X))
  end.


(*Wyckoff*)
(*This function just "pose proof"s a term with itself.
Used in conjunction with the one next below . . . *)
Ltac eq_refl_t x :=
  let h := fresh "he" in 
  pose proof (eq_refl x) as h.

(*Wyckoff*)
(*. . . which unfolds a global constant in an equality proof like 
[he] (derived from the above function, so that 
Ltac can return a list of arguments in a function application folded 
globally*)
Ltac unfold_larg h x :=
  let la' := fresh "la" in
  unfold x in h; 
    let larg := larg_list_h h in pose larg as la'.

(*Example:say you have a term [x := P a1 a2 a3 a4].
Enter >  eq_refl_t x.
to generate a term [he:x = x]
Then type [unfold_larg he x], to get 
the Boxer list >> a1 a2 a3 a4.
 *)


(*SF*)
Ltac find_head_match T :=
  match T with context [?E] =>
    match T with
    | E => fail 1
    | _ => constr:(E)
    end
  end.

(*SF*)
Ltac false_post :=
  solve [ assumption | discriminate | congruence ].

(*SF*)
Tactic Notation "false_goal" :=
  elimtype False.

(*SF*)
Tactic Notation "false" :=
  false_goal; try false_post.

(*SF*)
Tactic Notation "tryfalse" :=
  try solve [ false ].

(*SF*)
Ltac destruct_if_post := tryfalse.

(*SF*)
Ltac destruct_head_match_core cont :=
  match goal with
  | |- ?T1 = ?T2 => first [ let E := find_head_match T1 in cont E
                          | let E := find_head_match T2 in cont E ]
  | |- ?T1 => let E := find_head_match T1 in cont E
  end;
  destruct_if_post.

(*SF*)
Tactic Notation "destruct_head_match" "as" simple_intropattern(I) :=
  destruct_head_match_core ltac:(fun E => destruct E as I).

(*SF*)
Tactic Notation "destruct_head_match" :=
  destruct_head_match_core ltac:(fun E => destruct E).

(*End Tactics.*) (*Top Level*)