(* Copyright (C) 2014-2019, Daniel Wyckoff, except for the portions so labeleled
which I got from Daniel Schepler*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*Version 0.3 bool2 -- certified -- Coq 8.4 pl4.*)

Require Import ProofIrrelevance.
Require Import FunctionalExtensionality.
Require Export Ensembles.
Require Import NaryFunctions.
Require Import Image.
Require Import EqdepFacts.
Require Import ClassicalChoice.
Require Import Arith.
Require Import Basics.
Require Import Omega. 
Require Import Description.
Require QArith.
Require ZArith.
Require Import Equality.
Require Import Setoid.


Require Export EnsemblesImplicit.
Require Import FunctionProperties.
Require Export Finite_sets.
Require Import ImageImplicit.
Require Import DecidableDec.
Require Import SetUtilities.
Require Import CSB.
Require Import LogicUtilities.
Require Import ArithUtilities.
Require Import EqDec.


(*Wyckoff*)
Lemma type_eq_dec_false :
  type_eq_dec False.
  red; intro; contradiction.
Qed.

(*Wyckoff*)
Lemma unit_eq :
  forall (u u':unit),
    u = u'.
  intros u u'; destruct u, u'; auto.
Qed.


(*Wyckoff*)
(*Equinumerous relation on arbitrary types.*)
Inductive equi (T U:Type) : Prop :=
| equi_intro : forall f:T->U, invertible f -> equi T U.


(*Wyckoff*)
Lemma equi_refl :
  forall T,
    equi T T.
  intro T.
  econstructor.
  apply invertible_id.
Qed.

(*Wyckoff*)
Lemma equi_sym : 
  forall T U,
    equi T U ->
    equi U T.
  intros ? ? h1.
  destruct h1.
  econstructor.
  apply invertible_impl_inv_invertible.

  Grab Existential Variables.
  apply H.
Qed.

   
  
(*Wyckoff*)
Lemma equi_trans :
  forall T U V,
    equi T U ->
    equi U V ->
    equi T V.
  intros ? ? ?  h1 h2;
  destruct h1, h2;
  apply equi_intro with (compose f0 f); apply invertible_compose; auto.
Qed.


(*Wyckoff*)
Lemma some_neq_none :
  forall {T} (x:T),
    Some x <> None.
  intro; discriminate.
Qed.

(*Wyckoff*)
Lemma none_dec :
  forall {T} (o:option T),
    {o = None} + {o <> None}.
  intros T o.
  destruct o.
  right; intro; discriminate.
  left; auto.
Qed.

(*Wyckoff*)
Definition neq_none_inv {T} (o:option T) : o <> None -> T :=
  match o with
    | Some x => fun _ => x
    | None => fun pf => False_rect _ (pf (eq_refl _))
  end.

(*Wyckoff*)
Lemma inj_neq_none_inv :
  forall {T} (o o':option T)
         (pfo: o  <> None)
         (pfo':o' <> None),
    neq_none_inv o pfo = neq_none_inv o' pfo' ->
    o = o'.
  unfold neq_none_inv.
  intros ? o o'; destruct o, o'; intros ho ho' ?; subst; auto;
  [contradict ho'; auto | contradict ho; auto].
Qed.

(*Wyckoff*)
Lemma some_neq_none_inv :
  forall {T} (o:option T) (pf:o <> None),
    Some (neq_none_inv o pf) = o.
  intros T o; destruct o; auto.
  intro h1; contradict h1; auto.
Qed.




(*Wyckoff*)
Lemma simpl_sig : forall {T:Type} (P:T->Prop) (a:{x:T | P x})
                         (pf:P (proj1_sig a)),
  exist _ _ pf = a.
intros T P a h1.
destruct a as [a h2].
simpl.
apply
  proj1_sig_injective.
simpl.
reflexivity.
Qed.



         


(*Many of these functions are copied and pasted from Daniel Schepler's
  "Zorn's Lemma" user contribution.  Those which are are commented 
  "Schepler."  Those which aren't are commented "Wyckoff"*)
Section FiniteTypes.

(*Schepler*)
Inductive FiniteT : Type -> Prop :=
  | empty_finite: FiniteT False
  | add_finite: forall T:Type, FiniteT T -> FiniteT (option T)
  | bij_finite: forall (X Y:Type) (f:X->Y), FiniteT X ->
    invertible f -> FiniteT Y.

(*Wyckoff*)
Lemma FiniteT_functional :
  forall {T U},
    T = U ->
    FiniteT T ->
    FiniteT U.
  intros; subst; auto.
Qed.

(*Schepler*)
Lemma True_finite: FiniteT True.
Proof.
apply bij_finite with (option False)
  (fun _ => I).
constructor; constructor.
exists (True_rect None).
destruct x as [[]|].
remember (True_rect (@None False) I) as LHS.
destruct LHS as [[]|].
reflexivity.

exact (fun y:True => match y with
  | I => refl_equal I
  end).
Qed.


(*Schepler*)
Lemma finite_dec_exists: forall (X:Type) (P:X->Prop),
  FiniteT X -> (forall x:X, {P x} + {~ P x}) ->
  { exists x:X, P x } + { forall x:X, ~ P x }.
Proof.
intros.
apply exclusive_dec.
red; intro.
destruct H0.
destruct H0.
contradiction (H1 x).

revert P X0.
induction H.
right.
destruct x.

intros.
case (IHFiniteT (fun x:T => P (Some x))
  (fun x:T => X0 (Some x))).
left.
destruct H0.
exists (Some x).
assumption.
intro.
case (X0 None).
left.
exists None.
assumption.
right.
destruct x.
apply H0.
assumption.
destruct H0.
intros.
case (IHFiniteT (fun x:X => P (f x))
  (fun x:X => X0 (f x))).
left.
destruct H2.
exists (f x).
assumption.
right.
intro.
rewrite <- H1 with x.
apply H2.
Qed.
 
(*Schepler*)
Lemma finite_dec_forall: forall (X:Type) (P:X->Prop),
  FiniteT X -> (forall x:X, { P x } + { ~ P x }) ->
  { forall x:X, P x } + { exists x:X, ~ P x }.
Proof.
intros.
apply exclusive_dec.
intuition.
destruct H2.
contradiction (H1 x).

revert P X0.
induction H.
left.
destruct x.
intros.
case (IHFiniteT (fun x:T => P (Some x))
  (fun x:T => X0 (Some x))).
intro.
case (X0 None).
left.
destruct x.
apply H0.
assumption.
right.
exists None.
assumption.
right.
destruct H0.
exists (Some x).
assumption.

intros.
destruct H0.
case (IHFiniteT (fun x:X => P (f x))
  (fun x:X => X0 (f x))).
left.
intro y.
rewrite <- H1.
apply H2.
right.
destruct H2.
exists (f x).
assumption.
Qed.


(*Schepler*)
Lemma FiniteT_img: forall (X Y:Type) (f:X->Y),
  FiniteT X -> (forall y1 y2:Y, y1=y2 \/ y1<>y2) ->
  Finite (Im (Full_set _) f).
Proof.
intros.
induction H.
assert (Im (Full_set _) f = Empty_set _).
apply Extensionality_Ensembles.
red; split.
red; intros.
destruct H.
destruct x.
auto with sets.
rewrite H.
constructor.

assert ({exists x:T, f (Some x) = f None} +
        {forall x:T, f (Some x) <> f None}).
apply finite_dec_exists.
assumption.
intro.
apply decidable_dec.
apply H0.
case H1.
intro.
pose (g := fun (x:T) => f (Some x)).
assert (Im (Full_set _) f =
        Im (Full_set _) g).
apply Extensionality_Ensembles.
red; split.
red; intros.
destruct H2.
destruct x.
exists t.
constructor.
assumption.
destruct e.
exists x.
constructor.
transitivity (f None).
assumption.
symmetry; assumption.
red; intros.

destruct H2.
exists (Some x).
constructor.
assumption.
rewrite H2.
apply IHFiniteT.

intros.
pose (g := fun x:T => f (Some x)).
assert (Im (Full_set _) f =
  Add (Im (Full_set _) g) (f None)).
apply Extensionality_Ensembles.
red; split.
red; intros.
destruct H2.
destruct x.
left.
exists t.
constructor.
assumption.
right.
auto with sets.
red; intros.
destruct H2.
destruct H2.
exists (Some x).
constructor.
assumption.
destruct H2.
exists None.
constructor.
reflexivity.
rewrite H2.
constructor.
apply IHFiniteT.
red; intro.
destruct H3.
contradiction (n x).
symmetry; assumption.

pose (g := fun (x:X) => f (f0 x)).
assert (Im (Full_set _) f = Im (Full_set _) g).
apply Extensionality_Ensembles.
red; split.
red; intros.
destruct H2.
destruct H1.
rewrite H3.
rewrite <- H4 with x.
exists (g0 x).
constructor.
unfold g.
reflexivity.
red; intros.
destruct H2.
exists (f0 x).
constructor.
assumption.

rewrite H2.
apply IHFiniteT.
Qed.

(*Schepler*)
Lemma finite_subtype: forall (X:Type) (P:X->Prop),
  FiniteT X -> (forall x:X, P x \/ ~ P x) ->
  FiniteT {x:X | P x}.
Proof.
intros.
induction H.
apply bij_finite with False (False_rect _).
constructor.
exists (@proj1_sig _ _).
destruct x.
intro s; destruct s; destruct x.

destruct (H0 None).
pose (g := fun (x:option {x:T | P (Some x)}) =>
  match x return {x:option T | P x} with
  | Some (exist x0 i) => exist (fun x:option T => P x) (Some x0) i
  | None => exist (fun x:option T => P x) None H1
  end).
apply bij_finite with _ g.
apply add_finite.
apply IHFiniteT.
intro; apply H0.
pose (ginv := fun (s:{x0:option T | P x0}) =>
                       match s return option {x:T | P (Some x)} with
  | exist (Some x0) i => Some (exist (fun y:T => P (Some y)) x0 i)
  | exist None _ => None
  end).
exists ginv.
destruct x as [[x0]|].
simpl.
reflexivity.
simpl.
reflexivity.
destruct y as [[x0|]].
simpl.
reflexivity.
simpl.
destruct (proof_irrelevance _ H1 p).
reflexivity.

pose (g := fun (x:{x:T | P (Some x)}) =>
  match x return {x:option T | P x} with
    | exist x0 i => exist (fun x:option T => P x) (Some x0) i
  end).
apply bij_finite with _ g.
apply IHFiniteT.
intro; apply H0.
pose (ginv := fun s:{x0:option T | P x0} =>
  match s return {x:T | P (Some x)} with
    | exist (Some x0) i => exist (fun x:T => P (Some x)) x0 i
    | exist None i => False_rect _ (H1 i)
  end).
exists ginv.
destruct x; simpl.
reflexivity.
destruct y as [[x0|]].
simpl.
reflexivity.
contradiction H1.

pose (g := fun (x:{x:X | P (f x)}) =>
  match x with
  | exist x0 i => exist (fun x:Y => P x) (f x0) i
        end).
apply bij_finite with _ g.
apply IHFiniteT.
intro; apply H0.
destruct H1.
assert (forall y:Y, P y -> P (f (g0 y))).
intros; rewrite H2; assumption.
pose (ginv := fun (y:{y:Y | P y}) =>
  match y with
  | exist y0 i => exist (fun x:X => P (f x)) (g0 y0) (H3 y0 i)
  end).
exists ginv.
destruct x; simpl.
generalize (H3 (f x) p).
rewrite H1.
intro; destruct (proof_irrelevance _ p p0).
reflexivity.

destruct y; simpl.
generalize (H3 x p).
rewrite H2.
intro; destruct (proof_irrelevance _ p p0).
reflexivity.
Qed.

(*Wyckoff*)                   
Lemma FiniteT_Finite : forall (T:Type), FiniteT T -> 
  (forall (S:Ensemble T), Finite S).
  intros T h1 S.
assert (h2: forall x:T, In S x \/ ~In S x).
  intro. tauto.
pose proof (finite_subtype T _ h1 h2) as h3.
assert (h4: forall y1 y2 : T, y1 = y2 \/ y1 <> y2).
  intros. tauto.
pose proof 
  (FiniteT_img (sig_set S) T (@proj1_sig T _) h3 h4) as h5.
assert (h6: (Im (full_sig S) (proj1_sig (P:=fun x : T => In S x)))
  = S).
  apply Extensionality_Ensembles. red. unfold Included.
  split.
  (*left*)
  intros x h7.
  inversion h7 as [y h8 z h9].
  pose proof (proj2_sig y) as h10.
  rewrite <- h9 in h10.
  assumption.
  (*right*)
  intros x h11.
  apply Im_intro with (exist _ x h11).
  apply Full_intro.
  simpl. reflexivity.
unfold full_sig in h6. rewrite h6 in h5.
assumption.
Qed.

(*Wyckoff*)
Lemma FiniteT_sig_Finite : forall {T:Type} (S:Ensemble T), 
  FiniteT (sig_set S) -> Finite S.
intros T S h1.
pose proof (FiniteT_img _ _ (@proj1_sig _ _) h1) as h2.
assert (h3: forall y1 y2 : T, y1 = y2 \/ y1 <> y2).
  intros. apply classic.
apply h2 in h3.
assert (h4:(Im (full_sig S)
  (proj1_sig (P:=fun x : T => In S x))) = 
    S).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x h4.
  inversion h4 as [? ? ? h5].
  rewrite h5.
  apply proj2_sig.
  (* >= *)
  red. intros x h4.
  eapply Im_intro.
  apply (Full_intro _ (exist _ x h4)).
  reflexivity.
unfold full_sig in h4. unfold sig_set in h4.
rewrite h4 in h3; assumption.
Qed.

(*Schepler*)
Lemma Finite_ens_type: forall {X:Type} (S:Ensemble X),
  Finite S -> FiniteT (sig_set S).
Proof.
intros.
induction H.
apply bij_finite with False (False_rect _).
constructor.
assert (g:{x:X | In (Empty_set _) x}->False).
intro.
destruct X0.
destruct i.
exists g.
destruct x.
destruct y.
destruct g.

assert (Included A (Add A x)).
auto with sets.
assert (In (Add A x) x).
auto with sets.
pose (g := fun (y: option {x:X | In A x}) =>
  match y return {x0:X | In (Add A x) x0} with
                                          | Some (exist y0 i) => exist (fun x2:X => In (Add A x) x2) y0 (H1 y0 i)
  | None => exist (fun x2:X => In (Add A x) x2) x H2
  end).
apply bij_finite with _ g.
apply add_finite.
assumption.

assert (h:forall x0:X, In (Add A x) x0 ->
  { In A x0 } + { x0 = x }).
intros; apply exclusive_dec.
intuition.
destruct H6; auto.
destruct H3.
left; assumption.
right; destruct H3; reflexivity.

pose (ginv := fun s:{x0:X | In (Add A x) x0} =>
  match s return option {x:X | In A x} with
  | exist x0 i => match (h x0 i) with
                  | left iA => Some (exist _ x0 iA)
                  | right _ => None
   end
  end).
exists ginv.
intro; destruct x0.
destruct s.
simpl.
remember (h x0 (H1 x0 i)) as sum; destruct sum.
destruct (proof_irrelevance _ i i0).
reflexivity.
contradiction H0.
rewrite <- e; assumption.
simpl.
remember (h x H2) as sum; destruct sum.
contradiction H0.
reflexivity.
                   
intro.
unfold ginv.
destruct y.
destruct (h x0 i).
simpl.
generalize (H1 x0 i0); intro.
destruct (proof_irrelevance _ i i1).
reflexivity.
simpl.
destruct e.
destruct (proof_irrelevance _ H2 i).
reflexivity.
Qed.

(*Wyckoff*)
Lemma FiniteT_Full : forall T:Type, 
  FiniteT T <->  FiniteT {x:T | In (Full_set T) x}. 
intros T.
pose (fun (a:T) => exist _ a (Full_intro _ a)) as f.
assert (h2:invertible f).
  apply bijective_impl_invertible.
  red. split.
  (*inj*)
  red.
  intros x1 x2 h3.
  unfold f in h3.
  assert (h4: x1 = proj1_sig 
    (exist (In (Full_set T)) x1 (Full_intro T x1))).
    simpl. reflexivity.
  assert (h5: x2 = proj1_sig
   (exist (In (Full_set T)) x2 (Full_intro T x2))).
    simpl. reflexivity.
  congruence.
  (*surj*)
  red.
  intro y.
  exists (proj1_sig y).
  destruct y as [z h1].
  unfold f.
  simpl.
  pose proof 
    (proof_irrelevance _ h1 (Full_intro T z)) as h2.
  rewrite h2. 
  reflexivity.

pose (fun (a:{x:T | In (Full_set T) x}) => proj1_sig a) as f'.
assert (h3:invertible f').
  apply bijective_impl_invertible.
  red. split.
  (*inj*)
  red.
  intros x1 x2 h3.
  unfold f' in h3.
  apply proj1_sig_injective; trivial.
  (*surj*)
  red.
  intro y.
  exists (exist _ y (Full_intro _ y)).
  unfold f'.
  simpl.
  reflexivity.

split.
intro h1.  apply (bij_finite _ _ f);  trivial.
intro h1.  apply (bij_finite _ _ f'); trivial.
Qed.

(*Wyckoff*)
Lemma Finite_FiniteT : forall (T:Type), 
  Finite (Full_set T) -> FiniteT T.
intros T h1.
rewrite FiniteT_Full.
apply Finite_ens_type; trivial.
Qed.

(* Wyckoff *)
Lemma Finite_FiniteT_iff : forall (T:Type), 
  Finite (Full_set T) <-> FiniteT T.
intro T. split. apply Finite_FiniteT.
intro h1.
apply FiniteT_Finite.
assumption.
Qed.


(* Wyckoff *)
Lemma finite_full_sig_iff : 
  forall {T:Type} (A:Ensemble T),
    Finite A <-> Finite (full_sig A).
intros T A.
split.
intro h1.
pose proof (Finite_ens_type _ h1) as h2.
rewrite <- Finite_FiniteT_iff in h2.
assumption.
intro h1.
rewrite Finite_FiniteT_iff in h1.
apply FiniteT_sig_Finite. assumption.
Qed.


(*Wyckoff*)
Lemma finite_subset_sig_iff : forall {T:Type} (A B:Ensemble T) 
                                 (pf:Included A B),
         Finite A <-> Finite (subset_sig _ _ pf).
intros T A B h1.
split.
intro h2.
red in h1.
unfold subset_sig.
apply finite_image.
rewrite <- finite_full_sig_iff.
assumption.
intro h2.  unfold subset_sig in h2. 
assert (h3:Injective  (fun x : {x : T | In A x} =>
             exist (In B) (proj1_sig x) (h1 (proj1_sig x) (proj2_sig x)))).
  red. intros x1 x2 h3.
  pose proof (exist_injective _ _ _ _ _ h3) as h4.
  apply proj1_sig_injective. assumption.
pose proof (finite_image_rev_inj _ _ h3 h2) as h4.
rewrite finite_full_sig_iff.
assumption.
Qed.

(*Wyckoff*)
Lemma finite_incl_sig_iff :
  forall {T:Type} (A B:Ensemble T) (pf:Included A B),
         Finite (incl_sig A B pf) <->
         Finite A.
intros T A B h1.
rewrite incl_sig_eq_subset_sig.
symmetry.
apply finite_subset_sig_iff.
Qed.


(*Wyckoff*)
Lemma card_fun_full_sig_eq :
  forall {T:Type} (A:Ensemble T) (pf:Finite A),
    card_fun A pf = card_fun (full_sig A) (iff1 (finite_full_sig_iff A) pf).
intros T A h1.
pose proof (card_fun_compat _ h1) as h2.
pose proof (card_fun_compat _ (iff1 (finite_full_sig_iff A) h1)) as h3.
pose proof (proj1_sig_injective (fun x=>In A x)) as h4.
eapply injective_preserves_cardinal. red. apply h4.
apply h3.
pose proof (im_full_sig_proj1_sig A) as h5. 
rewrite <- h5 at 1.
assumption.
Qed.


(*Schepler*)
Lemma finite_eq_dec: forall X:Type, FiniteT X ->
  forall x y:X, {x=y} + {x<>y}.
Proof.
intros.
apply decidable_dec.
induction H.
destruct x.
decide equality.

destruct H0.
case (IHFiniteT (g x) (g y)).
left.
rewrite <- H1.
rewrite <- H1 with x.
rewrite H2.
reflexivity.
right.
contradict H2.
rewrite H2.
reflexivity.
Qed.

(*Schepler*)
Lemma finite_inj_surj: forall (X:Type) (f:X->X),
  FiniteT X -> Injective f -> surjective f.
Proof.
intros.
induction H.
red.
destruct y.

remember (f None) as f0; destruct f0 as [a|].
assert (forall x:T, f (Some x) <> Some a).
unfold not; intros.
assert (Some x = None).
apply H0.
congruence.
discriminate H2.
pose (g := fun x:T => match f (Some x) with
  | Some y => y
  | None => a
end).
assert (surjective g).
apply IHFiniteT.
red; intros.
remember (f (Some x1)) as fx1; destruct fx1;
remember (f (Some x2)) as fx2; destruct fx2.
unfold g in H2.
rewrite <- Heqfx1 in H2; rewrite <- Heqfx2 in H2.
destruct H2; assert (f (Some x1) = f (Some x2)).
congruence.
apply H0 in H2.
injection H2; trivial.

unfold g in H2; rewrite <- Heqfx1 in H2; rewrite <- Heqfx2 in H2.
destruct H2.
contradiction (H1 x1).
symmetry; assumption.

unfold g in H2; rewrite <- Heqfx1 in H2; rewrite <- Heqfx2 in H2.
destruct H2.
contradiction (H1 x2).
symmetry; assumption.
assert (Some x1 = Some x2).
apply H0.
congruence.
injection H3; trivial.

red; intro.
destruct y.
case (finite_eq_dec _ H t a).
exists None.
congruence.
destruct (H2 t).
exists (Some x).
unfold g in H3.
destruct (f (Some x)).
congruence.
contradiction n.
symmetry; assumption.
destruct (H2 a).
exists (Some x).
unfold g in H3.
remember (f (Some x)) as fx; destruct fx.
destruct H3.
contradiction (H1 x).
symmetry; assumption.
reflexivity.

assert (forall x:T, { y:T | f (Some x) = Some y }).
intros.
remember (f (Some x)) as fx; destruct fx.
exists t; reflexivity.
assert (Some x = None).
apply H0.
congruence.
discriminate H1.
pose (g := fun x:T => proj1_sig (X x)).
assert (surjective g).
apply IHFiniteT.
red; intros.
unfold g in H1.
repeat destruct X in H1.
simpl in H1.
assert (Some x1 = Some x2).
apply H0.
congruence.
injection H2; trivial.

red; intro.
destruct y.
destruct (H1 t).
unfold g in H2; destruct X in H2.
simpl in H2.
exists (Some x).
congruence.
exists None.
symmetry; assumption.

destruct H1.

pose (f' := fun (x:X) => g (f (f0 x))).
assert (surjective f').
apply IHFiniteT.
red; intros.
unfold f' in H3.
assert (f (f0 x1) = f (f0 x2)).
congruence.
apply H0 in H4.
congruence.

red; intro.
destruct (H3 (g y)).
unfold f' in H4.
exists (f0 x).
congruence.
Qed.

(*Schepler *)
Lemma surj_finite: forall (X Y:Type) (f:X->Y),
  FiniteT X -> surjective f ->
  (forall y1 y2:Y, y1=y2 \/ y1<>y2) ->
  FiniteT Y.
Proof.
intros.
apply bij_finite with {y:Y | In (Im (Full_set _) f) y}
  (@proj1_sig _ (fun y:Y => In (Im (Full_set _) f) y)).
apply Finite_ens_type.
apply FiniteT_img.
assumption.
assumption.
assert (forall y:Y, In (Im (Full_set _) f) y).
intro.
destruct (H0 y).
exists x; auto with sets.
constructor.

pose (proj1_sig_inv := fun y:Y =>
  exist (fun y0:Y => In (Im (Full_set _) f) y0) y (H2 y)).
exists proj1_sig_inv.
destruct x.
simpl.
unfold proj1_sig_inv.
destruct (proof_irrelevance _ (H2 x) i); trivial.
intros; simpl; reflexivity.
Qed.

(*Schepler*)
Lemma inj_finite: forall (X Y:Type) (f:X->Y),
  FiniteT Y -> Injective f ->
  (forall y:Y, (exists x:X, f x = y) \/
               (~ exists x:X, f x = y)) ->
  FiniteT X.
Proof.
intros.
assert (forall y:{y:Y | exists x:X, f x = y}, {x:X | f x = proj1_sig y}).
intro.
destruct y.
simpl.

apply constructive_definite_description.
destruct e.
exists x0.
red; split.
assumption.
intros.
apply H0.
transitivity x.
assumption.
symmetry; assumption.

pose (g := fun y:{y:Y | exists x:X, f x = y} =>
  proj1_sig (X0 y)).
apply bij_finite with _ g.
apply finite_subtype.
assumption.
assumption.

pose (ginv := fun (x:X) => exist (fun y:Y => exists x:X, f x = y)
  (f x) (ex_intro _ x (refl_equal _))).
exists ginv.
destruct x as [y [x e]].
unfold g; simpl.
match goal with |- context [X0 ?arg] => destruct (X0 arg) end.
simpl.
unfold ginv; simpl.
simpl in e0.
repeat match goal with |- context [ex_intro ?f ?x ?e] =>
  generalize (ex_intro f x e) end.
rewrite <- e0.
intros; destruct (proof_irrelevance _ e1 e2).
reflexivity.

intro; unfold ginv.
unfold g; simpl.
match goal with |- context [X0 ?arg] => destruct (X0 arg) end.
simpl.
simpl in e.
auto.
Qed. 

(*Wyckoff*)
Lemma finite_sig_sat_compat :
  forall {T:Type} (P:T->Prop),
    FiniteT {x:T | P x} <-> Finite [x:T | P x].
intros T P. split.
intro h1.
rewrite finite_full_sig_iff. unfold full_sig. unfold sig_set.
rewrite Finite_FiniteT_iff.
pose (fun a:{x:T| In [x0:T | P x0] x} => 
        (exist _ (proj1_sig a) 
               (match (proj2_sig a) with 
                  |intro_characteristic_sat pf => pf
                end))) as f.
assert (h2:Injective f).
  red. intros a b h2. destruct a; destruct b.
  destruct i; destruct i0. unfold f in h2. simpl in h2.
  apply proj1_sig_injective. simpl.
  apply exist_injective in h2. assumption.
eapply inj_finite; auto. apply h1. apply h2. intros; apply classic.  
intro h2.
rewrite finite_full_sig_iff in h2. unfold full_sig in h2. unfold sig_set in h2.
rewrite Finite_FiniteT_iff in h2.
pose (fun a:{x:T|P x} => 
        exist _ (proj1_sig a)
              (intro_characteristic_sat _ _ (proj2_sig a))) as f.
assert (h1:Injective f).
  red. intros a b h1. destruct a; destruct b. unfold f in h1.
  apply exist_injective in h1. simpl in h1.
  apply proj1_sig_injective. simpl. assumption.
eapply inj_finite; auto. apply h2. apply h1. intros; apply classic.
Qed.


(* Wyckoff *)
Lemma cart_prod_fin_comm : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
          Finite (cart_prod A B) ->
          Finite (cart_prod B A).
intros T U A B h1.
apply Finite_ens_type in h1.
pose (fun pr:{p:U*T | In (cart_prod B A) p} => 
        exist _ (snd (proj1_sig pr), fst (proj1_sig pr)) 
              (in_cart_prod_comm B A (proj1_sig pr)
               (proj2_sig pr))) as f.
assert (h2:Injective f).
  red. unfold f.
  intros pr1 pr2 h2.
  destruct pr1 as [p1 h3].
  destruct pr2 as [p2 h4].
  simpl in h2.
  apply existTexist.
  apply subsetT_eq_compat.
  inversion h2 as [h5].
  rewrite surjective_pairing.
  rewrite <- h5.
  rewrite <- H.
  apply surjective_pairing.
apply FiniteT_sig_Finite.
apply (inj_finite _ _ f h1 h2).
intro pr.
destruct pr as [p h4].
unfold f.
left.
exists (exist (fun p:U*T => In (cart_prod B A) p) (snd p, fst p) (in_cart_prod_comm A B p h4)).
simpl.
apply existTexist.
apply subsetT_eq_compat.
rewrite surjective_pairing.
reflexivity.
Qed.



(*Wyckoff*)
Lemma FiniteT_Finite_Union : forall {It T:Type} 
  (F:IndexedFamily It T),   (forall i:It, Finite (F i)) 
  -> FiniteT It -> Finite (IndexedUnion F).
Proof.
intros It T F h1 h2.
induction h2.
(*Empty*)
assert (h3:(IndexedUnion F) = Empty_set _ ).
  apply Extensionality_Ensembles.
  red. split. 
  (* <= *)
  red. intros x h4.
  inversion h4.
  contradiction.
  (* >= *)
  red. intros; contradiction.
rewrite h3. constructor.
(*Add*)
rename T0 into It.
pose (fun (i:It) => F (Some i)) as f.
assert (h4:(IndexedUnion F) = Union (F None) (IndexedUnion f)).
  apply Extensionality_Ensembles.
  red. split.
  (*<=*)
  red. intros x h5.   
  inversion h5 as [i ? h6].
  destruct i.
    (*Some*)
    right.
    unfold f.
    apply indexed_union_intro with i. assumption.
    (*None*)
    constructor. assumption.
  (*>=*)
  red. intros x h5.
  pose proof (Union_inv _ _ _ x h5) as h6.
  case h6 as [h7 | h8].
    apply indexed_union_intro with None; assumption.
  inversion h8.
    apply indexed_union_intro with (Some i); assumption.
assert (h5: forall i:It, Finite (f i)).
  intro i.
  apply (h1 (Some i)).
pose proof (IHh2 f h5) as h9.
assert (h10: Finite (F None)).
  apply (h1 None).
rewrite h4.
apply Union_preserves_Finite; assumption.
(*Bij*)
rename Y into It.
rename H into h3.
pose (fun (x:X) => F (f x)) as Ff.
assert (h4: forall x:X, Finite (Ff x)).
  intro x.
  apply (h1 (f x)).
pose proof (IHh2 Ff h4) as h5.
assert (h6:IndexedUnion Ff = IndexedUnion F).
  apply Extensionality_Ensembles.
  red. split.
  (*<=*)
  red.
  intros x h7.
  inversion h7 as [? ? h8].
  unfold Ff in h8.
  apply indexed_union_intro with (f i); assumption.
  (*>=*)  
  pose proof (invertible_impl_bijective _ h3) as h6.
  red. intros t h7.
  inversion h7 as [i y h8].
  unfold bijective in h6.
  destruct h6 as [h9 h10].
  red in h10.
  pose proof (h10 i) as h11.
  elim h11. intros x h12.
  rewrite <- h12 in h8.
  unfold Ff.
  apply indexed_union_intro with x; assumption.
rewrite <- h6.
assumption.
Qed.

(*Wyckoff*)
Lemma bool_finite : FiniteT bool.
Proof.
pose (fun x:(option (option False)) => 
        match x with
          | None => false
          | Some y => match y with 
                          | None => true
                          | Some z => (False_rec bool z)
                      end
        end) as f.
apply (bij_finite _ _ f); repeat constructor.
apply bijective_impl_invertible.
red. split.
(*injective*)
red.
intros a b h1.
unfold f in h1.
destruct a as [a2 | a3]; destruct b as [b2 | b3].
destruct a2 as [a2a | a2b]; destruct b2 as [b2a | b2b]; 
try contradiction; try reflexivity.
destruct a2 as [a2a | a2b]; try contradiction; try discriminate.
destruct b2 as [b2a | b2b]; try contradiction; try discriminate.
reflexivity.
(*surjective*)
red.
intro x.
unfold f.
destruct x. exists (Some None). reflexivity.
exists None. reflexivity.
Qed.

(*Wyckoff*)
Lemma bool_eq_true_iff :
  forall v,
    v = true <->
    v <> false.
  intro v; destruct v; intuition.
Qed.

(*Wyckoff*)
Lemma bool_eq_false_iff :
  forall v,
    v = false <->
    v <> true.
  intro v; destruct v; intuition.
Qed.

(*Wyckoff*)
Inductive bool_lt : bool -> bool -> Prop :=
| bool_lt_intro : bool_lt false true.

(*Wyckoff*)
Lemma tso_dec_bool_lt :
  tso_dec bool_lt.
  red; intros x y; destruct x, y.
  left; right; auto.
  right; constructor.
  left; left; constructor.
  left; right; auto.
Qed.

(*Wyckoff*)
Lemma total_strict_order_bool_lt :
  total_strict_order bool_lt.
  red; intros x y; destruct x, y.
  right; left; auto.
  right; right; constructor.
  left; constructor.
  right; left; auto.
Qed.

(*Wyckoff*)
Lemma transitive_bool_lt :
  transitive _ bool_lt.
  red; intros x y z h1 h2; destruct x, y, z; try inversion h1;
  try inversion h2.
Qed.

(*Wyckoff*)
Lemma irrefl_bool_lt :
  irrefl bool_lt.
  red; intro b; destruct b; intro h1; inversion h1.
Qed.

(*Wyckoff*)
Lemma sortable_dec_bool_lt :
  sort_dec bool_lt.
  constructor.
  apply tso_dec_bool_lt.
  apply transitive_bool_lt.
  apply irrefl_bool_lt.
Qed.


(*Wyckoff*)
Lemma sortable_bool_lt :
  sortable bool_lt.
  constructor.
  apply total_strict_order_bool_lt.
  apply transitive_bool_lt.
  apply irrefl_bool_lt.
Qed.
                          

(*Wyckoff*)
Fixpoint bool_le (p q:bool) : Prop :=
  if p then
    if q then
      True
    else
      False
  else
    True.

(*Wyckoff*)
Lemma false_bool_le :
  forall q,
    bool_le false q.
  intro q; destruct q; auto; simpl; auto.
Qed.

(*Wyckoff*)
Lemma bool_le_true :
  forall p,
    bool_le p true.
  intro p; destruct p; auto; simpl; auto.
Qed.

(*Wyckoff*)
Lemma bool_le_dec :
  forall v w,
    {bool_le v w} + {bool_le w v}.
  intros v w; destruct v, w; simpl.
  left; auto.
  right; auto.
  left; auto.
  right; auto.
Qed.

(*Wyckoff*)
Lemma total_strict_order_bool_le :
  total_strict_order bool_le.
  red.
  intros x y; destruct x, y; simpl.
  left; auto.
  right; auto.
  left; auto.
  left; auto.
Qed.


(*Wyckoff*)
Lemma transitive_bool_le :
  transitive _ bool_le.
  red; intros x y z; destruct x, y, z; simpl; auto.
Qed.


(*Wyckoff*)
Lemma eq_true_iff :
  forall b,
    b = true <->
    b <> false.
  intro b; destruct b; split; intro h1; try intro h2; try discriminate; auto.
Qed.

(*Wyckoff*)
Lemma eq_false_iff :
  forall b,
    b = false <->
    b <> true.
  intro b; destruct b; split; intro h1; try intro h2; try discriminate; auto.
  contradict h1; auto.
Qed.
  


(*Schepler*)  
Lemma finite_dep_choice: forall (A:Type) (B:forall x:A, Type)
  (R:forall x:A, B x->Prop),
  FiniteT A -> (forall x:A, exists y:B x, R x y) ->
  exists f:(forall x:A, B x), forall x:A, R x (f x).
Proof.
intros.
revert B R H0.
induction H.
intros.
exists (fun x:False => False_rect (B x) x).
destruct x.
intros.
pose proof (IHFiniteT (fun x:T => B (Some x))
  (fun x:T => R (Some x))
  (fun x:T => H0 (Some x))).
destruct H1.
pose proof (H0 None).
destruct H2.
exists (fun y:option T =>
  match y return (B y) with
  | Some y0 => x y0
  | None => x0
  end).
destruct x1.
apply H1.
assumption.

intros.
destruct H0.
pose proof (IHFiniteT (fun x:X => B (f x))
  (fun x:X => R (f x))
  (fun x:X => H1 (f x))).
destruct H3.
pose (f0 := fun y:Y => x (g y)).
pose (conv := fun (y:Y) (a:B (f (g y))) =>
  eq_rect (f (g y)) B a y (H2 y)).

exists (fun y:Y => conv y (x (g y))).
intro.
unfold conv; simpl.
generalize (H2 x0).
pattern x0 at 2 3 6.
rewrite <- H2.
intro.
rewrite <- eq_rect_eq.
apply H3.
Qed.

(*Schepler*)
Lemma finite_choice : forall (A B:Type) (R:A->B->Prop),
  FiniteT A -> (forall x:A, exists y:B, R x y) ->
  exists f:A->B, forall x:A, R x (f x).
Proof.
intros.
apply finite_dep_choice.
assumption.
assumption.
Qed.

(*Schepler*)
Lemma finite_sum: forall X Y:Type, FiniteT X -> FiniteT Y ->
  FiniteT (X+Y).
Proof.
intros.
induction H0.
apply bij_finite with _ inl.
assumption.
pose (g := fun (x:X+False) => match x with
  | inl x => x
  | inr f => False_rect X f
end).
exists g.
intro; simpl.
reflexivity.
destruct y.
simpl.
reflexivity.
destruct f.

pose (g := fun (x:option (X+T)) => match x with
  | Some (inl x) => inl _ x
  | Some (inr t) => inr _ (Some t)
  | None => inr _ None
  end).
apply bij_finite with _ g.
apply add_finite.
assumption.
pose (ginv := fun (x:X + option T) => match x with
  | inl x => Some (inl _ x)
  | inr (Some t) => Some (inr _ t)
  | inr None => None
  end).
exists ginv.
destruct x as [[x|t]|]; trivial.
destruct y as [x|[t|]]; trivial.

pose (g := fun (x:X+X0) => match x with
  | inl x0 => inl _ x0
  | inr x0 => inr _ (f x0)
  end).
destruct H1.
pose (ginv := fun (x:X+Y) => match x with
  | inl x0 => inl _ x0
  | inr y0 => inr _ (g0 y0)
  end).
apply bij_finite with _ g.
assumption.
exists ginv.
destruct x as [x0|x0]; trivial.
simpl.
rewrite H1; reflexivity.
destruct y as [x|y0]; trivial.
simpl.
rewrite H2; reflexivity.
Qed.


(*Schepler*)
Lemma finite_prod: forall (X Y:Type), FiniteT X -> FiniteT Y ->
  FiniteT (X*Y).
Proof.
intros.
induction H0.
apply bij_finite with _ (False_rect _).
constructor.
exists (@snd X False).
destruct x.
destruct y.
destruct f.

pose (g := fun (x:X*T + X) => match x with
  | inl (pair x0 t) => pair x0 (Some t)
  | inr x0 => pair x0 None
  end).
pose (ginv := fun (x:X * option T) => match x with
  | (x0, Some t) => inl _ (x0, t)
  | (x0, None) => inr _ x0
  end).
apply bij_finite with _ g.
apply finite_sum.
assumption.
assumption.
exists ginv.
destruct x as [[x0 t]|x0]; trivial.
destruct y as [x0 [t|]]; trivial.

pose (g := fun (y:X*X0) => match y with
  | pair x x0 => pair x (f x0)
  end).
destruct H1.
pose (ginv := fun (y:X*Y) => let (x,y0) := y in
  (x, g0 y0)).
apply bij_finite with _ g.
assumption.
exists ginv.
destruct x as [x x0]; unfold ginv, g; try rewrite H1; trivial.
destruct y as [x y]; unfold ginv, g; try rewrite H2; trivial.
Qed.


(*Schepler*)
Lemma finite_exp: forall X Y:Type, FiniteT X -> FiniteT Y ->
  FiniteT (X->Y).
Proof.
intros.
induction H.
pose (g := fun (x:True) (f:False) => False_rect Y f).
pose (ginv := fun (_:False->Y) => I).
apply bij_finite with _ g.
apply True_finite.
exists ginv.
destruct x as [].
trivial.
intro y.
apply functional_extensionality.
intro; contradiction.

pose (g := fun (p:(T->Y)*Y) (x:option T) =>
  let (f,y0) := p in
  match x with
  | Some x0 => f x0
  | None => y0
  end).
pose (ginv := fun (f:option T->Y) =>
  (fun x:T => f (Some x), f None)).
apply bij_finite with _ g.
apply finite_prod.
assumption.
assumption.
exists ginv.
destruct x as [f y0]; try extensionality t;
try destruct t as [t0|]; trivial.
intro.
extensionality t; destruct t as [t0|]; trivial.

destruct H1.
pose (g0 := fun (h:X->Y) (y:Y0) => h (g y)).
apply bij_finite with _ g0.
assumption.
pose (g0inv := fun (h:Y0->Y) (x:X) => h (f x)).
exists g0inv.
intro.
extensionality x0; unfold g0; unfold g0inv; simpl.
rewrite H1; reflexivity.
intro.
extensionality y0; unfold g0; unfold g0inv; simpl.
rewrite H2; reflexivity.
Qed.


(*Wyckoff*)
(*Uses excluded middle*)
Lemma power_set_finitet : forall {T:Type} (A:Ensemble T),
                           Finite A -> 
                           FiniteT {S:Ensemble T | Included S A}.
intros T A h1.
assert (h2: (forall (S:Ensemble T) (x:T), 
                              In A x -> {In S x} + {~In S x})).
  intros; apply classic_dec.
pose (fun S:{S':(Ensemble T) | Included S' A} => 
        (fun x:{t:T | In A t} => 
           if (h2 (proj1_sig S) (proj1_sig x) (proj2_sig x)) 
           then true 
           else false)) as f.
pose proof (Finite_ens_type _ h1) as h3.
pose proof bool_finite as h4.
pose proof (finite_exp _ _ h3 h4) as h5.
apply (inj_finite _ _ f). assumption.
red.
intros S1 S2 h6.
apply proj1_sig_injective.
apply Extensionality_Ensembles.
red. split.
  (* -> *)
  red.
  intros x h7.
  pose proof (proj2_sig S1) as h11.
  simpl in h11.
  assert (h12:In A x). auto with sets.
  pose (exist _ _ h12) as x'.
  assert (h13: f S1 x' = f S2 x').  rewrite h6. reflexivity.
  unfold f in h13.
  pose proof (conditional_correspondence _ _ _ _ h13) as h14.
  simpl in h14.
  tauto.
  (* <- *)
  red.
  intros x h7.
  pose proof (proj2_sig S2) as h11.
  simpl in h11.
  assert (h12:In A x). auto with sets.
  pose (exist _ _ h12) as x'.
  assert (h13: f S1 x' = f S2 x').  rewrite h6. reflexivity.
  unfold f in h13.
  pose proof (conditional_correspondence _ _ _ _ h13) as h14.
  simpl in h14.
  tauto.

intros; tauto.
Qed.

(*Wyckoff*)
Lemma power_set_finite : 
  forall {T:Type} (A:Ensemble T),
    Finite A -> Finite (power_set A).
intros T A h1.
pose proof (power_set_finitet A h1) as h2.
unfold power_set.
rewrite <- finite_sig_sat_compat.
assumption.
Qed.

(*Use is made of Full_set rather than T to exploit previous
  theorems easier.*)
(*Wyckoff*)
Lemma bij_seg_finite_full_iff : 
  forall {T:Type} 
         (f:sig_set (seg (card_fun1 (Full_set T))) -> 
            sig_set (Full_set _)),
    bijective f ->
    forall (P:T->Prop),
      (Finite [x:T | P x] <-> 
       Finite [n : nat
        | exists pf : In (seg (card_fun1 (Full_set T))) n,
            P
              (proj1_sig
                 (f (exist _ _ pf)))]).
  intros T f h1 P.
  match goal with
      |- _ <-> Finite ?S' =>
      pose S' as S end.
  fold S.
assert (h2:forall n:nat,
             In S n ->  In (seg (card_fun1 (Full_set T))) n).
  intros n h2. destruct h2 as [h2].
  destruct h2; auto.
assert (h3:forall (n:nat) 
                  (pf:In S n),
             P
               (proj1_sig
                  (f (exist (In (seg (card_fun1 (Full_set T)))) n (h2 n pf))))).
  intros n h3.
  destruct h3 as [h3]. destruct h3 as [h3 h4].
  gterm0.
  redterm0 c.
  redterm0 c0.
  redterm0 c1.
  clear c1 c0 c.
  pose proof (proof_irrelevance _ h3 c2) as h5.
  fold c2.
  rewrite <- h5 at 1.
  assumption.
  pose (fun a:sig_set S =>
          exist _ _ (intro_characteristic_sat P _ (h3 _ (proj2_sig a))))
    as ff.
apply bijective_impl_invertible in h1.
pose proof (proj2_sig (function_inverse f h1)) as h7.
destruct h7 as [h7 h8].
pose (proj1_sig (function_inverse _ h1)) as f'.
assert (h6:forall a:sig_set [x:T | P x],
             P (proj1_sig (f (exist _ _ (proj2_sig (f' (exist _ _ (Full_intro _ (proj1_sig a))))))))).
  intro a. 
  destruct a as [a h0].
  unfold f'. simpl.
  rewrite <- unfold_sig.
  rewrite h8. simpl.
  destruct h0. assumption.
assert (h9:forall a:sig_set [x:T | P x],
             In S (proj1_sig
                     (f' (exist _ _ (Full_intro _ (proj1_sig a)))))).
  intro a. constructor.
  exists (proj2_sig _).
  apply h6.
pose (fun a:sig_set [x:T | P x] =>
        (exist _ _ (h9 a))) as ff'.
assert (h4:Im (full_sig S) ff = full_sig [x:T | P x]).
  apply Extensionality_Ensembles.
  red; split; red; intros x h5. 
  unfold full_sig. constructor.
  apply Im_intro with (ff' x).
  constructor.
  unfold ff, ff'. simpl.
  apply proj1_sig_injective.
  simpl.
  unfold f'.
  gtermr.
  redterm0 y.
  redterm0 c.
  redterm0 c0.
  clear c0 c y.
  fold c1.
  pose proof (proof_irrelevance _ c1 (proj2_sig _)) as h10.
  rewrite h10 at 1.
  rewrite <- unfold_sig.
  rewrite h8.
  simpl. reflexivity.
assert (h5:Im (full_sig [x:T | P x]) ff' = 
           full_sig S).
  apply Extensionality_Ensembles.
  red; split; red; intros x h5.
  unfold full_sig. constructor.
  apply Im_intro with (ff x).
  constructor.
  unfold ff, ff'. simpl.
  apply proj1_sig_injective.
  simpl.
  unfold f'.
  gtermr.
  redterm0 y.
  redterm0 c.
  redterm0 c0.
  clear c0 c y.
  pose proof (proof_irrelevance _ c1 (proj2_sig _)) as h10.
  fold c1.
  rewrite h10 at 1.
  rewrite <- unfold_sig.
  rewrite h7.
  simpl. reflexivity.
pose proof (invertible_impl_bijective _ h1) as h13.
destruct h13 as [h13 h14].
assert (h11:Injective ff).
  red. intros x y h12.
  destruct x as [x h15], y as [y h16].
  apply proj1_sig_injective. simpl.
  unfold ff in h12. simpl in h12.
  apply exist_injective in h12.
  apply proj1_sig_injective in h12.
  apply h13 in h12. apply exist_injective in h12.
  assumption.
pose proof (invertible_impl_inv_invertible _ h1) as h16.
apply invertible_impl_bijective in h16.
destruct h16 as [h16 h17].
assert (h12:Injective ff').
  red. intros x y h12.
  destruct x as [x h15], y as [y h18].
  apply proj1_sig_injective. simpl.
  unfold ff' in h12. simpl in h12.
  apply exist_injective in h12.
  apply proj1_sig_injective in h12.
  apply h16 in h12. apply exist_injective in h12.
  assumption.
split.
intro h10. 
apply FiniteT_sig_Finite. apply Finite_ens_type in h10.
apply (inj_finite _ _ ff h10 h11).
intros. apply classic.
intro h10.
apply FiniteT_sig_Finite. apply Finite_ens_type in h10.
apply (inj_finite _ _ ff' h10 h12).
intros; apply classic.
Qed.

Lemma finite_constructive_definite_description_full : 
  forall (T:Type),
    FiniteT T ->
    forall f:sig_set (seg (card_fun1 (Full_set T))) ->
             sig_set (Full_set T),
      bijective f ->
      forall (P:T->Prop),
        (exists x:T, P x) ->
        {x:T | P x}.
intros T h1 f h2 P h3.
pose proof (ex_inh _  h3) as h4.
pose proof (bij_seg_inh_full_iff f h2 P) as h5.
rewrite h5 in h4.
pose [n:nat | exists pf:In (seg (card_fun1 (Full_set T))) n,
                        P (proj1_sig (f (exist _ _ pf)))] as S.
pose proof (FiniteT_Finite _ h1 [x:T | P x]) as h6.
rewrite bij_seg_finite_full_iff in h6.
pose (max_set_nat _ h6 h4).  
pose proof (max_set_nat_compat _ h6 h4) as h10.
simpl in h10.
destruct h10 as [h10 h11]. 
destruct h10 as [h10].
pose proof (destruct_exists_pf_set _ _ _ h10) as h12.
simpl in h12.
refine (exist _ _ h12).
assumption.
Qed.


Section SubtractT.

Definition subtractT 
           {U} T
           (f:option U->T) : Type :=
  {x:T | x <> f None}.

Lemma FiniteT_subtract : 
  forall {T U} (pft:FiniteT T) (f:option U->T),
    FiniteT (subtractT T f).
  unfold subtractT; intros; apply finite_subtype; auto.
  intro x.
  destruct (finite_eq_dec _ pft x (f None)); subst.
  right; intro h1; contradict h1; auto.
  left; auto.
Qed.

Definition de_option {T} (o:option T) : o <> None -> T :=
  match o with
    | None => fun pf => False_rect _ (pf (eq_refl None))
    | Some o' => fun _ => o'
  end.

Lemma inj_de_option :
  forall {T} (o o':option T)
         (pfo: o <> None)
         (pfo':o' <> None),
    de_option o pfo = de_option o' pfo' ->
    o = o'.
  intros T o o'.
  destruct o as [x|], o' as [x'|]; simpl; intros; subst; auto.
  contradict pfo'.
  contradict pfo.
Qed.


Lemma subtract_inv_neq_none :
  forall {T U} (f:T->option U) (pfinv:invertible f)
         (x:subtractT T (proj1_sig (function_inverse f pfinv))),
    f (proj1_sig x) <> None.
  intros T U f h1 x.
  destruct x as [x h2]; simpl.
  pose proof (proj2_sig (function_inverse f h1)) as h3; simpl in h3.
  destruct h3 as [h3 h4].
  specialize (h3 x).
  contradict h2; rewrite <- h3, h2; auto.
Qed.

Lemma f_some_neq_none :
  forall {T U} (f:option T -> U),
    injective f ->
    forall x, 
      f (Some x) <> f None.
  intros T U f h1 x h2; apply h1 in h2; discriminate.
Qed.

Lemma f_some_neq_none' :
  forall {T U} (f:option T -> U),
    invertible f ->
    forall x, 
      f (Some x) <> f None.
  intros T U f h1 x.
  apply invertible_impl_bijective in h1.
  destruct h1.
  apply f_some_neq_none; auto.
Qed.

Definition f_from_option1 {T U} (f:option T-> U)
           (pfinv:invertible f) : T -> subtractT U f :=
  fun x => exist _ (f (Some x)) (f_some_neq_none' f pfinv x).


Definition f_from_option2
           {T U}
           (f:T->option U)
           (pfinv:invertible f) :
  subtractT T (proj1_sig (function_inverse f pfinv)) -> U :=
  fun x => de_option (f (proj1_sig x)) (subtract_inv_neq_none f pfinv x).

Lemma inv_f_from_option1 :
  forall {T U} (f:option T-> U)
         (pfinv:invertible f),
    invertible (f_from_option1 f pfinv).
  intros T U f h1.
  pose proof (invertible_impl_bijective _ h1) as h2.
  apply bijective_impl_invertible.
  unfold f_from_option1.
  destruct h2 as [h2 h3].
  red.
  split.
  red in h2; red.
  intros x y h4.
  apply exist_injective in h4.
  apply h2 in h4.
  inversion h4; auto.
  red in h3; red.
  intro y.
  destruct y as [y h4].
  specialize (h3 y).
  destruct h3 as [x h3]; subst.
  destruct x as [x|].
  exists x.
  apply proj1_sig_injective; simpl; auto.
  contradict h4; auto.
Qed.

Lemma inv_f_from_option2 :
  forall {T U} (f:T-> option U)
         (pfinv:invertible f),
    invertible (f_from_option2 f pfinv).
  intros T U f h1.
  pose proof (invertible_impl_bijective _ h1) as h2. 
  apply bijective_impl_invertible.
  unfold f_from_option2.
  destruct h2 as [h2 h3].
  red.
  split.
  red in h2; red.
  intros x y h4.
  apply inj_de_option in h4.
  apply h2 in h4.
  apply proj1_sig_injective in h4; auto.
  pose proof h3 as h3'.
  red in h3; red.
  intro y.
  specialize (h3 (Some y)). 
  destruct h3 as [x h3]; subst. 
  assert (h4:x <> (proj1_sig (function_inverse f h1)) None).
    intro; subst.
    pose proof (proj2_sig (function_inverse f h1)) as h4.
    simpl in h4.
    destruct h4 as [h4 h5].
    rewrite  h5 in h3; discriminate.
  exists (exist _ x h4).
  simpl.
  gen0.
  simpl.
  rewrite h3.
  intro.
  simpl; auto.
Qed.


Lemma equi_subtract :
  forall {T U}
         (f:option T -> U),
    invertible f ->
    equi T (subtractT U f).
  intros T U f h1.    
  pose proof (inv_f_from_option1 f h1) as h2.
  econstructor.
  apply h2.
Qed.


Definition f_option_subtractT
           {T U} (f:option T->U) :
  option (subtractT U f) -> U :=
  fun x =>
    match x with
      | None => f None
      | Some x' => proj1_sig x'
    end.



Lemma inv_f_option_subtractT :
  forall {T U} (f:option T->U),
    invertible f ->
    invertible (f_option_subtractT f).
  intros T U f hi.
  pose proof (invertible_impl_bijective _ hi) as h1.
  destruct h1 as [hinj hsurj].
  apply bijective_impl_invertible.
  unfold f_option_subtractT.
  split.
  intros x y h3.
  destruct x as [x|], y as [y|].
  apply proj1_sig_injective in h3; subst; auto.
  pose proof (proj2_sig x) as h9; simpl in h9.
  rewrite h3 in h9; contradict h9; auto.
  pose proof (proj2_sig y) as h10; simpl in h10.
  rewrite h3 in h10 at 1.
  contradict h10; auto.
  reflexivity.
  red.
  intro y.
  red in hsurj.
  specialize (hsurj y).
  destruct hsurj; subst.
  destruct x as [x|].
  pose proof (f_some_neq_none f hinj x) as h9.
  exists (Some (exist (fun c => c <> f None) (f (Some x)) h9)); simpl;
  auto.
  exists None; auto.
Qed.

(*Think of this, as the bijective opposite of [map option] -- 
in the most "natural" way -- the nicest way to turn [f] into 
the proper signature!*)
Definition f_from_option {T U}
           (f:option T->option U) (pf:invertible f) : T -> U :=
  match (none_dec (f None)) with
    | left pfe =>
      fun x =>
        let pff := f_some_neq_none' f pf x in 
        match (none_dec (f (Some x))) with
          | left pfe' => False_rect U (pff (eq_trans pfe' (eq_sym pfe)))
          | right pfn' => neq_none_inv _ pfn'
        end
    | right pfn =>
      fun x =>
        let pff := f_some_neq_none' f pf x in 
        match (none_dec (f (Some x))) with
          | left pfe' => neq_none_inv _ pfn
          | right pfn' => neq_none_inv _ pfn'
        end
  end.

Lemma f_from_option_none :
  forall {T U} (f:option T -> option U) (pf:invertible f)
         (pfe:f None = None) x,
      let pfn := neq_eq_trans (f_some_neq_none' f pf x) pfe in
      f_from_option f pf x = de_option (f (Some x)) pfn.
  unfold f_from_option; intros;
  do 2 lr_match_goal'; unfold neq_none_inv, de_option; try contradiction.
  gterml.
  redterm0 x0; contradiction.
  f_equal; apply proof_irrelevance.
Qed.

Lemma f_from_option_some_eq :
  forall {T U} (f:option T -> option U) (pf:invertible f)
         (pfn:f None <> None) x,
    f (Some x) = None ->
    f_from_option f pf x = de_option (f None) pfn.
  unfold f_from_option, de_option; intros;
  do 2 lr_match_goal'; unfold neq_none_inv; try contradiction; 
  f_equal; apply proof_irrelevance.
Qed.

Lemma f_from_option_some_neq :
  forall {T U} (f:option T -> option U) (pf:invertible f),
    f None <> None ->
    forall x (pfn:f (Some x) <> None),
      f_from_option f pf x = de_option (f (Some x)) pfn.
  unfold f_from_option, de_option; intros; do 2 lr_match_goal';
  unfold neq_none_inv; try contradiction; f_equal;
  apply proof_irrelevance.
Qed.  

Lemma inv_f_from_option :
  forall {T U} (f:option T -> option U) (pf:invertible f),
    invertible (f_from_option f pf).
  unfold f_from_option; intros T U f h1.
  pose proof (invertible_impl_bijective _ h1) as h2.
  apply bijective_impl_invertible.
  destruct h2 as [h2 h3]; split.
  red in h2; red.
  intros x y h4.
  lr_match' h4.
  lr_match' h4.
  terml h4.
  redterm0 x0.
  contradiction.
  lr_match' h4.
  termr h4.
  redterm0 y0.
  contradiction.
  apply inj_neq_none_inv in h4.
  apply h2 in h4.
  inversion h4; auto.
  lr_match' h4.
  lr_match' h4.
  rewrite <- hl in hl0.
  apply h2 in hl0.
  inversion hl0; auto.
  apply inj_neq_none_inv in h4.
  apply h2 in h4; discriminate.
  lr_match' h4.
  apply inj_neq_none_inv in h4.
  apply h2 in h4; discriminate.
  apply inj_neq_none_inv in h4.
  apply h2 in h4.
  inversion h4; auto.
  red in h3; red.
  intro y. 
  specialize (h3 (Some y)).
  destruct h3 as [x h3].
  lr_match_goal'.
  destruct x as [x|].
  exists x.
  lr_match_goal'.
  rewrite hl0 in h3; discriminate.
  unfold neq_none_inv.
  destruct (f (Some x)).
  inversion h3; auto.
  contradict hr; auto.
  rewrite hl in h3; discriminate.
  destruct x as [x|].
  exists x. 
  lr_match_goal'.
  unfold neq_none_inv.
  rewrite h3 in hl; discriminate.
  unfold neq_none_inv.
  pose proof (some_neq_none y) as h4.
  pose proof h4 as h4'.
  rewrite <- h3 in h4'.
  pose proof (subsetT_eq_compat _ (fun c => c <> None) _ _ hr0 h4 h3)
    as h5.
  dependent rewrite h5; auto.
  unfold neq_none_inv.
  pose proof hr as h4.
  rewrite h3 in h4.
  pose proof (subsetT_eq_compat _ (fun c => c <> None) _ _ hr h4 h3)
    as h5.
  dependent rewrite h5. 
  destruct (none_dec (proj1_sig (function_inverse _ h1) None))
    as [h6 | h6].
  pose proof (proj2_sig (function_inverse _ h1)) as h7; simpl in h7;
  destruct h7 as [h7 h8]. 
  specialize (h8 None).
  rewrite h6 in h8.
  contradiction.
  exists (neq_none_inv _ h6).
  lr_match_goal'; auto.
  pose proof hr0 as h7.
  rewrite some_neq_none_inv in h7.
  pose proof (proj2_sig (function_inverse _ h1)) as h8; simpl in h8;
  destruct h8 as [h8 h9].
  specialize (h9 None).
  contradiction.
Qed.

  
End SubtractT.

Lemma equi_option :
  forall U V,
    equi U V ->
    equi (option U) (option V).
  intros U V h1.
  destruct h1.
  econstructor.
  apply invertible_option_map.
  apply H.
Qed.

Lemma inj_option :
  forall T U,
    equi (option T) (option U) ->
    equi T U.
  intros T U h1.
  inversion h1.
  apply equi_intro with (f_from_option f H).
  apply inv_f_from_option.
Qed.

Lemma equi_sig_false :
  forall T,
    equi False {x:T | False}.
  intro T.
  pose (fun x:False => exist (fun c:T => False) (False_rect _ x) x) as f.
  assert (h1:invertible f).
    apply bijective_impl_invertible.
    red; split; red.
    intros; contradiction.
    intro y; destruct y; contradiction.
  pose proof (equi_intro _ _ _ h1); auto.
Qed.

  

Lemma false_option :
  forall T,
    ~ (eq False (option T) (A:=Type)).
  intros T h1.
  pose proof (@None T).
  rewrite <- h1 in X; auto.
Qed.

Lemma finite_segT :
  forall n,
    FiniteT (n @).
  intro n; induction n as [|n ih]. 
  apply (bij_finite _ _  (fun (pff:False) => False_rect (0 @) pff)).
  apply empty_finite.
  apply bijective_impl_invertible; red; split; red;
  [intros; contradiction|intro y; destruct y; omega].
  apply (bij_finite _ _ (fun x:option (n @) =>
                           match x with
                             | Some e =>
                               let (a, pfa) := e in
                               exist (fun c => c < S n)
                                     (S a)
                                     (lt_n_S _ _ pfa)
                             | None =>
                               exist (fun c => c < S n)
                                     0
                                     (lt_0_Sn _)
                           end)).
  apply add_finite; auto.
  apply bijective_impl_invertible; red; split; red.
  intros x y h1; destruct x, y.
  try destruct s, s0.
  apply exist_injective in h1.
  apply S_inj in h1; subst.
  f_equal; apply proj1_sig_injective; simpl; auto.
  destruct s.
  apply exist_injective in h1.
  discriminate.
  destruct s.
  apply exist_injective in h1.
  discriminate.
  reflexivity.
  intro y.
  destruct y as [y hy].
  destruct y as [|y].
  exists None.
  f_equal; apply proof_irrelevance.
  exists (Some (exist (fun c => c < n) y (lt_S_n _ _ hy))).
  f_equal; apply proof_irrelevance.
Qed.


End FiniteTypes.


(*Whole section copied and pasted from file of same name in 
Schepler's Zorn's Lemma.*)
Section InfiniteTypes.
  
Open Scope nat.
(*Schepler*)
Lemma finite_nat_initial_segment: forall n:nat,
  FiniteT { m:nat | m < n }.
Proof.
intros.  
apply Finite_ens_type.
rewrite <- characteristic_function_to_ensemble_is_identity.
induction n.
assert ([x:nat | x < 0] = (Empty_set _)).
apply Extensionality_Ensembles; split; auto with sets.
red; intros.
destruct H.
contradict H.
auto with arith. 
rewrite H; constructor.

assert ([x:nat | S x <= S n] = Add [x:nat | x < n] n).
apply Extensionality_Ensembles; split.
red; intros.
destruct H.
assert (x <= n); auto with arith. 
apply le_lt_or_eq in H0.
case H0. 
left; constructor; trivial.
right; auto with sets.
red; intros.
case H.
intros.
destruct H0; constructor.
auto with arith.
intros.
destruct H0.
constructor.
auto with arith.
assert (h1:forall x, S x <= S n <-> x < S n).
intros; tauto; trivial with arith.
rewrite (sat_iff _ _ h1) in H.
rewrite H; constructor; trivial.
red; intro.
destruct H0.
contradict H0.
auto with arith.
Qed.

(*Schepler*)
Lemma nat_infinite: ~ FiniteT nat.
Proof.
red; intro.
assert (surjective S).
apply finite_inj_surj; trivial.
red; intros.
injection H0; trivial.

destruct (H0 0).
discriminate H1.
Qed.

(*Wyckoff*)
Lemma Z_infinite : ~FiniteT Z.
intro h1.
assert (h2:forall y : Z,
             (exists x : nat, Z.of_nat x = y) \/
             ~ (exists x : nat, Z.of_nat x = y)).
  intros; apply classic.
pose proof (inj_finite _ _ Z.of_nat h1 Nat2Z.inj h2) as h3.
pose proof nat_infinite.
contradiction.
Qed.



(*Wyckoff*)
Lemma Q_infinite : ~FiniteT QArith_base.Q.
intro h1.

assert (h2: Injective QArith_base.inject_Z).
  red. intros ? ? h3.
  assert (h3': QArith_base.Qeq(QArith_base.inject_Z x1)
                              (QArith_base.inject_Z x2)).
    rewrite h3. constructor.
  rewrite QArith_base.inject_Z_injective in h3'.
  assumption.

assert (h3: forall y : QArith_base.Q,
       (exists x : Z, QArith_base.inject_Z x = y) \/
       ~ (exists x : Z, QArith_base.inject_Z x = y)).
  intros; apply classic.
pose proof (inj_finite _ _ QArith_base.inject_Z h1 h2 h3) as h4.
pose proof Z_infinite.
contradiction.
Qed.


(*Schepler*)
Lemma infinite_nat_inj: forall X:Type, ~ FiniteT X ->
  exists f:nat->X, Injective f.
Proof.
intros.
assert (inhabited (forall S:Ensemble X, Finite S ->
  { x:X | ~ In S x})).
pose proof (choice (fun (x:{S:Ensemble X | Finite S}) (y:X) =>
  ~ In (proj1_sig x) y)).
simpl in H0.
match type of H0 with | ?A -> ?B => assert B end.
apply H0.
intros.
apply NNPP.
red; intro.
pose proof (not_ex_not_all _ _ H1); clear H1.
destruct x.
assert (x = (Full_set _)).
apply Extensionality_Ensembles; red; split; auto with sets.
intro x0; constructor.
symmetry in H1; destruct H1.
contradiction H.
clear H2.
apply bij_finite with (f:=@proj1_sig _ (fun x:X => In (Full_set _) x)).
apply Finite_ens_type; assumption.
exists (fun x:X => exist _ x (Full_intro _ x)).
destruct x; simpl.
generalize (Full_intro X x).
intro i0; destruct (proof_irrelevance _ i i0); trivial.
trivial.
clear H0.
destruct H1.
exists.
intros.
exists (x (exist _ S H1)).
exact (H0 (exist _ S H1)).
destruct H0.
 
assert (forall (n:nat) (g:forall m:nat, m<n -> X),
  { x:X | forall (m:nat) (Hlt:m<n), g m Hlt <> x }).
intros.
assert (Finite (fun x:X => exists m:nat, exists Hlt:m<n,
           g m Hlt = x)).
pose (h := fun x:{m:nat | m<n} =>
  g (proj1_sig x) (proj2_sig x)).

match goal with |- Finite ?S => assert (S =
  Im (Full_set _) h) end.
apply Extensionality_Ensembles; red; split; red; intros.
destruct H0.
destruct H0.
exists (exist (fun m:nat => m < n) x0 x1).
constructor.
unfold h; simpl.
symmetry; assumption.
destruct H0.
destruct x.
unfold h in H1; simpl in H1.
exists x; exists l; symmetry; assumption.

rewrite H0; apply FiniteT_img.
apply finite_nat_initial_segment.
intros; apply classic.

destruct (X0 _ H0).
unfold In in n0.
exists x.
intros; red; intro.
contradiction n0; exists m; exists Hlt; exact H1.

pose (f := Fix lt_wf (fun n:nat => X)
  (fun (n:nat) (g:forall m:nat, m<n->X) => proj1_sig (X1 n g))).
simpl in f.
assert (forall n m:nat, m<n -> f m <> f n).
pose proof (Fix_eq lt_wf (fun n:nat => X)
  (fun (n:nat) (g:forall m:nat, m<n->X) => proj1_sig (X1 n g))).
fold f in H0.
simpl in H0.
match type of H0 with | ?A -> ?B => assert (B) end.
apply H0.
intros.
assert (f0 = g).
Require Import FunctionalExtensionality.
extensionality y; extensionality p; apply H1.
destruct H2; trivial.
intros.
pose proof (H1 n).
destruct X1 in H3.
simpl in H3.
destruct H3.
auto.

exists f.
red; intros m n ?.
destruct (lt_eq_lt_dec m n) as [[Hlt|Heq]|Hlt]; trivial.
contradiction (H0 n m).
contradiction (H0 m n).
symmetry; assumption.
Qed.


(*Schepler*)
Lemma inf_comp_fin : forall {T:Type}, ~FiniteT T -> 
  forall U:Ensemble T, Finite (Comp U) -> 
    ~ Finite U.
intros T h0 U h1.
case (classic (Finite U)); trivial.
intro h2.
pose proof (Union_preserves_Finite T U (Comp U) h2 h1) as h3.
rewrite excl_middle_full in h3.
pose proof (Finite_ens_type (Full_set T) h3) as h4.
pose (fun (a:{x:T | In (Full_set T) x}) => proj1_sig a) as f.
assert (h5:invertible f).
  apply bijective_impl_invertible.
  red. split.
  (*left*)
  red.
  intros x1 x2 h6.
  unfold f in h6.
  apply proj1_sig_injective; trivial.
  (*right*)
  red.
  intro y.
  exists (exist _ y (Full_intro _ y)).
  unfold f.
  simpl.
  reflexivity.
pose proof (bij_finite _ _ f h4 h5). 
contradiction.
Qed.

End InfiniteTypes.


Section CountableT.

(*Schepler*)
Inductive CountableT (X:Type) : Prop :=
  | intro_nat_injection: forall f:X->nat, Injective f -> CountableT X.

(*Wyckoff*)
Lemma countable_nat : 
  CountableT nat.
apply intro_nat_injection with id.
apply inj_id.
Qed.


(*Schepler*)
Lemma CountableT_is_FiniteT_or_countably_infinite:
  forall X:Type, CountableT X ->
      {FiniteT X} + {exists f:X->nat, bijective f}.
Proof.
intros.
apply exclusive_dec.
red; intro.
destruct H0 as [? [f ?]].
contradiction nat_infinite.
apply bij_finite with _ f; trivial.
apply bijective_impl_invertible; trivial.

case (classic (FiniteT X)).
left; trivial.
right.
apply infinite_nat_inj in H0.
destruct H.
destruct H0 as [g].
apply CSB with f g; trivial.
Qed.

(*Wyckoff*)
Lemma inf_countable_index : 
  forall T:Type, 
    CountableT T -> ~FiniteT T ->
    exists f:nat->T, bijective f.
intros T h1 h2.
pose proof (CountableT_is_FiniteT_or_countably_infinite T h1) as h3.
destruct h3 as [h3 | h3]; try contradiction.
destruct h3 as [f h4].
apply bijective_impl_invertible in h4.
pose (proj1_sig (function_inverse f h4)) as f'.
exists f'.
pose proof (invertible_impl_inv_invertible f h4) as h5.
apply invertible_impl_bijective in h5.
assumption.
Qed.


(*Schepler*)
Lemma countable_nat_product: CountableT (nat*nat).
Proof.
pose (sum_1_to_n := fix sum_1_to_n n:nat := match n with
  | O => O
  | S m => (sum_1_to_n m) + n
end).
exists (fun p:nat*nat => let (m,n):=p in
  (sum_1_to_n (m+n)) + n).
assert (forall m n:nat, m<n ->
  sum_1_to_n m + m < sum_1_to_n n).
intros.
induction H.
simpl.
auto with arith.
apply lt_trans with (sum_1_to_n m0).
assumption.
simpl.
assert (0 < S m0); auto with arith.
assert (sum_1_to_n m0 + 0 < sum_1_to_n m0 + S m0); auto with arith.
assert (sum_1_to_n m0 + 0 = sum_1_to_n m0); auto with arith.
rewrite H2 in H1; assumption.

red; intros.
destruct x1 as [x1 y1].
destruct x2 as [x2 y2].
Require Import Compare_dec.
case (lt_eq_lt_dec (x1+y1) (x2+y2)); intro.
case s; intro.
assert (sum_1_to_n (x1+y1) + y1 < sum_1_to_n (x2+y2) + y2).
apply le_lt_trans with (sum_1_to_n (x1+y1) + (x1+y1)).
assert (sum_1_to_n (x1+y1) + (x1+y1) =
  (sum_1_to_n (x1+y1) + y1) + x1).
Require Import ArithRing.
ring.
auto with arith.
apply lt_le_trans with (sum_1_to_n (x2+y2)).
apply H; trivial.
auto with arith.
rewrite H0 in H1.
contradict H1.
auto with arith.

assert (y1=y2).
rewrite e in H0.
Require Import Arith.
apply plus_reg_l in H0.
assumption.
f_equal; trivial.
rewrite H1 in e.
rewrite plus_comm in e.
rewrite (plus_comm x2 y2) in e.
apply plus_reg_l in e.
assumption.

assert (sum_1_to_n (x2+y2) + y2 < sum_1_to_n (x1+y1) + y1).
apply le_lt_trans with (sum_1_to_n (x2+y2) + (x2+y2)).
auto with arith.
apply lt_le_trans with (sum_1_to_n (x1+y1)); auto with arith.
rewrite H0 in H1.
contradict H1.
auto with arith.
Qed.

(*Schepler*)
Lemma countable_sum: forall X Y:Type,
  CountableT X -> CountableT Y -> CountableT (X+Y).
Proof.
intros.
destruct H as [f].
destruct H0 as [g].
destruct countable_nat_product as [h].
exists (fun s:X+Y => match s with
  | inl x => h (0, f x)
  | inr y => h (1, g y)
end).
red; intros s1 s2 ?.
destruct s1 as [x1|y1]; destruct s2 as [x2|y2];
  apply H1 in H2; try discriminate H2;
  intros; f_equal; (apply H || apply H0); injection H2; trivial.
Qed.

(*Schepler*)
Lemma countable_product: forall X Y:Type,
  CountableT X -> CountableT Y -> CountableT (X*Y).
Proof.
intros.
destruct H as [f].
destruct H0 as [g].
pose (fg := fun (p:X*Y) => let (x,y):=p in (f x, g y)).
destruct countable_nat_product as [h].
exists (fun p:X*Y => h (fg p)).
red; intros.
apply H1 in H2.
destruct x1 as [x1 y1].
destruct x2 as [x2 y2].
unfold fg in H2.
injection H2; intros.
apply H0 in H3.
apply H in H4.
f_equal; trivial.
Qed.

Require Import FunctionalExtensionality.

(*Schepler*)
Lemma countable_exp: forall X Y:Type,
  FiniteT X -> CountableT Y -> CountableT (X->Y).
Proof.
intros.
induction H.
exists (fun _ => 0).
red; intros.
extensionality f.
destruct f.

destruct (countable_product (T->Y) Y); trivial.

exists (fun (g:option T->Y) =>
  f (fun x:T => g (Some x), g None)).
red; intros g1 g2 ?.
apply H1 in H2.
extensionality o.
destruct o.
injection H2; intros.
pose proof (equal_f H4).
simpl in H5.
apply H5.
injection H2; trivial.

destruct H1.
destruct IHFiniteT.
exists (fun (h:Y0->Y) => f0 (fun x:X => h (f x))).
red; intros h1 h2 ?.
apply H3 in H4.
pose proof (equal_f H4).
simpl in H5.
extensionality y.
rewrite <- (H2 y).
apply H5.
Qed.

(*Schepler*)
Definition Countable {X:Type} (S:Ensemble X) : Prop :=
  CountableT {x:X | In S x}.

(*Schepler*)
Lemma inj_countable: forall {X Y:Type} (f:X->Y),
  CountableT Y -> Injective f -> CountableT X.
Proof.
intros.
destruct H as [g].
exists (fun x:X => g (f x)).
red; intros; auto.
Qed.

(*Schepler*)
Lemma surj_countable: forall {X Y:Type} (f:X->Y),
  CountableT X -> surjective f -> CountableT Y.
Proof.
intros.
Require Import ClassicalChoice.

pose proof (choice (fun (y:Y) (x:X) => f x = y)).
destruct H1 as [finv].
exact H0.

apply inj_countable with finv.
assumption.
red; intros.
congruence.
Qed.

(*Schepler*)
Lemma countable_downward_closed: forall {X:Type} (S T:Ensemble X),
  Countable T -> Included S T -> Countable S.
Proof.
intros.
destruct H.
exists (fun x:{x:X | In S x} => match x with
  | exist x0 i => f (exist _ x0 (H0 _ i))
  end).
red; intros.
destruct x1 as [x1].
destruct x2 as [x2].
apply H in H1.
injection H1; intros.
destruct H2.
destruct (proof_irrelevance _ i i0).
trivial.
Qed.

(*Schepler*)
Lemma countable_img: forall {X Y:Type} (f:X->Y) (S:Ensemble X),
  Countable S -> Countable (Im S f).
Proof.
intros.
assert (forall x:X, In S x -> In (Im S f) (f x)).
auto with sets.
pose (fS := fun x:{x:X | In S x} =>
  match x return {y:Y | In (Im S f) y} with
  | exist x0 i => exist _ (f x0) (H0 x0 i)
  end).
apply surj_countable with fS; trivial.
red; intros.
destruct y.
destruct i.
exists (exist _ x i).
simpl.
generalize (H0 x i); intro.
generalize (Im_intro X Y S f x i y e); intro.
destruct e.
destruct (proof_irrelevance _ i0 i1).
trivial.
Qed.

(*slightly modified from Schepler*)
Lemma countable_type_ensemble: forall {X:Type} (S:Ensemble X),
  CountableT X -> Countable S.
Proof.
intros.
red.
apply inj_countable with (@proj1_sig _ (fun x:X => In S x)).
assumption.
red.
intros a1 a2 h1.
destruct a1. destruct a2.
apply existTexist.
apply subsetT_eq_compat.
simpl in h1. assumption.
Qed.

(*Wyckoff*)
Lemma inf_countable_sig_index : 
  forall T:Type,
    CountableT T -> ~FiniteT T ->
    forall S:Ensemble nat,
      ~Finite S ->
      exists f:sig_set S->T,
        bijective f.
intros T h1 h2 S h3.
pose proof countable_nat as h4.
pose proof (countable_type_ensemble S h4) as h5.
pose proof (CountableT_is_FiniteT_or_countably_infinite _ h5) as h6.
destruct h6 as [h6 | h6].
apply FiniteT_sig_Finite in h6.
contradiction.
destruct h6 as [f h6].
pose proof (inf_countable_index _ h1 h2) as h7.
destruct h7 as [g h7].
exists (compose g f).
apply bij_compose; auto.
Qed.

Require Import Infinite_sets.

(*Wyckoff*)
Lemma inf_countable_index_iff_inf_countable : 
  forall (T:Type),
    (exists S:Ensemble nat, ~Finite S /\ exists f:sig_set S->T, bijective f) <->
    CountableT T /\ ~FiniteT T.
intro T. split.
intro h1.
destruct h1 as [S h1]. destruct h1 as [h1 h2].
destruct h2 as [f h2].
split. 
eapply surj_countable.
assert (h3:CountableT (sig_set S)).
  apply countable_type_ensemble.
  apply countable_nat.
apply h3.
red in h2. destruct h2 as [h2a h2b]. apply h2b.
intro h3.
rewrite finite_full_sig_iff in h1.
assert (h4: Finite (Im (full_sig S) f)).
  eapply Finite_downward_closed.
  rewrite <- Finite_FiniteT_iff in h3.
  apply h3.
  red; intros; constructor.
pose proof (Pigeonhole_bis _ _ _ f h1 h4) as h5.
contradict h5.
destruct h2; auto.

intro h1.
destruct h1 as [h1 h2].
exists (Full_set nat). split.
rewrite Finite_FiniteT_iff.
apply nat_infinite.
pose proof (inf_countable_index _ h1 h2) as h3.
destruct h3 as [f h3].
pose (fun n:sig_set (Full_set nat) => (f (proj1_sig n))) as f'.
exists f'.
destruct h3 as [h3 h4].
red. split.
red; intros x y h5.
destruct x as [x h6], y as [y h7]. 
unfold f' in h5. simpl in h5.
apply proj1_sig_injective; simpl.
apply h3; auto.
red in h4. red.
intro y.
specialize (h4 y).
destruct h4 as [n h4].
exists (exist _ _ (Full_intro _ n)).
unfold f'. simpl.
assumption.
Qed.


(*Schepler*)
Lemma FiniteT_impl_CountableT: forall X:Type,
  FiniteT X -> CountableT X.
Proof.
intros.
induction H.
exists (False_rect nat).
red; intros.
destruct x1.
destruct IHFiniteT.
exists (fun x:option T => match x with
  | Some x0 => S (f x0)
  | None => 0
end).
red; intros.
destruct x1; destruct x2; try (injection H1 || discriminate H1); trivial.
intro.
apply H0 in H2.
destruct H2; trivial.

destruct IHFiniteT as [g].
destruct H0 as [finv].
exists (fun y:Y => g (finv y)).
red; intros y1 y2 ?.
apply H1 in H3.
congruence.
Qed.

(*Schepler*)
Lemma Finite_impl_Countable: forall {X:Type} (S:Ensemble X),
  Finite S -> Countable S.
Proof.
intros.
apply FiniteT_impl_CountableT.
apply Finite_ens_type; trivial.
Qed.

(* Schepler *)
Lemma countable_union: forall {X A:Type}
  (F:IndexedFamily A X), CountableT A ->
    (forall a:A, Countable (F a)) ->
    Countable (IndexedUnion F).
Proof.
intros.
destruct (choice_on_dependent_type (fun (a:A)
                               (f:{x:X | In (F a) x} -> nat) =>
  Injective f)) as [choice_fun_inj].
intro.
destruct (H0 a).
exists f; trivial.

destruct (choice (fun (x:{x:X | In (IndexedUnion F) x}) (a:A) =>
  In (F a) (proj1_sig x))) as [choice_fun_a].
destruct x as [x [a]].
exists a.
assumption.

destruct countable_nat_product as [g].
destruct H as [h].
exists (fun x:{x:X | In (IndexedUnion F) x} =>
  g (h (choice_fun_a x), choice_fun_inj (choice_fun_a x)
                                   (exist _ (proj1_sig x) (H2 x)))).
red; intros.
apply H3 in H4.
injection H4; intros.
apply H in H6.
revert H5.
generalize (H2 x1).
generalize (H2 x2).
rewrite H6.
intros.
apply H1 in H5.
injection H5.
apply proj1_sig_injective.
Qed.


(*Wyckoff*)
Lemma countable_union2 : forall (T:Type) (S1 S2:Ensemble T), 
  Countable S1 -> Countable S2 -> Countable (Union S1 S2).
intros T S1 S2 h1 h2.
pose (option (option False)) as twoset.
pose (fun (t:twoset) => 
  match t with | None => S1 | Some _ => S2 end) as F.
assert (h3:FiniteT twoset).
  apply add_finite. apply add_finite. apply empty_finite.
pose proof (FiniteT_impl_CountableT _ h3) as h4.
assert (h5: forall t:twoset, Countable (F t)).
  destruct t; simpl; assumption.
pose proof (countable_union F h4 h5) as h6.
assert (h7:IndexedUnion F =  Union S1 S2).
  apply Extensionality_Ensembles.
  red. split.
  red. intros x h7. 
  (* <= *) 
  inversion h7 as [i t h8].
  destruct i. simpl in h8. apply Union_intror; assumption.
  simpl in h8. apply Union_introl; assumption.
  (* >= *)
  red. intros x h7.
  pose proof (Union_inv _ _ _ _ h7) as h8.
  case h8. 
    unfold F. intro h9. 
      apply indexed_union_intro with None; assumption.
    unfold F. intro h9. 
      apply indexed_union_intro with (Some None); assumption.
rewrite h7 in h6. assumption.
Qed.
   

(*Wyckoff*)
Lemma countable_intersection: forall (T:Type) (S1 S2:Ensemble T), 
  Countable S1 -> Countable S2 -> Countable (Intersection S1 S2).
intros T S1 S2 h1 h2.
apply (countable_downward_closed (Intersection S1 S2) S1).
assumption.
auto with sets.
Qed.

(*Schepler*)
Lemma Countable_Full : 
  forall {T:Type}, 
    CountableT T->Countable (Full_set T).
intros T h1.
red.
pose (fun (x:T) => (exist _ x (Full_intro _ x))) as f.
apply surj_countable with f. assumption.
red. intro y.
exists (proj1_sig y).
unfold f.
destruct y.
simpl.
pose proof (proof_irrelevance _ i (Full_intro T x)) as h2.
rewrite h2. reflexivity.
Qed.

(*Wyckoff*)
Lemma countable_full_iff :
  forall {T:Type}, 
    CountableT T <-> Countable (Full_set T).
intro T. split; try apply Countable_Full.
intro h1.
red in h1.
pose (fun x:T => (exist _ _ (Full_intro _ x))) as f.
apply (inj_countable f); auto.
unfold f.
red; intros ? ? h2.
apply exist_injective in h2.
assumption.
Qed.


(*Schepler*)
Lemma Countable_Sig : forall {T:Type} (P:T->Prop), CountableT T -> 
  CountableT {t:T|P t}.
intros T P h1.
pose [t:T|P t] as S.
pose proof (countable_downward_closed S (Full_set _)) as h2.
pose proof (Countable_Full h1) as h3.
assert (h4:Included S (Full_set _)).
  red. intros. constructor.
pose proof (h2 h3 h4) as h5.
red in h5.
assert (h6:forall x:T, In S x <-> P x).
  intro x.
  split.
  (*->*)
  intro h7.
  inversion h7. assumption.
  (*<-*)
  intro h8.
  constructor. assumption.
assert (h7:forall x:T, In S x -> P x).
  intros x h8.
  rewrite h6 in h8. assumption.

pose (fun a:{x:T | In S x} => 
  exist _ (proj1_sig a) (h7 (proj1_sig a) (proj2_sig a))) as f.
apply surj_countable with f.
assumption.

red.
intros y.
destruct y as [t h8].
assert (h9:In S t).
  rewrite h6. assumption.
exists (exist _ t (h9)).
unfold f. simpl.
pose proof (proof_irrelevance _ (h7 t h9) h8) as h10.
rewrite h10. reflexivity.
Qed.

(*Wyckoff*)
Corollary CountableT_Countable : forall (T:Type), CountableT T -> 
  (forall (S:Ensemble T), Countable S).
intros T h1 S.
red.
apply Countable_Sig; assumption.
Qed.

(*Schepler*)
Lemma positive_countable: CountableT BinPos.positive.
Proof.
exists nat_of_P.
red; intros.
apply nat_of_P_inj; trivial.
Qed.

(*Schepler*)
Lemma Z_countable: CountableT Z.
Proof.
destruct (countable_nat_product) as [f].
destruct positive_countable as [g].
exists (fun n:Z => match n with
  | Z0 => f (0, 0)
  | Zpos p => f (1, g p)
  | Zneg p => f (2, g p)
end).
red; intros n1 n2 ?.
destruct n1 as [|p1|p1]; destruct n2 as [|p2|p2]; apply H in H1;
  try discriminate H1.
trivial.
injection H1; intro; f_equal; auto.
injection H1; intro; f_equal; auto.
Qed.

(*Schepler*)
Lemma Q_countable: CountableT QArith_base.Q.
Proof.
destruct countable_nat_product as [f].
destruct positive_countable as [g].
destruct Z_countable as [h].
exists (fun q:QArith_base.Q => match q with
  (QArith_base.Qmake n d) => f (h n, g d)
end).
red; intros q1 q2 ?.
destruct q1 as [n1 d1]; destruct q2 as [n2 d2].
apply H in H2.
injection H2; intros.
f_equal; auto.
Qed.



End CountableT.

(*Wyckoff*)
Inductive point_true {T} (P:T->Prop) : Prop :=
| point_true_intro : forall x, P x -> point_true P.

  
Section FiniteT_card.

(*Schepler*)
Lemma FiniteT_has_nat_cardinal: forall X:Type, FiniteT X ->
  exists! n:nat, cardinal _ (@Full_set X) n.
Proof.
intros.
apply -> unique_existence; split.
apply finite_cardinal.
pose (idX := fun x:X => x).
assert (Im (Full_set _) idX = Full_set _).
apply Extensionality_Ensembles.
red; split.
red; intros; constructor.
red; intros.
exists x.
constructor.
trivial.

rewrite <- H0.
apply FiniteT_img with (f:=fun x:X => x).
assumption.
intros.
case (finite_eq_dec X H y1 y2); tauto.

red; intros.
apply cardinal_unicity with X (Full_set _); trivial.
Qed.


(*Schepler*)
Definition FiniteT_card (X:Type) (H:FiniteT X) : nat :=
  proj1_sig (constructive_definite_description _
              (FiniteT_has_nat_cardinal X H)).

(*Schepler*)
Lemma FiniteT_card_def: forall (X:Type) (H:FiniteT X),
  cardinal _ (@Full_set X) (FiniteT_card X H).
Proof.
intros; unfold FiniteT_card.
destruct constructive_definite_description.
assumption.
Qed.

(*Schepler*)
Lemma FiniteT_card_cond: forall (X:Type) (H:FiniteT X)
  (n:nat),
  cardinal _ (@Full_set X) n ->
  FiniteT_card X H = n.
Proof.
intros.
pose proof (FiniteT_has_nat_cardinal X H).
destruct H1.
red in H1.
destruct H1.
transitivity x.
symmetry; apply H2.
apply FiniteT_card_def.
apply H2; trivial.
Qed.

(*Schepler*)
Lemma FiniteT_card_False:
  FiniteT_card False empty_finite = 0.
Proof.
apply FiniteT_card_cond.
assert (@Full_set False = @Empty_set False).
apply Extensionality_Ensembles; red; split; auto with sets.
red; intros.
destruct x.
rewrite H.
constructor.
Qed.


(*Schepler*)
Lemma injection_preserves_cardinal: forall (X Y:Type)
  (f:X->Y) (n:nat) (S:Ensemble X), cardinal _ S n ->
  Injective f -> cardinal _ (Im S f) n.
Proof.
intros.
induction H.
assert (Im (Empty_set _) f = (Empty_set _)).
apply Extensionality_Ensembles; split; auto with sets.
red; intros.
destruct H.
destruct H.
rewrite H; constructor.
assert (Im (Add A x) f = Add (Im A f) (f x)).
apply Extensionality_Ensembles; split.
red; intros.
destruct H2.
symmetry in H3; destruct H3.
destruct H2.
left; exists x0; auto with sets.
destruct H2; right; auto with sets.
red; intros.
destruct H2.
destruct H2.
exists x0.
left; auto with sets.
assumption.
destruct H2.
exists x; trivial; right; auto with sets.
rewrite H2.
constructor; trivial.
red; intro H3; inversion H3.
apply H0 in H5; destruct H5.
contradiction H1.
Qed.

(*Schepler*)
Lemma FiniteT_card_option:
  forall (X:Type) (H:FiniteT X),
  FiniteT_card (option X) (add_finite X H) =
  S (FiniteT_card X H).
Proof.
intros.
apply FiniteT_card_cond.
assert ((Full_set _) =
        Add (Im (Full_set _) (@Some X)) None).
apply Extensionality_Ensembles; split.
red; intros.
destruct x.
left; exists x; constructor.
right; constructor.
red; intros; constructor.
rewrite H0.
constructor.
apply injection_preserves_cardinal.
apply FiniteT_card_def.
red; intros x1 x2 Heq; injection Heq; trivial.
red; intro.
inversion H1.
discriminate H3.
Qed.


(*Schepler*)
Lemma FiniteT_card_bijection:
  forall (X Y:Type) (H:FiniteT X) (g:X->Y) (Hinv:invertible g),
    FiniteT_card Y (bij_finite X Y g H Hinv) =
    FiniteT_card X H.
Proof.
intros.
apply FiniteT_card_cond.
apply invertible_impl_bijective in Hinv.
destruct Hinv as [g_inj g_surj].
assert ((Full_set _)= Im (Full_set _) g).
apply Extensionality_Ensembles; split; red; intros;
  try constructor.
destruct (g_surj x).
exists x0; try constructor; auto.
rewrite H0; apply injection_preserves_cardinal; trivial.
apply FiniteT_card_def.
Qed.

(*Wyckoff*)
Lemma unit_finite : FiniteT unit.
pose (fun x:(option False) => tt) as f.
assert (h1:invertible f).
  apply bijective_impl_invertible.
  red; split. 
  red. intros x1 x2. destruct x1; destruct x2; try contradiction; auto.
  red. intros x1. exists None. unfold f. destruct x1. reflexivity.
eapply bij_finite; auto.
eapply add_finite.
apply empty_finite.
apply h1.
Qed.


(*Wyckoff*)
Lemma FiniteT_card_unit : 
  FiniteT_card _ unit_finite = 1.
assert (h2:FiniteT_card _ (add_finite _ empty_finite) = 1).
  pose proof (FiniteT_card_False) as h2.
  rewrite <- h2.
  apply FiniteT_card_option.
rewrite <- h2.
pose (fun x:(option False) => tt) as f.
assert (h3:invertible f).
  apply bijective_impl_invertible.
  red; split. 
  red. intros x1 x2. destruct x1; destruct x2; try contradiction; auto.
  red. intros x1. exists None. unfold f. destruct x1. reflexivity.
generalize unit_finite.
intro h4.
pose proof (proof_irrelevance
              _ h4
              (bij_finite _ _ f (add_finite _ empty_finite) h3)); subst.
eapply FiniteT_card_bijection.
Qed.

(*Wyckoff*)
Lemma FiniteT_card_fun1_compat : 
  forall {T:Type} (pf:FiniteT T),
    FiniteT_card _ pf = card_fun1 (Full_set T).
intros T h1.
pose proof (card_fun1_compat (Full_set T)) as h2.
destruct h2 as [h2l h2r].
pose proof (Finite_FiniteT_iff T) as h3.
pose proof h1 as h1'.
rewrite <- h3 in h1'.
specialize (h2l h1').
pose proof (FiniteT_card_cond T h1 _ h2l) as h4.
assumption.
Qed.

(*Wyckoff*)
Lemma FiniteT_card_add_empty :
  FiniteT_card _ (add_finite _ empty_finite) = 1.
  rewrite (FiniteT_card_option _ empty_finite).
  f_equal; apply FiniteT_card_False.
Qed.


(*Wyckoff*)
Lemma FiniteT_card_eq0 :
  forall T (pf:FiniteT T),
    FiniteT_card T pf = 0 ->
    equi T False.
  intros T h1. 
  pose proof card_empty T as h2.
  intros h3.
  pose proof (FiniteT_card_def T h1) as h4.   
  rewrite h3 in h4. 
  clear h3 h2.
  revert h4. 
  induction h1; auto.
  intros h2.
  exists id.
  apply invertible_id.
  intro h2.
  inversion h2.
  pose proof (Full_intro (option T) None) as h3.
  rewrite <- H0 in h3.
  inversion h3.
  intro h2; auto.
  inversion h2. 
  pose (fun y:Y => (Full_intro _ y)) as g. 
  rewrite <- H1 in g.
  pose proof (equi_intro _ False (fun y => match (g y) with | end)) as h3.
  apply h3.
  econstructor.
  intro x.
  pose proof (g x) as h4.
  contradiction.
  intro; contradiction.

  Grab Existential Variables.
  intro; contradiction.
Qed.


End FiniteT_card. 


(*Wyckoff*)
Lemma FiniteT_iff :
  forall T, FiniteT T <->
            (equi T False \/
             (exists U, FiniteT U /\ equi T (option U)) \/
             (exists U  (f:U->T), invertible f /\ FiniteT U)).
  intros T; split; intro h1.  
  induction h1.
  left; auto.
  apply equi_refl. 
  destruct IHh1 as [|h2]; subst.
  right.
  left.
  exists False.
  split; auto.
  apply empty_finite.
  inversion H.
  apply equi_intro with (option_map f).
  apply invertible_option_map; auto.
  destruct h2 as [h2 | h2].
  destruct h2 as [U [h2]]. 
  inversion H.
  right.
  left.
  exists (option U).
  split.
  constructor; auto.
  apply equi_intro with (option_map f).
  apply invertible_option_map; auto.
  destruct h2 as [U [f h2]].
  right; right.
  destruct h2 as [h2 h3].
  exists (option U), (option_map f).
  split.
  apply invertible_option_map; auto.
  constructor; auto.
  destruct IHh1 as [h2|h2].
  left.
  apply (equi_trans
           _ _ _
           (equi_intro _ _ _ (invertible_impl_inv_invertible _ H)) h2).
  right.
  destruct h2 as [h2|h2].
  destruct h2 as [U [h2 h3]].
  left.
  exists U.
  split; auto.
  apply (equi_trans
           _ _ _
           (equi_intro _ _ _ (invertible_impl_inv_invertible _ H)) h3).
  right. 
  destruct h2 as [U [g [h3 h4]]].
  exists U.
  exists (compose f g).
  split; auto.
  apply invertible_compose; auto.
  destruct h1 as [h1 | [h1 | h1]].
  inversion h1.
  apply (bij_finite _ _ _ empty_finite (invertible_impl_inv_invertible _ H)).
  destruct h1 as [U [h1 h2]].
  inversion h2.
  apply (bij_finite _ _ _ (add_finite _ h1) (invertible_impl_inv_invertible _ H)).
  destruct h1 as [U [f [h1 h2]]].
  apply (bij_finite _ _ _ h2 h1).
Qed.

(*Wyckoff*)
Inductive FiniteT_prop (P:forall T, FiniteT T -> Prop) : Prop :=
| empty_finite_prop : P False empty_finite -> FiniteT_prop P
| add_finite_prop : (forall U (pfu:FiniteT U),
                         P U pfu ->
                         P (option U) (add_finite U pfu)) ->
                      FiniteT_prop P
| bij_finite_prop : (forall U V (pfu:FiniteT U) (f:U->V)
                              (pfinv:invertible f),
                         P U pfu -> P V (bij_finite _ _ f pfu pfinv)) ->
                      FiniteT_prop P.                    

(*Wyckoff*)
Inductive FiniteT_prop_dep (P:forall T, FiniteT T -> Prop) : 
  forall T, FiniteT T -> Prop :=
| empty_finite_prop_dep : P False empty_finite ->
                          FiniteT_prop_dep P False empty_finite
| add_finite_prop_dep : (forall U (pfu:FiniteT U),
                         P U pfu ->
                         P (option U) (add_finite U pfu)) ->
                        forall T (pft:FiniteT T),
                          P T pft ->
                          FiniteT_prop_dep P (option T) (add_finite T pft)
| bij_finite_prop_dep : (forall U V (pfu:FiniteT U) (f:U->V)
                              (pfinv:invertible f),
                         P U pfu -> P V (bij_finite _ _ f pfu pfinv)) ->
                        forall T U (pft:FiniteT T) (f:T->U)
                               (pfinv:invertible f),
                          P T pft ->
                          FiniteT_prop_dep P U (bij_finite _ _ f pft pfinv).

(*Wyckoff*)
Lemma FiniteT_ind' :
  forall {T} (pft:FiniteT T) (P:forall T, FiniteT T->Prop),
    FiniteT_prop_dep P T pft ->
    P T pft.
  intros ? ? ? h1; induction h1; auto.
Qed.

(*Wyckoff*)
(*This section is included as a curiosity only; its use of 
  providing a canonical index on elements in a finite type is
  better served by [FinT] in [TypeUtilities2].*)
Section FiniteTIndex.

  (*option (option ( ... False*)
  Fixpoint option_false n : Type :=
    match n with
      | 0 => False
      | S n' => option (option_false n')
    end.

  Inductive FiniteT_index :
    forall {T} (pff:FiniteT T),
      (T -> (FiniteT_card _ pff) @) -> Prop :=
| FiniteT_ind_empty :
    FiniteT_index empty_finite
                (fun x => False_rect _ x)
| FiniteT_ind_add :
    forall {T} (pff:FiniteT T) (f:T->(FiniteT_card _ pff) @),
      FiniteT_index pff f ->
      FiniteT_index (add_finite _ pff)
                  (fun x =>
                     let pf0 :=
                         lt_eq_trans
                           _ _ _
                           (lt_0_Sn (FiniteT_card _ pff))
                           (eq_sym
                              (FiniteT_card_option _ pff)) in
                     match x with
                       | Some x =>
                         let z := f x in
                         let pflt :=
                             lt_eq_trans
                               _ _ _
                               (lt_n_S _ _ (proj2_sig z))
                               (eq_sym
                                  (FiniteT_card_option _ pff)) in
                         exist
                           (fun i => i < FiniteT_card
                                           _
                                           (add_finite T pff))
                           (S (proj1_sig z))
                           pflt
                           
                       | None => exist _ 0 pf0
                     end)
| FiniteT_ind_bij :
    forall {T U}
           (g:T->U) (pff:FiniteT T) (pfg:invertible g)
           (f:T->(FiniteT_card _ pff) @),
      FiniteT_index pff f ->
      FiniteT_index
        (bij_finite T U g pff pfg)
        (fun y =>
           let z := f (proj1_sig (function_inverse _ pfg) y) in
           let pfz := lt_eq_trans
                            _ _ _
                            (proj2_sig z)
                            (eq_sym
                               (FiniteT_card_bijection T U
                                  pff g pfg)) in
        
        exist
          (fun i => i < FiniteT_card _
                                 (bij_finite T U g pff pfg))
          (proj1_sig z)
          pfz).

  Lemma injective_FiniteT_index :
    forall {T} (pff:FiniteT T) (f:T->(FiniteT_card _ pff) @),
      FiniteT_index pff f ->
      injective f.
    intros T hff f h1.
    induction h1. 
    apply inj_false.
    red in IHh1; red. 
    intros x y; destruct x as [x|], y as [y|]; intro h2; apply exist_injective in h2; f_equal; try discriminate; auto.
    apply S_inj in h2.
    apply proj1_sig_injective in h2; auto.
    red; intros x y h2. 
    apply exist_injective in h2.
    apply proj1_sig_injective in h2.
    apply IHh1 in h2.
    pose proof (invertible_impl_inv_invertible _ pfg) as h3.
    pose proof (invertible_impl_bijective _ h3) as h4.
    destruct h4 as [h4 h5].
    apply h4 in h2; auto.
  Qed.
  
  Lemma surjective_FiniteT_index :
    forall {T} (pff:FiniteT T) (f:T->(FiniteT_card _ pff) @),
      FiniteT_index pff f ->
      surjective f.
    intros T hf f h1. 
    induction h1.
    red.
    intro y.
    destruct y as [? hy].
    pose proof hy as hy'.
    rewrite FiniteT_card_False in hy'; omega.
    red.
    intro y.
    destruct y as [y hy].
    destruct y as [|y].
    exists None. 
    apply proj1_sig_injective; auto.
    pose proof hy as hy'. 
    rewrite FiniteT_card_option in hy'. 
    apply lt_S_n in hy'.  
    red in IHh1. 
    specialize (IHh1 (exist (fun c => c < FiniteT_card T pff) _ hy')).
    destruct IHh1 as [x h3]. 
    pose proof (injective_FiniteT_index pff _ h1) as h4.
    pose proof (f_equal (@proj1_sig _ _) h3) as h5.
    simpl in h5; subst.
    pose proof (in_im x f) as h6.
    exists (Some x).
    f_equal; apply proof_irrelevance.
    red in IHh1.
    red.
    intro y; destruct y as [y hy].
    pose proof hy as hy'.
    rewrite FiniteT_card_bijection in hy'.
    specialize (IHh1
                  (exist
                     (fun c => c < FiniteT_card _ pff) y hy')).
    destruct IHh1 as [x hx].
    pose proof (f_equal (@proj1_sig _ _) hx) as h2.
    simpl in h2; subst.
    exists (g x).
    apply proj1_sig_injective; simpl.
    pose proof (proj2_sig (function_inverse g pfg)) as h2.
    simpl in h2.
    destruct h2 as [h2 h3].
    rewrite h2; auto.
  Qed.
   
  Lemma bijective_FiniteT_index :
    forall {T} (pff:FiniteT T) (f:T->(FiniteT_card _ pff) @),
      FiniteT_index pff f ->
      bijective f.
    intros; red; split;
    [apply injective_FiniteT_index |
     apply surjective_FiniteT_index]; auto.
  Qed. 


End FiniteTIndex.

(*Wyckoff*)
(*[Set]-version of [sumor], which is usually not necessary.*)
Inductive sumor' (A:Set) (B:Prop) : Set :=
|  inleft' : A -> sumor' A B 
| inright' : B -> sumor' A B.

(*Wyckoff*)
Fixpoint nsumorl (l:list Prop) : Set :=
  match l return Set with
    | nil => False
    | cons P l' => sumor (nsumorl l') P
  end.

Section NProd.

(*Wyckoff*)
Lemma nprod0_eq :
  forall {T} (x y:T^0),
    x = y.
  intros T x y; destruct x, y; auto.
Qed.


(*Wyckoff*)
Fixpoint in_nprod {T n} y : T^n -> Prop :=
  match n with
    | 0 => fun _ => False
    | S n' => fun x =>
                y = fst x \/ in_nprod y (snd x)
  end.

(*Wyckoff*)
Fixpoint hd_nprod {T n} : n > 0 -> forall (x:T^n), T :=
  match n with
    | 0 => fun pf => False_rect _ (lt_irrefl _ pf)
    | S _ => fun _ x => fst x
  end.

(*Wyckoff*)
Fixpoint tl_nprod
         {T n} : n > 0 -> forall (x:T^n), T^pred n :=
  match n with
    | 0 => fun pf => False_rect _ (lt_irrefl _ pf)
    | S _ => fun _ x => snd x
  end.

(*Wyckoff*)
Lemma in_nprod_fst :
  forall {T n} (x:T^S n),
    in_nprod (fst x) x.
  intros; simpl; left; auto.
Qed.



(*Wyckoff*)
Definition all_tuples (T:Type) := {n:nat & T^n}.

(*Wyckoff*)
Lemma full_set_all_tuples_eq : 
  forall T:Type,
    Full_set (all_tuples T) = 
    IndexedUnion (fun n:nat => [x:all_tuples T | projT1 x = n]).
intro T.
apply Extensionality_Ensembles; red; split; red; intros x h1.
clear h1. destruct x as [n x].
apply indexed_union_intro with n.
constructor. simpl.
reflexivity.
constructor.
Qed.


(*Wyckoff*)
Definition match_all_tuples_type T n : Type :=
  match n with
    | 0 => unit
    | S _ => (T * all_tuples T)%type
  end.

(*Wyckoff*)
Definition match_all_tuples {T} (tpl:all_tuples T) :
  match_all_tuples_type T (projT1 tpl) :=
  let (n, x) as pr return (match_all_tuples_type T (projT1 pr)) := tpl in
  match n as m return (T ^ m -> match_all_tuples_type T m) with
    | 0 => fun _ => tt
    | S m' =>
      fun tpl': T ^ S m' =>
        (fst tpl', existT _ m' (snd tpl'))
  end x.



(*Wyckoff*)
Lemma finite_nprod :
  forall T n,
    FiniteT T ->
    FiniteT (T^n).
  intros T n.
  induction n as [|n ih]; simpl.
  intros.
  apply unit_finite.
  intro h1.
  specialize (ih h1).
  apply finite_prod; auto.
Qed.


(*Wyckoff*)
Lemma countable_nprod : 
  forall (T:Type) (n:nat),
    CountableT T ->
    CountableT (T^n).
intros T n. 
induction n as [|n h1]; simpl; auto.
intros. 
apply FiniteT_impl_CountableT.
apply unit_finite.
intro h2.
specialize (h1 h2).
apply countable_product; auto.
Qed.


(*Wyckoff*)
Lemma countable_all_tuples_of_countable_type : 
  forall T:Type,
    CountableT T ->
    CountableT (all_tuples T).
intros T h1.
rewrite countable_full_iff.
rewrite full_set_all_tuples_eq.
apply countable_union.
apply countable_nat.
intro n.
pose (fun x:T^n => existT _ n x) as f.
assert (h3: [x : all_tuples T | projT1 x = n] = 
            Im (Full_set (T ^ n)) f).
  apply Extensionality_Ensembles; red; split; red; intros x h3.
  destruct h3 as [h3]. 
  destruct x as [n' x]. simpl in h3. subst.
  apply Im_intro with x. constructor.
  unfold f.
  f_equal.
  destruct h3 as [x h3]. subst. clear h3.
  constructor. unfold f. simpl. reflexivity.
rewrite h3.
apply countable_img.
rewrite <- countable_full_iff.
apply countable_nprod; auto.
Qed.

(*Wyckoff*)
Lemma nprod_to_list_eq_nil_iff :
  forall {T n} (x:T^n),
    nprod_to_list _ _ x = nil <->
    n = 0.
  intros T n.
  destruct n as [|n]; simpl.
  tauto.
  intro x; destruct x as [fx sx].
  split; intro h1; discriminate.
Qed.

(*Wyckoff*)
Lemma length_nprod_to_list :
  forall {T n} (x:T^n),
    length (nprod_to_list _ _ x)  = n.
  intros ? n.
  induction n as [|n ih]; simpl; auto.
  intro x.
  destruct x as [x y].
  simpl; f_equal.
  apply ih.
Qed.

(*Wyckoff*)
Fixpoint nprod_deg {n T} :
  T^n -> nat :=
  match n with
    | 0 => fun _ => 0
    | S n' => fun x => S (nprod_deg (snd x))
  end.

(*Wyckoff*)
Lemma length_nprod_of_list :
  forall {T} (l:list T),
    nprod_deg (nprod_of_list _ l) = length l.
  intros ? l; induction l; simpl; auto.
Qed.


(*Wyckoff*)
Definition hd_prod {T n} : T^S n  -> T :=
  fun x => fst x.


(*Wyckoff*)
Definition tl_prod {T n} : T^S n->T ^ n :=
  fun x =>
    snd x.

(*Wyckoff*)
(*A tuple of [n] identical copies of [x]*)
Fixpoint nprod_sing {T} (x:T) n : T^n :=
  match n with 
    | 0 => tt
    | S n' => (x, nprod_sing x n')
  end.

(*Wyckoff*)
Definition None_n {T:Type} n : (option T)^n :=
  nprod_sing None n.

(*Wyckoff*)
Fixpoint removelast_prod {T:Type} {n:nat} : T ^ S n -> T^n :=
  match n return (T ^ S n -> T ^ n) with
    | O => fun  y => tt
    | S n' => fun y => (fst y, removelast_prod (snd y))
  end.

(*Wyckoff*)
Fixpoint nth_prod {T:Type} {n:nat} i : i < n -> T^n -> T :=
  match n return (i < n -> T^n -> T) with 
    | 0 => fun pf _ => False_rect _ (lt_n_0 _ pf) 
    | S n' =>  match i with
                  | 0 =>
                    fun _ x => fst x
                  | S i' =>
                    fun pflt x => 
                      let pflt' := lt_S_n _ _ pflt in
                      @nth_prod T _ _ pflt' (snd x)
                end
  end.


(*Wyckoff*)
Lemma nth_prod0 :
  forall {T:Type} {n} (pf:0 < S n) (x:T^S n),
    nth_prod 0 pf x =
    fst x.
  intros ? n; destruct n; simpl; auto.
Qed.

(*Wyckoff*)
Lemma hd_nprod_eq :
  forall {T n} (pf:n > 0) (x:T^n),
    hd_nprod pf x =
    nth_prod 0 pf x.
  intros T n; destruct n; auto.
  intros; omega.
Qed.


(*Wyckoff*)
Lemma nth_prod_S :
  forall {T n} i (pfsi:S i < S n) (x:T^S n),
    nth_prod (S i) pfsi x =
    nth_prod i (lt_S_n _ _ pfsi) (snd x).
  auto.
Qed.


(*Wyckoff*)
Lemma in_nth_prod :
  forall {T n} i (pfi:i < n) (x:T^n),
    in_nprod (nth_prod i pfi x) x.
  intros T n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intro i.
  destruct i as [|i].
  intros; left; auto.
  intros hi x.
  specialize (ih i (lt_S_n _ _ hi) (snd x)); auto.
Qed.

(*Wyckoff*)
Lemma in_nprod_iff :
  forall {T n} y (x:T^n),
    in_nprod y x <->
    exists i (pfi:i < n),
      nth_prod i pfi x = y.
  intros T n.
  induction n as [|n ih]; simpl.
  intros; split; intro h1.
  contradiction.
  destruct h1 as [? [? ?]]; omega.
  intros y x; split; intro h1.  
  destruct h1 as [|h1]; subst.
  exists 0, (lt_0_Sn _); auto. 
  rewrite ih in h1. 
  destruct h1 as [i [hi ?]]; subst.
  exists (S i), (lt_n_S _ _ hi).
  f_equal; apply proof_irrelevance.
  destruct h1 as [i [hi h2]].
  destruct i as [|i]; subst.
  left; auto.
  right.
  rewrite ih.
  exists i, (lt_S_n _ _ hi); auto.
Qed.

(*Wyckoff*)
Lemma nth_prod_seg_fun_pf_eq :
  forall {T n}
         (x:T^n)
         (lt_fun : forall i, i < n -> i < n),
    (fun i (pfi:i < n) =>
       (nth_prod i pfi x)) =
    (fun i (pfi:i < n) =>
       (nth_prod i (lt_fun i pfi) x)).
  intros; apply functional_extensionality_dep;
  intro; apply functional_extensionality_dep;
  intros; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma nth_prod_seg_fun_pf2_eq :
  forall {T n}
         (x:T^n)
         (lt_fun1 : forall i, i < n -> i < n)
         (lt_fun2 : forall i, i < n -> i < n),
    (fun i (pfi:i < n) =>
       (nth_prod i (lt_fun1 i pfi) x)) =
    (fun i (pfi:i < n) =>
       (nth_prod i (lt_fun2 i pfi) x)).
  intros; apply functional_extensionality_dep;
  intro; apply functional_extensionality_dep;
  intros; f_equal; apply proof_irrelevance.
Qed.


              


(*Wyckoff*)
Fixpoint plus_nprod {n} : forall (x:nat^n), nat :=
  match n with
    | 0 => fun _ => 0
    | S n' => fun x => (fst x) + (plus_nprod (snd x))
  end.

(*Wyckoff*)
Fixpoint times_nprod {n} : forall (x:nat^n), nat :=
  match n with
    | 0 => fun _ => 1
    | S n' => fun x => (fst x) * (times_nprod (snd x))
  end.

(*Wyckoff*)
Definition all_sing_nprod {T n} (x:T^n) (a:T) :=
  forall b, in_nprod b x -> b = a.





(*Wyckoff*)
Definition in_pre_w {X S n} (s:S) (v:X->S) (w:S^n) : Prop :=
  forall i (pfi:i < n),
    in_pre (nth_prod i pfi w) v.


(*Wyckoff*)
Definition last_prod {T n} : T^S n -> T :=
  fun x => nth_prod n (lt_n_Sn _) x.


(*Wyckoff*)
Fixpoint im_nprod {T U n} (f:T->U) : T^n -> U^n :=
  match n with
    | 0 => fun _ => tt
    | S n' => fun x => ((f (fst x)), im_nprod f (snd x))
  end.

(*Wyckoff*)
Lemma im_nprod_im_nprod : 
  forall {T U V:Type} {n:nat} (t:T^n) (f:T->U) (g:U->V),
    im_nprod g (im_nprod f t) = im_nprod (fun x=> g (f x)) t.
intros T U V n.
induction n as [|n h1].
intros t f g; simpl. reflexivity.
intros t f g; simpl.
apply injective_projections; simpl. reflexivity.
destruct t as [x t]. simpl.
apply h1.
Qed.

(*Wyckoff*)
Lemma im_nprod_inj : 
  forall {T U:Type} {n:nat} (t1 t2:T^n) (f:T->U),
    Injective f -> im_nprod f t1 = im_nprod f t2 ->
    t1 = t2.
intros T U.
induction n as [|n h1]; intros t1 t2 f h2 h3; simpl.
simpl in h3.
destruct t1. destruct t2. reflexivity.
destruct t1 as [ft1 st1].
destruct t2 as [ft2 st2].
simpl in h3.
inversion h3 as [h4].
specialize (h1 _ _ _ h2 H).
red in h2.
specialize (h2 _ _ h4).
apply injective_projections; auto.
Qed.


(*Wyckoff*)
Fixpoint map_nprod {U n} : seg_fun n U -> U^n :=
  match n with
    | 0 => fun _ => tt
    | S n' =>
      fun f =>
        let f' := seg_fun_to_S f in
        (f 0 (lt_0_Sn _), map_nprod f')
  end.

(*Wyckoff*)
Lemma nth_map_nprod :
  forall {U n} (f:seg_fun n U) i (pfi:i < n),
    nth_prod i pfi (map_nprod f) =
    f i pfi.
  intros U n. revert U.
  induction n as [|n ih].
  intros; omega.
  intros U f i.
  destruct i as [|i].
  simpl.
  intros; f_equal; apply proof_irrelevance. 
  intro h1.
  specialize (ih U (seg_fun_to_S f) i (lt_S_n _ _ h1)).
  simpl.
  rewrite ih.
  unfold seg_fun_to_S; f_equal.
  apply proof_irrelevance.
Qed.

Definition map_nprod_functional :
  forall {U n} (f g:seg_fun n U),
    f = g ->
    map_nprod f =
    map_nprod g.
  intros; subst; auto.
Qed.


Lemma map_nprod_eq_iff :
  forall {U n} (f:seg_fun n U) (y:U^n),
    map_nprod f = y <->
    forall i (pfi:i < n),
      f i pfi = nth_prod i pfi y.
  intros U n; induction n as [|n ih]; simpl.
  intros; split; subst.
  intros; omega.
  intros; apply unit_eq.
  unfold seg_fun_to_S.
  intros f y; destruct y as [y y']; split; intro h1. 
  inversion h1; subst. 
  intro i; destruct i as [|i]; simpl; intros; repeat f_equal. 
  apply proof_irrelevance.
  rewrite nth_map_nprod; f_equal; apply proof_irrelevance.
  pose proof (h1 0 (lt_0_Sn _)) as h2; simpl in h2; subst.
  f_equal.
  rewrite ih.
  intros i hi.
  rewrite (h1 (S i) (lt_n_S _ _ hi)); f_equal; apply proof_irrelevance.
Qed.
  


  

(*This transfer is done pointwise, rather than as a whole, to avoid 
universe inconsistencies, which in general arise when you're assuming 
too much type-theoretical information, which in this case is that,  
type-substitutions would destruct correspdondingly.*)  
(*Wyckoff*)
Definition transfer_nprod {T n n'} (pfe:n = n') (x:T^n) : T^n' :=
  map_nprod (fun i (pfi:i < n') => nth_prod i (lt_eq_trans _ _ _ pfi
                                                           (eq_sym pfe))
                                            x).

(*Wyckoff*)
Definition transfer_nprod_r {T n n'} (pfe:n = n') (x:T^n') : T^n :=
  map_nprod (fun i (pfi:i < n) =>
               nth_prod i (lt_eq_trans _ _ _ pfi pfe) x).


(*Wyckoff*)
Lemma transfer_nprod_eq_refl :
  forall {T n} (pfe:n = n) (x:T^n),
    transfer_nprod pfe x = x.
  unfold transfer_nprod; intros T n. revert T.
  induction n as [|n ih]; simpl.
  intros; apply unit_eq.
  intros T he x; simpl.
  destruct x as [x w]; simpl.
  f_equal.
  specialize (ih T (S_inj _ _ he) w).
  rewrite <- ih.
  f_equal.
  apply functional_extensionality_dep.
  intro.
  unfold seg_fun_to_S.
  apply functional_extensionality_dep.
  intro.
  f_equal.
  apply proof_irrelevance.
  simpl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_nprod_r_eq_refl :
  forall {T n} (pfe:n = n) (x:T^n),
    transfer_nprod_r pfe x = x.
  unfold transfer_nprod_r; intros T n. revert T.
  induction n as [|n ih]; simpl.
  intros; apply unit_eq.
  intros T he x; simpl.
  destruct x as [x w]; simpl.
  f_equal.
  specialize (ih T (S_inj _ _ he) w).
  rewrite <- ih.
  f_equal.
  apply functional_extensionality_dep.
  intro.
  unfold seg_fun_to_S.
  apply functional_extensionality_dep.
  intro.
  f_equal.
  apply proof_irrelevance.
  simpl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_nprod_eq :
  forall {T m n} (pfe:m = n) (x:T^m),
    transfer_nprod pfe x =
    transfer_nprod_r (eq_sym pfe) x.
  intros; subst; rewrite transfer_nprod_eq_refl, transfer_nprod_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_nprod_r_eq :
  forall {T m n} (pfe:m = n) (x:T^n),
    transfer_nprod_r pfe x =
    transfer_nprod (eq_sym pfe) x.
  intros; subst; rewrite transfer_nprod_eq_refl, transfer_nprod_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_nprod_r_undoes_transfer_nprod :
    forall {T m n} (pfe pfe':m = n) (x:T^m),
      transfer_nprod_r pfe' (transfer_nprod pfe x) = x.
  intros; subst; rewrite transfer_nprod_eq_refl, transfer_nprod_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_nprod_undoes_transfer_nprod_r :
    forall {T m n} (pfe pfe':m = n) (x:T^n),
      transfer_nprod pfe' (transfer_nprod_r pfe x) = x.
  intros; subst; rewrite transfer_nprod_eq_refl, transfer_nprod_r_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_nprod_S :
  forall {T n m} (pfe:S n = S m) (x:T^S n),
    transfer_nprod pfe x =
    (fst x, transfer_nprod (S_inj _ _ pfe) (snd x)).
  intros ? ? ? h1 x; pose proof (S_inj _ _ h1); subst;
  intros; do 2 rewrite transfer_nprod_eq_refl; destruct x; auto.
Qed.

(*Wyckoff*)
Lemma nprod_transfer_plus_pr :
  forall {T m n} (pf:S m = S m + n) (x:T^m) (a:T),
    let pf' := S_eq_plus _ _ pf in
    transfer_nprod pf (a, x) =
    (a, transfer_nprod pf' x).
  intros T m n h1.  
  pose proof (eq_plus_l _ _ h1); subst. 
  intros x a.
  simpl.
  generalize (S_eq_plus m 0 h1).
  revert h1.
  simpl.
  rewrite <- plus_n_O.
  intros.
  do 2 rewrite transfer_nprod_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_nprod_r_S :
  forall {T n m} (pfe:S n = S m) (x:T^S m),
    transfer_nprod_r pfe x =
    (fst x, transfer_nprod_r (S_inj _ _ pfe) (snd x)).
  intros ? ? ? h1 x; pose proof (S_inj _ _ h1); subst;
  intros; do 2 rewrite transfer_nprod_r_eq_refl; destruct x; auto.
Qed.

(*Wyckoff*)
Lemma nth_prod_transfer_prod :
  forall {T m n} {U:T->Type}
         (pfe:m = n)
         (w:T^m)
         i (pfi:i < n),
    let pfi' := lt_eq_trans _ _ _ pfi (eq_sym pfe) in
    nth_prod i pfi (transfer_nprod pfe w) =
    nth_prod i pfi' w.
  simpl; intros; subst; rewrite transfer_nprod_eq_refl; f_equal;
  apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma nth_prod_transfer_prod_r :
  forall {T m n} {U:T->Type}
         (pfe:m = n)
         (w:T^n)
         i (pfi:i < m),
    let pfi' := lt_eq_trans _ _ _ pfi pfe in
    nth_prod i pfi (transfer_nprod_r pfe w) =
    nth_prod i pfi' w.
  simpl; intros; subst; rewrite transfer_nprod_r_eq_refl; f_equal;
  apply proof_irrelevance.
Qed.



(*Wyckoff*)
Lemma nth_prod_functional0 :
  forall {T n n'} i (pfi:i < n) (pfi':i < n') (w:T^n)
         (pfe:n = n'),
    @nth_prod _ n i pfi w =
    @nth_prod _ n' i pfi' (transfer_nprod pfe w).
  intros; subst; rewrite transfer_nprod_eq_refl; f_equal;
  apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma nth_prod_functional1 :
  forall {T n} i i' (pfi:i < n) (pfi':i' < n) (w:T^n)
         (pfe:i = i'),
    nth_prod i  pfi  w =
    nth_prod i' pfi' w.
  intros; subst; f_equal;
  apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma nth_prod_functional2 :
  forall {T n} i (pfi pfi':i < n) (w:T^n),
    nth_prod i pfi  w =
    nth_prod i pfi' w.
  intros; f_equal;
  apply proof_irrelevance.
Qed.

(**Wyckoff*)
Lemma nth_prod_functional3 :
  forall {T n} i (pfi pfi':i < n) (w w':T^n),
    w = w' ->
    nth_prod i pfi w =
    nth_prod i pfi' w'.
  intros; subst; f_equal;
  apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma nprod_ext :
  forall {T n} (x y:T^n),
    (forall i (pfi:i < n),
       nth_prod i pfi x =
       nth_prod i pfi y) ->
    x = y.
  intros T n; induction n as [|n ih]; simpl.
  intros; apply unit_eq.
  intros x y.
  destruct x as [t x], y as [u y]; simpl.
  intro h1.
  pose proof (h1 0 (lt_0_Sn _)) as h2; simpl in h2; subst; f_equal.
  apply ih.
  intros i hi.
  specialize (h1 (S i) (lt_n_S _ _ hi)); simpl in h1.
  erewrite nth_prod_functional2.
  rewrite h1 at 1; f_equal; apply proof_irrelevance.
Qed.
  



(*Wyckoff*)
(*This is just the first [n] entries, as opposed to the last [n], 
as in [tl_prod].*)
Fixpoint fst_tl_prod {T n} : T^S n->T ^ n :=
  fun x => map_nprod (fun i (pf:i < n) =>
                       let pf' := lt_trans _ _ _ pf (lt_n_Sn _) in
                       nth_prod i pf' x).



(*Wyckoff*)
(*This function allows you to specify a (presumably surjective) 
mapping from [seg n] to [seg r] that corresponds indices
of the returned tuple in [T^n] to indices of the given tuple,
[x:T^r].*)  
Definition nprod_identify {T n r}
           {t:seg_fun n nat} :
  seg_fun_endo t r -> T^r -> T^n :=
  fun pfs x =>
    map_nprod (fun i (pfi:i < n) =>
                 let pflt := lt_seg_fun_endo t pfs i pfi in
                 nth_prod (t i pfi) pflt x).
                 
(*This is the analogue of [nprod] in [Set].  Usaully unnecessary.*)
Fixpoint setprod (U:Set) (n:nat) : Set :=
  match n with
    | 0 => unit
    | S n' => (U * setprod U n')%type
  end.


(*Wyckoff*)
Definition full_segT_ind (n:nat) : (n @)^n :=
  map_nprod (fun i (pfi:i < n) => exist (fun c => c < n) i pfi).

(*Wyckoff*)
Lemma nth_full_segT_ind :
  forall {n} i (pfi:i < n),
    proj1_sig (nth_prod i pfi (full_segT_ind n)) =
    i.
  intros n i hi.
  unfold full_segT_ind.
  rewrite nth_map_nprod; simpl; auto.
Qed.

(*Wyckoff*)
Lemma nth_full_segT_ind_S :
  forall {n} i (pfi:i < n),
    proj1_sig (nth_prod i pfi (snd (full_segT_ind (S n)))) =
    S i.
  intro n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros i hi.
  destruct i as [|i]; simpl; auto.
  specialize (ih i (lt_S_n _ _ hi)).
  rewrite nth_map_nprod.
  unfold seg_fun_to_S; simpl; auto.
Qed.


(*Wyckoff*)
Definition seg_fun_type_to_ind {n} (f:seg_fun n Type) :
  n @ -> Type :=
  fun i => f (proj1_sig i) (proj2_sig i).


(*Wyckoff*)
Lemma seg_fun_type_to_ind0 :
  forall {n} (f:seg_fun (S n) Type),
    seg_fun_type_to_ind f (fst (full_segT_ind (S n))) =
    f 0 (lt_0_Sn n).
  unfold seg_fun_type_to_ind; auto.
Qed.

(*Wyckoff*)
Definition pointwise_ind_eq
           {T U V n}
           (P:T->V) (Q:U->V)
           (w:T^n)
           (w':U^n) : Prop :=
  forall i (pfi:i < n),
    P (nth_prod i pfi w) =
    Q (nth_prod i pfi w').

(*Wyckoff*)
Lemma eq_pointwise_ind_eq :
  forall {T U n}
         {w x:T^n},
    w = x ->
    forall (P:T->U),
      pointwise_ind_eq P P w x.
  intros; subst; red; intros; auto.
Qed.

  
(*Wyckoff*)
Definition pointwise_ind_eq_to_S
           {T U V n}
           {P:T->V} {Q:U->V}
           {w:T^(S n)}
           {w':U^(S n)}
           (pfp:pointwise_ind_eq P Q w w') :
  pointwise_ind_eq P Q (snd w) (snd w').
  red in pfp; red; intros i hi; specialize (pfp (S i) (lt_n_S _ _ hi));
  simpl in pfp; erewrite nth_prod_functional2.
  rewrite pfp at 1.
  repeat f_equal; apply proof_irrelevance.
Defined.

(*Wyckoff*)
Lemma pointwise_ind_eq_seg_fun_type_to_ind :
  forall {n} (f:seg_fun (S n) Type),
    pointwise_ind_eq
      (seg_fun_type_to_ind (seg_fun_to_S f))
      (seg_fun_type_to_ind f)
      (full_segT_ind n)
      (snd (full_segT_ind (S n))).
  intros n f; red; intros i hi.
  unfold seg_fun_type_to_ind, seg_fun_to_S; f_equal.
  gen0.
  intro h1.
  symmetry.
  gen0.
  revert h1. 
  rewrite nth_full_segT_ind.
  rewrite nth_full_segT_ind_S.
  intros; f_equal.
  apply proof_irrelevance.
Qed.

(*Wyckoff*)
(*A binary relation establishing corresponding types of tuples.*)
Definition pointwise_ind_eq_rel 
           {T V}
           (P:T->V) n :
  Relation (T ^ n) :=
  fun w w' => 
    forall i (pfi:i < n),
      P (nth_prod i pfi w) =
      P (nth_prod i pfi w').

(*Wyckoff*)
Lemma refl_pointwise_ind_eq_rel : 
  forall {T V}
         (P:T->V) n,
    Reflexive (pointwise_ind_eq_rel P n).
  intros; red; intro; red; intros; auto.
Qed.

(*Wyckoff*)
Lemma symm_pointwise_ind_eq_rel : 
  forall {T V}
         (P:T->V) n,
    Symmetric (pointwise_ind_eq_rel P n).
  intros; red; intro; red; intros; auto.
Qed.

(*Wyckoff*)
Lemma trans_pointwise_ind_eq_rel : 
  forall {T V}
         (P:T->V) n,
    Transitive (pointwise_ind_eq_rel P n).
  intros; red; intro; red; intros; auto.
  rename H into h1, H0 into h2.
  red in h1, h2.
  specialize (h1 i pfi).
  specialize (h2 i pfi).
  congruence.
Qed.

(*Wyckoff*)
Add Parametric Relation
    {T V n} (P:T->V) : (T ^ n) (pointwise_ind_eq_rel P n)
    reflexivity proved by (refl_pointwise_ind_eq_rel P n)
    symmetry proved by (symm_pointwise_ind_eq_rel P n)
    transitivity proved by (trans_pointwise_ind_eq_rel P n)
      as pointwise_ind_eq_rel_equiv.




           
Section TupleProd.

  (*For "many sorts"*)
  Fixpoint tuple_prod {T n} (A:T->Type) : T^n -> Type :=
    match n with
      | 0 => fun _ => unit
      | S n' => fun x:T^S n' =>
                  ((A (fst x)) * (tuple_prod A (snd x)))%type
  end.
  
  Fixpoint nth_tuple_prod {T n} {A:T->Type} :
    forall {t:T^n}, tuple_prod A t ->
                    forall i (pf:i < n), A (nth_prod i pf t)  :=
  match n with
    | 0 => fun _ _ _ pf0 => False_rect _ (lt_n_0 _ pf0)
    | S n' => fun (t:T^(S n')) (y:tuple_prod A t) i =>
                   match i with
                     | 0 => fun _ => (fst y)
                     | S i' =>
                       fun pfi => 
                         let pfi' := lt_S_n _ _ pfi in
                         @nth_tuple_prod T n' A (snd t) (snd y) i' pfi'
                   end
  end.
  
  
  Lemma tuple_prod_const :
    forall {n} T U (x:T^n),
      tuple_prod (fun _ => U) x = (U^n)%type.
    intros n; induction n as [|n ih]; auto.
    intros ? ? x.
    destruct x as [fx sx].
    specialize (ih T U sx).
    simpl.
    f_equal; auto.
  Defined.

  Lemma tuple_prod0 :
    forall {S} {A:S->Type} (x:S^0),
      tuple_prod A x = unit.
    auto.
  Qed.
 
  Lemma tuple_prod_S :
    forall {T n} (A:T->Type) (x:T^S n),
      tuple_prod A x =
      (A (fst x) * tuple_prod A (snd x))%type.
    auto.
  Qed.

  Definition transfer_tuple_prod1 
             {T} {A:T->Type} {x:T^1}
             (p:tuple_prod A x) : A (fst x) :=
    fst p.

  Definition transfer_to_tuple_prod1 
             {T} {A:T->Type} {x:T}
             (a:A x) :
    tuple_prod A (x, tt) (n := 1) :=
    (a, tt).

  
  Definition tuple_prod_all 
             {S} (A:S->Type) (x:all_tuples S) : Type :=
    tuple_prod A (projT2 x).

  
  Lemma tuple_prod_all0 :
    forall {S} (A:S->Type) (x:S^0),
      tuple_prod_all A (existT _ _ x) = unit.
    unfold tuple_prod_all; intros; simpl; auto.
  Qed.

  Lemma tuple_prod_all_tuple : 
    forall {S} (A:S->Type) (x:all_tuples S),
      tuple_prod_all A x =
      tuple_prod A (projT2 x).
    unfold tuple_prod_all, tuple_prod;
    intros ? ? x; destruct x as [n x].
    revert x A. revert S.
    induction n as [|n ih]; simpl; intros; simpl; try simpl in ih; auto.
  Qed.

    
  Lemma tuple_prod_functional :
    forall {T n} {A B:T->Type} {w w':T^n},
      pointwise_ind_eq A B w w' ->
      tuple_prod A w =
      tuple_prod B w'.
    intros T n; revert T.
    induction n as [|n ih]; simpl; auto.
    intros T A B w w' h1.
    pose proof (h1 0 (lt_0_Sn _)) as h2; simpl in h2.
    f_equal; auto.
    apply ih.
    intros i hi.
    specialize (h1 (S i) (lt_n_S _ _ hi)).
    simpl in h1.
    erewrite f_equal_dep.
    rewrite h1 at 1.
    reflexivity.
    reflexivity.
  Qed.


  Lemma tuple_prod_seg_functional :
    forall {n r s} (A:r @->Type) (B:s @ ->Type)
           (w: (r @)^n)
           (w':(s @)^n),
      pointwise_ind_eq A B w w' ->
      tuple_prod A w =
      tuple_prod B w'.
    intro n.
    induction n as [|n ih]; simpl; auto;
    intros r s A B w w' h1.
    f_equal.
    specialize (h1 0 (lt_0_Sn _)); simpl in h1; auto.
    specialize (ih r s A B (snd w) (snd w')).
    apply ih.
    intros i hi.
    specialize (h1 (S i) (lt_n_S _ _ hi)); simpl in h1.
    erewrite f_equal_dep.
    rewrite h1 at 1.
    repeat f_equal.
    apply proof_irrelevance.
    reflexivity.
  Qed.

  
  Lemma tuple_prod_ext :
    forall {T n} {A:T->Type} {w:T^n}
           (p q:tuple_prod A w),
      (forall i (pfi:i < n),
         nth_tuple_prod p i pfi =
         nth_tuple_prod q i pfi) ->
      p = q.
    intros T n.
    induction n as [|n ih]; simpl.
    intros; apply unit_eq.
    intros A w p q.
    destruct w as [w w'], p as [p p'], q as [q q']; simpl;
    simpl in w', p, q, p', q'.
    intro h1.
    pose proof (h1 0 (lt_0_Sn _)) as h2; simpl in h2; subst; f_equal.
    apply ih.
    intros i hi.
    specialize (h1 _ (lt_n_S _ _ hi)); simpl in h1.
    rewrite (proof_irrelevance _ hi (lt_S_n _ _ (lt_n_S _ _ hi))); auto.
  Qed.

  
  
  Definition Sfun {S} (A B:S -> Type) : Type :=
    forall s, A s -> B s.

  
  Definition Sfun' {S U} (A B:S -> Ensemble U) : Type :=
    forall s, fun_in_ens (A s) (sig_set (B s)).
  
  Definition Sfun_dep {S} {U V:S->Type}
             (A:forall s:S, Ensemble (U s))
             (B:forall s:S, Ensemble (V s)) : Type :=
    forall s, fun_in_ens (A s) (sig_set (B s)).
  

End TupleProd.


Fixpoint snth_prod_aux {U n} : setprod U (S n) -> nat -> U :=
  match n with
    | 0 => fun x _ => fst x
    | n' => fun x i =>
              match i with
                | 0 => fst x
                | S i' => snth_prod_aux (snd x) i'
              end
  end.


Definition snth_prod {U n} :
  setprod U n -> (match n with
                     | 0 => unit
                     | S _ => nat -> U
                   end) :=
  match n with
    | 0 => fun _ => tt
    | S n' => fun x => snth_prod_aux x
  end.

Fixpoint setprod_to_nprod {U n} : setprod U n -> U^n :=
  match n with
    | 0 => fun _ => tt
    | S n' => fun x =>
                let f := fst x in
                let s := snd x in
                let stn := setprod_to_nprod s in
                (f, stn)
  end.


Lemma snth_prod_compat :
  forall {U n} (x:setprod U (S n)) i
         (pfi:i < S n),
    snth_prod x i =
    nth_prod i pfi (setprod_to_nprod x).
  intros U n. 
  induction n as [|n ih].
  intros x i h1.
  pose proof (lt1 _ h1); subst; simpl; auto.
  intros x i h1.
  destruct i as [|i].
  simpl; auto.
  destruct i as [|i].
  specialize (ih (snd x) 0 (lt_0_Sn _)).
  simpl in ih; simpl; auto.
  specialize (ih (snd x) (S i) (lt_S_n _ _ h1)).
  simpl in ih; simpl; auto.
Qed.

Lemma snth_prod_compat' :
  forall {U n} (x:setprod U (S n)) i
         (pfi:i >= S n),
    snth_prod x i =
    nth_prod n (lt_n_Sn _) (setprod_to_nprod x).
  intros U n.
  induction n as [|n ih].
  intros; auto.
  intros x i h1.
  destruct i as [|i].
  destruct n; omega.
  specialize (ih (snd x) i (le_S_n _ _ h1)).
  simpl in ih; simpl; auto.
  rewrite ih.
  f_equal.
  apply proof_irrelevance.
Qed.

End NProd.

Definition ind_prod {T n} (x:T^n) (i:nat) : option T :=
  match (lt_le_dec i n) with
    | left pflt => Some (nth_prod i pflt x)
    | right _ => None
  end.

(*Chosen as [right associativity] as a dual to "exponentation."*)
Notation "x '__' i" := (ind_prod x i) (at level 30, right associativity).


Section RandomFacts.
  
(*Wyckoff*)
Lemma finite_seg_one_t : 
  forall n:nat,
    FiniteT (seg_one_t n).
intro n. 
pose proof (finite_seg_one n) as h1.
apply Finite_ens_type in h1.
assert (h2:forall m:seg_one_t n,
             In (seg_one n) (proj1_sig m)).
  intro m. destruct m as [m h3].
  constructor. simpl. assumption.
pose (fun i => (exist (fun m => In (seg_one n) m) _ (h2 i))) as f.
assert (h3:Injective f).
  red. intros x y h4.
  destruct x as [x h5], y as [y h6].
  unfold f in h4. simpl in h4.
  apply exist_injective in h4.
  apply proj1_sig_injective; auto.
eapply inj_finite.
apply h1.
apply h3.
intros; apply classic.
Qed.

(*Wyckoff*)
Lemma iff_pred_sig_eq : 
  forall {T:Type} (P Q:T->Prop),
    iff_pred P Q ->
    {x|P x}={x|Q x}.
  intros T P Q h1; destruct h1 as [h1 h2]; red in h1; do 2 red in h2.
  f_equal. 
  apply functional_extensionality.
  intro x.
  specialize (h1 x); specialize (h2 x).
  apply prop_ext; tauto.
Qed.

End RandomFacts.


(*Reserach standard library [eq_dep] to interface with [transfer]!*)
Section Transfer.
(* Wyckoff *)
Definition transfer {T U:Type} (pf:T=U) (x:T) : U.
revert x. rewrite pf. intro u. refine u.
Defined.

(*Wyckoff*)
Definition transfer_r {T U:Type} (pf:T = U) (y:U) : T.
revert y. rewrite pf. intro u. refine u.
Defined.

(*Wyckoff*)
Lemma transfer_eq : 
  forall {T U:Type} (pf:T=U) (x:T),
    transfer pf x = eq_rect _ id x _ pf.
intros; subst; simpl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_r_eq : 
  forall {T U:Type} (pf:T=U) (x:U),
    transfer_r pf x = eq_rect_r id x pf.
intros; subst; simpl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_eq_refl : 
  forall {T:Type} (x:T)
         (pfe:T = T),
    transfer pfe x = x.
  intros T x h1;
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  unfold transfer, eq_rec_r, eq_rec, eq_rect; simpl; auto.
Defined.

(*Wyckoff*)
Lemma transfer_r_eq_refl : 
  forall {T:Type} (x:T)
         (pfe:T = T),
    transfer_r pfe x = x.
  intros T x h1;
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  unfold transfer_r, eq_rec_r, eq_rec, eq_rect; simpl; auto.
Defined.

(*Wyckoff*)
Lemma match_eq_refl_eq :
  forall {T} {U:T->Type}
         {x y}
         (pfxy:x = y)
         (pfeu:U x = U y)
         (ux:U x),
    match pfeu in (_ = y) return y with
      | eq_refl => ux
    end =
    transfer pfeu ux.
  intros; subst; rewrite transfer_eq_refl;
  pose proof (proof_irrelevance _ pfeu (eq_refl _)); subst; auto.
Qed.




(*Wyckoff*)
Lemma transfer_transfer_r_compat : 
  forall {T U:Type} (pf:T=U) (x:T),
    transfer pf x = transfer_r (eq_sym pf) x.
  intros; subst; rewrite transfer_eq_refl, transfer_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_r_transfer_compat : 
  forall {T U:Type} (pf:T=U) (y:U),
    transfer_r pf y = transfer (eq_sym pf) y.
  intros; subst; rewrite transfer_eq_refl, transfer_r_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_undoes_transfer_r : 
  forall {T U:Type} (pf:T=U) (y:U),
    transfer pf (transfer_r pf y) = y.
  intros; subst; rewrite transfer_eq_refl, transfer_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_r_undoes_transfer : 
  forall {T U:Type} (pf:T=U) (x:T),
    transfer_r pf (transfer pf x) = x.
  intros; subst; rewrite transfer_eq_refl, transfer_r_eq_refl; auto.
Qed.



(*Wyckoff*)
Lemma transfer_undoes_transfer_r' : 
  forall {T U:Type} (pf pf':T=U) (y:U),
    transfer pf (transfer_r pf' y) = y.
  intros T U h1 h2 y; pose proof (proof_irrelevance _ h1 h2); subst.
  rewrite transfer_eq_refl, transfer_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_r_undoes_transfer' : 
  forall {T U:Type} (pf pf':T=U) (x:T),
    transfer_r pf (transfer pf' x) = x.
  intros T U h1 h2 y; pose proof (proof_irrelevance _ h1 h2); subst.
  rewrite transfer_eq_refl, transfer_r_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_inj : 
  forall {T U:Type} (pf:T=U) (x y:T),
    transfer pf x = transfer pf y ->
    x = y.
intros T U pf x y h1.
subst. 
unfold transfer in h1. unfold eq_rect_r in h1.
rewrite <- eq_rect_eq in h1.
assumption.
Qed.

(*Wyckoff*)
Lemma transfer_hetero_inj : 
  forall {T U V:Type} (pft:T=V) (pfu:U=V) (x:T) (y:U),
    transfer pft x = transfer pfu y ->
    existT id _ x = existT id _ y.
intros T U V h1 h2 x y. subst.
intro h1.
apply transfer_inj in h1. subst.
reflexivity.
Qed.



                    
(*Wyckoff*)
Lemma transfer_r_inj : 
  forall {T U:Type} (pf:T=U) (x y:U),
    transfer_r pf x = transfer_r pf y ->
    x = y.
intros T U pf x y h1.
subst. 
unfold transfer_r in h1. unfold eq_rect_r in h1.
rewrite <- eq_rect_eq in h1.
assumption.
Qed.


(*Wyckoff*)
Lemma transfer_r_hetero_inj : 
  forall {T U V:Type} (pft:V=T) (pfu:V=U) (x:T) (y:U),
    transfer_r pft x = transfer_r pfu y ->
    existT id _ x = existT id _ y.
intros T U V h1 h2 x y.  subst. subst.
intro h1.
apply transfer_r_inj in h1. subst.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_existT_iff :
  forall {T U:Type} (pf:T = U) (x:T) (y:U),
    transfer pf x = y <->
    existT id _ x = existT id _ y.
  intros T U h1 x y; subst.
  rewrite transfer_eq_refl.
  split.
  intros; subst; auto.
  intro h1.
  apply inj_pair2 in h1; auto.
Qed.


(*Wyckoff*)
Lemma transfer_r_existT_iff : 
  forall {T U:Type} (pf:T = U) (x:T) (y:U),
    transfer_r pf y = x <->
    existT id _ y = existT id _ x.
  intros T U h1 x y; subst.
  rewrite transfer_r_eq_refl.
  split.
  intros; subst; auto.
  intro h1.
  apply inj_pair2 in h1; auto.
Qed.



(*Wyckoff*)
Lemma transfer_transfer_hetero : 
  forall {T U V:Type} (pf:T = U) 
         (pf':U = V) (a:T),
    transfer pf' (transfer pf a) =
         transfer (eq_trans pf pf') a.
intros T U V h1 h2 a.
subst. subst.
unfold transfer. unfold transfer_r. unfold eq_rect_r. simpl.
reflexivity.
Qed.

 
(*Wyckoff*)
Lemma transfer_transfer_r_hetero : 
  forall {T U V:Type} (pf:T = U) 
         (pf':T = V) (a:U),
    transfer pf' (transfer_r pf a) =
         transfer (eq2 pf pf') a.
intros T U V h1 h2 a.
subst. subst.
unfold transfer. unfold transfer_r. unfold eq_rect_r. simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_r_transfer_hetero : 
  forall {T U V:Type} (pf:T = U) 
         (pf':V = U) (a:T),
    transfer_r pf' (transfer pf a) =
         transfer (eq2' pf pf') a.
intros T U V h1 h2 a.
subst. subst.
unfold transfer. unfold transfer_r. unfold eq_rect_r. simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_r_transfer_r_hetero : 
  forall {T U V:Type} (pf:T = U) 
         (pf':V = T) (a:U),
    transfer_r pf' (transfer_r pf a) =
         transfer_r (eq_trans' pf pf') a.
intros T U V h1 h2 a.
subst. subst.
unfold transfer. unfold transfer_r. unfold eq_rect_r. simpl.
reflexivity.
Qed.

(*Wyckoff*)
Definition transfer_pred {T} {P Q:T->Prop} (pfe:pred_eq P Q)
           (x:{t:T | P t}) : {t:T | Q t} :=
  exist _ _ (iff1 (pfe (proj1_sig x)) (proj2_sig x)).

(*Wyckoff*)
Definition transfer_pred_r {T} {P Q:T->Prop} (pfe:pred_eq P Q)
           (x:{t:T | Q t}) : {t:T | P t} :=
  exist _ _ (iff2 (pfe (proj1_sig x)) (proj2_sig x)).

(*Wyckoff*)
Lemma transfer_pred_undoes_transfer_pred_r :
  forall {T} {P Q:T->Prop} (pfe:pred_eq P Q)
         (x:{t:T | Q t}),
    transfer_pred pfe (transfer_pred_r pfe x) = x.
  unfold transfer_pred_r, transfer_pred; simpl; intros;
  apply proj1_sig_injective; simpl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_pred_r_undoes_transfer_pred :
  forall {T} {P Q:T->Prop} (pfe:pred_eq P Q)
         (x:{t:T | P t}),
    transfer_pred_r pfe (transfer_pred pfe x) = x.
  unfold transfer_pred_r, transfer_pred; simpl; intros;
  apply proj1_sig_injective; simpl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_pred_refl :
  forall {T} {P:T->Prop}
         (x:{t:T | P t})
         (pf:pred_eq P P),
    transfer_pred pf x = x.
  unfold transfer_pred; intros; apply proj1_sig_injective; auto.
Qed.

(*Wyckoff*)
Lemma transfer_pred_refl_r :
  forall {T} {P:T->Prop}
         (x:{t:T | P t})
         (pf:pred_eq P P),
    transfer_pred_r pf x = x.
  unfold transfer_pred_r; intros; apply proj1_sig_injective; auto.
Qed.

(*Wyckoff*)
Lemma transfer_pred_sigT : forall {T T':Type} (pf:T = T')
                             (P:{U:Type & U} -> Prop),
                        forall x:T,
                          P (existT id T x) <->
                          P (existT id T' (transfer pf x)).
intros T T' h1 P x.
subst.
unfold transfer.
unfold eq_rect_r.
simpl.
tauto.
Qed.


(*Wyckoff*)
Lemma transfer_fun_pred : 
  forall {T T' U U':Type} (pft:T = T') (pfu:U = U') (f:T->U)
    (P:{pT:Type*Type & ((fst pT)->(snd pT))}->Prop),
    P (existT _ (T,U) f)  <-> 
    P (existT _ (T', U')
              (fun x:(fst (T', U')) =>
                 transfer (eq_refl (snd (T', U')))
                          (transfer pfu (f (transfer_r pft x))))).
simpl.
intros T T' U U' h1 h2 f P.
subst.
unfold transfer, transfer_r, eq_rect_r. simpl.
tauto.
Qed.

(*Wyckoff*)
Lemma transfer_fun_pred' : 
  forall {T T' U U':Type} (pft:T' = T) (pfu:U = U') (f:T->U)
    (P:{pT:Type*Type & ((fst pT)->(snd pT))}->Prop),
    P (existT _ (T,U) f)  <-> 
    P (existT _ (T', U')
              (fun x:(fst (T', U')) =>
                 transfer (eq_refl (snd (T', U')))
                          (transfer pfu (f (transfer pft x))))).
simpl.
intros T T' U U' h1 h2 f P.
subst.
unfold transfer, eq_rect_r. simpl.
tauto.
Qed.


(*Wyckoff*)
Definition transfer_dep {T:Type} {U:T->Type} {x y:T} (pf:x = y)
           (a:U x):U y.
subst. refine a.
Defined.

(*Wyckoff*)
Definition transfer_dep_r {T:Type} {U:T->Type} {x y:T} (pf:y = x)
           (a:U x):U y.
subst. refine a.
Defined.


(*Wyckoff*)
Definition transfer_dep2
           {T U:Type} {V:T->U->Type}
           {x x':T}
           {y y':U}
           (pfx:x = x')
           (pfy:y = y')
           (a:V x y) : V x' y'.
subst. refine a.
Defined.


(*Wyckoff*)
Definition transfer_dep2_r
           {T U:Type} {V:T->U->Type}
           {x x':T}
           {y y':U}
           (pfx:x = x')
           (pfy:y = y')
           (a:V x' y') : V x y.
subst. refine a.
Defined.


(*Wyckoff*)
Lemma transfer_dep_eq_refl : forall {T:Type} (U:T->Type) (x:T)
                                    (pfe:x = x)
                                    (p:U x),
                               transfer_dep pfe p = p.
  intros ? ? ? h1 ?.
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  unfold transfer_dep, eq_rect_r, eq_rect; simpl; auto.
Defined.

(*Wyckoff*)
Lemma transfer_dep_r_eq_refl : forall {T:Type} (U:T->Type) (x:T)
                                      (pfe:x = x)
                                      (p:U x),
                                 transfer_dep_r pfe p = p.
  intros ? ? ? h1 ?.
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  unfold transfer_dep_r, eq_rect_r, eq_rect; simpl; auto.
Defined.


(*Wyckoff*)
Lemma transfer_dep2_eq_refl :
  forall 
    {T U:Type} {V:T->U->Type}
    {x:T}
    {y:U}
    (pfx:x = x)
    (pfy:y = y)
    (a:V x y),
    transfer_dep2 pfx pfy a = a.
  unfold transfer_dep2, eq_rect_r, eq_rect.
  intros T U V x y hx hy a.
  pose proof (proof_irrelevance _ (eq_sym hy) (eq_refl _)) as h1.
  rewrite h1.
  pose proof (proof_irrelevance _ (eq_sym hx) (eq_refl _)) as h2.
  rewrite h2; auto.
Qed.


(*Wyckoff*)
Lemma transfer_dep2_r_eq_refl :
  forall 
    {T U:Type} {V:T->U->Type}
    {x:T}
    {y:U}
    (pfx:x = x)
    (pfy:y = y)
    (a:V x y),
    transfer_dep2_r pfx pfy a = a.
  unfold transfer_dep2_r, eq_rect_r, eq_rect.
  intros T U V x y hx hy a.
  pose proof (proof_irrelevance _ (eq_sym hy) (eq_refl _)) as h1.
  rewrite h1.
  pose proof (proof_irrelevance _ (eq_sym hx) (eq_refl _)) as h2.
  rewrite h2; auto.
Qed.

Lemma transfer_dep2_transfer_dep2_r_compat :
  forall {T U} {V:T->U->Type} {x x':T} {y y':U}
         (pfex:x = x')
         (pfey:y = y')
         (pfex':x' = x)
         (pfey':y' = y)
         (a:V x y),
    transfer_dep2 pfex pfey a =
    transfer_dep2_r pfex' pfey' a.
  intros T U V x x' y y' hx hy hx' hy' a.
  pose proof (proof_irrelevance _ hx' (eq_sym hx)) as h1.
  rewrite h1.
  pose proof (proof_irrelevance _ hy' (eq_sym hy)) as h2.
  rewrite h2.
  clear hx' hy' h1 h2.
  subst.
  rewrite transfer_dep2_eq_refl, transfer_dep2_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep_transfer_dep_r_compat : 
  forall {T:Type} {U:T->Type} {x y:T} (pf:x = y) (pf':y = x) (a:U x),
    transfer_dep pf a = transfer_dep_r pf' a.
intros T U x y h1 h2 a.
unfold transfer_dep, transfer_dep_r.
destruct h2. unfold eq_rect_r. simpl.
pose proof (proof_irrelevance _ (eq_sym h1) (eq_refl _)) as h2.
rewrite h2.
simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_dep_r_undoes_transfer_dep : 
  forall {T:Type} {U:T->Type} {x y:T} (pf:x = y) (a:U x),
    transfer_dep_r pf (transfer_dep pf a) = a.
  intros; subst; rewrite transfer_dep_eq_refl,
                 transfer_dep_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep_undoes_transfer_dep_r : 
  forall {T:Type} {U:T->Type} {x y:T} (pf:x = y) (a:U y),
    transfer_dep pf (transfer_dep_r pf a) = a.
  intros; subst; rewrite transfer_dep_eq_refl,
                 transfer_dep_r_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_dep_r_undoes_transfer_dep' : 
  forall {T:Type} {U:T->Type} {x y:T} (pf pf':x = y) (a:U x),
    transfer_dep_r pf (transfer_dep pf' a) = a.
  intros T U x y h1 h2 a; pose proof (proof_irrelevance _ h1 h2); subst.
  rewrite transfer_dep_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep_undoes_transfer_dep_r' : 
  forall {T:Type} {U:T->Type} {x y:T} (pf pf':x = y) (a:U y),
    transfer_dep pf (transfer_dep_r pf' a) = a.
  intros T U x y h1 h2 a; pose proof (proof_irrelevance _ h1 h2); subst.
  rewrite transfer_dep_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep2_undoes_transfer_dep2_r :
  forall {T U} {V:T->U->Type} {x x':T} {y y':U}
         (pfex pfex':x' = x)
         (pfey pfey':y' = y)
         (a:V x y),
    transfer_dep2 pfex pfey (transfer_dep2_r pfex' pfey' a) = a.
  intros; subst;
  rewrite transfer_dep2_eq_refl, transfer_dep2_r_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_dep2_r_undoes_transfer_dep2 :
  forall {T U} {V:T->U->Type}
         {x x':T} {y y':U}
         (pfex pfex':x' = x)
         (pfey pfey':y' = y)
         (a:V x y),
    transfer_dep2 pfex pfey (transfer_dep2_r pfex' pfey' a) = a.
  intros; subst;
  rewrite transfer_dep2_eq_refl, transfer_dep2_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep_trivial :
  forall {T U} {x y:T}
         (pfe:x = y)
         (u:U),
    transfer_dep (U := fun _ => U) pfe u = u.
  intros; subst; rewrite transfer_dep_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_dep_r_trivial :
  forall {T U} {x y:T}
         (pfe:x = y)
         (u:U),
    transfer_dep_r (U := fun _ => U) pfe u = u.
  intros; subst; rewrite transfer_dep_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep2_trivial1 :
  forall {T U}
         {V:U->Type}
         {x x' x0 x0':T} {y y':U}
         (pfex:x = x')
         (pfex0:x0 = x0')
         (pfey:y = y')
         (a:V y),
    transfer_dep2 (V := fun _ q => V q) pfex pfey a =
    transfer_dep2 (V := fun _ q => V q) pfex0 pfey a.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep2_trivial2 :
  forall {T U}
         {V:T->Type}
         {x x':T} {y y' y0 y0':U}
         (pfex:x = x')
         (pfey:y = y')
         (pfey0:y0 = y0')
         (a:V x),
    transfer_dep2 (V := fun q _ => V q) pfex pfey a =
    transfer_dep2 (V := fun q _ => V q) pfex pfey0 a.
  intros; subst; auto.
Qed.


(*Wyckoff*)
Lemma transfer_dep2_r_trivial1 :
  forall {T U}
         {V:U->Type}
         {x x' x0 x0':T} {y y':U}
         (pfex:x = x')
         (pfex0:x0 = x0')
         (pfey:y = y')
         (a:V y'),
    transfer_dep2_r (V := fun _ q => V q) pfex pfey a =
    transfer_dep2_r (V := fun _ q => V q) pfex0 pfey a.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep2_r_trivial2 :
  forall {T U}
         {V:T->Type}
         {x x':T} {y y' y0 y0':U}
         (pfex:x = x')
         (pfey:y = y')
         (pfey0:y0 = y0')
         (a:V x'),
    transfer_dep2_r (V := fun q _ => V q) pfex pfey a =
    transfer_dep2_r (V := fun q _ => V q) pfex pfey0 a.
  intros; subst; auto.
Qed.



(*Wyckoff*)
Lemma transfer_dep_inj : 
  forall {T:Type} {U:T->Type} {x y:T} (pf:x = y) (a b:U x),
    transfer_dep pf a = transfer_dep pf b ->
    a = b.
intros T U x y h1 a b h2.
subst.
unfold transfer_dep in h2.
unfold eq_rect_r in h2. simpl in h2.
assumption.
Qed.

(*Wyckoff*)
Lemma transfer_dep_r_inj : 
  forall {T:Type} {U:T->Type} {x y:T} (pf:x = y) (a b:U y),
    transfer_dep_r pf a = transfer_dep_r pf b ->
    a = b.
intros T U x y h1 a b h2.
subst.
unfold transfer_dep_r, eq_rect_r in h2.
simpl in h2.
assumption.
Qed.


(*Wyckoff*)
Lemma transfer_dep_existT :
  forall {T:Type} {U:T->Type} (x y:T) (pf:x = y)
         (a:U x), 
    existT _ _ a = existT _ _ (transfer_dep pf a).
intros T U x y h1 a.
subst.
f_equal.
Qed.

(*Wyckoff*)
Lemma projT_injective :
  forall {T:Type} (A:T->Type) 
         (x y:{a:T & (A a)})
         (pf:projT1 x = projT1 y),
    (transfer_dep pf (projT2 x)) =
    projT2 y->
    x = y.
intros T A x y h1 h2.
destruct x; destruct y.
simpl in h1. simpl in h2.
subst.
apply transfer_dep_existT.
Qed.

(*Wyckoff*)
Definition transfer_pred_dep
           {T} {P Q:T->Prop} {U:forall R:T->Prop, {x:T | R x} -> Type}
           {p:{x:T | P x}}
           (pfeq:pred_eq P Q)
           (u:U P p) : U Q (transfer_pred pfeq p).
  pose proof (pred_ext _ _ pfeq); subst.
  rewrite transfer_pred_refl; auto.
Defined.


(*Wyckoff*)
Definition transfer_pred_dep_r
           {T} {P Q:T->Prop} {U:forall R:T->Prop, {x:T | R x} -> Type}
           {q:{x:T | Q x}}
           (pfeq:pred_eq P Q)
           (u:U Q q) : U P (transfer_pred_r pfeq q).
  pose proof (pred_ext _ _ pfeq); subst.
  rewrite transfer_pred_refl_r; auto.
Defined.



(*Wyckoff*)
Lemma existT_injective1 : forall {T:Type} (U:T->Type) (x y:T)
                                 (p:U x) (q:U y),
  existT _ x p = existT _ y q -> x = y.
intros T U x y p q h1.
pose proof (f_equal (@projT1 _ _) h1) as h2. 
simpl in h2.
assumption.
Qed.

(*Wyckoff*)
Lemma existT_injective2 :
  forall {T:Type} (U:T->Type) (x:T) (p q:U x),
    existT _ x p = existT _ x q -> p = q.
  intros; apply inj_pair2; auto.
Qed.

(*Wyckoff*)
Lemma existT_injective2' :
  forall {T:Type} (U:T->Type) (x x':T)
         (p:U x) (q:U x')
         (pfx:x = x'),
    existT _ x p = existT _ x' q ->
    transfer_dep pfx p = q.
  intros; subst; rewrite transfer_dep_eq_refl;
  apply existT_injective2 in H; auto.
Qed.


(*Wyckoff*)  
Lemma transfer_dep_type : 
  forall {T:Type} {U:Type->Type} (y y':U T)
         (P:T->U T->Prop)
         (f:forall t:T, P t y -> Prop)
         (pfe:y = y')
         (x:T)
         (pfxy:P x y'),
    @transfer_dep (U T) (fun l:U T => forall t:T, P t l -> Prop) 
                  y y' pfe f x pfxy =
    f x (@transfer_dep (U T) (fun l:U T => P x l) y' y (eq_sym pfe) pfxy).
  intros; subst; rewrite transfer_dep_eq_refl; reflexivity.
Qed.

Definition transfer_dep_pf 
           {T} {P:T->Prop} {U:forall x:T, P x -> Type}
           {x} (pf pf':P x)
           (u:U x pf) : U x pf' :=
  transfer_dep (proof_irrelevance (P x) pf pf') u.

Definition transfer_dep_pf_r 
           {T} {P:T->Prop} {U:forall x:T, P x -> Type}
           {x} (pf pf':P x)
           (u:U x pf') : U x pf :=
  transfer_dep_r (proof_irrelevance (P x) pf pf') u.


Lemma transfer_dep_pf_eq :
  forall {T U} {P:T->Prop}
         (f:prop_dep_fun P U)
         x x'
         (pfe:x = x')
         (pfpx:P x)
         (pfpx':P x')
         (pfe':pfpx = transfer_dep_r pfe pfpx'),
    transfer_dep pfe' (f x pfpx)
                 (U := fun pf:P x => U) =
    f x' (transfer_dep pfe pfpx).
  intros; subst; rewrite transfer_dep_eq_refl; auto.
Qed.

Lemma transfer_dep_r_pf_eq :
  forall {T U} {P:T->Prop}
         (f:prop_dep_fun P U)
         x x'
         (pfe:x = x')
         (pfpx:P x)
         (pfpx':P x')
         (pfe':transfer_dep pfe pfpx = pfpx'),
    transfer_dep_r pfe' (f x' pfpx')
                 (U := fun pf:P x' => U) =
    f x (transfer_dep_r pfe pfpx').
  intros; subst; rewrite transfer_dep_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_dep_prop : 
  forall {T:Type} {U:T->Type} (x y:T) (pf:x = y)
         (a:U x) (P:{t:T & (U t)}->Prop),
    P (existT _ _ a) <-> P (existT _ _ (transfer_dep pf a)).
intros T U x y h1 a P.
rewrite (transfer_dep_existT _ _ h1 a).
tauto.
Qed.

(*Wyckoff*)
Definition transfer_dep_prop1 
           {T V:Type} {P:T->Prop}
           {U: forall x:T, P x -> Type}
           {x y:T}
           (pfe:x = y)
           (z:{pfpx:P x & U x pfpx}) :
  {pfpy:P y & U y pfpy} :=
  let z' := transfer_dep pfe (projT1 z) in
  let u := transfer (f_equal_dep U _ _ (projT1 z) z' pfe) (projT2 z) in
  existT (fun pfpy:P y => U y pfpy) z' u.


(*Wyckoff*)
Lemma transfer_dep_eq_iff : 
  forall {T:Type} {U:T->Type} (x y:T) (pf:x = y)
         (a:U x) (b:U y), 
    existT _ _ a = existT _ _ b <->
    transfer_dep pf a = b.
intros T U x y h1 a b.
subst.
split.
intro h1.
apply inj_pair2 in h1. 
unfold transfer_dep.
unfold eq_rect_r.
rewrite <- eq_rect_eq.
assumption.
intro h1.
unfold transfer_dep in h1. unfold eq_rect_r in h1.
rewrite <- eq_rect_eq in h1.
subst.
reflexivity.
Qed.



(*Wyckoff*)
Lemma transfer_dep_r_eq_iff : 
  forall {T:Type} {U:T->Type} (x y:T) (pf:y = x)
         (a:U x) (b:U y), 
    existT _ _ a = existT _ _ b <->
    transfer_dep_r pf a = b.
intros T U x y h1 a b.
subst.
split.
intro h1.
apply inj_pair2 in h1. 
unfold transfer_dep_r.
unfold eq_rect_r.
rewrite <- eq_rect_eq.
assumption.
intro h1.
unfold transfer_dep_r in h1. unfold eq_rect_r in h1.
rewrite <- eq_rect_eq in h1.
subst.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_dep2_eq_iff : 
  forall {T U} {V:T->U->Type}
         {x x':T}
         {y y':U}
         (pfe:x = x')
         (pfe':y = y')
         (a:V x y) (b:V x' y'), 
    existT _ _ (existT _ _ a) = existT _ _ (existT _ _ b)<->
    transfer_dep2 pfe pfe' a = b.
  intros; subst; rewrite transfer_dep2_eq_refl.
  split; intro h1.
  apply inj_pair2 in h1.
  apply inj_pair2 in h1; auto.
  subst; auto.
Qed.


(*Wyckoff*)
Lemma transfer_dep2_r_eq_iff : 
  forall {T U} {V:T->U->Type}
         {x x':T}
         {y y':U}
         (pfe:x' = x)
         (pfe':y' = y)
         (a:V x y) (b:V x' y'), 
    existT _ _ (existT _ _ a) = existT _ _ (existT _ _ b)<->
    transfer_dep2_r pfe pfe' a = b.
  intros; subst; rewrite transfer_dep2_r_eq_refl.
  split; intro h1.
  apply inj_pair2 in h1.
  apply inj_pair2 in h1; auto.
  subst; auto.
Qed.


(*Wyckoff*)
Definition transfer_prop_app2
           {T} {P:T->Prop} {f:forall x:T, P x -> Type}
           {x x'} (pfe:x = x') (pfx:P x)
           (y:f x pfx) :
  f x' (transfer_dep pfe pfx).
  subst; rewrite transfer_dep_eq_refl; auto.
Defined.


(*Wyckoff*)
Lemma existT_inv :
  forall {T:Type}
         {U:T->Type}
         (x y:T)
         (pf:x = y)
         (f:U x)
         (g:U y),
    @existT T U x f =
    @existT T U y g ->
    f = transfer_dep_r pf g.
  intros T U x y h1 f g h2.
  symmetry in h2.
  rewrite transfer_dep_r_eq_iff in h2.
  symmetry.
  apply h2.
Qed.

(*Wyckoff*)
Lemma existT_s_inv :
  forall {T:Set} {V:Type}
         {U:T->Type}
         (x y:T)
         (pf:x = y)
         (f:U x)
         (g:U y),
    @existT T U x f =
    @existT T U y g ->
    f = transfer_dep_r pf g.
  intros T V U x y h1 f g h2.
  symmetry in h2.
  rewrite transfer_dep_r_eq_iff in h2.
  symmetry.
  apply h2.
Qed.



(*Wyckoff*)
Lemma transfer_dep_fun1_eq : 
  forall {T U:Type} (pf:T = U)
         (f:T->T) (x:U),
    (transfer_dep (U:=fun x=>x->x) pf f) x =
         (transfer pf (f (transfer_r pf x))).
intros T U h1 f x.
subst.
rewrite transfer_dep_eq_refl.
unfold transfer, transfer_r, eq_rect_r. simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_dep_fun2_eq : 
  forall {T U:Type} (pf:T = U)
         (f:T->T->T) (x y:U),
    (transfer_dep (U:=fun x=>x->x->x) pf f) x y =
         (transfer pf (f (transfer_r pf x) (transfer_r pf y))).
intros T U h1 f x y.
subst.
rewrite transfer_dep_eq_refl.
unfold transfer, transfer_r, eq_rect_r. simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_dep_hetero : 
  forall {T:Type} {U:T->Type} (x x' y:T) (pf:x = y) 
         (pf':x' = y) (a:U x) (a':U x'),
    transfer_dep pf a = transfer_dep pf' a' <->
         a = transfer_dep (eq_trans pf' (eq_sym pf)) a'.
intros; subst;
do 3 rewrite transfer_dep_eq_refl; tauto.
Qed.

(*Wyckoff*)
Lemma transfer_dep_r_fun1_eq : 
  forall {T U:Type} (pf:T = U)
         (f:U->U) (x:T),
    (transfer_dep_r (U:=fun x=>x->x) pf f) x =
         (transfer_r pf (f (transfer pf x))).
intros T U h1 f x.
subst.
rewrite transfer_dep_r_eq_refl.
unfold transfer, transfer_r, eq_rect_r. simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_dep_r_fun2_eq : 
  forall {T U:Type} (pf:T = U)
         (f:U->U->U) (x y:T),
    (transfer_dep_r (U:=fun x=>x->x->x) pf f) x y =
         (transfer_r pf (f (transfer pf x) (transfer pf y))).
intros T U h1 f x y.
subst.
rewrite transfer_dep_r_eq_refl.
unfold transfer, transfer_r, eq_rect_r. simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_dep_r_hetero : 
  forall {T:Type} {U:T->Type} (x y y':T) (pf:x = y) 
         (pf':x = y') (a:U y) (a':U y'),
    transfer_dep_r pf a = transfer_dep_r pf' a' <->
         a = transfer_dep_r (eq2 pf pf') a'.
intros T U x y y' h1 h2 a a'.
subst. subst.
do 2 rewrite transfer_dep_r_eq_refl. 
unfold eq2. unfold eq_ind_r. unfold eq_ind. simpl.
rewrite transfer_dep_r_eq_refl.
tauto.
Qed.


(*Wyckoff*)
Definition transfer_fun {T U V:Type} (pf:T=U) (f:T->V) : U->V.
revert f. rewrite pf. intro f. refine f.
Defined.

(*Wyckoff*)
Definition transfer_fun' {T U:Type} (pf:T=U) (f:T->T) : U->U.
revert f. rewrite pf. intro f. refine f.
Defined.


(*Wyckoff*)
Definition transfer_fun2 {T U V:Type} (pf:T=U) (f:T->T->V) : U->U->V.
revert f. rewrite pf. intro f. refine f.
Defined.


(*Wyckoff*)
Definition transfer_fun2' {T U:Type} (pf:T=U) (f:T->T->T) : U->U->U.
revert f. rewrite pf. intro f. refine f.
Defined.


(*Wyckoff*)
Definition transfer_fun_r {T U V:Type} (pf:T=U) (f:U->V) : T->V.
revert f. rewrite pf. intro f. refine f.
Defined.


(*Wyckoff*)
Definition transfer_fun_r' {T U:Type} (pf:T=U) (f:U->U) : T->T.
revert f. rewrite pf. intro f. refine f.
Defined.


(*Wyckoff*)
Definition transfer_fun2_r {T U V:Type} (pf:T=U) (f:U->U->V) : T->T->V.
revert f. rewrite pf. intro f. refine f.
Defined.

(*Wyckoff*)
Definition transfer_fun2_r' {T U:Type} (pf:T=U) (f:U->U->U) : T->T->T.
revert f. rewrite pf. intro f. refine f.
Defined.


(*Wyckoff*)
Lemma transfer_fun_eq_refl : 
  forall {T U:Type} (f:T->U)
         (pfe:T = T),
    transfer_fun pfe f = f.
intros. apply functional_extensionality. intro.
unfold transfer_fun, eq_rect_r, eq_rect; simpl.
pose proof (proof_irrelevance _ pfe (eq_refl _)) as h1.
rewrite h1; simpl; auto.
Defined.

(*Wyckoff*)
Lemma transfer_fun2_eq_refl : 
  forall {T U:Type} (f:T->T->U)
         (pfe:T = T),
    transfer_fun2 pfe f = f.
intros. 
apply functional_extensionality. intro.
apply functional_extensionality. intro.
unfold transfer_fun2, eq_rect_r, eq_rec, eq_rect. simpl.
pose proof (proof_irrelevance _ pfe (eq_refl _)) as h1.
rewrite h1; simpl; auto.
Defined.

(*Wyckoff*)
Lemma transfer_fun_eq_refl' : 
  forall {T:Type} (f:T->T),
    transfer_fun' eq_refl f = f.
intros. apply functional_extensionality. intro.
unfold transfer_fun'. unfold eq_rect_r. simpl.
reflexivity.
Defined.


(*Wyckoff*)
Lemma transfer_fun2_eq_refl' : 
  forall {T:Type} (f:T->T->T),
    transfer_fun2' eq_refl f = f.
intros. 
apply functional_extensionality. intro.
apply functional_extensionality. intro.
unfold transfer_fun2'. unfold eq_rect_r. simpl.
reflexivity.
Defined.


(*Wyckoff*)
Lemma transfer_fun_r_eq_refl : 
  forall {T U:Type} (f:T->U),
    transfer_fun_r eq_refl f = f.
intros. apply functional_extensionality. intro.
unfold transfer_fun. unfold eq_rect_r. simpl.
reflexivity.
Defined.

(*Wyckoff*)
Lemma transfer_fun2_r_eq_refl : 
  forall {T U:Type} (f:T->T->U),
    transfer_fun2_r eq_refl f = f.
intros. 
apply functional_extensionality. intro.
apply functional_extensionality. intro.
unfold transfer_fun2. unfold eq_rect_r. simpl.
reflexivity.
Defined.

(*Wyckoff*)
Lemma transfer_fun_r_eq_refl' : 
  forall {T:Type} (f:T->T),
    transfer_fun_r' eq_refl f = f.
intros. apply functional_extensionality. intro.
unfold transfer_fun_r'. unfold eq_rect_r. simpl.
reflexivity.
Defined.

(*Wyckoff*)
Lemma transfer_fun2_r_eq_refl' : 
  forall {T:Type} (f:T->T->T),
    transfer_fun2_r' eq_refl f = f.
intros. 
apply functional_extensionality. intro.
apply functional_extensionality. intro.
unfold transfer_fun2_r'. unfold eq_rect_r. simpl.
reflexivity.
Defined.


(*Wyckoff*)
Lemma transfer_fun_eq : 
  forall {T U V:Type} (pf:T = U) (f:T->V),
    transfer_fun pf f =
    (fun x:U => f (transfer_r pf x)).
intros T U V h1 f. subst.
rewrite transfer_fun_eq_refl.
apply functional_extensionality. intro x.
rewrite transfer_r_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_fun_eq' : 
  forall {T U:Type} (pf:T=U) (f:T->T),
    transfer_fun' pf f =
    (fun x:U => (transfer pf (f (transfer_r pf x)))).
intros T U h1 f. subst.
rewrite transfer_fun_eq_refl'.
apply functional_extensionality. intro x.
rewrite transfer_eq_refl.
rewrite transfer_r_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_eq : 
  forall {T U V:Type} (pf:T = U) (f:T->T->V),
    transfer_fun2 pf f =
    (fun (x:U) (y:U) => f (transfer_r pf x) (transfer_r pf y)).
intros T U V h1 f. subst.
rewrite transfer_fun2_eq_refl.
apply functional_extensionality. intro x.
rewrite transfer_r_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_eq' : 
  forall {T U:Type} (pf:T=U) (f:T->T->T),
    transfer_fun2' pf f =
    (fun (x:U) (y:U) => (transfer pf (f (transfer_r pf x)
                                        (transfer_r pf y)))).
intros T U h1 f. subst.
rewrite transfer_fun2_eq_refl'. 
apply functional_extensionality. intro x.
apply functional_extensionality. intro y.
rewrite transfer_eq_refl.
do 2 rewrite transfer_r_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun_r_eq : 
  forall {T U V:Type} (pf:T = U) (f:U->V),
    transfer_fun_r pf f =
    (fun x:T => f (transfer pf x)).
intros T U V h1 f. subst.
rewrite transfer_fun_r_eq_refl.
apply functional_extensionality. intro x.
rewrite transfer_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_fun_r_eq' : 
  forall {T U:Type} (pf:T=U) (f:U->U),
    transfer_fun_r' pf f =
    (fun x:T => (transfer_r pf (f (transfer pf x)))).
intros T U h1 f. subst.
rewrite transfer_fun_r_eq_refl'.
apply functional_extensionality. intro x.
rewrite transfer_eq_refl.
rewrite transfer_r_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_r_eq : 
  forall {T U V:Type} (pf:T = U) (f:U->U->V),
    transfer_fun2_r pf f =
    (fun (x:T) (y:T) => f (transfer pf x) (transfer pf y)).
intros T U V h1 f. subst.
rewrite transfer_fun2_r_eq_refl.
apply functional_extensionality. intro x.
rewrite transfer_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_r_eq' : 
  forall {T U:Type} (pf:T=U) (f:U->U->U),
    transfer_fun2_r' pf f =
    (fun (x:T) (y:T) => (transfer_r pf (f (transfer pf x)
                                          (transfer pf y)))).
intros T U h1 f. subst.
rewrite transfer_fun2_r_eq_refl'. 
apply functional_extensionality. intro x.
apply functional_extensionality. intro y.
do 2 rewrite transfer_eq_refl.
rewrite transfer_r_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun_compat : 
  forall {T U V:Type} (pf:T = U) (f:T->V) (x:T),
    (transfer_fun pf f) (transfer pf x) = f x.
intros T U V h1 f x.
unfold transfer_fun. unfold transfer.
destruct h1.   compute.
reflexivity.
Qed.

Lemma transfer_fun_inj_iff :
  forall {T U V:Type} (pf: T = U) (f:T->V),
    Injective f <->
    Injective (transfer_fun pf f).
  intros T U V h1 f.
  subst.
  rewrite transfer_fun_eq_refl.
  tauto.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_compat : 
  forall {T U V:Type} (pf:T = U) (f:T->T->V) (x y:T),
    (transfer_fun2 pf f) (transfer pf x) (transfer pf y) = f x y.
intros T U V h1 f x y.
unfold transfer_fun2. unfold transfer, eq_rect_r. subst.
simpl. reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun_compat' : 
  forall {T U:Type} (pf:T = U) (f:T->T) (x:T),
    (transfer_fun' pf f) (transfer pf x) =
    transfer pf (f x). 
intros T U h1 f x. subst.
unfold transfer_fun'. unfold transfer. unfold eq_rect_r.
simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_compat' : 
  forall {T U:Type} (pf:T = U) (f:T->T->T) (x y:T),
    (transfer_fun2' pf f) (transfer pf x) (transfer pf y) =
    transfer pf (f x y).
intros T U h1 f x y.
unfold transfer_fun2'. unfold transfer, eq_rect_r. subst.
simpl. reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun_r_compat : 
  forall {T U V:Type} (pf:T = U) (f:U->V) (x:U),
    f x = (transfer_fun_r pf f) (transfer_r pf x).
intros T U V h1 f x.
unfold transfer_fun_r. unfold transfer_r.
destruct h1.   compute.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_r_compat : 
  forall {T U V:Type} (pf:T = U) (f:U->U->V) (x y:U),
    f x y = (transfer_fun2_r pf f) (transfer_r pf x) (transfer_r pf y).
intros T U V h1 f x y.
unfold transfer_fun2_r. unfold transfer, eq_rect_r. subst.
simpl. reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun_r_compat' : 
  forall {T U:Type} (pf:T = U) (f:U->U) (x:U),
    (transfer_fun_r' pf f) (transfer_r pf x) = transfer_r pf (f x).
intros T U h1 f x.
unfold transfer_fun_r'. unfold transfer_r.
destruct h1.   compute.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_r_compat' : 
  forall {T U:Type} (pf:T = U) (f:U->U->U) (x y:U),
    (transfer_fun2_r' pf f) (transfer_r pf x) (transfer_r pf y) =
    transfer_r pf (f x y).
intros T U h1 f x y.
unfold transfer_fun2_r'. unfold transfer_r, eq_rect_r. subst.
simpl. reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun_transfer_dep_compat : 
  forall {T U V:Type} (pf:T = U) (f:T->V),
    transfer_fun pf f = 
    transfer_dep pf (U:=fun A:Type=>A->V) f.
intros T U V h1 f. subst.
simpl.
rewrite transfer_fun_eq_refl, transfer_dep_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_fun2_transfer_dep_compat : 
  forall {T U V:Type} (pf:T = U) (f:T->T->V),
    transfer_fun2 pf f = 
    transfer_dep pf (U:=fun A:Type=>A->A->V) f.
intros T U V h1 f. subst.
simpl.
rewrite transfer_fun2_eq_refl, transfer_dep_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_fun_r_transfer_dep_r_compat : 
  forall {T U V:Type} (pf:T = U) (f:U->V),
    transfer_fun_r pf f = 
    transfer_dep_r pf (U:=fun A:Type=>A->V) f.
intros T U V h1 f. subst.
simpl.
rewrite transfer_fun_r_eq_refl, transfer_dep_r_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_r_transfer_dep_r_compat : 
  forall {T U V:Type} (pf:T = U) (f:U->U->V),
    transfer_fun2_r pf f = 
    transfer_dep_r pf (U:=fun A:Type=>A->A->V) f.
intros T U V h1 f. subst.
simpl.
rewrite transfer_fun2_r_eq_refl, transfer_dep_r_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun_transfer_dep_compat' : 
  forall {T U:Type} (pf:T = U) (f:T->T),
    transfer_fun' pf f = 
    transfer_dep pf (U:=fun A:Type=>A->A) f.
intros T U h1 f. subst.
simpl.
rewrite transfer_fun_eq_refl', transfer_dep_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_fun2_transfer_dep_compat' : 
  forall {T U:Type} (pf:T = U) (f:T->T->T),
    transfer_fun2' pf f = 
    transfer_dep pf (U:=fun A:Type=>A->A->A) f.
intros T U h1 f. subst.
simpl.
rewrite transfer_fun2_eq_refl', transfer_dep_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun_r_transfer_dep_r_compat' : 
  forall {T U:Type} (pf:T = U) (f:U->U),
    transfer_fun_r' pf f = 
    transfer_dep_r pf (U:=fun A:Type=>A->A) f.
intros T U h1 f. subst.
simpl.
rewrite transfer_fun_r_eq_refl', transfer_dep_r_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_r_transfer_dep_r_compat' : 
  forall {T U:Type} (pf:T = U) (f:U->U->U),
    transfer_fun2_r' pf f = 
    transfer_dep_r pf (U:=fun A:Type=>A->A->A) f.
intros T U h1 f. subst.
simpl.
rewrite transfer_fun2_r_eq_refl', transfer_dep_r_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_fun_transfer_compat : 
  forall {T U V:Type} (pf:T = U) (f:T->V),
    transfer_fun pf f = transfer (f_equal (fun A=>A->V) pf) f.
intros T U V h1 f.
subst.
rewrite transfer_fun_eq_refl. simpl.
rewrite transfer_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun_transfer_compat' : 
  forall {T U:Type} (pf:T = U) (f:T->T),
    transfer_fun' pf f = transfer (f_equal (fun A=>A->A) pf) f.
intros T U h1 f.
subst.
rewrite transfer_fun_eq_refl'. simpl.
rewrite transfer_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_fun_r_transfer_r_compat : 
  forall {T U V:Type} (pf:T = U) (f:U->V),
    transfer_fun_r pf f = transfer_r (f_equal (fun A=>A->V) pf) f.
intros T U V h1 f.
subst.
rewrite transfer_fun_r_eq_refl. simpl.
rewrite transfer_r_eq_refl.
reflexivity.
Qed.



(*Wyckoff*)
Lemma transfer_fun_r_transfer_r_compat' : 
  forall {T U:Type} (pf:T = U) (f:U->U),
    transfer_fun_r' pf f = transfer_r (f_equal (fun A=>A->A) pf) f.
intros T U h1 f.
subst.
rewrite transfer_fun_r_eq_refl'. simpl.
rewrite transfer_r_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_fun2_transfer_compat : 
  forall {T U V:Type} (pf:T = U) (f:T->T->V),
    transfer_fun2 pf f = transfer (f_equal (fun A=>A->A->V) pf) f.
intros T U V h1 f.
subst.
rewrite transfer_fun2_eq_refl. simpl.
rewrite transfer_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_transfer_compat' : 
  forall {T U:Type} (pf:T = U) (f:T->T->T),
    transfer_fun2' pf f = transfer (f_equal (fun A=>A->A->A) pf) f.
intros T U h1 f.
subst.
rewrite transfer_fun2_eq_refl'. simpl.
rewrite transfer_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_fun2_r_transfer_r_compat : 
  forall {T U V:Type} (pf:T = U) (f:U->U->V),
    transfer_fun2_r pf f = transfer_r (f_equal (fun A=>A->A->V) pf) f.
intros T U V h1 f.
subst.
rewrite transfer_fun2_r_eq_refl. simpl.
rewrite transfer_r_eq_refl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_fun2_r_transfer_r_compat' : 
  forall {T U:Type} (pf:T = U) (f:U->U->U),
    transfer_fun2_r' pf f = transfer_r (f_equal (fun A=>A->A->A) pf) f.
intros T U h1 f.
subst.
rewrite transfer_fun2_r_eq_refl'. simpl.
rewrite transfer_r_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
(*[transfer_fun_sig], is useful when the type equality 
 in transfer_fun (applied to a sigma function) is too coarse, in that
you can't extract more fundamental information from it, like 
equality of underlying sets in the sigma type.*)
Definition transfer_fun_sig {T U:Type} {A B:Ensemble T} (pf:A = B)
      (f:sig_set A->U) : sig_set B->U.
  subst. refine f.
Defined.

(*Wyckoff*)
Lemma transfer_fun_sig_eq_refl :
  forall {T U:Type} {A:Ensemble T}
         (f:sig_set A->U),
    transfer_fun_sig eq_refl f = f.
  intros T U A f.
  apply functional_extensionality.
  intro x.
  unfold transfer_fun_sig, eq_rect_r, eq_rect.
  simpl.
  reflexivity.
Defined.


(*Wyckoff*)
Lemma transfer_fun_sig_compat :
  forall {T U:Type} {A B:Ensemble T} (pf : A = B)
         (f:sig_set A->U) (x:sig_set B),
    transfer_fun_sig pf f x = f (transfer_dep_r pf x).
  intros T U A B h1 f x.
  subst.
  rewrite transfer_dep_r_eq_refl, transfer_fun_sig_eq_refl.
  reflexivity.
Qed.

(*Wyckoff*)
Definition transfer_fun_sig_r {T U:Type} {A B:Ensemble T} (pf : A = B)
      (f:sig_set B->U) : sig_set A->U.
  subst. refine f.
Defined.


(*Wyckoff*)
Lemma transfer_fun_sig_r_eq_refl :
  forall {T U:Type} {A:Ensemble T}
         (f:sig_set A->U),
    transfer_fun_sig_r eq_refl f = f.
  intros T U A f.
  apply functional_extensionality.
  intro x.
  unfold transfer_fun_sig_r, eq_rect_r, eq_rect.
  simpl.
  reflexivity.
Defined.


(*Wyckoff*)
Lemma transfer_fun_sig_r_compat :
  forall {T U:Type} {A B:Ensemble T} (pf : A = B)
         (f:sig_set B->U) (x:sig_set A),
    transfer_fun_sig_r pf f x = f (transfer_dep pf x).
  intros T U A B h1 f x.
  subst.
  rewrite transfer_dep_eq_refl, transfer_fun_sig_r_eq_refl.
  reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_prop : forall {T U:Type} (pf:T=U) (f:Type->Prop), 
  f T <-> f U.
  intros T U h1. subst.
tauto.
Qed.

(*Wyckoff*)
Lemma transfer_prop_dependent : 
  forall {T:Type} {PT:T->Type} (x y:T) (P:PT x->Prop)
         (pf:PT x = PT y) (a:PT x),
    P a <-> (transfer_fun pf P) (transfer pf a).
intros T PT x y P h1 a. split.
intro h2.
rewrite <- (transfer_fun_compat h1 P a) in h2.
assumption.
intro h2.
rewrite <- (transfer_fun_compat h1 P a).
assumption.
Qed.


(*Wyckoff*)
Lemma transfer_prop' {T U:Type} (pf:T= U) (P:T->Prop) : U->Prop.
  subst. refine P.
Defined.


(*Wyckoff*)
Definition transfer_sig_set {T:Type} (A B:Ensemble T)
           (pf:sig_set A = sig_set B) (x:sig_set A) : sig_set B.
  rewrite <- pf.
  refine x.
Defined.

(*Wyckoff*)
Definition transfer_sig_set_r {T:Type} (A B:Ensemble T)
           (pf:sig_set A = sig_set B) (x:sig_set B) : sig_set A.
  rewrite pf.
  refine x.
Defined.

(*Wyckoff*)
Lemma transfer_dep_sig_set_eq :
  forall {T:Type} (A B:Ensemble T) (pfeq:A = B) (x:T)
         (pfin:Inn A x),
    transfer_dep pfeq (exist _ _ pfin) =
    exist _ _ (in_eq_set _ _ pfeq x pfin).
  intros T A B h1 x h2.
  subst.
  rewrite transfer_dep_eq_refl.
  apply proj1_sig_injective; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep_r_sig_set_eq :
  forall {T:Type} (A B:Ensemble T) (pfeq:A = B) (x:T)
         (pfin:Inn B x),
    transfer_dep_r pfeq (exist _ _ pfin) =
    exist _ _ (in_eq_set _ _ (eq_sym pfeq) x pfin).
  intros T A B h1 x h2.
  subst.
  rewrite transfer_dep_r_eq_refl.
  apply proj1_sig_injective; auto.
Qed.

(*Wyckoff*)
Lemma f_equal_nprod_fst_fun : 
  forall {T n} {U:T->Type}
         {w:T}
         {w':T^n}
         {x:T^(S n)}
         (pfe:x = (w, w'))
         (p:U (fst x)),
    transfer_dep (U := fun c => U (fst c)) pfe p =
    transfer_dep (U := U) (f_equal (@fst _ _) pfe) p.
  intros; subst; rewrite transfer_dep_eq_refl; simpl;
  rewrite transfer_dep_eq_refl; auto.
Qed.

  
(*These are [Set]-analogues of much of the above.*)
Section SetSortTransfer.


(*Wyckoff*)
Definition s_transfer {T U:Set} (pf:T=U) (x:T) : U.
revert x. rewrite pf. intro u. refine u.
Defined.

(*Wyckoff*)
Definition s_transfer_r {T U:Set} (pf:T = U) (y:U) : T.
revert y. rewrite pf. intro u. refine u.
Defined.

(*Wyckoff*)
Lemma s_transfer_eq :
  forall {T U:Set} (pf:T=U) (x:T),
    s_transfer pf x = eq_rec _ id x _ pf.
intros; subst; simpl; auto.
Qed.

(*Wyckoff*)
Lemma s_transfer_eq_refl :
  forall {T:Set} (x:T)
         (pfe:T = T),
    s_transfer pfe x = x.
  intros T x h1; pose proof (proof_irrelevance _ h1 (eq_refl _));
  subst.
  unfold s_transfer, eq_rec_r, eq_rec, eq_rect; simpl; auto.
Defined.

(*Wyckoff*)
Lemma s_transfer_r_eq_refl :
  forall {T:Set} (x:T)
         (pfe:T = T),
    s_transfer_r pfe x = x.
  intros T x h1; pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  unfold s_transfer_r, eq_rec_r, eq_rec, eq_rect; simpl; auto.
Defined.


(*Wyckoff*)
Lemma s_transfer_r_eq :
  forall {T U:Set} (pf:T=U) (x:U),
    s_transfer_r pf x = eq_rec_r id x pf.
intros; subst; simpl; auto.
Qed.


(*Wyckoff*)
Lemma s_transfer_transfer_r_compat :
  forall {T U:Set} (pf:T=U) (pf':U=T) (x:T),
    s_transfer pf x = s_transfer_r pf' x.
intros T U ? ? x.
unfold s_transfer. unfold s_transfer_r.
destruct pf.
pose proof (proof_irrelevance _ pf' (eq_refl _)); subst.
reflexivity.
Qed.

(*Wyckoff*)
Lemma s_transfer_undoes_transfer_r :
  forall {T U:Set} (pf:T=U) (y:U),
    s_transfer pf (s_transfer_r pf y) = y.
intros T U h1 y. subst.
unfold s_transfer, s_transfer_r. simpl. unfold eq_rect_r.
simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma s_transfer_r_undoes_transfer :
  forall {T U:Set} (pf:T=U) (x:T),
    s_transfer_r pf (s_transfer pf x) = x.
intros T U h1 y. subst.
unfold s_transfer, s_transfer_r. simpl. unfold eq_rect_r.
simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma s_transfer_inj :
  forall {T U:Set} (pf:T=U) (x y:T),
    s_transfer pf x = s_transfer pf y ->
    x = y.
intros T U pf x y h1.
subst.
unfold s_transfer in h1. unfold eq_rec_r in h1.
rewrite <- eq_rec_eq in h1.
assumption.
Qed.

(*Wyckoff*)
Lemma s_transfer_hetero_inj :
  forall {T U V:Set} (pft:T=V) (pfu:U=V) (x:T) (y:U),
    s_transfer pft x = s_transfer pfu y ->
    existT id _ x = existT id _ y.
intros T U V h1 h2 x y. subst.
intro h1.
apply s_transfer_inj in h1. subst.
reflexivity.
Qed.
 



(*Wyckoff*)
Lemma s_transfer_r_inj :
  forall {T U:Set} (pf:T=U) (x y:U),
    s_transfer_r pf x = s_transfer_r pf y ->
    x = y.
intros T U pf x y h1.
subst.
unfold s_transfer_r in h1. unfold eq_rec_r in h1.
rewrite <- eq_rec_eq in h1.
assumption.
Qed.


(*Wyckoff*)
Lemma s_transfer_r_hetero_inj :
  forall {T U V:Set} (pft:V=T) (pfu:V=U) (x:T) (y:U),
    s_transfer_r pft x = s_transfer_r pfu y ->
    existT id _ x = existT id _ y.
intros T U V h1 h2 x y.  subst. subst.
intro h1.
apply s_transfer_r_inj in h1. subst.
reflexivity.
Qed.



(*Wyckoff*)
Lemma s_transfer_transfer_hetero :
  forall {T U V:Set} (pf:T = U)
         (pf':U = V) (a:T),
    s_transfer pf' (s_transfer pf a) =
         s_transfer (eq_trans pf pf') a.
intros T U V h1 h2 a.
subst. subst.
unfold s_transfer. unfold s_transfer_r. unfold eq_rect_r. simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma s_transfer_transfer_r_hetero :
  forall {T U V:Set} (pf:T = U)
         (pf':T = V) (a:U),
    s_transfer pf' (s_transfer_r pf a) =
         s_transfer (eq2 pf pf') a.
intros T U V h1 h2 a.
subst. subst.
unfold s_transfer. unfold s_transfer_r. unfold eq_rec_r. simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma s_transfer_r_transfer_hetero :
  forall {T U V:Set} (pf:T = U)
         (pf':V = U) (a:T),
    s_transfer_r pf' (s_transfer pf a) =
         s_transfer (eq2' pf pf') a.
intros T U V h1 h2 a.
subst. subst.
unfold s_transfer. unfold s_transfer_r. unfold eq_rec_r. simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma s_transfer_r_transfer_r_hetero :
  forall {T U V:Set} (pf:T = U)
         (pf':V = T) (a:U),
    s_transfer_r pf' (s_transfer_r pf a) =
         s_transfer_r (eq_trans' pf pf') a.
intros T U V h1 h2 a.
subst. subst.
unfold s_transfer. unfold s_transfer_r. unfold eq_rec_r. simpl.
reflexivity.
Qed.



(*Wyckoff*)
Definition s_transfer_dep {T:Set} {U:T->Set} {x y:T} (pf:x = y)
           (a:U x):U y.
subst. refine a.
Defined.

(*Wyckoff*)
Definition s_transfer_dep_r {T:Set} {U:T->Set} {x y:T} (pf:y = x)
           (a:U x):U y.
subst. refine a.
Defined.

(*Wyckoff*)
Lemma s_transfer_dep_transfer_dep_r_compat :
  forall {T:Set} {U:T->Set} {x y:T} (pf:x = y) (pf':y = x) (a:U x),
    s_transfer_dep pf a = s_transfer_dep_r pf' a.
intros T U x y h1 h2 a.
unfold s_transfer_dep, s_transfer_dep_r.
destruct h2. unfold eq_rec_r. simpl.
pose proof (proof_irrelevance _ (eq_sym h1) (eq_refl _)) as h2.
rewrite h2.
simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma s_transfer_dep_r_undoes_transfer_dep :
  forall {T:Set} {U:T->Set} {x y:T} (pf:x = y) (a:U x),
    s_transfer_dep_r pf (s_transfer_dep pf a) = a.
intros T U x y h1 a.
subst.
unfold transfer_dep_r, transfer_dep, eq_rect_r. simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma s_transfer_dep_undoes_transfer_dep_r :
  forall {T:Set} {U:T->Set} {x y:T} (pf:x = y) (a:U y),
    s_transfer_dep pf (s_transfer_dep_r pf a) = a.
intros T U x y h1 a.
subst.
unfold s_transfer_dep_r, s_transfer_dep, eq_rect_r. simpl.
reflexivity.
Qed.




(*Wyckoff*)
Lemma s_transfer_dep_inj :
  forall {T:Set} {U:T->Set} {x y:T} (pf:x = y) (a b:U x),
    s_transfer_dep pf a = s_transfer_dep pf b ->
    a = b.
intros T U x y h1 a b h2.
subst.
unfold transfer_dep in h2.
unfold eq_rect_r in h2. simpl in h2.
assumption.
Qed.

(*Wyckoff*)
Lemma s_transfer_dep_r_inj :
  forall {T:Set} {U:T->Set} {x y:T} (pf:x = y) (a b:U y),
    s_transfer_dep_r pf a = s_transfer_dep_r pf b ->
    a = b.
intros T U x y h1 a b h2.
subst.
unfold s_transfer_dep_r, eq_rect_r in h2.
simpl in h2.
assumption.
Qed.


(*Wyckoff*)
Lemma s_transfer_dep_existT :
  forall {T:Set} {U:T->Set} (x y:T) (pf:x = y)
         (a:U x),
    existT _ _ a = existT _ _ (transfer_dep pf a).
intros T U x y h1 a.
subst.
f_equal.
Qed.

(*Wyckoff*)
Lemma s_projT_injective :
  forall {T:Set} (A:T->Set)
         (x y:{a:T & (A a)})
         (pf:projT1 x = projT1 y),
    s_transfer_dep pf (projT2 x) = projT2 y ->
    x = y.
intros T A x y h1 h2.
destruct x; destruct y.
simpl in h1. simpl in h2.
subst.
apply s_transfer_dep_existT.
Qed.




(*Wyckoff*)
Lemma s_transfer_dep_eq_refl : forall {T:Set} (U:T->Set) (x:T)
                                      (p:U x)
                                      (pfe:x = x),
                               transfer_dep pfe p = p.
  intros ? ? ? ? h1.
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  unfold s_transfer_dep, eq_rect_r, eq_rect; simpl; auto.
Defined.


(*Wyckoff*)
Lemma s_transfer_dep_r_eq_refl : forall {T:Set} (U:T->Set) (x:T)
                                        (p:U x)
                                        (pfe:x = x),
                                   s_transfer_dep_r pfe p = p.
  intros ? ? ? ? h1.
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  unfold s_transfer_dep_r, eq_rect_r, eq_rect; simpl; auto.
Defined.


(*Wyckoff*)
Lemma s_existT_injective1 : forall {T:Set} (U:T->Set)
                                   (x y:T) (p:U x) (q:U y),
  existT _ x p = existT _ y q -> x = y.
intros T U x y p q h1.
pose proof (f_equal (@projT1 _ _) h1) as h2.
simpl in h2.
assumption.
Qed.


(*Wyckoff*)
Lemma s_transfer_dep_prop :
  forall {T:Set} {U:T->Set} (x y:T) (pf:x = y)
         (a:U x) (P:{t:T & (U t)}->Prop),
    P (existT _ _ a) <-> P (existT _ _ (transfer_dep pf a)).
intros T U x y h1 a P.
rewrite (s_transfer_dep_existT _ _ h1 a).
tauto.
Qed.



(*Wyckoff*)
Lemma s_transfer_dep_eq_iff :
  forall {T:Set} {U:T->Set} (x y:T) (pf:x = y)
         (a:U x) (b:U y),
    existT _ _ a = existT _ _ b <->
        s_transfer_dep pf a = b.
intros T U x y h1 a b.
subst.
split.
intro h1.
apply inj_pair2 in h1.
unfold s_transfer_dep.
unfold eq_rec_r.
rewrite <- eq_rect_eq.
assumption.
intro h1.
unfold s_transfer_dep in h1. unfold eq_rec_r in h1.
rewrite <- eq_rec_eq in h1.
subst.
reflexivity.
Qed.

(*Wyckoff*)
Lemma s_transfer_dep_r_eq_iff :
  forall {T:Set} {U:T->Set} (x y:T) (pf:y = x)
         (a:U x) (b:U y),
    existT _ _ a = existT _ _ b <->
        s_transfer_dep_r pf a = b.
intros T U x y h1 a b.
subst.
split.
intro h1.
apply inj_pair2 in h1.
unfold s_transfer_dep_r.
unfold eq_rec_r.
rewrite <- eq_rec_eq.
assumption.
intro h1.
unfold s_transfer_dep_r in h1. unfold eq_rec_r in h1.
rewrite <- eq_rec_eq in h1.
subst.
reflexivity.
Qed.

End SetSortTransfer.

(*Wyckoff*)
Lemma transfer_in : 
  forall {T T':Type} (pf:T = T') (A:Ensemble T) (x:T),
    In A x <-> In (transfer_dep pf A) (transfer pf x).
intros T T' h1 A x. subst.
unfold transfer_dep. unfold eq_rect_r. simpl.
unfold transfer. unfold eq_rect_r. simpl.
tauto.
Qed. 


(*Wyckoff*)
Lemma transfer_in' : 
  forall {T T':Type} (pf:T = T') (A:Ensemble T') (x:T),
    In A (transfer pf x) <-> In (transfer_dep_r pf A) x.
intros T T' h1 A x. subst.
rewrite transfer_eq_refl, transfer_dep_r_eq_refl.
tauto.
Qed. 


(*Wyckoff*)
Lemma transfer_sig_set_eq_in : 
  forall {T:Type} (A B:Ensemble T) 
         (pfa:A = B) (D:Ensemble (sig_set B)) (x:sig_set A),
    In D (transfer (sig_set_eq _ _ pfa) x) <->
    Inn 
      (transfer_dep_r 
         (U:=(fun C : Ensemble T => Ensemble (sig_set C))) pfa D) x.
intros T A B h1 D x. subst.
pose proof (proof_irrelevance _ (sig_set_eq B B eq_refl) eq_refl) as h1.
rewrite h1.
rewrite transfer_eq_refl, transfer_dep_r_eq_refl.
tauto.
Qed.    




(*Wyckoff*)
Lemma transfer_in_r : 
  forall {T T':Type} (pf:T' = T) (A:Ensemble T) (x:T),
    In A x <-> In (transfer_dep_r pf A) (transfer_r pf x).
intros T T' h1 A x. subst.
tauto.
Qed.


(*Wyckoff*)
Lemma transfer_in_r' : 
  forall {T T':Type} (pf:T = T') (A:Ensemble T) (x:T'),
    In A (transfer_r pf x) <-> In (transfer_dep pf A) x.
intros T T' h1 A x. subst.
rewrite transfer_r_eq_refl, transfer_dep_eq_refl.
tauto.
Qed. 


(*Wyckoff*)
Lemma transfer_r_sig_set_eq_in : 
  forall {T:Type} (A B:Ensemble T) 
         (pfa:A = B) (D:Ensemble (sig_set A)) (x:sig_set B),
    In D (transfer_r (sig_set_eq _ _ pfa) x) <->
    Inn 
      (transfer_dep 
         (U:=(fun C : Ensemble T => Ensemble (sig_set C))) pfa D) x.
intros T A B h1 D x. subst.
pose proof (proof_irrelevance
              _ (sig_set_eq B B eq_refl)
              eq_refl) as h1.
rewrite h1, transfer_r_eq_refl, transfer_dep_eq_refl.
tauto.
Qed.    



(*Wyckoff*)
Lemma transfer_sig_set_eq : 
  forall {T:Type} (A A':Ensemble T) 
         (pf:A = A') (pfs:sig_set A = sig_set A')
         (x:sig_set A),
    transfer pfs x = exist (fun x=>In A' x) (proj1_sig x) (@transfer_dep _ (fun S=>In S (proj1_sig x)) _ _ pf (proj2_sig x)).
intros T A A' h1 h2 x.
subst.
pose proof (proof_irrelevance _ h2 (eq_refl _)); subst.
simpl.
destruct x as [x h1].
apply proj1_sig_injective.
simpl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_r_sig_set_eq : 
  forall {T:Type} (A A':Ensemble T) 
         (pf:A = A') (pfs:sig_set A = sig_set A')
         (x:sig_set A'),
    transfer_r pfs x = exist (fun x=>In A x) (proj1_sig x) (@transfer_dep_r _ (fun S=>In S (proj1_sig x)) _ _ pf (proj2_sig x)).
intros T A A' h1 h2 x.
subst.
pose proof (proof_irrelevance _ h2 (eq_refl _)); subst.
simpl.
destruct x as [x h1].
apply proj1_sig_injective.
simpl.
reflexivity.
Qed.


(*Wyckoff*)
Lemma transfer_dep_transfer_dep_sig_set_eq_compat : 
  forall {T:Type} (A B:Ensemble T) (pf:A = B)
         (U: Type->Type)  (f:U (sig_set A)),
    transfer_dep pf f (U:=fun C:Ensemble T=> U (sig_set C))  = transfer_dep (sig_set_eq _ _ pf) f.
intros T A B h1 U f.
subst.
rewrite transfer_dep_eq_refl.
pose proof (proof_irrelevance _ (sig_set_eq B B eq_refl) eq_refl) as h1.
rewrite h1.
rewrite transfer_dep_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_dep_r_transfer_dep_r_sig_set_eq_compat : 
  forall {T:Type} (A B:Ensemble T) (pf:A = B)
         (U: Type->Type)  (f:U (sig_set B)),
    transfer_dep_r pf f (U:=fun C:Ensemble T=> U (sig_set C))  = transfer_dep_r (sig_set_eq _ _ pf) f.
intros T A B h1 U f.
subst.
rewrite transfer_dep_r_eq_refl.
pose proof (proof_irrelevance _ (sig_set_eq B B eq_refl)
            eq_refl) as h1.
rewrite h1, transfer_dep_r_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_dep_transfer_compat : 
  forall {T:Type} (A B:T) (pf:A = B)
         (U: T->Type)  (f:U A),
    transfer_dep pf f (U:=fun C:T=> U C)  = 
    transfer (f_equal U pf) f.
intros T A B h1 U f.
subst.
simpl.
rewrite transfer_dep_eq_refl.
rewrite transfer_eq_refl.
reflexivity.
Qed.                                

(*Wyckoff*)
Lemma transfer_dep_r_transfer_r_compat : 
  forall {T:Type} (A B:T) (pf:A = B)
         (U: T->Type)  (f:U B),
    transfer_dep_r pf f (U:=fun C:T=> U C)  = 
    transfer_r (f_equal U pf) f.
intros T A B h1 U f.
subst.
simpl.
rewrite transfer_dep_r_eq_refl.
rewrite transfer_r_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_transfer_dep_compat : 
  forall {T:Type} {U: T->Type}
         {A B:T}
         (pfe:A = B)
         (pfeu:U A = U B)
         (f:U A),
    transfer pfeu f =
    transfer_dep pfe f (U:=fun C:T=> U C).
  intros; subst; rewrite transfer_eq_refl, transfer_dep_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_transfer_dep_r_compat : 
  forall {T:Type} {U: T->Type}
         {A B:T}
         (pfe:A = B)
         (pfeu:U A = U B)
         (f:U B),
    transfer_r pfeu f =
    transfer_dep_r pfe f (U:=fun C:T=> U C).
  intros; subst; rewrite transfer_r_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.



(*Wyckoff*)
Lemma transfer_dep_id_transfer_compat : 
  forall (T U:Type) (pf:T = U)
         (a:T),
         transfer_dep pf a (U:=@id Type) = transfer pf a.
intros T U h1 a. subst.
rewrite transfer_dep_eq_refl, transfer_eq_refl.
reflexivity.
Qed.

(*Wyckoff*)
Lemma transfer_dep_r_id_transfer_r_compat : 
  forall (T U:Type) (pf:T = U)
         (a:U),
         transfer_dep_r pf a (U:=@id Type) = transfer_r pf a.
intros T U h1 a. subst.
rewrite transfer_dep_r_eq_refl, transfer_r_eq_refl.
reflexivity.
Qed.



(*Wyckoff*)
Lemma extends_prim_impl_extends_sig : 
  forall {T T' U:Type}
         (f:T->U) (g:T'->U),
    extends_prim f g ->
    exists (A:Ensemble T) (pf:T' = sig_set A),
      extends_sig1 f (transfer_dep (U:=fun V=>V->U) pf g).
intros T T' U f g h1.
red in h1. destruct h1 as [A h1]. destruct h1 as [h1 h2].
subst.
exists A. exists eq_refl. 
red.
intro x.
rewrite transfer_dep_eq_refl.
specialize (h2 (proj1_sig x) x (proj2_sig x)).
rewrite <- unfold_sig in h2.
specialize (h2 eq_refl).
rewrite h2.
reflexivity.
Qed.


(*Wyckoff*)
Lemma proj1_sig_transfer_dep :
  forall {T:Type} {U:Type->Type} {P:T->(U T)->Prop}
         (l l':U T) (pfx:l = l') (x:{c:T | P c l}),
    proj1_sig (transfer_dep (U:=(fun xl => {t:T | P t xl}))
                              pfx x) = 
               proj1_sig x.
  intros; subst; rewrite transfer_dep_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma proj1_sig_transfer_dep_r :
  forall {T:Type} {U:Type->Type} {P:T->(U T)->Prop}
         (l l':U T) (pfx:l = l') (x:{c:T | P c l'}),
    proj1_sig (transfer_dep_r (U:=(fun xl => {t:T | P t xl}))
                              pfx x) = 
               proj1_sig x.
  intros; subst; rewrite transfer_dep_r_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma proj1_sig_transfer_dep_s :
  forall {T:Type} {U:Type->Set} {P:T->(U T)->Prop}
         (l l':U T) (pfx:l = l') (x:{c:T | P c l}),
    proj1_sig (transfer_dep (U:=(fun xl => {t:T | P t xl}))
                              pfx x) = 
               proj1_sig x.
  intros; subst; rewrite transfer_dep_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma proj1_sig_transfer_dep_r_s :
  forall {T:Type} {U:Type->Set} {P:T->(U T)->Prop}
         (l l':U T) (pfx:l = l') (x:{c:T | P c l'}),
    proj1_sig (transfer_dep_r (U:=(fun xl => {t:T | P t xl}))
                              pfx x) = 
               proj1_sig x.
  intros; subst; rewrite transfer_dep_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep_fun_compat :
  forall {T V:Type} {U:T->Type} {a b:T}
         (pf:a = b) (f:U a->V) (x:U b),
    transfer_dep (U:=(fun x:T => U x -> V)) pf f x =
    f (transfer_dep_r pf x).
  intros; subst.
  rewrite transfer_dep_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma transfer_dep_r_fun_compat :
  forall {T V:Type} {U:T->Type} {a b:T}
         (pf:a = b) (f:U b->V) (x:U a),
    transfer_dep_r (U:=(fun x:T => U x -> V)) pf f x =
    f (transfer_dep pf x).
  intros; subst.
  rewrite transfer_dep_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.


(*Wyckoff*)
Definition transfer_seg_fun {n n':nat} {U:Type}
           (pf:n = n') (f:seg_fun n U) : seg_fun n' U :=
  transfer_dep (U:=fun a => seg_fun a U) pf f.

(*Wyckoff*)
Lemma transfer_seg_fun_eq_refl :
  forall {n:nat} {U:Type}
         (f:seg_fun n U),
    transfer_seg_fun eq_refl f = f.
  intros; unfold transfer_seg_fun; rewrite transfer_dep_eq_refl; auto.
Qed.
  
(*Wyckoff*)
Lemma transfer_seg_fun_compat :
  forall {n n':nat} {U:Type} (pfeq:n = n') (f:seg_fun n U)
         (m:nat) (pflt:m < n'),
    transfer_seg_fun pfeq f m pflt =
    f m (transfer_dep_r (U:= fun c => m < c) pfeq pflt).
  intros n n' U h1 f m h2.
  subst.
  rewrite transfer_seg_fun_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_seg_fun_app :
  forall {n} {U:Type}
         {i i'}
         (pfe:i = i')
         (f:seg_fun n U)
         (pfi:i < n),
    f i pfi =
    f i' (eq_lt_trans _ _ _ (eq_sym pfe) pfi).
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma transfer_seg_fun_app_r :
  forall {n} {U:Type}
         {i i'}
         (pfe:i = i')
         (f:seg_fun n U)
         (pfi':i' < n),
    f i' pfi' =
    f i (eq_lt_trans _ _ _ pfe pfi').
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma transfer_seg_fun_app' :
  forall {n} {U:Type}
         {i i'}
         (pfe:i = i')
         (f:seg_fun n U)
         (pfi: i  < n)
         (pfi':i' < n),
    f i pfi =
    f i' pfi'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma transfer_seg_fun_app_r' :
  forall {n} {U:Type}
         {i i'}
         (pfe:i = i')
         (f:seg_fun n U)
         (pfi:i < n)
         (pfi':i' < n),
    f i' pfi' =
    f i pfi.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.


(*Wyckoff*)
Definition transfer_seg_fun_depT2
           {T n} {U:T->Type}
           {f:seg_fun n T}
           {i i'}
           {pfi: i  < n}
           (pfe:i = i')
           (pfi':i' < n)
           (u:U (f i pfi)) :
  U (f i' pfi').
  subst; assert (pfi = pfi'); try apply proof_irrelevance; subst; auto.
Defined.


(*Wyckoff*)
Definition transfer_seg_fun_depT2_r
           {T n} {U:T->Type}
           {f:seg_fun n T}
           {i i'}
           {pfi': i'  < n}
           (pfe:i = i')
           (pfi:i < n)
           (u:U (f i' pfi')) :
  U (f i pfi).
  subst; assert (pfi = pfi'); try apply proof_irrelevance; subst; auto.
Defined.



(*Wyckoff*)
Lemma transfer_seg_fun_depT2_eq_refl : 
  forall {T n} {U:T->Type}
         {f:seg_fun n T}
         {i}
         (pfi:i < n)
         (pfe: i = i)
         (u:U (f i pfi)),
    transfer_seg_fun_depT2 pfe pfi u = u.
  unfold transfer_seg_fun_depT2, eq_rect_r, eq_rect; intros. 
  pose proof (proof_irrelevance _ (eq_sym pfe) (eq_refl _)) as h1.
  rewrite h1.
  pose proof (proof_irrelevance
                _ (eq_sym (proof_irrelevance (i < n) pfi pfi))
                (eq_refl _)) as h2.
  rewrite h2; auto.
Qed.



(*Wyckoff*)
Lemma transfer_seg_fun_depT2_r_eq_refl : 
  forall {T n} {U:T->Type}
         {f:seg_fun n T}
         {i}
         (pfi:i < n)
         (pfe: i = i)
         (u:U (f i pfi)),
    transfer_seg_fun_depT2_r pfe pfi u = u.
  unfold transfer_seg_fun_depT2_r, eq_rect_r, eq_rect; intros. 
  pose proof (proof_irrelevance
                _ (eq_sym pfe) (eq_refl _)) as h1.
  pose proof (proof_irrelevance
                _ (eq_sym (proof_irrelevance (i < n) pfi pfi))
                (eq_refl _)) as h2.
  rewrite h1, h2; auto.
Qed.

(*Wyckoff*)
Lemma transfer_seg_fun_depT2_eq_iff :
  forall {T n} {U:T->Type}
         {f:seg_fun n T}
         {i i'}
         {pfi: i  < n}
         (pfe:i = i')
         (pfi':i' < n)
         (u:U (f i pfi))
         (v:U (f i' pfi')),
    transfer_seg_fun_depT2 pfe pfi' u = v <->
    existT id _ u =
    existT id _ v.
  intros; subst; pose proof (proof_irrelevance _ pfi pfi'); subst;
  rewrite transfer_seg_fun_depT2_eq_refl;
  split; intro h1; subst; auto;
  apply existT_injective2 in h1; auto.
Qed.


(*Wyckoff*)
Lemma transfer_seg_fun_depT2_r_eq_iff :
  forall {T n} {U:T->Type}
         {f:seg_fun n T}
         {i i'}
         {pfi':i' < n}
         (pfe:i = i')
         (pfi:i < n)
         (u:U (f i' pfi'))
         (v:U (f i pfi)),
    transfer_seg_fun_depT2_r pfe pfi u = v <->
    existT id _ u =
    existT id _ v.
  intros; subst; pose proof (proof_irrelevance _ pfi pfi'); subst;
  rewrite transfer_seg_fun_depT2_r_eq_refl;
  split; intro h1; subst; auto;
  apply existT_injective2 in h1; auto.
Qed.




(*Wyckoff*)
Lemma transfer_seg_fun_depT2_transfer_dep :
  forall {T n} {U:T->Type}
         {f:seg_fun n T}
         {i i' t}
         {pfi: i < n}
         (pfi':i' < n)
         (pfe: i = i')
         (u:U t)
         (pfe':t = f i pfi),
    transfer_seg_fun_depT2
      pfe pfi'
      (transfer_dep pfe' u) =
    transfer_dep (eq_trans pfe' (transfer_seg_fun_app' pfe f pfi pfi')) u.
  intros; subst;
  rewrite transfer_dep_eq_refl;
  pose proof (proof_irrelevance _ pfi pfi'); subst; 
  rewrite transfer_seg_fun_depT2_eq_refl,
  transfer_dep_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_seg_fun_depT2_r_transfer_dep :
  forall {T n} {U:T->Type}
         {f:seg_fun n T}
         {i i' t}
         {pfi: i < n}
         (pfi':i' < n)
         (pfe: i = i')
         (u:U t)
         (pfe':t = f i' pfi'),
    transfer_seg_fun_depT2_r
      pfe pfi
      (transfer_dep pfe' u) =
    transfer_dep
      (eq_trans pfe' (transfer_seg_fun_app_r' pfe f pfi pfi')) u.
  intros; subst;
  rewrite transfer_dep_eq_refl;
  pose proof (proof_irrelevance _ pfi pfi'); subst;
  rewrite transfer_seg_fun_depT2_r_eq_refl,
  transfer_dep_eq_refl; auto.
Qed.

(*Wyckoff*)
Definition transfer_seg_fun_depT_pf
           {T n} {U:T->Type} {f:seg_fun n T}
           {i} {pfi:i < n}
           (pfi':i < n) 
           (x:U (f i pfi)) :
  U (f i pfi') :=
  transfer_dep (proof_irrelevance _ pfi pfi')
               x (U := fun pfc => U (f i pfc)).

(*Wyckoff*)
Definition transfer_sig {T} {P Q:T->Prop} (pfi:pred_impl P Q)
           (x:{t:T | P t}) :
  {x:T | Q x} :=
  exist Q (proj1_sig x) (pfi _ (proj2_sig x)).

(*Wyckoff*)
Definition transfer_sig' {T} {P Q:T->Prop} (pfi:iff_pred P Q)
           (x:{t:T | P t}) :
  {x:T | Q x} :=
  transfer_sig (iff_pred1 _ _ pfi) x.

(*Wyckoff*)
Lemma in_transfer_dep :
  forall {T} {U:T->Type} {x y:T}
         {A:forall t, Ensemble (U t)}
         (pfe:x = y)
         (z:U x),
    Inn (A x) z ->
    Inn (A y) 
        (transfer_dep pfe z).
  intros; subst; rewrite transfer_dep_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma in_transfer_dep_r :
  forall {T} {U:T->Type} {x y:T}
         {A:forall t, Ensemble (U t)}
         (pfe:x = y)
         (z:U y),
    Inn (A y) z ->
    Inn (A x) 
        (transfer_dep_r pfe z).
  intros; subst; rewrite transfer_dep_r_eq_refl; auto.
Qed.


(*Wyckoff*)
Lemma in_transfer_dep_iff :
  forall {T U} (pfe:T = U) (A:Ensemble T) (y:U),
    Inn (transfer_dep pfe A) y <->
    Inn A (transfer_r pfe y).
  intros; subst; tauto.
Qed.

(*Wyckoff*)
Lemma in_transfer_dep_r_iff :
  forall {T U} (pfe:T = U) (A:Ensemble U) (x:T),
    Inn (transfer_dep_r pfe A) x <->
    Inn A (transfer pfe x).
  intros; subst; tauto.
Qed.


(*Wyckoff*)
Lemma in_transfer_ens :
  forall {T U} {A:T->Ensemble U}
         {x y} {u}
         (pfe:x = y),
    Inn (A x) u ->
    Inn (A y) u.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma in_transfer_ens_r :
  forall {T U} {A:T->Ensemble U}
         {x y} {u}
         (pfe:x = y),
    Inn (A y) u ->
    Inn (A x) u.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma in_transfer_dep_ens :
  forall {T} {U:T->Type} {A:forall x:T, Ensemble (U x)}
         {x y} {u:U x}
         (pfe:x = y),
      Inn (A x) u ->
      Inn (A y) (transfer_dep pfe u).
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma in_transfer_dep_ens_r :
  forall {T} {U:T->Type} {A:forall t:T, Ensemble (U t)}
         {x y} {u:U y}
         (pfe:x = y),
    Inn (A y) u ->
    Inn (A x) (transfer_dep_r pfe u).
  intros; subst; auto.
Qed.



(*Wyckoff*)
Definition transfer_dep_ens' {T} {U:T->Type} {x y:T}
           (pfe:x = y) (A:Ensemble (U x)) :
  Ensemble (U y).
  subst; auto.
Defined.

(*Wyckoff*)
Definition transfer_dep_ens_r' {T} {U:T->Type} {x y:T}
           (pfe:x = y) (A:Ensemble (U y)) :
  Ensemble (U x).
  subst; auto.
Defined.

(*Wyckoff*)
Lemma in_transfer_dep_ens' :
  forall {T} {U:T->Type} {x y:T}
         {A:Ensemble (U x)}
         {z:U x}
         (pfe:x = y),
    Inn A z ->
    Inn (transfer_dep_ens' pfe A)
        (transfer_dep pfe z).
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma in_transfer_dep_ens_r' :
  forall {T} {U:T->Type} {x y:T}
         {A:Ensemble (U y)}
         {z:U y}
         (pfe:x = y),
    Inn A z ->
    Inn (transfer_dep_ens_r' pfe A)
        (transfer_dep_r pfe z).
  intros; subst; auto.
Qed.

(*Wyckoff*)
Definition transfer_prod {U U' V V'} (pfu:U = U') (pfv:V = V')
           (pr:U*V) : U' * V' :=
  (transfer pfu (fst pr),
   transfer pfv (snd pr)).


(*Wyckoff*)
Fixpoint transfer_tuple_prod
         {T V n}
         {U:T->Type}
         {W:V->Type} :
  forall {w:T^n}
         {v:V^n}
         (pfe:pointwise_ind_eq U W w v)
         (t:tuple_prod U w),
    tuple_prod W v :=
  match n with
    | 0 => fun _ _ _ _ => tt
    | S n' => fun w v pfe t =>
                let pfe' := pointwise_ind_eq_to_S pfe in
                (transfer (pfe 0 (lt_0_Sn _)) (fst t),
                 @transfer_tuple_prod _ _ n'
                                      _ _
                                      (snd w) (snd v) pfe' (snd t))
  end.

(*Wyckoff*)
Fixpoint transfer_tuple_prod_r
         {T V n}
         {U:T->Type}
         {W:V->Type} :
  forall {w:T^n}
         {v:V^n}
         (pfe:pointwise_ind_eq U W w v)
         (t:tuple_prod W v),
    tuple_prod U w :=
  match n with
    | 0 => fun _ _ _ _ => tt
    | S n' => fun w v pfe t =>
                let pfe' := pointwise_ind_eq_to_S pfe in
                (transfer_r (pfe 0 (lt_0_Sn _)) (fst t),
                 @transfer_tuple_prod_r _ _ n'
                                      _ _
                                      (snd w) (snd v) pfe' (snd t))
  end.



(*Wyckoff*)
Fixpoint transfer_tuple_prod_const {T U n} : 
  forall {x:T^n} 
         (p:tuple_prod (fun _:T => U) x),
    U ^ n :=
  match n with
    | 0 => fun _ _ => tt 
    | S n' =>
      fun x p =>
        let (fx, sx) := x in
        let (fp, sp) := p in
        let t := transfer_tuple_prod_const sp in
        (fp, t)
  end.

(*Wyckoff*)
Fixpoint transfer_to_tuple_prod_const {T U n} : 
  forall
    (x:T^n)
    (u:U^n),
    tuple_prod (fun _:T => U) x :=
  match n
  with
    | 0 => fun _ _ => tt
    | S n' =>
      fun x u =>
        let (fu, su) := u in
        let t := transfer_to_tuple_prod_const _ su in
        (fu, t) 
  end.


(*Wyckoff*)
Lemma transfer_tuple_prod_const_to_compat :
  forall {T U n} (x:T^n) (u:U^n),
    transfer_tuple_prod_const
      (transfer_to_tuple_prod_const x u) = u.
  intros T U n; induction n; simpl.
  intros; apply unit_eq.
  intros x u; destruct x, u; f_equal; auto.
Qed.

(*Wyckoff*)
Lemma transfer_to_tuple_prod_const_compat :
  forall {T U n} {x:T^n} (p:tuple_prod (fun _:T => U) x),
    transfer_to_tuple_prod_const x
      (transfer_tuple_prod_const p) = p.
  intros T U n; induction n; simpl.
  intros; apply unit_eq.
  intros x u; destruct x, u; f_equal; auto.
Qed.


(*Wyckoff*)
Lemma inj_transfer_tuple_prod_const :
  forall {T U n} (x:T^n) (p p':tuple_prod (fun _:T => U) x),
    transfer_tuple_prod_const p =
    transfer_tuple_prod_const p' ->
    p = p'.
  intros T U n.
  induction n as [|n ih]; simpl.
  intros; apply unit_eq.
  intros x p p' h1; simpl in h1.
  destruct x as [fx sx],
                p as [fp sp],
                     p' as [fp' sp'].
  inversion h1; subst; f_equal; auto.
Qed.


(*Wyckoff*)
Lemma inj_transfer_to_tuple_prod_const :
  forall {T U n} (x:T^n) (u u':U^n),
    transfer_to_tuple_prod_const x u =
    transfer_to_tuple_prod_const x u' ->
    u = u'.
  intros T U n. 
  induction n as [|n ih]; simpl.
  intros; apply unit_eq. 
  intros x u u' h1; simpl in h1.
  destruct x as [fx sx],
                u as [fu su],
                     u' as [fu' su'].
  inversion h1; subst; f_equal.
  apply ih in H1; auto.
Qed.


                             
(*Wyckoff*)
Lemma nth_tuple_prod_transfer_to_tuple_prod_const :
  forall {T U n} (x:T^n) (y:U^n) i (pfi:i < n),
    nth_tuple_prod (transfer_to_tuple_prod_const x y) i pfi =
    nth_prod i pfi y.
  intros ? ? n; induction n; simpl.
  intros; omega.
  intros x y i; destruct x, y, i; simpl; auto.
Qed.


(*Wyckoff*)
Lemma nth_prod_transfer_tuple_prod_const :
  forall {T U n} {x:T^n}
         (y:tuple_prod (fun _:T => U) x) i (pfi:i < n),
    nth_prod i pfi (transfer_tuple_prod_const y) =
    nth_tuple_prod y i pfi.
  intros ? ? n; induction n; simpl.
  intros; omega.
  intros x y i; destruct x, y, i; simpl; auto.
Qed.

(*Wyckoff*)
Lemma tuple0_eq :
  forall {T n} (x:T^n)
         (pfe:n = 0),
    x = transfer_dep_r pfe tt.
  intros; subst; rewrite transfer_dep_r_eq_refl; apply unit_eq.
Qed.


(*Wyckoff*)
Definition transfer_tt0 T : T^0 :=
  tt.


(*Wyckoff*)
Definition transfer_tt {n} (pfeq:n = 0) T : T^n :=
  transfer_dep_r pfeq (transfer_tt0 T).

(*Wyckoff*)
Lemma transfer_tt_eq_refl :
  forall T (pf0:0 = 0), 
    transfer_tt pf0 T = tt.
  unfold transfer_tt; intros. rewrite transfer_dep_r_eq_refl.
  apply nprod0_eq.
Defined.

  
(*Wyckoff*)
Definition transfer_tt0_tuple_prod
           {T} (A:T->Type) :
  tuple_prod A (transfer_tt0 T) :=
  tt.


(*Wyckoff*)
Definition transfer_tt0_tuple_prod'
           {T} (A:T->Type) (x:T^0) : 
  tuple_prod A x :=
  tt.


(*Wyckoff*)
Definition transfer_tt_tuple_prod
           {T n} (A:T->Type) (pfeq:n = 0) (x:T^n) :
  tuple_prod A x.
  intros; subst; destruct x; refine tt.
Defined.

(*Wyckoff*)
Lemma transfer_tt_tuple_prod_eq_refl :
  forall {T} (A:T->Type) (pfeq:0 = 0) (x:T^0),
    transfer_tt_tuple_prod A pfeq x =
    tt.
  unfold transfer_tt_tuple_prod, eq_rect_r, eq_rect; intros.
  pose proof (proof_irrelevance _ pfeq (eq_refl _)); subst.
  simpl.
  destruct x; auto.
Defined.

(*Wyckoff*)
Definition transfer_to_tuple_prod0 {T} (A:T->Type) (x:T^0)
           (p:tuple_prod A x) :
  tuple_prod A (transfer_tt0 T) :=
  p.

(*Wyckoff*)
Definition transfer_tuple_prod0 {T} (A:T->Type) (x:T^0)
           (p:tuple_prod A (transfer_tt0 T)) :
  tuple_prod A x := p.

(*Wyckoff*)
Lemma tuple_prod0_eq :
  forall {T} (A:T->Type) (x:T^0)
         (p q:tuple_prod A x),
    p = q.
  simpl; intros; apply unit_eq.
Qed.

(*Wyckoff*)
Lemma tuple_prod0_eq' :
  forall {T} (A:T->Type) (x y:T^0)
         (p:tuple_prod A x)
         (q:tuple_prod A y)
         (pfe:x = y),
    p = transfer_dep_r pfe q.
  simpl; intros; apply unit_eq.
Qed.


(*Wyckoff*)
Lemma tuple_prod_n0_eq :
  forall {T n} (A:T->Type) (x:T^n)
         (p q:tuple_prod A x),
    n = 0 ->
    p = q.
  simpl; intros; subst; apply tuple_prod0_eq.
Qed.


(*Wyckoff*)
Lemma tuple_prod_n0_eq' :
  forall {T n} (A:T->Type) (x y:T^n)
         (p:tuple_prod A x)
         (q:tuple_prod A y)
         (pfe:x = y),
    n = 0 ->
    p = transfer_dep_r pfe q.
  simpl; intros; subst; rewrite transfer_dep_r_eq_refl; apply tuple_prod0_eq.
Qed.


(*Wyckoff*)
Lemma tuple_prod_tt :
  forall {T n} {U:T->Type} {x:T^n} 
         (w:tuple_prod U x)
         (pfe:n = 0),
    w = transfer_tt_tuple_prod U pfe x.
  intros; subst.
  rewrite transfer_tt_tuple_prod_eq_refl.
  destruct x.
  red in w.
  apply unit_eq.
Qed.




(*Wyckoff*)
Definition transfer_seg_fun_to_S_app
           {n} (f:seg_fun (S n) Type) i
           (pfi:i < n)
           (pfi':S i < S n) 
           (x:seg_fun_to_S f i pfi) :
  f (S i) pfi'.
  unfold seg_fun_to_S in x.
  erewrite f_equal_dep.
  apply x.
  auto.
Defined.

(*Wyckoff*)
Definition transfer_to_seg_fun_to_S_app
           {n} (f:seg_fun (S n) Type) i
           (pfi:i < n)
           (pfi':S i < S n)
           (x:f (S i) pfi') : 
  seg_fun_to_S f i pfi. 
  unfold seg_fun_to_S.
  erewrite f_equal_dep in x.
  apply x.
  auto.
Defined.



(*Wyckoff*)  
Lemma transfer_seg_fun_to_S_app_undoes :
  forall {n} (f:seg_fun (S n) Type) i
         (pfi:i < n)
         (pfi':S i < S n)
         (x:f (S i) pfi'),
    transfer_seg_fun_to_S_app
      f i pfi pfi' 
      (transfer_to_seg_fun_to_S_app f i pfi pfi' x) =
    x.
  unfold transfer_to_seg_fun_to_S_app,
  transfer_seg_fun_to_S_app; intros;
  unfold eq_rect_r, eq_rect.
  pose proof (proof_irrelevance _ pfi' (lt_n_S i n pfi)); subst.
  rewrite eq_eq_refl at 1.
  rewrite eq_eq_refl at 1.
  reflexivity.
Qed.


(*Wyckoff*)  
Lemma transfer_to_seg_fun_to_S_app_undoes :
  forall {n} (f:seg_fun (S n) Type) i
         (pfi:i < n)
         (pfi':S i < S n)
         (x:seg_fun_to_S f i pfi),
    transfer_to_seg_fun_to_S_app
      f i pfi pfi'
      (transfer_seg_fun_to_S_app f i pfi pfi' x) =
    x.
  unfold transfer_to_seg_fun_to_S_app,
  transfer_seg_fun_to_S_app; intros;
  unfold eq_rect_r, eq_rect.
  pose proof (proof_irrelevance _ pfi' (lt_n_S i n pfi)); subst.
  rewrite eq_eq_refl at 1.
  rewrite eq_eq_refl at 1.
  reflexivity.
Qed.


(*Wyckoff*)  
Definition transfer_seg_fun_to_S_app_iff : 
  forall {n} (f:seg_fun (S n) Type) i
         (pfi:i < n)
         (pfi':S i < S n) 
         (x:seg_fun_to_S f i pfi)
         (y:f (S i) pfi'),
    transfer_seg_fun_to_S_app f i pfi pfi' x = y <->
    x = transfer_to_seg_fun_to_S_app f i pfi pfi' y.
  intros n f i hi hi' x y; split; intro h1; subst;
  [rewrite transfer_to_seg_fun_to_S_app_undoes; auto |
   rewrite transfer_seg_fun_to_S_app_undoes; auto].
Qed.

(*Wyckoff*)
Lemma transfer_seg_fun_to_S_app_subst :
  forall {n} {f:seg_fun (S n) Type}
         (k:seg_fun_dep (seg_fun_to_S f))
         i (pfi:i < n),
    transfer_seg_fun_to_S_app f i pfi (lt_n_S _ _ pfi)
                              (k i pfi) =
    k i pfi.
  intros n f k i hi.
  unfold transfer_seg_fun_to_S_app, eq_rect_r, eq_rect.
  match goal with
      |- context [match ?pf with | _ => _ end] =>
      pose pf as h3 end.
  pose proof (proof_irrelevance _ h3 (eq_refl _)) as h4.
  subst.
  rewrite h4; auto.
Qed.

(*Wyckoff*)
Definition transfer_dep_funT_app
           {T:Type} {P:T->Prop}
           {f:prop_dep_funT P}
           {x y:T}
           (pfe:x = y)
           (pfpx:P x) (pfpy:P y) : 
  f x pfpx ->
  f y pfpy.
  intros; subst; pose proof (proof_irrelevance _ pfpx pfpy);
  subst; auto.
Defined.

(*Wyckoff*)
Definition transfer_dep_funT_app_r
           {T:Type} {P:T->Prop}
           {f:prop_dep_funT P}
           {x y:T}
           (pfe:x = y)
           (pfpx:P x) (pfpy:P y) :
  f y pfpy ->
  f x pfpx.
  intros; subst; pose proof (proof_irrelevance _ pfpx pfpy);
  subst; auto.
Defined.

(*Wyckoff*)
Lemma transfer_dep_funT_app_eq_refl : 
  forall {T:Type} {P:T->Prop}
         {f:prop_dep_funT P}
         {x:T}
         (pfe:x = x)
         (pfx pfy:P x)
         (fx:f x pfx),
    transfer_dep_funT_app pfe pfx pfy fx =
    transfer_dep (proof_irrelevance _ pfx pfy) fx (U := fun pf:P x => f x pf).
  unfold transfer_dep_funT_app, eq_rect_r, eq_rect;
  intros ? ? ? ? h1.
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  simpl.
  intros pfx pfy h1.
  pose proof (proof_irrelevance _ pfx pfy); subst.
  pose proof (proof_irrelevance
                _
                (proof_irrelevance (P x) pfy pfy)
                (eq_refl _)) as h2.
  rewrite h2; simpl; rewrite transfer_dep_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_dep_funT_app_r_eq_refl : 
  forall {T:Type} {P:T->Prop}
         {f:prop_dep_funT P}
         {x:T}
         (pfe:x = x)
         (pfx pfy:P x)
         (fy:f x pfy),
    transfer_dep_funT_app_r pfe pfx pfy fy =
    transfer_dep (proof_irrelevance _ pfy pfx) fy (U := fun pf:P x => f x pf).
  unfold transfer_dep_funT_app_r, eq_rect_r, eq_rect;
  intros ? ? ? ? h1.
  pose proof (proof_irrelevance _ h1 (eq_refl _)); subst.
  simpl.
  intros pfx pfy h1.
  pose proof (proof_irrelevance _ pfx pfy); subst.
  pose proof (proof_irrelevance
                _ (proof_irrelevance (P x) pfy pfy)
                (eq_refl _)) as h2.
  rewrite h2; simpl; rewrite transfer_dep_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma f_equal_dep2 :
  forall {T:Type} {P:T->Prop} {U}
         {f:prop_dep_funT P}
         {g:prop_dep_fun_nest2 f U}
         {x y:T}
         {pfpx:P x} {pfpy:P y}
         {fx:f x pfpx}
         {fy:f y pfpy}
         (pfe: x  = y)
         (pffe:fx = transfer_dep_funT_app_r pfe pfpx pfpy fy),
    g x pfpx fx =
    g y pfpy fy.
  intros; subst; pose proof (proof_irrelevance _ pfpx pfpy);
  subst; f_equal.
  rewrite transfer_dep_funT_app_r_eq_refl,
  transfer_dep_eq_refl; auto.
Qed.

(*Wyckoff*)
Definition transfer_prop_dep 
           {T}
           {U:T->Type}
           {P:forall x:T, U x -> Prop}
           {x y:T}
           {ux:U x}
           {uy:U y}
           (pfe:x = y)
           (_:ux = transfer_dep_r pfe uy) : 
  P x ux ->
  P y uy.
  subst; rewrite transfer_dep_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma transfer_eq_iff :
  forall {T U}
         {x:T} {y:U}
         (pfe:T = U),
    transfer pfe x = y <->
    existT id T x =
    existT id U y.
  intros; subst; rewrite transfer_eq_refl;
  split; intro h1; subst; try apply existT_injective2 in h1; subst; auto.
Qed.

(*Wyckoff*)
Lemma transfer_r_eq_iff :
  forall {T U}
         {x:T} {y:U}
         (pfe:T = U),
    transfer_r pfe y = x <->
    existT id U y =
    existT id T x.
  intros; subst; rewrite transfer_r_eq_refl;
  split; intro h1; subst; try apply existT_injective2 in h1; subst; auto.
Qed.

(*Wyckoff*)
Lemma f_equal_depT :
  forall {T} {U:T->Type}
         {x y}
         (pfe:x = y)
         (ux:U x),
    existT _ x ux =
    existT _ y (transfer_dep pfe ux).
  intros; subst; f_equal; apply proof_irrelevance.
Qed.
(*Wyckoff*)
Lemma f_equal_depT' :
  forall {T} {U:T->Type}
         x y
         (ux:U x)
         (uy:U y)
         (pfe:x = y),
    ux = transfer_dep_r pfe uy ->
    existT (fun x => U x) x ux =
    existT (fun x => U x) y uy.
  intros; subst; f_equal.
Qed.

(*Wyckoff*)
Lemma f_equal_nprod :
  forall {T m n}
         (pfe:m = n)
         (x:T^m),
    existT _ m x =
    existT _ n (transfer_nprod pfe x).
  intros; subst; f_equal; rewrite transfer_nprod_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma f_equal_nprod' :
  forall {T m n}
         {x:T^m}
         {y:T^n}
         (pfe:m = n),
    x = transfer_nprod_r pfe y ->
    existT _ m x =
    existT _ n y.
  intros; subst; f_equal; rewrite transfer_nprod_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma f_equal_prop_depT :
  forall {T} {U:T->Type}
         (P:forall x, U x -> Prop)
         x y
         (ux:U x)
         (uy:U y)
         (pfe:x = y),
    ux = transfer_dep_r pfe uy ->
    P x ux ->
    P y uy.
  intros ? ? ? ? ? ? ? ? ? h1; subst;
  rewrite transfer_dep_r_eq_refl in h1; auto.
Qed.

(*Wyckoff*)
Lemma f_equal_fun_dep' : 
  forall {T V}
         {U:T->Type}
         {P:forall x:T, U x -> Prop}
         {x y:T}
         {ux:U x}
         {uy:U y}
         (f:prop_dep_fun_dep P V)
         (pfe:x = y)
         (pfe':ux = transfer_dep_r pfe uy)
         (px:P x ux),
    let py := f_equal_prop_depT P x y ux uy pfe pfe' px in
    f x ux px = f y uy py.
  intros T V U P x y  ux uy f px h1; subst.
  simpl. 
  intro h1.
  symmetry.
  gen0.
  revert h1.
  rewrite transfer_dep_r_eq_refl.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma f_equal_fun_dep :
  forall {T V} {U:T->Type}
         {P:forall x, U x -> Prop}
         (f:prop_dep_fun_dep P V)
         (x y:T)
         (ux:U x)
         (uy:U y)
         (pfpx:P x ux)
         (pfpy:P y uy)
         (pfe:x = y)
         (pfeu:ux = transfer_dep_r pfe uy),
    f x ux pfpx =
    f y uy pfpy.
  intros; rewrite (f_equal_fun_dep' f pfe pfeu);
  repeat f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma f_equal_dep2T' :
  forall {T U:Type} {P:T->U->Prop} {V:T->U->Type}
         (f:prop_dep_fun2T' P V)
         (x x':T)
         (y y':U)
         (pfp:P x y) (pfp':P x' y')
         (pfex:x = x')
         (pfey:y = y'),
  f x y pfp = transfer_dep2_r pfex pfey (f x' y' pfp').
  intros; subst; f_equal; rewrite transfer_dep2_r_eq_refl; f_equal;
  apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma proj1_sig_transfer_sig_set :
  forall {T} {A B:Ensemble T}
         (x:sig_set A)
         (pfe:A = B)
         (pfe':sig_set A = sig_set B),
    proj1_sig (P := B)
              (transfer pfe' x) =
    proj1_sig x.
  intros; subst; rewrite transfer_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma proj1_sig_transfer_r_sig_set :
  forall {T} {A B:Ensemble T}
         (x:sig_set B)
         (pfe:A = B)
         (pfe':sig_set A = sig_set B),
    proj1_sig (P := A)
              (transfer_r pfe' x) =
    proj1_sig x.
  intros; subst; rewrite transfer_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma proj1_sig_set_transfer_dep :
  forall {T} {U:T->Type}
         {A:forall t:T, Ensemble (U t)}
         {t t'}
         (pfe:t = t')
         (x:sig_set (A t)),
    proj1_sig (transfer_dep (U := fun c => sig_set (A c)) pfe x) =
    transfer_dep pfe (proj1_sig x).
  intros; subst; rewrite transfer_dep_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma proj1_sig_set_transfer_dep_r :
  forall {T} {U:T->Type}
         {A:forall t:T, Ensemble (U t)}
         {t t'}
         (pfe:t = t')
         (x:sig_set (A t')),
    proj1_sig (transfer_dep_r (U := fun c => sig_set (A c)) pfe x) =
    transfer_dep_r pfe (proj1_sig x).
  intros; subst; rewrite transfer_dep_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma Sfun_dep_functional :
  forall {S} {U V:S->Type}
         {A:forall s:S, Ensemble (U s)}
         {B:forall s:S, Ensemble (V s)}
         (f:Sfun_dep A B)
         s s'
         (u:U s) (u':U s')
         (pfa: Inn (A  s) u)
         (pfa':Inn (A s') u')
         (pfe:s = s')
         (pfe':u = transfer_dep_r pfe u'),
    f s u pfa =
    transfer_dep_r pfe (f s' u' pfa') (U := fun s => sig_set (B s)).
  intros; subst; rewrite transfer_dep_r_eq_refl.
  revert pfa.
  rewrite transfer_dep_r_eq_refl.
  intro pfa.
  pose proof (proof_irrelevance _ pfa pfa'); subst; auto.
Qed.

(*Wyckoff*)
Lemma nprod_to_list_functional :
  forall T n n' (x:T^n) (x':T^n')
         (pfe:n = n'),
    x = transfer_dep_r pfe x' ->
    nprod_to_list T n x =
    nprod_to_list T n' x'.
  intros; subst; rewrite transfer_dep_r_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma nprod_to_list_S :
  forall T n (x:T^n) (a:T),
    nprod_to_list _ (S n) (a, x) =
    cons a (nprod_to_list _ _ x).
  auto.
Qed.

(*Wyckoff*)
Lemma nprod_of_list_cons :
  forall {T} (l:list T) a,
    nprod_of_list _ (a::l) =
    (a, nprod_of_list _ l).
  auto.
Qed.


(*Wyckoff*)
Lemma nprod_of_list_functional :
  forall T l l'
         (pfe:l = l'),
    nprod_of_list T l =
    transfer_dep_r (U := fun l => T^(length l)) pfe (nprod_of_list T l').
  intros; subst; rewrite transfer_dep_r_eq_refl; auto.
Qed.



(*Wyckoff*)
Lemma match_eq_refl_eq_transfer_dep_r_iff :
  forall {T} {U:T->Type}
         {x y z}
         (pfxy:x = y)
         (pfxz:x = z)
         (ux:U x)
         (uz:U z)
         (pfeu:U y = U x),
    match pfeu with
      | eq_refl => ux
    end =
    transfer_dep_r (U := U) pfxz uz <->
    ux = transfer_dep_r pfxz uz.
  intros; subst;
  pose proof (proof_irrelevance _ pfeu (eq_refl _)); subst;
  rewrite transfer_dep_r_eq_refl; intuition.
Qed.

(*Wyckoff*)
Lemma match_eq_refl_eq_transfer_dep_iff :
  forall {T} {U:T->Type}
         {x y z}
         (pfxz:x = z)
         (pfyz:y = z)
         (ux:U x)
         (uz:U z)
         (pfeu:U x = U z),
    match pfeu with
      | eq_refl => uz
    end =
    transfer_dep (U := U) pfxz ux <->
    uz = transfer_dep pfxz ux.
  intros; subst;
  pose proof (proof_irrelevance _ pfeu (eq_refl _)); subst;
  rewrite transfer_dep_eq_refl; intuition.
Qed.

End Transfer.

Section TupleProdAgain.

  Fixpoint in_tuple_prod
             {T n} {A:T->Type} {s:T}
  : forall {tpl:T^n} (pfdt:type_eq_dec T)
           (x:A s), tuple_prod A tpl -> Prop :=
    match n with
      | 0 => fun _ _ _ _ => False
      | S n' =>
        fun tpl pfdt =>
          match (pfdt s (fst tpl)) with
            | left pfe =>
              fun x p =>
                let P := in_tuple_prod pfdt x (snd p) in
                let x' := transfer_dep pfe x in 
                x' = fst p \/ P
            | right _ =>
              fun x p =>
                in_tuple_prod pfdt x (snd p)
          end
    end.

  

  Lemma in_nth_tuple_prod :
    forall {T n} {A:T->Type}
           {tpl:T^n} (pfdt:type_eq_dec T)
           i (pfi:i < n)
           (p:tuple_prod A tpl),
      in_tuple_prod pfdt (nth_tuple_prod p i pfi) p. 
    intros T n.
    induction n as [|n ih].
    intros; omega.
    intros A tpl.
    destruct tpl as [ft st].
    intros hdt i.
    destruct i as [|i]; simpl.
    intros.
    pose proof (eq_dec_same hdt ft (hdt ft ft)) as h1.
    rewrite h1, transfer_dep_eq_refl.
    left; auto.
    intros hi p.
    destruct p as [fp sp]. 
    lr_match_goal'.
    subst; rewrite transfer_dep_eq_refl; simpl. 
    right; auto.  
    specialize (ih A st hdt i (lt_S_n _ _ hi) sp).
    simpl; auto.
  Qed.  

  Lemma in_tuple_prod_iff :
    forall {T n} {A:T->Type} {s:T}
           {tpl:T^n} (pfdt:type_eq_dec T)
           (x:A s) (p:tuple_prod A tpl),
      in_tuple_prod pfdt x p <->
      exists i (pfi:i < n) (pfe:s = nth_prod i pfi tpl),
        transfer_dep pfe x = nth_tuple_prod p i pfi.
    intros T n.
    induction n as [|n ih]; simpl.
    intros; split; [intro; contradiction |
                    intro h1; destruct h1 as [? [? h1]]; omega].
    intros A s tpl hdt x p. 
    destruct tpl as [ft st], p as [fp sp] ; simpl; simpl in sp, fp, ih.
    specialize (ih A s st hdt x sp).
    lr_match_goal'; subst; simpl.
    rewrite transfer_dep_eq_refl; split; intro h1.
    destruct h1 as [|h1]; subst.
    exists 0, (lt_0_Sn _); simpl.
    exists eq_refl; rewrite transfer_dep_eq_refl; auto.
    rewrite ih in h1. 
    destruct h1 as [i [hi [h2 h3]]]; subst.
    exists (S i), (lt_n_S _ _ hi); simpl.
    ex_goal. 
    f_equal; apply proof_irrelevance.  
    exists hex.
    rewrite transfer_dep_eq_refl in h3; subst.
    termr hex.
    redterm1 y.
    pose proof (proof_irrelevance _ c hi) as he.
    fold c in hex; fold c.
    rewrite <- transfer_dep_eq_iff.
    gterml.
    redterm1 x.
    redterm1 c.
    fold c0.
    rewrite he; auto.
    destruct h1 as [i [hi [h2 h3]]].
    destruct i as [|i]; simpl in h2, h3; subst.
    pose proof (proof_irrelevance _ h2 (eq_refl _)); subst.
    rewrite transfer_dep_eq_refl.
    left; auto.
    rewrite transfer_dep_eq_refl in h3; subst.
    right.
    apply in_nth_tuple_prod.
    rewrite ih; split; intro h1.
    destruct h1 as [i [h1 [h2 h3]]]; subst.
    rewrite transfer_dep_eq_refl in h3; subst.
    exists (S i), (lt_n_S _ _ h1).
    ex_goal.
    f_equal; apply proof_irrelevance.
    exists hex.
    simpl.
    rewrite <- transfer_dep_eq_iff.
    pose proof (proof_irrelevance
                  _ h1
                  (lt_S_n _ _ (lt_n_S _ _ h1))) as h2.
    rewrite <- h2; auto.
    destruct h1 as [i [h1 [h2 h3]]]; subst.
    rewrite transfer_dep_eq_refl in h3; subst.
    destruct i as [|i].
    simpl in hr.
    contradict hr; auto.
    exists i, (lt_S_n _ _ h1); simpl.
    exists (eq_refl _).
    rewrite transfer_dep_eq_refl; auto.
  Qed.

  Lemma in_tuple_prod_r_iff :
    forall {T n} {A:T->Type} {s:T}
           {tpl:T^n} (pfdt:type_eq_dec T)
           (x:A s) (p:tuple_prod A tpl),
      in_tuple_prod pfdt x p <->
      exists i (pfi:i < n) (pfe:s = nth_prod i pfi tpl),
        x = transfer_dep_r pfe (nth_tuple_prod p i pfi).
    intros; rewrite in_tuple_prod_iff; split; intro h1;
    destruct h1 as [i [hi [h2 h3]]]; subst.
    rewrite transfer_dep_eq_refl in h3; subst.
    exists i, hi, eq_refl; rewrite transfer_dep_r_eq_refl; auto.
    exists i, hi, eq_refl; rewrite transfer_dep_eq_refl; auto.
  Qed.
   
  Fixpoint in_tuple_prod'
             {T n} {A:T->Type} {s:T}
  : forall {tpl:T^n}
           (pfdt:type_eq_dec T)
           (pfda:dep_type_eq_dec A)
           (x:A s), tuple_prod A tpl -> Prop :=
    match n
    with
      | 0 => fun _ _ _ _ _ => False
      | S n' =>
        fun tpl pfdt pfda =>
          match (pfdt s (fst tpl)) with
            | left pfe =>
              fun x p =>
                let x' := transfer_dep pfe x in
                let P := in_tuple_prod' pfdt pfda x (snd p) in
                match (pfda _ x' (fst p)) with
                  | left _ => True
                  | right _ => P
                end
            | right _ =>
              fun x p =>
                in_tuple_prod' pfdt pfda x (snd p) 
          end
    end.
  
  Lemma in_tuple_prod_compat' :
    forall {T n} {A:T->Type} {s:T}
           {tpl:T^n}
           (pfdt:type_eq_dec T)
           (pfda:dep_type_eq_dec A)
           (x:A s)
           (p:tuple_prod A tpl),
      in_tuple_prod' pfdt pfda x p <->
      in_tuple_prod pfdt x p.
    intros T n; induction n as [|n ih]; simpl.
    tauto.
    intros.
    lr_match_goal'; subst.
    rewrite transfer_dep_eq_refl.
    lr_if_goal'; subst.
    tauto.
    split; intro h1.
    right; auto.
    rewrite <- ih; auto.
    apply h1.
    destruct h1 as [|h1]; subst.
    contradict hr; auto.
    rewrite ih; auto.
    auto.
  Qed.

  Lemma in_tuple_prod_iff' :
    forall {T n} {A:T->Type} {s:T}
           {tpl:T^n} (pfdt:type_eq_dec T)
           (pfda:dep_type_eq_dec A)
           (x:A s) (p:tuple_prod A tpl),
      in_tuple_prod' pfdt pfda x p <->
      exists i (pfi:i < n) (pfe:s = nth_prod i pfi tpl),
        transfer_dep pfe x = nth_tuple_prod p i pfi.
    intros; rewrite in_tuple_prod_compat', in_tuple_prod_iff;
    tauto.
  Qed.

  
  Lemma in_tuple_prod_r_iff' :
    forall {T n} {A:T->Type} {s:T}
           {tpl:T^n}
           (pfdt:type_eq_dec T)
           (pfda:dep_type_eq_dec A) 
           (x:A s) (p:tuple_prod A tpl),
      in_tuple_prod' pfdt pfda x p <->
      exists i (pfi:i < n) (pfe:s = nth_prod i pfi tpl),
        x = transfer_dep_r pfe (nth_tuple_prod p i pfi).
    intros; rewrite in_tuple_prod_compat', in_tuple_prod_r_iff;
    tauto.
  Qed.

  Lemma in_tuple_prod_dec :
    forall {T n} {A:T->Type} {s:T}
           {tpl:T^n}
           (pfdt:type_eq_dec T)
           (pfda:dep_type_eq_dec A)
           (x:A s)
           (p:tuple_prod A tpl),
      {  in_tuple_prod' pfdt pfda x p} +
      {~ in_tuple_prod' pfdt pfda x p}.
    intros T n.
    induction n as [|n ih]; simpl.
    intros; right; intro; contradiction.
    intros A s tpl hdt hda x p.
    lr_match_goal'; subst.
    lr_if_goal'; subst.
    left; auto.
    auto.
    auto.
  Qed.

  Definition in_tuple_prod_all
             {T} {A:T->Type} {s:T}
             {tpl:all_tuples T}
             (pfdt:type_eq_dec T)
             (x:A s) (p:tuple_prod_all A tpl) : Prop :=
    in_tuple_prod pfdt x p.

  Definition in_tuple_prod_all'
             {T} {A:T->Type} {s:T}
             {tpl:all_tuples T}
             (pfdt:type_eq_dec T)
             (pfda:dep_type_eq_dec A)
             (x:A s) (p:tuple_prod_all A tpl) : Prop :=
    in_tuple_prod' pfdt pfda x p.

  Lemma in_tuple_prod_all_dec :
    forall {T} {A:T->Type} {s:T}
           {tpl:all_tuples T}
           (pfdt:type_eq_dec T)
           (pfda:dep_type_eq_dec A)
           (x:A s) (p:tuple_prod_all A tpl),
      {  in_tuple_prod_all' pfdt pfda x p} +
      {~ in_tuple_prod_all' pfdt pfda x p}.
    intros; apply in_tuple_prod_dec.
  Qed.
  
Fixpoint in_tuple_prod0 {T n} {A:T->Type} :
  forall {x:T^n} {t:T} (z:A t) (y:tuple_prod A x), Prop :=
  match n with
    | 0 => fun _ _ _ _ => False
        | S n' => fun x t z y =>
                    (exists (pf:t = fst x),
                       transfer_dep pf z =
                       fst y) \/
                    (@in_tuple_prod0 T n' A (snd x) t z (snd y))
  end.
         

Lemma in_nth_tuple_prod0 :
  forall {T n} {A:T->Type} {x:T^n} 
         i (pfi:i < n) (y:tuple_prod A x), 
    in_tuple_prod0 (nth_tuple_prod y i pfi) y.
  intros T n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros A x.
  destruct x as [xf xs].
  intro i.
  destruct i as [|i]; simpl.
  intros h1 h2.
  left.
  exists (eq_refl _).
  rewrite transfer_dep_eq_refl; auto.
  intros hi y.
  destruct y as [yf ys]; simpl.
  right; auto.
Qed.

  
Lemma nth_tuple_prod_transfer :
  forall {S T n}
         {U:S->Type}
         {V:T->Type}
         {x:S^n}
         {y:T^n}
         (pfpw:pointwise_ind_eq U V x y)
         i (pfi:i < n)
         (p:tuple_prod U x),
    nth_tuple_prod
      (transfer_tuple_prod
         pfpw p) i pfi =
    transfer (pfpw i pfi) (nth_tuple_prod p i pfi).
  intros S T n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros U V x y.
  destruct x as [x hx].
  destruct y as [y hy].
  intros h1 i.
  destruct i as [|i]; simpl.
  intros h2 p.
  destruct p as [fp sp]; simpl; auto.
  f_equal; apply proof_irrelevance.
  intros hi p.
  destruct p as [fp sp].
  rewrite ih; f_equal; apply proof_irrelevance.
Qed.


Lemma nth_tuple_prod_functional1 :
  forall {T n} {A:T->Type} {t:T^n}
         (p p':tuple_prod A t)
         k k' (pfk:k < n)
         (pfep:p = p')
         (pfe:k = k'),
    nth_tuple_prod p k pfk =
    transfer_dep (nth_prod_functional1
                k' k
                (eq_lt_trans _ _ _ (eq_sym pfe) pfk) pfk t (eq_sym pfe))
                 (nth_tuple_prod p' k' _)
                 (U := fun c => A c).
  intros; subst.
  pose proof (proof_irrelevance _ pfk (eq_lt_trans _ _ _ (eq_sym eq_refl) pfk)) as h2.
  rewrite <- h2, transfer_dep_eq_refl; auto.
Qed.



Lemma nth_tuple_prod_functional2 :
  forall {T n} {A:T->Type} {t:T^n}
         (p:tuple_prod A t)
         k k' (pfk:k < n)
         (pfe:k = k'),
    nth_tuple_prod p k pfk =
    transfer_dep (nth_prod_functional1
                k' k
                (eq_lt_trans _ _ _ (eq_sym pfe) pfk) pfk t (eq_sym pfe))
                 (nth_tuple_prod p k' _)
                 (U := fun c => A c).
  intros; subst;
  pose proof (proof_irrelevance _ pfk (eq_lt_trans _ _ _ (eq_sym eq_refl) pfk)) as h2; 
  rewrite <- h2, transfer_dep_eq_refl; auto.
Qed.

Lemma nth_tuple_prod_functional3 :
  forall {T n} {U:T->Type} {t:T^n}
         (p:tuple_prod U t)
         k (pfk pfk':k < n),
    nth_tuple_prod p k pfk =
    transfer_dep (U := fun c => U c) (nth_prod_functional2 k pfk' pfk t)
                 (nth_tuple_prod p k pfk').
  intros T n U t p k hk hk'.
  pose proof (proof_irrelevance _ hk hk'); subst.
  rewrite transfer_dep_eq_refl; auto.
Qed.


Lemma transfer_tuple_prod_eq_iff :
  forall {T V n}
         {U:T->Type}
         {W:V->Type} 
         {w z:T^n}
         {v:V^n}
         (pfew:pointwise_ind_eq U W w v)
         (pfez:pointwise_ind_eq U W z v)
         (t:tuple_prod U w)
         (u:tuple_prod U z)
         (pfewz:w = z),
    transfer_tuple_prod pfew t =
    transfer_tuple_prod pfez u <->
    (forall i (pfi:i < n),
       transfer (pfew i pfi) (nth_tuple_prod t i pfi) =
       transfer (pfez i pfi) (nth_tuple_prod u i pfi)). 
  intros T V n. 
  induction n as [|n ih]; simpl.
  intros; subst.
  split; intros; try omega; auto.
  intros U W w z v hew hez t u h1; subst.
  pose proof (proof_irrelevance _ hez hew); subst.
  destruct z as [z z'], v as [v v'], u as [u u'], t as [t t']; simpl;
  simpl in hew, u, t, u', t'.
  pose proof hew as h1.
  apply pointwise_ind_eq_to_S in h1; simpl in h1. 
  specialize (ih U W z' z' v'
                 (pointwise_ind_eq_to_S hew)
                 (pointwise_ind_eq_to_S hew)
                 t' u' (eq_refl _)). 
  split; intro h2.
  inversion h2; clear h2.
  rewrite ih in H1.
  intros i hi.
  destruct i as [|i].
  pose proof (proof_irrelevance _ hi (lt_0_Sn _)); subst; auto.
  rename H1 into h2.
  specialize (h2 i (lt_S_n _ _ hi)). 
  terml h2.
  redterm1 x.
  fold c in h2.
  gterml.
  redterm1 x0.
  fold c0. 
  pose proof (proof_irrelevance _ c c0) as h3; rewrite <- h3.
  rewrite h2 at 1; auto. 
  pose proof (h2 0 (lt_0_Sn _)) as h3; simpl in h3. 
  rewrite h3; f_equal. 
  apply ih.
  intros i hi.    
  specialize (h2 _ (lt_n_S _ _ hi)); simpl in h2.
  do 2 erewrite (nth_tuple_prod_functional3
                   _ _ hi
                   (lt_S_n _ _ (lt_n_S _ _ hi))).
  apply transfer_inj in h2.
  rewrite h2; auto.
Qed.




Fixpoint seg_fun_prod_type_to_tuple
         {n} :
  forall {f:seg_fun n Type}
         (g:seg_fun_prod_type f),
    tuple_prod (seg_fun_type_to_ind f) (full_segT_ind n) :=
  match n
  with
    | 0 => fun _ _ => tt
    | S n' => fun f g =>
                let f' := seg_fun_to_S f in
                let fg := transfer_r (seg_fun_type_to_ind0 f) (fst g) in
                let sg :=
                    transfer_tuple_prod
                      (U := seg_fun_type_to_ind f')
                      (pointwise_ind_eq_seg_fun_type_to_ind f)
                      (seg_fun_prod_type_to_tuple
                         (snd g)
                         (n := n')
                         (f := f')) in
                (fg, sg)
  end.




Definition seg_fun_depT_nth {T n} (U:T->Type) (x:T^n) : Type :=
  seg_fun_depT U (fun i (pfi:i < n) => nth_prod i pfi x).


Definition seg_fun_depT_nth_to_S {T n} {U:T->Type} {x:T^S n}
           (f:seg_fun_depT_nth U x) :
  seg_fun_depT_nth U (snd x) :=
  fun i (pfi:i < n) =>
    transfer_dep (U := (fun c => U c))
                 (nth_prod_functional2
                    i
                    (lt_S_n _ _ (lt_n_S _ _ pfi))
                    pfi (snd x)) (seg_fun_depT_to_S f i pfi).

  
Lemma seg_fun_depT_nth_to_S_compat :
  forall {T n} {U:T->Type} (x:T^(S n))
         (f:seg_fun_depT_nth U x)
         i
         (hi:S i < S n),
    seg_fun_depT_nth_to_S f i (lt_S_n i n hi) = f (S i) hi.
  unfold seg_fun_depT_nth_to_S; intros T n.
  intros U x f i hi.
  rewrite <- transfer_dep_eq_iff.
  unfold seg_fun_depT_to_S.
  pose proof (proof_irrelevance
                _ hi
                (lt_n_S _ _ (lt_S_n _ _ hi))) as h2.
  rewrite <- h2; auto.
Qed.  


Fixpoint seg_fun_depT_nth_to_tuple_prod
           {T n} 
           {U:T->Type} : 
  forall (x:T^n)
         (g:seg_fun_depT_nth U x),
    tuple_prod U x :=
  match n with
    | 0 => fun x g => tt
    | S n' => fun x g =>
                let g' := seg_fun_depT_nth_to_S g in
                let sx :=
                    @seg_fun_depT_nth_to_tuple_prod T n' U (snd x) g' in
                
                (g 0 (lt_0_Sn _), sx)
  end.

Lemma seg_fun_depT_nth_to_tuple_prod_S : 
  forall {T n} 
         {U:T->Type}
         (x:T^S n)
         (g:seg_fun_depT_nth U x),
    seg_fun_depT_nth_to_tuple_prod x g =
    (g 0 (lt_0_Sn _),
     seg_fun_depT_nth_to_tuple_prod (snd x) (seg_fun_depT_nth_to_S g)).
  auto.
Qed.


(*Wyckoff*)
Definition transfer_tuple_prod_arity
           {T m n} {U:T->Type}
           {w:T^m}
           (pfe:m = n)
           (p:tuple_prod U w) :
  tuple_prod U (transfer_nprod pfe w) :=
  seg_fun_depT_nth_to_tuple_prod
    (transfer_nprod pfe w)
    (fun i (pfi:i < n) =>
       let pfi' := lt_eq_trans _ _ _ pfi (eq_sym pfe) in
       (transfer_dep_r
          (nth_prod_transfer_prod pfe w i pfi (U := U))
          (nth_tuple_prod p i pfi'))).
 
(*Wyckoff*)
Definition transfer_tuple_prod_arity_r
           {T m n} {U:T->Type}
           {w:T^n}
           (pfe:m = n)
           (p:tuple_prod U w) :
  tuple_prod U (transfer_nprod_r pfe w) :=
  seg_fun_depT_nth_to_tuple_prod
    (transfer_nprod_r pfe w)
    (fun i (pfi:i < m) =>
       let pfi' := lt_eq_trans _ _ _ pfi pfe in
       (transfer_dep_r
          (nth_prod_transfer_prod_r pfe w i pfi (U := U))
          (nth_tuple_prod p i pfi'))).


(*Wyckoff*)
Definition transfer_tuple_prod_eq_refl
           {T n} {U:T->Type}
           {w:T^n}
           (pfe:n = n)
           (p:tuple_prod U (transfer_nprod pfe w)) :
  tuple_prod U w :=
  transfer_tuple_prod (eq_pointwise_ind_eq
                         (transfer_nprod_eq_refl pfe w) U) p.

(*Wyckoff*)
Definition transfer_tuple_prod_arity_to_S_cast 
           {T m n} {U:T->Type}
           {w:T^S m}
           (pfe:S m = S n)
           (p:tuple_prod U (transfer_nprod (S_inj _ _ pfe) (snd w))) :
  tuple_prod
    U
    (map_nprod
         (seg_fun_to_S
            (fun (i : nat) (pfi : i < S n) =>
               match i as i0 return (i0 < S m -> T * T ^ m -> T) with
                 | 0 => fun (_ : 0 < S m) (x : T * T ^ m) => fst x
                 | S i' =>
                   fun (pflt : S i' < S m) (x : T * T ^ m) =>
                     nth_prod i' (lt_S_n i' m pflt) (snd x)
               end (lt_eq_trans i (S n) (S m) pfi (eq_sym pfe)) w))).
  unfold seg_fun_to_S;
  destruct w as [w w']; unfold transfer_nprod in p; simpl in p.
  erewrite f_equal.
  refine p.
  repeat f_equal;
    apply functional_extensionality_dep;
    intro;
    apply functional_extensionality_dep;
    intro; f_equal; apply proof_irrelevance.
Defined.

(*Wyckoff*)
Definition transfer_tuple_prod_map_nprod_S_cast
           {T m n} {U:T->Type}
           {w:T^S m}
           (pfe:S m = S n)
           (p:tuple_prod U
                (map_nprod
                   (seg_fun_to_S
                      (fun (i : nat) (pfi : i < S n) =>
                         match i as i0 return (i0 < S m -> T * T ^ m -> T) with
                           | 0 => fun (_ : 0 < S m) (x : T * T ^ m) => fst x
                           | S i' =>
                             fun (pflt : S i' < S m) (x : T * T ^ m) =>
                               nth_prod i' (lt_S_n i' m pflt) (snd x)
                         end (lt_eq_trans i (S n) (S m) pfi (eq_sym pfe)) w)))) : tuple_prod U (transfer_nprod (S_inj _ _ pfe) (snd w)).
  unfold seg_fun_to_S;
  destruct w as [w w']; unfold transfer_nprod in p; simpl in p.
  erewrite f_equal. 
  refine p.
  unfold transfer_nprod.
  f_equal; apply functional_extensionality_dep;
    intro;
    apply functional_extensionality_dep;
    intro; unfold seg_fun_to_S; f_equal;
    apply proof_irrelevance.
Defined.

(*Wyckoff*)
Lemma map_nprod_nth_prod_snd_eq_refl : 
  forall {T n}
         (w:T^S n)
         (pfe:S n = S n),
    map_nprod
         (seg_fun_to_S
            (fun (i : nat) (pfi : i < S n) =>
               match i as i0 return (i0 < S n -> T * T ^ n -> T) with
                 | 0 => fun (_ : 0 < S n) (x : T * T ^ n) => fst x
                 | S i' =>
                   fun (pflt : S i' < S n) (x : T * T ^ n) =>
                     nth_prod i' (lt_S_n i' n pflt) (snd x)
               end (lt_eq_trans i (S n) (S n) pfi (eq_sym pfe)) w)) =
    snd w.
  intros; rewrite map_nprod_eq_iff;
  unfold seg_fun_to_S; intros; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma transfer_dep_transfer_nprod_eq_refl_S :
  forall {T n} {U:T->Type}
         {w:T}
         {w':T^n}
         (pfe:S n = S n)
         (pfe':transfer_nprod pfe (w, w') = (w, w'))
         (p:U w),
    transfer_dep (U := fun c => U (fst c)) pfe' p =
    p.
  intros; erewrite f_equal_nprod_fst_fun;
  rewrite transfer_dep_eq_refl; auto.
Qed.


(*Wyckoff*)
Definition transfer_tuple_prod_arity_to_S 
           {T m n} {U:T->Type}
           {w:T^S m}
           (pfe:S m = S n)
           (p:tuple_prod U w) :
  tuple_prod U (transfer_nprod (S_inj _ _ pfe) (snd w)) :=
  transfer_tuple_prod_arity (S_inj _ _ pfe) (snd p).


(*Wyckoff*)
Definition transfer_tuple_prod_arity_r_to_S 
           {T m n} {U:T->Type}
           {w:T^S n}
           (pfe:S m = S n)
           (p:tuple_prod U w) :
  tuple_prod U (transfer_nprod_r (S_inj _ _ pfe) (snd w)) :=
  transfer_tuple_prod_arity_r (S_inj _ _ pfe) (snd p).


(*Wyckoff*)
Lemma transfer_seg_fun_depT_nth_to_tuple_prod_nth_prod :
  forall {S T n} (w:S^n) (t:T^n),
    transfer_tuple_prod_const
      (T:=S) (U:=T)
      (seg_fun_depT_nth_to_tuple_prod
         (U := fun _ => T) w
         (fun (j : nat) (pfj : j < n) =>
            nth_prod j pfj t)) = t.
  intros S T n.
  induction n as [|n ih]; simpl.
  intros; apply unit_eq.
  intros w t.
  destruct w as [fw sw],
                t as [ft st]; simpl.
  f_equal.
  specialize (ih sw st).
  rewrite <- ih at 1.
  repeat f_equal.
  apply functional_extensionality_dep.
  intro;
    apply functional_extensionality;
    intro.
  unfold seg_fun_depT_nth_to_S, seg_fun_depT_to_S.
  simpl.
  pose proof (proof_irrelevance
                _ x0
                (lt_S_n _ _ (lt_n_S _ _ x0))) as h1.
  rewrite <- h1.
  gterml.
  redterm1 x1.
  rewrite transfer_dep_eq_refl; auto.
Qed.

(*Wyckoff*)
Lemma seg_fun_depT_nth_to_tuple_prod_nth_prod :
  forall {S T n} (w:S^n) (t:T^n),
      (seg_fun_depT_nth_to_tuple_prod
         (U := fun _ => T) w
         (fun (j : nat) (pfj : j < n) =>
            nth_prod j pfj t)) =
      transfer_to_tuple_prod_const w t.
  intros S T n w t.
  apply inj_transfer_tuple_prod_const.
  rewrite transfer_tuple_prod_const_to_compat.
  apply transfer_seg_fun_depT_nth_to_tuple_prod_nth_prod.
Qed.

(*Wyckoff*)
Lemma seg_fun_depT_nth_to_tuple_prod_eq_iff :
  forall {T n} {U:T->Type} {x:T^n}
         (f:seg_fun_depT_nth U x)
           (w:tuple_prod U x),
    seg_fun_depT_nth_to_tuple_prod x f = w <->
    (forall i (pfi:i < n),
       f i pfi = nth_tuple_prod w i pfi).
  intros T n.
  induction n as [|n ih]; simpl.
  intros; split; intros; subst.
  omega.
  apply unit_eq.
  intros U x f w.
  destruct x as [fx sx],
                w as [fw sw]; simpl.
  simpl in f, fw, sw.
  specialize (ih U sx (seg_fun_depT_nth_to_S f) sw).
  split; intro h1.
  inversion h1; subst.
  intro i.
  destruct i as [|i].
  intros; f_equal; apply proof_irrelevance.
  intro hi.
  hl ih.
  auto.
  rewrite ih in hl.
  specialize (hl i (lt_S_n _ _ hi)).
  rewrite <- hl.
  unfold seg_fun_depT_nth_to_S; simpl.
  symmetry.
  rewrite <- transfer_dep_eq_iff.
  unfold seg_fun_depT_to_S.
  pose proof (proof_irrelevance
                _ hi
                (lt_n_S _ _ (lt_S_n _ _ hi))) as h2.
  rewrite <- h2; auto.
  pose proof (h1 0 (lt_0_Sn _)) as h2; simpl in h2.
  subst.
  f_equal.
  rewrite ih.
  intros i hi.  
  specialize (h1 (S i) (lt_n_S _ _ hi)).
  simpl in h1.
  pose proof (proof_irrelevance
                _ hi
                (lt_S_n _ _ (lt_n_S _ _ hi))) as h3.
  rewrite h3, <- h1, seg_fun_depT_nth_to_S_compat; auto.
Qed.
   
End TupleProdAgain.
